/**
 * @file      kdp_host.h
 * @brief     kdp host lib header file
 * @version   0.1 - 2019-04-19
 * @copyright (c) 2019 Kneron Inc. All right reserved.
 */

#ifndef __KDP_HOST__H__
#define __KDP_HOST__H__

#if defined (__cplusplus) || defined (c_plusplus)
extern "C"{
#endif

#define KDP_UART_DEV   0
#define KDP_USB_DEV    1

#define FD_RES_LENGTH  8
#define FR_RES_LENGTH  2048
#define LM_RES_LENGTH  20
#define LV_RES_LENGTH  4

#define IMG_FORMAT_RGBA8888     0x80000000
#define IMG_FORMAT_RAW8         0x80000020

/* The data byte sequence in DDR memory is [lowest byte]Y0CbY1CrY2CbY3Cr...[highest byte] */
#define IMG_FORMAT_YCbCr422     0x80000037
#define IMG_FORMAT_RGB565       0x80000060

#define THRESHOLD_DEFAULT 0.4f
#define NIR_THRESHOLD_DEFAULT 0.0f
#define RGB_THRESHOLD_DEFAULT 0.0f

#include <stdint.h>


//-----------------------library init ----------------------

/**
 * init the host library
 * return 0 on succeed, -1 on failure
 */
int kdp_lib_init();

/**
 * start the host library to wait for messages
 * return 0 on succeed, -1 on failure
 */
int kdp_lib_start(); //kdp lib start 

/**
 * free the resources used by host lib
 * return 0 on succeed, -1 on failure
 */
int kdp_lib_de_init(); //kdp lib de init

/**
 * init the host lib internal log
 * dir: the directory name of the log file
 * name: the log file name
 * return 0 on succeed, -1 on failure
 */
int kdp_init_log(const char* dir, const char* name); //init log

/**
 * add com device to the host lib
 * type: the device type, only KDP_UART_DEV supported now
 * name: the UART device name
 * return dev idx on succeed, -1 on failure
 */
int kdp_add_dev(int type, const char* name); //add device to communicate
//---------------------- lib init end------------------------


//-----------------------system operation APIs --------------------------
/**
 * request for doing system reset
 * dev_idx: connected device ID. A host can connect several devices
 * reset_mode: specifies the reset mode

 *             0 - no operation
               1 - reset message protocol
               3 - switch to suspend mode
               4 - switch to active mode
               255 - reset whole system
               256 - system shutdown(RTC)
               0x1000xxxx - reset debug output level

 * return value: 0 on succeed, else for error code
 */
int kdp_reset_sys(int dev_idx, uint32_t reset_mode);

/**
 * request for system status 
 * dev_idx: connected device ID. A host can connect several devices
 * device: the name of the device
 * firmware_id: the id of the firmware in device
 * sys_status: system status
 * app_status: application status
 * return value: 0 on succeed, else -1
 */
int kdp_report_sys_status(int dev_idx, uint32_t* device, uint32_t* firmware_id, 
                uint16_t* sys_status, uint16_t* app_status);

/**
 * request for update firmware 
 * dev_idx: connected device ID. A host can connect several devices
 * module_id: the module id of which the firmware to be updated
              0 - no operation
              1 - scpu module
              2 - ncpu module
 * img_buf, buf_len: the fw image buffer and file size 
 * return value: 0 if succeed, else error code
 */
int kdp_update_fw(int dev_idx, uint32_t* module_id, char* img_buf, int buf_len);

/**
 * request for update model 
 * dev_idx: connected device ID. A host can connect several devices
 * model_id: the model id to be updated
 * model_size: the size of the model
 * img_buf, buf_len: the fw image buffer and file size 
 * return value: 0 if succeed, else error code
 */
int kdp_update_model(int dev_idx, uint32_t* model_id, uint32_t model_size, \
                char* img_buf, int buf_len);

//-----------------------system operation APIs end--------------------------

//-----------------------app level APIs--------------------------
/**
 * start the user verification mode 
 * dev_idx: connected device ID. A host can connect several devices
 * img_size: the required image file size will be returned.
 * return 0 on succeed, error code on failure
 */
int kdp_start_verify_mode(int dev_idx, uint32_t* img_size);

/**
 * start the user verification mode with specified threshold 
 * dev_idx: connected device ID. A host can connect several devices
 * img_size: the required image file size will be returned.
 * thresh:  the threshold used to match face recognition result.
            if 0, the default threshold is used.
 * return 0 on succeed, error code on failure
 */
int kdp_start_verify_mode_thresh(int dev_idx, uint32_t* img_size, float thresh);

/**
 * start the user sfid mode with specified threshold, image format 
 * dev_idx: connected device ID. A host can connect several devices
 * img_size: the required image file size will be returned.
 * thresh:  the threshold used to match face recognition result.
            if 0, the default threshold is used.
 * img_width: the width of input image
 * img_height: the height of input image
 * format: the input image format
 * return 0 on succeed, error code on failure
 */
int kdp_start_sfid_mode(int dev_idx, uint32_t* img_size, float thresh, 
        uint16_t width, uint16_t height, uint32_t format);

/**
 * extract the face feature from input image, and compare it with DB,
 * return the matched user ID, and face detection, face recognition result.
 * dev_idx: connected device ID. A host can connect several devices
 * img_buf, buf_len: the image buffer and file size 
 * user_id: matched user ID in DB will be returned.
 * fd_flag: indicate whether fd result is requested and whether fd result is available.
 * fd_res: contains fd result, 8 bytes
 * fr_flag: indicate whether fr result is requested and whether fr result is available.
 * fr_res: contains fr result, 2048 bytes
 * return 0 on succeed, error code on failure
 */
int kdp_verify_user_id_fdfr_res(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                bool* fd_flag, char* fd_res, bool* fr_flag, char* fr_res);

/**
 * extract the face feature from input image, and compare it with DB,
 * return the matched user ID.
 * dev_idx: connected device ID. A host can connect several devices
 * img_buf, buf_len: the image buffer and file size 
 * user_id: matched user ID in DB will be returned.
 * return 0 on succeed, error code on failure
 */
int kdp_verify_user_id(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len);

/**
 * extract the face feature from input image, and compare it with DB,
 * return the matched user ID, and face analysis related result.
 * dev_idx: connected device ID. A host can connect several devices
 * img_buf, buf_len: the image buffer and file size 
 * user_id: matched user ID in DB will be returned.
 * fd_flag: indicate whether fd result is requested and whether fd result is available.
 * fd_res: contains fd result, 8 bytes
 * fr_flag: indicate whether fr result is requested and whether fr result is available.
 * fr_res: contains fr result, 2048 bytes
 * lm_flag: indicate whether lm result is requested and whether lm result is available
 * lm_res:  landmark result 20 bytes
 * live_flag: indicate whether liveness result is requested and whether liveness result is available
 * live_res: liveness detection result
 * return 0 on succeed, error code on failure
 */
int kdp_verify_user_id_all_res(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                bool* fd_flag, char* fd_res, bool* fr_flag, char* fr_res, \
                bool* lm_flag, char* lm_res, bool* live_flag, char* live_res);

/**
 * extract the face feature from input image, and compare it with DB,
 * return the matched user ID, and face analysis related result.
 * dev_idx: connected device ID. A host can connect several devices
 * user_id: matched user ID in DB will be returned.
 * img_buf, buf_len: the image buffer and file size 
 * mask: indicate the requested and responsed flags
   bit 0 - FD result
   bit 1 - LM data
   bit 2 - FM feature map
   bit 3 - liveness
 * res: contains all the related result
 * FD result: 8 bytes
   LM result: 20 bytes
   FR result: 2048 bytes
   liveness:  two uint16_t
 * return 0 on succeed, error code on failure
 */
int kdp_verify_user_id_generic(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                uint16_t* mask, char* res);


/**
 * start the user register mode 
 * dev_idx: connected device ID. A host can connect several devices
 * user_id: the user id that will be registered.
 * img_idx: the image idx that will be saved. (a user could have 5 images)
 * return 0 on succeed, -1 else
 */
int kdp_start_reg_user_mode(int dev_idx, uint16_t usr_id, uint16_t img_idx);

/**
 * extract face feature from inut image and save it in device,
 * returns face detection, face recoginition result.
 * dev_idx: connected device ID. A host can connect several devices
 * img_buf, buf_len: the image buffer and file size 
 * fd_flag: indicate whether fd result is requested and whether fd result is available.
 * fd_res: contains fd result, 8 bytes
 * fr_flag: indicate whether fr result is requested and whether fr result is available.
 * fr_res: contains fr result, 2048 bytes
 * before calling this API, host must call kdp_start_reg_user_mode to enable register mode,
 * and specifed for which user and which index to be extracted.
 * return 0 on succeed, error code on failure
 */
int kdp_extract_feature_fdfr_res(int dev_idx, char* img_buf, int buf_len, \
                bool* fd_flag, char* fd_res, bool* fr_flag, char* fr_res);
/**
 * extract face feature from inut image and save it in device
 * dev_idx: connected device ID. A host can connect several devices
 * img_buf, buf_len: the image buffer and file size 
 * before calling this API, host must call kdp_start_reg_user_mode to enable register mode,
 * and specifed for which user and which index to be extracted.
 * return 0 on succeed, error code on failure
 */
int kdp_extract_feature(int dev_idx, char* img_buf, int buf_len);

/**
 * extract face feature from inut image and save it in device,
 * returns face detection, face recoginition result.
 * dev_idx: connected device ID. A host can connect several devices
 * img_buf, buf_len: the image buffer and file size 
 * fd_flag: indicate whether fd result is requested and whether fd result is available.
 * fd_res: contains fd result, 8 bytes
 * fr_flag: indicate whether fr result is requested and whether fr result is available.
 * fr_res: contains fr result, 2048 bytes
 * lm_flag: indicate whether lm result is requested and whether lm result is available
 * lm_res:  landmark result 20 bytes
 * live_flag: indicate whether liveness result is requested and whether liveness result is available
 * live_res: liveness detection result
 * before calling this API, host must call kdp_start_reg_user_mode to enable register mode,
 * and specifed for which user and which index to be extracted.
 * return 0 on succeed, error code on failure
 */
int kdp_extract_feature_all_res(int dev_idx, char* img_buf, int buf_len, \
                bool* fd_flag, char* fd_res, bool* fr_flag, char* fr_res, \
                bool* lm_flag, char* lm_res, bool* live_flag, char* live_res);

/**
 * extract face feature from inut image and save it in device,
 * returns face detection, face recoginition result.
 * dev_idx: connected device ID. A host can connect several devices
 * img_buf, buf_len: the image buffer and file size 
 * mask: indicate the requested and responsed flags
   bit 0 - FD result
   bit 1 - LM data
   bit 2 - FM feature map
   bit 3 - liveness
 * res: contains all the related result
 * FD result: 8 bytes
   LM result: 20 bytes
   FR result: 2048 bytes
   liveness:  two uint16_t
 * return 0 on succeed, error code on failure
 */
int kdp_extract_feature_generic(int dev_idx, char* img_buf, int buf_len, \
                uint16_t* mask, char* res);

/**
 * register the face features to device DB
 * dev_idx: connected device ID. A host can connect several devices
 * user_id: the user id that be registered. must be same as the kdp_start_reg_user_mode.
 * before calling this API, host must have called kdp_extract_feature
 * at least once successfully, otherwise, no features could be saved.
 * return 0 on succeed, error code on failure
 */
int kdp_register_user(int dev_idx, uint32_t user_id);

/**
 * remove user from device DB
 * It needs to be called after start lw3d mode or start verify mode
 * dev_idx: connected device ID. A host can connect several devices
 * user_id: the user to be removed. 0 for all users
 * return 0 on succeed, error code on failure
 */
int kdp_remove_user(int dev_idx, uint32_t user_id);

/**
 * start the light weight 3D mode  with specified threshold, img format 
 * dev_idx: connected device ID. A host can connect several devices
 * rgb_size: the required rgb image file size will be returned.
 * nir_size: the required nir image file size will be returned.
 * rgb_thresh:  the threshold used to match rgb face recognition result.
            if 0, the default threshold is used.
 * nir_thresh:  the threshold used to match nir face recognition result.
            if 0, the default threshold is used.
 * rgb_width: the width of rgb image
 * rgb_height: the height of rgb image
 * nir_width: the width of nir image
 * nir_height: the height of nir image
 * rgb_fmt:   the format of rgb image
 * nir_fmt:  the format of nir image
 * return 0 on succeed, error code on failure
 */
int kdp_start_lw3d_mode(int dev_idx, uint32_t* rgb_size, uint32_t* nir_size, 
        float rgb_thresh, float nir_thresh, uint16_t rgb_width, uint16_t rgb_height, 
        uint16_t nir_width, uint16_t nir_height, uint32_t rgb_fmt, uint32_t nir_fmt);

/**
 * start the light weight 3D mode  with specified threshold 
 * dev_idx: connected device ID. A host can connect several devices
 * rgb_size: the required rgb image file size will be returned.
 * nir_size: the required nir image file size will be returned.
 * rgb_thresh:  the threshold used to match rgb face recognition result.
            if 0, the default threshold is used.
 * nir_thresh:  the threshold used to match nir face recognition result.
            if 0, the default threshold is used.
 * return 0 on succeed, error code on failure
 */
int kdp_start_lw3d_mode_thresh(int dev_idx, uint32_t* rgb_size, uint32_t* nir_size, 
                float rgb_thresh, float nir_thresh);


/**
 * extract the face feature from input rgb/nir image, and save it in device,
 * return the face analysis related result if required.
 * dev_idx: connected device ID. A host can connect several devices
 * img_buf, buf_len: the rgb image buffer and file size 
 * img_buf_nir, buf_len_nir: the nir image buffer and file size 
 * return 0 on succeed, error code on failure
 */
int kdp_extract_lw3D_feature(int dev_idx, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir);


/**
 * extract the face feature from input rgb/nir image, and save it in device,
 * return the face analysis related result if required.
 * dev_idx: connected device ID. A host can connect several devices
 * img_buf, buf_len: the rgb image buffer and file size 
 * img_buf_nir, buf_len_nir: the nir image buffer and file size 
 * fd_flag/fd_flag_nir: indicate whether fd result is requested and whether fd result is available.
 * fd_res/fd_res_nir: contains fd result, 8 bytes
 * fr_flag/fr_flag_nir: indicate whether fr result is requested and whether fr result is available.
 * fr_res/fr_res_nir: contains fr result, 2048 bytes
 * lm_flag/lm_flag_nir: indicates whether lankmwark result is requested and whether result is available.
 * lm_res/lm_res_nir: contains landmark result, 20 bytes
 * live_flag: indicates whether liveness detection is requested and whether result is available
 * live_res: contains liveness detection result, two uint16_t
 * return 0 on succeed, error code on failure
 */
int kdp_extract_lw3D_feature_all_res(int dev_idx, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir, bool* fd_flag, char* fd_res, \
                bool* fr_flag, char* fr_res, bool* lm_flag, char* lm_res, \
                bool* fd_flag_nir, char* fd_res_nir, bool* fr_flag_nir, char* fr_res_nir, \
                bool* lm_flag_nir, char* lm_res_nir, bool* live_flag, char* live_res);

/**
 * extract the face feature from input rgb/nir image, and save it in device,
 * return the face analysis related result if required.
 * dev_idx: connected device ID. A host can connect several devices
 * img_buf, buf_len: the rgb image buffer and file size 
 * img_buf_nir, buf_len_nir: the nir image buffer and file size 

 * mask: indicate the requested and responsed flags
   bit 0 - FD result
   bit 1 - FM feature map
   bit 2 - LM data
   bit 4 - NIR FD result
   bit 5 - NIR feature map
   bit 6 - NIR LM data
   bit 8 - Final liveness
 * res: contains all the related result
 * FD result: 8 bytes
   LM result: 20 bytes
   FR result: 2048 bytes
 * NIR FD result: 8 bytes
   NIR LM result: 20 bytes
   NIR FR result: 2048 bytes
   liveness:  two uint16_t

 * return 0 on succeed, error code on failure
 */
int kdp_extract_lw3D_feature_generic(int dev_idx, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir, uint16_t* mask, char* res);



/**
 * extract the face feature from input rgb/nir image, and compare it with DB,
 * return the matched user ID, and face analysis related result.
 * dev_idx: connected device ID. A host can connect several devices
 * user_id: matched user ID in DB will be returned.
 * img_buf, buf_len: the rgb image buffer and file size 
 * img_buf_nir, buf_len_nir: the nir image buffer and file size 
 * return 0 on succeed, error code on failure
 */
int kdp_verify_lw3D_user(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir);

/**
 * extract the face feature from input rgb/nir image, and compare it with DB,
 * return the matched user ID, and face analysis related result.
 * dev_idx: connected device ID. A host can connect several devices
 * user_id: matched user ID in DB will be returned.
 * img_buf, buf_len: the rgb image buffer and file size 
 * img_buf_nir, buf_len_nir: the nir image buffer and file size 
 * fd_flag/fd_flag_nir: indicate whether fd result is requested and whether fd result is available.
 * fd_res/fd_res_nir: contains fd result, 8 bytes
 * fr_flag/fr_flag_nir: indicate whether fr result is requested and whether fr result is available.
 * fr_res/fr_res_nir: contains fr result, 2048 bytes
 * lm_flag/lm_flag_nir: indicates whether lankmwark result is requested and whether result is available.
 * lm_res/lm_res_nir: contains landmark result, 20 bytes
 * live_flag: indicates whether liveness detection is requested and whether result is available
 * live_res: contains liveness detection result, two uint16_t
 * return 0 on succeed, error code on failure
 */
int kdp_verify_lw3D_image_all_res(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir, bool* fd_flag, char* fd_res, \
                bool* fr_flag, char* fr_res, bool* lm_flag, char* lm_res, \
                bool* fd_flag_nir, char* fd_res_nir, bool* fr_flag_nir, char* fr_res_nir, \
                bool* lm_flag_nir, char* lm_res_nir, bool* live_flag, char* live_res);

/**
 * extract the face feature from input rgb/nir image, and compare it with DB,
 * return the matched user ID, and face analysis related result.
 * dev_idx: connected device ID. A host can connect several devices
 * user_id: matched user ID in DB will be returned.
 * img_buf, buf_len: the rgb image buffer and file size 
 * img_buf_nir, buf_len_nir: the nir image buffer and file size 

 * mask: indicate the requested and responsed flags
   bit 0 - FD result
   bit 1 - FM feature map
   bit 2 - LM data
   bit 4 - NIR FD result
   bit 5 - NIR feature map
   bit 6 - NIR LM data
   bit 8 - Final liveness
 * res: contains all the related result
 * FD result: 8 bytes
   LM result: 20 bytes
   FR result: 2048 bytes
 * NIR FD result: 8 bytes
   NIR LM result: 20 bytes
   NIR FR result: 2048 bytes
   liveness:  two uint16_t

 * return 0 on succeed, error code on failure
 */
int kdp_verify_lw3D_image_generic(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir, uint16_t* mask, char* res);

//-----------------------app level APIs end-------------------------------------


//-----------------------dynamic model execution APIs --------------------------
/**
 * request for starting dynamic model execution
 * dev_idx: connected device ID. A host can connect several devices
 * model_size: size of inference model
 * data: firmware setup data
 * dat_size: setup data size
 * ret_size: returned model size
 * img_buf, buf_len: the model file buffer and file size 
 * return 0 on succeed, error code on failure
 */
int kdp_start_dme(int dev_idx, uint32_t model_size, char* data, int dat_size, \
                uint32_t* ret_size, char* img_buf, int buf_len);

/**
 * request for configuring dme 
 * dev_idx: connected device ID. A host can connect several devices
 * data: inference setup data
 * dat_size: the size of setup data
 * return 0 on succeed, error code on failure
 */
int kdp_dme_configure(int dev_idx, char* data, int dat_size);

/**
 * request for configuring dme 
 * dev_idx: connected device ID. A host can connect several devices
 * setup_conf: specifies how the setup data to be used.
 * data: inference setup data
 * dat_size: the size of setup data
 * return 0 on succeed, error code on failure
 */
int kdp_dme_configure_param(int dev_idx, uint32_t setup_conf, char* data, int dat_size);

/**
 * calling KL520 to do inference with provided model 
 * dev_idx: connected device ID. A host can connect several devices
 * img_buf, buf_len: the image buffer and file size 
 * inf_size: the size of inference result
 * res_flag: indicate whether result is requested and available
 * inf_res: contains the returned inference result
 * before calling this API, host must call kdp_start_dme and kdp_dme_configure to configure the dme model.
 * return 0 on succeed, error code on failure
 */
int kdp_dme_inference(int dev_idx, char* img_buf, int buf_len, uint32_t* inf_size, \
                bool* res_flag, char* inf_res);

/**
 * request for retrieving dme result 
 * dev_idx: connected device ID. A host can connect several devices
 * inf_size: inference data size
 * inf_res: inference result data
 * return 0 on succeed, error code on failure
 */
//int kdp_get_dme_res(int dev_idx, uint32_t* inf_size, char* inf_res);

//-----------------------dynamic model execution APIs end --------------------------

#if defined (__cplusplus) || defined (c_plusplus)
}
#endif

#endif
