/**
 * @file      KdpHostApi.cpp
 * @brief     kdp host api source file
 * @version   0.1 - 2019-04-19
 * @copyright (c) 2019 Kneron Inc. All right reserved.
 */

#include <string.h>
#include <unistd.h>
#include <math.h>

#include "KdpHostApi.h"
#include "err_code.h"
#include "MsgMgmt.h"
#include "KnLog.h"
#include "KnUtil.h"
#include "UARTDev.h"

#include <libusb-1.0/libusb.h>

//#include <thread>
#include <vector>
using namespace std;

#define FLOAT_ZERO_VAL 0.000001
#define MAX_THRESH_VAL 1.0f
#define MIN_THRESH_VAL 0.0f

MsgMgmt*                _p_mgr = NULL; //msg manager
AsyncMsgHandler*        _async_hdlr = NULL;
SyncMsgHandler*         _sync_hdlr = NULL;

static bool is_valid_dev(int dev_idx) {
    if(!_p_mgr) return false;
    return _p_mgr->is_valid_dev(dev_idx);
}

//init the host lib
int kdp_lib_init() {
    //_async_hdlr = new AsyncMsgHandler();
    libusb_init(NULL); //init usb lib
    _sync_hdlr = new SyncMsgHandler();
    _p_mgr = new MsgMgmt(_async_hdlr, _sync_hdlr);

    KnPrint(KN_LOG_DBG, "msg handler and msg mgr created.\n");
    return 0;
}

//start the host lib
int kdp_lib_start() {
    if(!_p_mgr) return -1;
    //start receiver, sender thread
    
    _p_mgr->start_receiver();
    
    //_p_mgr->start_sender();

    //start msg handling thread
    _p_mgr->start_handler();
    KnPrint(KN_LOG_DBG, "handler thread and receiver thread started.\n");
    return 0;
}

//init the lib log
//dir: directory of the log file
//name: file name
int kdp_init_log(const char* dir, const char* name) {
    string dirs(dir);
    string names(name);
    KnLogInit(dirs, names);
    KnSetLogLevel(KN_LOG_ALL);

    KnPrint(KN_LOG_DBG, "\n\n\n************----**********\n");
    KnPrint(KN_LOG_DBG, " -- kdp host lib restarted -- \n");
    KnPrint(KN_LOG_DBG, "************----**********\n\n\n");

    KnPrint(KN_LOG_DBG, "init host lib log to %s%s.\n", dir, name);
    return 0;
}

//add device to lib
//currently only UART and USB are supported.
//name: device name
int kdp_add_dev(int type, const char* name) {
    if(type == KDP_UART_DEV) {
    
        string u_name(name);
        BaseDev* pDev = new UARTDev(u_name);
        if(pDev->get_dev_fd() < 0) {
            KnPrint(KN_LOG_ERR, "dev:%s could not be opened.\n", name);
            delete pDev;
            return -1;
        }

        pDev->set_baudrate(115200);
        pDev->set_parity();

        KnPrint(KN_LOG_DBG, "adding dev for:%s...\n", name);
        int ret = _p_mgr->add_device(pDev);
        if(ret < 0) delete pDev;
        return ret;
    } else if (type == KDP_USB_DEV) {
        USBDev* pDev = new USBDev(VENDOR_ID, PRODUCT_ID);
        KnPrint(KN_LOG_DBG, "created usb dev %x,%x.\n", VENDOR_ID, PRODUCT_ID);

        KnPrint(KN_LOG_DBG, "trying to init usb port..\n");
        if(pDev->u_init_port() < 0) {
            KnPrint(KN_LOG_ERR, "USB dev init port failed.\n");
            delete pDev;
            return -1;
        } else {
            KnPrint(KN_LOG_DBG, "trying to open usb device..\n");
            if(pDev->u_open() < 0) {
                KnPrint(KN_LOG_ERR, "USB dev name open failed.\n");
                delete pDev;
                return -1;
            }
        }

        KnPrint(KN_LOG_DBG, "adding usb dev ...\n");
        int ret = _p_mgr->add_usb_device(pDev);
        if(ret < 0) delete pDev;
        return ret;
    } else {
    	KnPrint(KN_LOG_DBG, "No identifiable device\n");
        return -1;
    }
}

//free the resources
int kdp_lib_de_init() {
    if(_p_mgr) {
        _p_mgr->stop_receiver();
        //_p_mgr->stop_sender();
        _p_mgr->stop_handler();

        for(int i = 0; i < 5000; i++) {
            usleep(1000);
            if(_p_mgr->is_receiver_running() == false &&
              //_p_mgr->is_sender_running() == false && 
                _p_mgr->is_handler_running() == false) break;
        }
    }

    usleep(1000);
    KnPrint(KN_LOG_DBG, "threads stopped, starting deconstructing objs...\n");
    if(_p_mgr) delete _p_mgr;
    if(_async_hdlr) delete _async_hdlr;
    if(_sync_hdlr) delete _sync_hdlr;

    libusb_exit(NULL);
    //close log
    KnLogClose();
    return 0;
}

int kdp_query_app_sync(int dev_idx, uint32_t* app_ids, int* size) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_query_app(dev_idx, app_ids, size);
}

int kdp_select_app_sync(int dev_idx, uint32_t app_id, uint32_t* rsp_code, uint32_t* buf_size) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_select_app(dev_idx, app_id, rsp_code, buf_size);
}

int kdp_sfid_start_sync(int dev_idx, uint32_t* rsp_code, uint32_t* img_size, float thresh,
                uint16_t width, uint16_t height, uint32_t format) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_sfid_start(dev_idx, rsp_code, img_size, thresh, 
                width, height, format);
}

int kdp_sfid_start_lw3d_sync(int dev_idx, uint32_t* rsp_code, uint32_t* rgb_size, \
            uint32_t* nir_size, float rgb_thresh, float nir_thresh, uint16_t rgb_width, \
            uint16_t rgb_height, uint16_t nir_width, uint16_t nir_height, \
            uint32_t rgb_fmt, uint32_t nir_fmt) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_sfid_start_lw3d(dev_idx, rsp_code, 
            rgb_size, nir_size, rgb_thresh, nir_thresh, rgb_width, rgb_height,
            nir_width, nir_height, rgb_fmt, nir_fmt);
}

int kdp_sfid_new_user_sync(int dev_idx, uint16_t usr_id, uint16_t img_idx) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_sfid_new_user(dev_idx, usr_id, img_idx);
}

int kdp_sfid_register_sync(int dev_idx, uint32_t usr_id, uint32_t* rsp_code) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_sfid_register(dev_idx, usr_id, rsp_code);
}

int kdp_sfid_del_db_sync(int dev_idx, uint32_t usr_id, uint32_t* rsp_code) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_sfid_del_db(dev_idx, usr_id, rsp_code);
}

int kdp_sfid_send_img_sync(int dev_idx, char* img_buf, int buf_len, uint32_t* rsp_code, \
        uint16_t* user_id, uint16_t* mode, bool* fd_flag, char* fd_res, \
        bool* fr_flag, char* fr_res, \
        bool* lm_flag, char* lm_res, bool* live_flag, char* live_res) {
    if(!_sync_hdlr) return -1;
    
    return _sync_hdlr->kdp_sfid_send_img(dev_idx, img_buf, buf_len, rsp_code, user_id, mode,
                fd_flag, fd_res, fr_flag, fr_res, lm_flag, lm_res, live_flag, live_res);
}

int kdp_sfid_send_lw3d_img_sync(int dev_idx, char* img_buf, int buf_len, \
        char* img_buf_nir, int buf_len_nir, uint32_t* rsp_code, \
        uint16_t* user_id, uint16_t* mode, bool* fd_flag, char* fd_res, \
        bool* fr_flag, char* fr_res, bool* lm_flag, char* lm_res, \
        bool* fd_flag_nir, char* fd_res_nir, bool* fr_flag_nir, \
        char* fr_res_nir, bool* lm_flag_nir, char* lm_res_nir, \
        bool* live_flag, char* live_res) {
    if(!_sync_hdlr) return -1;
    
    return _sync_hdlr->kdp_sfid_send_lw3d_img(dev_idx, img_buf, buf_len, \
                img_buf_nir, buf_len_nir, rsp_code, user_id, mode,
                fd_flag, fd_res, fr_flag, fr_res, lm_flag, lm_res, 
                fd_flag_nir, fd_res_nir, fr_flag_nir, fr_res_nir,
                lm_flag_nir, lm_res_nir, live_flag, live_res);
}

int kdp_sfid_send_img_gen(int dev_idx, char* img_buf, int buf_len, uint32_t* rsp_code, \
        uint16_t* user_id, uint16_t* mode, uint16_t* mask, char* res) {
    if(!_sync_hdlr) return -1;
    
    return _sync_hdlr->kdp_sfid_send_img_gen(dev_idx, img_buf, buf_len, rsp_code, \
                user_id, mode, mask, res);
}

int kdp_sfid_send_lw3d_img_gen(int dev_idx, char* img_buf, int buf_len, \
        char* img_buf_nir, int buf_len_nir, uint32_t* rsp_code, \
        uint16_t* user_id, uint16_t* mode, uint16_t* mask, char* res) {
    if(!_sync_hdlr) return -1;
    
    return _sync_hdlr->kdp_sfid_send_lw3d_gen(dev_idx, img_buf, buf_len, img_buf_nir, \
                buf_len_nir, rsp_code, user_id, mode, mask, res);
}

// -------------------system operation APIs
int kdp_reset_sys(int dev_idx, uint32_t reset_mode) {
    KnPrint(KN_LOG_DBG, "\n\ncalling reset sys...:%d,%x.\n", dev_idx, reset_mode);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_reset_sys_sync(dev_idx, reset_mode);
}

int kdp_report_sys_status(int dev_idx, uint32_t* device, uint32_t* firmware_id,
            uint16_t* sys_status, uint16_t* app_status) {
    KnPrint(KN_LOG_DBG, "\n\ncalling report sys status...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_report_sys_status_sync(dev_idx, device, \
                firmware_id, sys_status, app_status);
}

int kdp_update_fw(int dev_idx, uint32_t* module_id, char* img_buf, int buf_len) {
    KnPrint(KN_LOG_DBG, "\n\ncalling updating firmware...:%d, %d.\n", dev_idx, *module_id);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_update_fw_sync(dev_idx, module_id, img_buf, buf_len);
}


int kdp_update_model(int dev_idx, uint32_t* model_id, uint32_t model_size, \
                    char* img_buf, int buf_len) {
    KnPrint(KN_LOG_DBG, "\n\ncalling updating model...:%d, %d.\n", dev_idx, *model_id);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_update_model_sync(dev_idx, model_id, \
                model_size, img_buf, buf_len);
}


//-------------------------------app level APIs
int kdp_start_verify_mode(int dev_idx, uint32_t* img_size) {
    return kdp_start_verify_mode_thresh(dev_idx, img_size, 0.0f);
}

int kdp_start_verify_mode_thresh(int dev_idx, uint32_t* img_size, float thresh) {
    return kdp_start_sfid_mode(dev_idx, img_size, thresh, 640, 480, IMG_FORMAT_RGB565);
}

int kdp_start_sfid_mode(int dev_idx, uint32_t* img_size, float thresh,
            uint16_t width, uint16_t height, uint32_t format) {
    KnPrint(KN_LOG_DBG, "\n\ncalling start verify mode...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }
    if(!img_size) {
        KnPrint(KN_LOG_ERR, "img size is null.\n");
        return -1;
    }

    if(fabsf(thresh) < FLOAT_ZERO_VAL) {
        //0.0f
    } else {
        if(thresh > MAX_THRESH_VAL || thresh < MIN_THRESH_VAL) {
            KnPrint(KN_LOG_ERR, "threshold val is invalid, range:%5.2f ~ %5.2f.\n", 
                    MIN_THRESH_VAL, MAX_THRESH_VAL);
            return -1;
        }
    }

    if(width == 0) width = 640;
    if(height == 0) height = 480;

    uint32_t rsp_code = 0;
    int ret = kdp_sfid_start_sync(dev_idx, &rsp_code, img_size, 
                thresh, width, height, format);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "start verify mode failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "start verify mode failed with error:%d.\n", rsp_code);
        return rsp_code;
    }
    return 0;
}


int kdp_verify_user_id_all_res(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                bool* fd_flag, char* fd_res, bool* fr_flag, char* fr_res, \
                bool* lm_flag, char* lm_res, bool* live_flag, char* live_res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling verify user id...:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!user_id || !img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "verify user id input parameter is invalid.\n");
        return -1;
    }

    if(fd_flag && (*fd_flag == true)) {
        if(fd_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for fd flag, result.\n");
            return -1;
        }
    }

    if(fr_flag &&(*fr_flag == true)) {
        if(fr_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for fr flag, result.\n");
            return -1;
        }
    }

    if(lm_flag && (*lm_flag == true)) {
        if(lm_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for lm flag, result.\n");
            return -1;
        }
    }

    if(live_flag &&(*live_flag == true)) {
        if(live_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for live flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t mode = 1;
    int ret = kdp_sfid_send_img_sync(dev_idx, img_buf, buf_len, &rsp_code, user_id, &mode,
                fd_flag, fd_res, fr_flag, fr_res, lm_flag, lm_res, live_flag, live_res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "verify user id failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "verify user id failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    if(mode != 0) {
        KnPrint(KN_LOG_ERR, "device is not in user verify mode.\n");
        return -1;
    }

    KnPrint(KN_LOG_DBG, "verify user id completed, user id:%d.\n", *user_id);
    if(*user_id < 1 || *user_id > 20) {
        KnPrint(KN_LOG_ERR, "invalid user id received:%d.\n", *user_id);
        return -1;
    }
    return 0;
}
int kdp_verify_user_id_generic(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                uint16_t* mask, char* res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling verify user id...:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!user_id || !img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "verify user id input parameter is invalid.\n");
        return -1;
    }

    if( mask && (*mask & 0x0f) ) {
        if(res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for res flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t mode = 1;
    int ret = kdp_sfid_send_img_gen(dev_idx, img_buf, buf_len, &rsp_code, user_id, &mode,
                mask, res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "verify user id failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "verify user id failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    if(mode != 0) {
        KnPrint(KN_LOG_ERR, "device is not in user verify mode.\n");
        return -1;
    }

    KnPrint(KN_LOG_DBG, "verify user id completed, user id:%d.\n", *user_id);
    if(*user_id < 1 || *user_id > 20) {
        KnPrint(KN_LOG_ERR, "invalid user id received:%d.\n", *user_id);
        return -1;
    }
    return 0;
}

int kdp_verify_user_id_fdfr_res(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                bool* fd_flag, char* fd_res, bool* fr_flag, char* fr_res) {
    return kdp_verify_user_id_all_res(dev_idx, user_id, img_buf, buf_len,
                    fd_flag, fd_res, fr_flag, fr_res, NULL, NULL, NULL, NULL);
}

int kdp_verify_user_id(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len) {
    return kdp_verify_user_id_fdfr_res(dev_idx, user_id, img_buf, buf_len,
                    NULL, NULL, NULL, NULL);
}

int kdp_start_reg_user_mode(int dev_idx, uint16_t usr_id, uint16_t img_idx) {
    KnPrint(KN_LOG_DBG, "\n\ncalling start reg user mode...:%d,%d.\n", usr_id, img_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }
    if(usr_id < 1 || usr_id > 20 || img_idx < 1 || img_idx > 5) {
        KnPrint(KN_LOG_ERR, "new user input parameter is invalid.\n");
        return -1;
    }

    int ret =  kdp_sfid_new_user_sync(dev_idx, usr_id, img_idx);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "start reg mode failed with error:%d.\n", ret);
        return -1;
    }

    return 0;
}


int kdp_extract_feature_all_res(int dev_idx, char* img_buf, int buf_len, \
                bool* fd_flag, char* fd_res, bool* fr_flag, char* fr_res, \
                bool* lm_flag, char* lm_res, bool* live_flag, char* live_res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling extract feature...:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }
    if(!img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "extract feature input parameter is invalid.\n");
        return -1;
    }

    if(fd_flag && (*fd_flag == true)) {
        if(fd_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for fd flag, result.\n");
            return -1;
        }
    }

    if(fr_flag &&(*fr_flag == true)) {
        if(fr_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for fr flag, result.\n");
            return -1;
        }
    }


    if(lm_flag && (*lm_flag == true)) {
        if(lm_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for lm flag, result.\n");
            return -1;
        }
    }

    if(live_flag &&(*live_flag == true)) {
        if(live_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for live flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t img_idx = 0;
    uint16_t mode = 0;
    int ret = kdp_sfid_send_img_sync(dev_idx, img_buf, buf_len, &rsp_code, &img_idx, &mode, 
                fd_flag, fd_res, fr_flag, fr_res, lm_flag, lm_res, live_flag, live_res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "extract feature failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "extract feature failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    if(mode == 0) {
        KnPrint(KN_LOG_ERR, "device is not in register mode.\n");
        return -1;
    }

    KnPrint(KN_LOG_DBG, "extract feature completed, img index:%d.\n", img_idx);
    if(img_idx < 1 || img_idx > 5) {
        KnPrint(KN_LOG_ERR, "invalid image received:%d.\n", img_idx);
        return -1;
    }
    return 0;
}


int kdp_extract_feature_generic(int dev_idx, char* img_buf, int buf_len, \
                uint16_t* mask, char* res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling extract feature generic..:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "extract feature input parameter is invalid.\n");
        return -1;
    }

    if( mask && (*mask & 0x0f) ) {
        if(res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for res flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t img_idx = 0;
    uint16_t mode = 0;
    int ret = kdp_sfid_send_img_gen(dev_idx, img_buf, buf_len, &rsp_code, &img_idx, &mode,
                mask, res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "extract feature failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "extract feature failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    if(mode == 0) {
        KnPrint(KN_LOG_ERR, "device is not in user register mode.\n");
        return -1;
    }

    KnPrint(KN_LOG_DBG, "extract feature completed, img index:%d.\n", img_idx);
    if(img_idx < 1 || img_idx > 5) {
        KnPrint(KN_LOG_ERR, "invalid image received:%d.\n", img_idx);
        return -1;
    }
    return 0;
}

int kdp_extract_feature_fdfr_res(int dev_idx, char* img_buf, int buf_len, \
                bool* fd_flag, char* fd_res, bool* fr_flag, char* fr_res) {
    return kdp_extract_feature_all_res(dev_idx, img_buf, buf_len,
                    fd_flag, fd_res, fr_flag, fr_res, NULL, NULL, NULL, NULL);
}

int kdp_extract_feature(int dev_idx, char* img_buf, int buf_len) {
    return kdp_extract_feature_fdfr_res(dev_idx, img_buf, buf_len, NULL, NULL, NULL, NULL);
}

int kdp_register_user(int dev_idx, uint32_t user_id) {
    KnPrint(KN_LOG_DBG, "\n\ncalling register user...:%d,%d.\n", dev_idx, user_id);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }
    if(user_id < 1 || user_id > 20) {
        KnPrint(KN_LOG_ERR, "reg user input parameter is invalid.\n");
        return -1;
    }

    uint32_t rsp_code = 0;
    int ret = kdp_sfid_register_sync(dev_idx, user_id, &rsp_code);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "reg user failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "reg user failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    KnPrint(KN_LOG_DBG, "reg user completed, user id:%d.\n", user_id);
    return 0;
}

int kdp_remove_user(int dev_idx, uint32_t user_id) {
    KnPrint(KN_LOG_DBG, "\n\ncalling remove user...:%d,%d.\n", dev_idx, user_id);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }
    if(user_id < 0 || user_id > 20) {
        KnPrint(KN_LOG_ERR, "del user input parameter is invalid.\n");
        return -1;
    }

    uint32_t rsp_code = 0;
    int ret = kdp_sfid_del_db_sync(dev_idx, user_id, &rsp_code);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "remove all user failed for dev:%d, with error:%d.\n", dev_idx, ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "remove all user failed for dev:%d, with error:%d.\n", dev_idx, rsp_code);
        return rsp_code;
    }
    return 0;
}

int kdp_start_lw3d_mode_thresh(int dev_idx, uint32_t* rgb_size, uint32_t* nir_size,
        float rgb_thresh, float nir_thresh) {
    return kdp_start_lw3d_mode(dev_idx, rgb_size, nir_size, rgb_thresh, nir_thresh, 
            640, 480, 640, 480, IMG_FORMAT_RGB565, IMG_FORMAT_RAW8);
}

int kdp_start_lw3d_mode(int dev_idx, uint32_t* rgb_size, uint32_t* nir_size, float rgb_thresh, \
        float nir_thresh, uint16_t rgb_width, uint16_t rgb_height, uint16_t nir_width, \
        uint16_t nir_height, uint32_t rgb_fmt, uint32_t nir_fmt) {

    KnPrint(KN_LOG_DBG, "\n\ncalling start lw3d verify mode...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }
    if(!rgb_size || !nir_size) {
        KnPrint(KN_LOG_ERR, "img size is null.\n");
        return -1;
    }

    if(fabsf(rgb_thresh) < FLOAT_ZERO_VAL) {
        //0.0f
    } else {
        if(rgb_thresh > MAX_THRESH_VAL || rgb_thresh < MIN_THRESH_VAL) {
            KnPrint(KN_LOG_ERR, "rgb threshold val is invalid, range:%5.2f ~ %5.2f.\n", 
                    MIN_THRESH_VAL, MAX_THRESH_VAL);
            return -1;
        }
    }

    if(fabsf(nir_thresh) < FLOAT_ZERO_VAL) {
        //0.0f
    } else {
        if(nir_thresh > MAX_THRESH_VAL || nir_thresh < MIN_THRESH_VAL) {
            KnPrint(KN_LOG_ERR, "nir threshold val is invalid, range:%5.2f ~ %5.2f.\n", 
                    MIN_THRESH_VAL, MAX_THRESH_VAL);
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    int ret = kdp_sfid_start_lw3d_sync(dev_idx, &rsp_code, rgb_size, 
                nir_size, rgb_thresh, nir_thresh, rgb_width, rgb_height,
                nir_width, nir_height, rgb_fmt, nir_fmt);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "start lw3d verify mode failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "start lw3d verify mode failed with error:%d.\n", rsp_code);
        return rsp_code;
    }
    return 0;
}


int kdp_verify_lw3D_image_all_res(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir, bool* fd_flag, char* fd_res, \
                bool* fr_flag, char* fr_res, bool* lm_flag, char* lm_res, \
                bool* fd_flag_nir, char* fd_res_nir, bool* fr_flag_nir, char* fr_res_nir, \
                bool* lm_flag_nir, char* lm_res_nir, bool* live_flag, char* live_res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling verify lw3d  user id...:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!user_id || !img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "verify user id lw3d input rgb parameter is invalid.\n");
        return -1;
    }

    if(!img_buf_nir || buf_len_nir <= 0) {
        KnPrint(KN_LOG_ERR, "verify user id lw3d input nir parameter is invalid.\n");
        return -1;
    }

    if(fd_flag && (*fd_flag == true)) {
        if(fd_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for rgb fd flag, result.\n");
            return -1;
        }
    }

    if(fr_flag &&(*fr_flag == true)) {
        if(fr_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for rgb fr flag, result.\n");
            return -1;
        }
    }

    if(lm_flag && (*lm_flag == true)) {
        if(lm_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for rgb lm flag, result.\n");
            return -1;
        }
    }

    if(fd_flag_nir && (*fd_flag_nir == true)) {
        if(fd_res_nir == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for nir fd flag, result.\n");
            return -1;
        }
    }

    if(fr_flag_nir &&(*fr_flag_nir == true)) {
        if(fr_res_nir == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for nir fr flag, result.\n");
            return -1;
        }
    }

    if(lm_flag_nir && (*lm_flag_nir == true)) {
        if(lm_res_nir == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for nir lm flag, result.\n");
            return -1;
        }
    }

    if(live_flag &&(*live_flag == true)) {
        if(live_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for live flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t mode = 1;
    int ret = kdp_sfid_send_lw3d_img_sync(dev_idx, img_buf, buf_len, img_buf_nir, buf_len_nir,
                &rsp_code, user_id, &mode, fd_flag, fd_res, fr_flag, fr_res, lm_flag, lm_res, 
                fd_flag_nir, fd_res_nir, fr_flag_nir, fr_res_nir, 
                lm_flag_nir, lm_res_nir, live_flag, live_res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "verify user id lw3d failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "verify user id lw3d failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    KnPrint(KN_LOG_DBG, "verify user id lw3d completed, user id:%d.\n", *user_id);
    if(*user_id < 1 || *user_id > 20) {
        KnPrint(KN_LOG_ERR, "invalid user id received:%d.\n", *user_id);
        return -1;
    }
    return 0;
}

int kdp_verify_lw3D_user(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir) {
    return kdp_verify_lw3D_image_generic(dev_idx, user_id, img_buf, buf_len, \
                    img_buf_nir, buf_len_nir, NULL, NULL);
}

int kdp_verify_lw3D_image_generic(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir, uint16_t* mask, char* res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling verify user id lw3d generic...:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!user_id || !img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "verify user id lw3d rgb input parameter is invalid.\n");
        return -1;
    }

    if(!img_buf_nir || buf_len_nir <= 0) {
        KnPrint(KN_LOG_ERR, "verify user id lw3d nir input parameter is invalid.\n");
        return -1;
    }

    if( mask && (*mask & 0x0f) ) {
        if(res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for lw3d res flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t mode = 1;
    int ret = kdp_sfid_send_lw3d_img_gen(dev_idx, img_buf, buf_len, img_buf_nir, buf_len_nir,
                    &rsp_code, user_id, &mode, mask, res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "verify user id lw3d failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "verify user id ls3d failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    KnPrint(KN_LOG_DBG, "verify user id lw3d completed, user id:%d.\n", *user_id);
    if(*user_id < 1 || *user_id > 20) {
        KnPrint(KN_LOG_ERR, "invalid user id received:%d.\n", *user_id);
        return -1;
    }
    return 0;
}

int kdp_extract_lw3D_feature_all_res(int dev_idx, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir, bool* fd_flag, char* fd_res, \
                bool* fr_flag, char* fr_res, bool* lm_flag, char* lm_res, \
                bool* fd_flag_nir, char* fd_res_nir, bool* fr_flag_nir, char* fr_res_nir, \
                bool* lm_flag_nir, char* lm_res_nir, bool* live_flag, char* live_res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling extract feature lw3d  user id...:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d input rgb parameter is invalid.\n");
        return -1;
    }

    if(!img_buf_nir || buf_len_nir <= 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d input nir parameter is invalid.\n");
        return -1;
    }

    if(fd_flag && (*fd_flag == true)) {
        if(fd_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for rgb fd flag, result.\n");
            return -1;
        }
    }

    if(fr_flag &&(*fr_flag == true)) {
        if(fr_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for rgb fr flag, result.\n");
            return -1;
        }
    }

    if(lm_flag && (*lm_flag == true)) {
        if(lm_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for rgb lm flag, result.\n");
            return -1;
        }
    }

    if(fd_flag_nir && (*fd_flag_nir == true)) {
        if(fd_res_nir == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for nir fd flag, result.\n");
            return -1;
        }
    }

    if(fr_flag_nir &&(*fr_flag_nir == true)) {
        if(fr_res_nir == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for nir fr flag, result.\n");
            return -1;
        }
    }

    if(lm_flag_nir && (*lm_flag_nir == true)) {
        if(lm_res_nir == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for nir lm flag, result.\n");
            return -1;
        }
    }

    if(live_flag &&(*live_flag == true)) {
        if(live_res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for live flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t mode = 1;
    uint16_t user_id = 0;
    int ret = kdp_sfid_send_lw3d_img_sync(dev_idx, img_buf, buf_len, img_buf_nir, buf_len_nir,
                &rsp_code, &user_id, &mode, fd_flag, fd_res, fr_flag, fr_res, lm_flag, lm_res, 
                fd_flag_nir, fd_res_nir, fr_flag_nir, fr_res_nir, 
                lm_flag_nir, lm_res_nir, live_flag, live_res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    KnPrint(KN_LOG_DBG, "extract feature lw3d completed, user id:%d.\n", user_id);
    if(user_id < 1 || user_id > 5) {
        KnPrint(KN_LOG_ERR, "invalid user id received:%d.\n", user_id);
        return -1;
    }
    return 0;
}

int kdp_extract_lw3D_feature(int dev_idx, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir) {
    return kdp_extract_lw3D_feature_generic(dev_idx, img_buf, buf_len, img_buf_nir,\
                buf_len_nir, NULL, NULL);
}

int kdp_extract_lw3D_feature_generic(int dev_idx, char* img_buf, int buf_len, \
                char* img_buf_nir, int buf_len_nir, uint16_t* mask, char* res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling extract lw3d feature generic...:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d rgb input parameter is invalid.\n");
        return -1;
    }

    if(!img_buf_nir || buf_len_nir <= 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d nir input parameter is invalid.\n");
        return -1;
    }

    if( mask && (*mask & 0x0f) ) {
        if(res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for lw3d res flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t mode = 1;
    uint16_t user_id = 0;
    int ret = kdp_sfid_send_lw3d_img_gen(dev_idx, img_buf, buf_len, img_buf_nir, buf_len_nir,
                    &rsp_code, &user_id, &mode, mask, res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    KnPrint(KN_LOG_DBG, "extract feature lw3d completed, user id:%d.\n", user_id);
    if(user_id < 1 || user_id > 5) {
        KnPrint(KN_LOG_ERR, "invalid user id received:%d.\n", user_id);
        return -1;
    }
    return 0;
}

//-------------------------------DME APIs
int kdp_start_dme(int dev_idx, uint32_t model_size, char* data, int dat_size, \
                  uint32_t* ret_size, char* img_buf, int buf_len) {
    KnPrint(KN_LOG_DBG, "\n\ncalling starting dme mode...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_start_dme(dev_idx, model_size, data, dat_size,
                    ret_size, img_buf, buf_len);
}

int kdp_dme_configure_param(int dev_idx, uint32_t setup_conf, char* data, int dat_size) {
    KnPrint(KN_LOG_DBG, "\n\ncalling dme config...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_dme_configure(dev_idx, setup_conf, data, dat_size);
}

int kdp_dme_configure(int dev_idx, char* data, int dat_size) {
    return kdp_dme_configure_param(dev_idx, 0, data, dat_size);
}

int kdp_dme_inference(int dev_idx, char* img_buf, int buf_len, uint32_t* inf_size, \
            bool* res_flag, char* inf_res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling dme inference...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_dme_inference(dev_idx, img_buf, buf_len, inf_size, res_flag, inf_res);
}

int kdp_get_dme_res(int dev_idx, uint32_t* inf_size, char* inf_res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling dme retrieving...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_get_dme_res(dev_idx, inf_size, inf_res);
}

void error_code_2_str(int code, char* buf, int len) {
    memset(buf, 0, len);

    switch(code) {
    case GEN_ERROR:
        snprintf(buf, len-1, "General error");
        break;

    case ERR_APP_NOT_LOADED:
        snprintf(buf, len-1, "SFID application not loaded");
        break;

    case ERR_INVALID_UID:
        snprintf(buf, len-1, "Invalid User ID or image index specified");
        break;

    case ERR_UID_EXIST:
        snprintf(buf, len-1, "User ID already exists");
        break;

    case ERR_UID_CHANGED:
        snprintf(buf, len-1, "User ID changed from the New User session");
        break;

    case ERR_IMG_IDX_START:
        snprintf(buf, len-1, "User ID changed from the New User session");
        break;

    case ERR_NO_USER_DATA:
        snprintf(buf, len-1, "No feature maps from New User session found");
        break;

    case ERR_USER_NOT_FOUND:
        snprintf(buf, len-1, "User ID not found in DB");
        break;

    case ERR_DB_DELETE_FAIL:
        snprintf(buf, len-1, "DB delete failure");
        break;

    case ERR_NO_MATCHED_USR:
        snprintf(buf, len-1, "No match in User DB");
        break;

    case ERR_INFER_FAILURE:
        snprintf(buf, len-1, "No successful inference (FD-FR)");
        break;

    default:
        snprintf(buf, len-1, "error code not found");
        break;
    }
    return;
}
