/**
 * @file      SyncMsgHandler.cpp
 * @brief     implementation file for sync message handler
 * @version   0.1 - 2019-04-23
 * @copyright (c) 2019 Kneron Inc. All right reserved.
 */

#include "UARTDev.h"
#include "USBDev.h"
#include "MsgMgmt.h"
#include "KnLog.h"
#include "KnUtil.h"
#include "SyncMsgHandler.h"

#include <string.h>
#include <unistd.h>

#include <vector>
using namespace std;

extern MsgMgmt*                _p_mgr; //msg manager

SyncMsgHandler::SyncMsgHandler() {
}

SyncMsgHandler::~SyncMsgHandler() {
    //loop for pending rsp msg
    for(int i = 0; i < MAX_COM_DEV; i++) {
        _map_mtx[i].lock();
        map<uint16_t, KnBaseMsg*>::iterator iter;
        for(iter = _runing_map[i].begin(); iter != _runing_map[i].end(); ++iter) {
            KnBaseMsg* p = iter->second;
            if(p) delete p;
        }
        _map_mtx[i].unlock();
    }
}

int SyncMsgHandler::wait_dev_restart(int dev_idx) {
    //get USB dev
    BaseDev* pDev = _p_mgr->getDevFromIdx(dev_idx);
    if(!pDev) {
        KnPrint(KN_LOG_ERR, "could not get device for idx:%d.\n", dev_idx);
        return -1;
    }

    USBDev* pUsb = NULL;
    if(pDev->dev_type == KDP_UART_DEV) {
        UARTDev* pUart = (UARTDev*)pDev;
        pUsb = pUart->get_helper_dev();
        if(!pUsb) {
            KnPrint(KN_LOG_ERR, "could not get helper device for idx:%d.\n", dev_idx);
            return -1;
        }
    } else {
        pUsb = (USBDev*)pDev;
    }

    for(int i = 0; i < KDP_RESET_NUM; i++) {
        usleep(KDP_RESET_TIME);
        KnPrint(KN_LOG_DBG, "retrying to reconnect usb...:%d.\n", dev_idx);
        int ret = pUsb->reconnect_usb();
        if(ret == 0) {
            KnPrint(KN_LOG_DBG, "usb reconnected...:%d.\n", dev_idx);
            return 0;
        }
    }

    KnPrint(KN_LOG_ERR, "usb reconnect failed...:%d.\n", dev_idx);
    return -1;
}

KnBaseMsg*  SyncMsgHandler::wait_rsp_msg(int dev_idx, uint16_t msg_type, int wait_num) {
    map<uint16_t, KnBaseMsg*>::iterator iter;
    KnBaseMsg* p_rsp = NULL;
    _p_mgr->set_cmd_going_flag(dev_idx);

    if(wait_num < 5) wait_num = 5; //min 1 sec

    for(int i = 0; i < wait_num; i++)  {
        //lock
        _map_mtx[dev_idx].lock();
        //get the iter
        iter = _runing_map[dev_idx].find(msg_type);

        //if could not find
        if(iter == _runing_map[dev_idx].end()) {
            _map_mtx[dev_idx].unlock(); //unlock
            KnTsPrint(KN_LOG_ERR, "no req msg in the local map, no need wait:%d.\n", msg_type);
            _p_mgr->clear_cmd_going_flag(dev_idx);
            return NULL;
        }

        //if no message coming
        if(iter->second == NULL) {
            _map_mtx[dev_idx].unlock(); //unlock

            _cond_mtx.lock();
            //wait for msg
             #if defined(THREAD_USE_CPP11)
            _cond_wait.wait_for(_cond_mtx, std::chrono::milliseconds(KDP_SYNC_WAIT_TIME));
            #else
    
            _cond_wait.wait_for(_cond_mtx, KDP_SYNC_WAIT_TIME);
           
            #endif
            _cond_mtx.unlock();
            continue; //wait
        } else {
            p_rsp = iter->second;
            _runing_map[dev_idx].erase(iter);
            _map_mtx[dev_idx].unlock();
            KnTsPrint(KN_LOG_DBG, "rsp msg from device received:0x%x.\n", msg_type);
            return p_rsp; //return msg
        }
    }

    KnTsPrint(KN_LOG_ERR, "waiting for rsp timeout.\n");
    _p_mgr->clear_cmd_going_flag(dev_idx);
    return NULL; //timeout
}

int SyncMsgHandler::kdp_query_app(int dev_idx, uint32_t* app_ids, int* size) {
    uint16_t msg_type = KN_OPR_QUERY_APPS_REQ;
    uint16_t msg_len = 0;

    //create req message
    KnOprQueryAppReq* p_req = new KnOprQueryAppReq(msg_type, msg_len);

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending query app to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending query app request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving query app req to local map.\n");
    save_req_to_map(msg_type);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for query app rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "query app response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "query app rsp from device received.\n");

    //extract msg content
    KnOprQueryAppRsp* query_msg = (KnOprQueryAppRsp*)p_rsp;
    *size = query_msg->_len;
    uint32_t* p = query_msg->_p_app_id;

    for(int i = 0; i < *size; i++) {
        KnTsPrint(KN_LOG_DBG, "app id:%d from query app rsp. \n", p[i]);
        app_ids[i] = p[i];
    }

    //free resposne memory
    delete p_rsp;
    return 0;
}

int SyncMsgHandler::kdp_select_app(int dev_idx, uint32_t app_id, uint32_t* rsp_code, uint32_t* buf_size) {
    uint16_t msg_type = KN_OPR_SELECT_APP_REQ;
    uint16_t msg_len = sizeof(uint32_t);

    //create req message
    KnOprSelectAppReq* p_req = new KnOprSelectAppReq(msg_type, msg_len, app_id);

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending select app to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending select app request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving select app req to local map.\n");
    save_req_to_map(msg_type);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for select app rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "select app response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "select app rsp from device received.\n");

    //extract msg content
    KnOprSelectAppRsp* sel_app_msg = (KnOprSelectAppRsp*)p_rsp;
    *rsp_code = sel_app_msg->_rsp_code;
    *buf_size = sel_app_msg->_img_size;

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, buf size:%d from select app rsp.\n", 
            sel_app_msg->_rsp_code, sel_app_msg->_img_size);
    //free resposne memory
    delete p_rsp;
    return 0;
}

int SyncMsgHandler::kdp_sfid_start_lw3d(int dev_idx, uint32_t* rsp_code, uint32_t* rgb_size, \
            uint32_t* nir_size, float rgb_thresh, float nir_thresh, uint16_t rgb_width, \
            uint16_t rgb_height, uint16_t nir_width, uint16_t nir_height, \
            uint32_t rgb_fmt, uint32_t nir_fmt) {
    uint16_t msg_type = KN_APP_START_LW3D;
    uint16_t msg_len = sizeof(float) + sizeof(float) + sizeof(uint16_t)*4 + sizeof(uint32_t)*2;

    //create req message
    KnStartLw3dReq* p_req = new KnStartLw3dReq(msg_type, msg_len, rgb_thresh, nir_thresh, rgb_width,
                rgb_height, nir_width, nir_height, rgb_fmt, nir_fmt);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create sfid start lw3d req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending sfid start lw3d to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending sfid start lw3d request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving sfid start lw3d req to local map.\n");
    save_req_to_map(msg_type);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for sfid start lw3d rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "sfid start lw3d response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "sfid start lw3d rsp from device received.\n");

    //extract msg content
    KnStartLw3dRsp* sfid_msg = (KnStartLw3dRsp*)p_rsp;
    *rsp_code = sfid_msg->_rsp_code;
    *rgb_size = sfid_msg->_rgb_size;
    *nir_size = sfid_msg->_nir_size;

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, img size:%d,%d from sfid start lw3d rsp.\n", 
            sfid_msg->_rsp_code, sfid_msg->_rgb_size, sfid_msg->_nir_size);
    //free resposne memory
    delete p_rsp;
    return 0;
}


int SyncMsgHandler::kdp_sfid_start(int dev_idx, uint32_t* rsp_code, uint32_t* img_size, float thresh,
                uint16_t width, uint16_t height, uint32_t format) {
    uint16_t msg_type = KN_APP_SFID_START_REQ;
    uint16_t msg_len = sizeof(float)+sizeof(uint32_t)*2 + sizeof(uint16_t)*2;

    //create req message
    KnSFIDStartReq* p_req = new KnSFIDStartReq(msg_type, msg_len, thresh,
                width, height, format);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create sfid start req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending sfid start to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending sfid start request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving sfid start req to local map.\n");
    save_req_to_map(msg_type);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for sfid start rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "sfid start response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "sfid start rsp from device received.\n");

    //extract msg content
    KnSFIDStartRsp* sfid_msg = (KnSFIDStartRsp*)p_rsp;
    *rsp_code = sfid_msg->_rsp_code;
    *img_size = sfid_msg->_img_size;

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, img size:%d from sfid img rsp.\n", sfid_msg->_rsp_code, sfid_msg->_img_size);
    //free resposne memory
    delete p_rsp;
    return 0;
}

#if 0
int SyncMsgHandler::kdp_sfid_new_user_req(int dev_idx, uint32_t usr_id, uint32_t img_idx) {
    uint16_t msg_type = KN_APP_SFID_NEW_USER;
    uint16_t msg_len = sizeof(uint32_t) + sizeof(uint32_t);

    //create req message
    KnSFIDNewUserReq* p_req = new KnSFIDNewUserReq(msg_type, msg_len, usr_id, img_idx);

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending sfid new user req to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_DBG, "sending sfid new user request to device failed:%d.\n", res);
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "sending sfid new user to dev completed. \n");
    return 0;
}
#endif

int SyncMsgHandler::kdp_sfid_new_user(int dev_idx, uint16_t usr_id, uint16_t img_idx) {
    uint16_t msg_type = KN_APP_SFID_NEW_USER;
    uint16_t msg_len = sizeof(uint32_t) + sizeof(uint32_t);

    //create req message
    KnSFIDNewUserReq* p_req = new KnSFIDNewUserReq(msg_type, msg_len, usr_id, img_idx);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create sfid new user req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending sfid new user to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending sfid new user request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving sfid new user req to local map.\n");
    save_req_to_map(msg_type);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for sfid new user rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "sfid new user response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "sfid new user rsp from device received.\n");

    //free resposne memory
    delete p_rsp;
    return 0;
}

int SyncMsgHandler::kdp_sfid_register(int dev_idx, uint32_t usr_id, uint32_t* rsp_code) {
    uint16_t msg_type = KN_APP_SFID_REGISTER;
    uint16_t msg_len = sizeof(uint32_t);

    //create req message
    KnSFIDRegisterReq* p_req = new KnSFIDRegisterReq(msg_type, msg_len, usr_id);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create sfid register req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending sfid register to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_DBG, "sending sfid register request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving sfid register req to local map.\n");
    save_req_to_map(msg_type);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for sfid register rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "sfid register response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "sfid register rsp from device received.\n");

    //extract msg content
    KnSFIDRegisterRsp* sfid_msg = (KnSFIDRegisterRsp*)p_rsp;
    *rsp_code = sfid_msg->_rsp_code;

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, from sfid register rsp.\n", sfid_msg->_rsp_code);

    //free resposne memory
    delete p_rsp;
    return 0;
}

int SyncMsgHandler::kdp_sfid_del_db(int dev_idx, uint32_t usr_id, uint32_t* rsp_code) {
    uint16_t msg_type = KN_APP_SFID_DEL_DB;
    uint16_t msg_len = sizeof(uint32_t);

    //create req message
    KnSFIDDelDBReq* p_req = new KnSFIDDelDBReq(msg_type, msg_len, usr_id);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create sfid del db req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending sfid del db to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending sfid del db request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving sfid del db req to local map.\n");
    save_req_to_map(msg_type);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for sfid del db rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type, 100);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "sfid del db response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "sfid del db rsp from device received.\n");

    //extract msg content
    KnSFIDDelDBRsp* sfid_msg = (KnSFIDDelDBRsp*)p_rsp;
    *rsp_code = sfid_msg->_rsp_code;

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d from sfid del db rsp.\n", sfid_msg->_rsp_code);
    //free resposne memory
    delete p_rsp;
    return 0;
}

int SyncMsgHandler::kdp_sfid_send_img_gen(int dev_idx, char* img_buf, int buf_len, \
                uint32_t* rsp_code, uint16_t* user_id, uint16_t* mode, \
                uint16_t* mask, char* res) {
    uint16_t msg_type = KN_APP_SFID_SEND_IMG;
    uint16_t msg_len = sizeof(uint32_t) + sizeof(uint16_t)*2;

    BaseDev* pDev = _p_mgr->getDevFromIdx(dev_idx);
    if(!pDev) {
        KnPrint(KN_LOG_ERR, "could not get device for idx:%d.\n", dev_idx);
        return -1;
    }

    USBDev* pUsb = NULL;
    if(pDev->dev_type == KDP_UART_DEV) {
        UARTDev* pUart = (UARTDev*)pDev;
        pUsb = pUart->get_helper_dev();
        if(!pUsb) {
            KnPrint(KN_LOG_ERR, "could not get helper device for idx:%d.\n", dev_idx);
            return -1;
        }
    } else {
        pUsb = (USBDev*)pDev;
    }

    //create req message
    uint16_t _msk = 0;
    if (mask != NULL)
        _msk = *mask;
    KnSFIDSendImageReq* p_req = new KnSFIDSendImageReq(msg_type, msg_len, buf_len, _msk);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create sfid send img req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending sfid img to dev ........\n");
    int ret = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "sending sfid send img request to device failed:%d.\n", ret);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving sfid img req to local map.\n");
    save_req_to_map(msg_type, dev_idx);

    //to call usb interface to transfer image file.
    KnTsPrint(KN_LOG_DBG, "calling usb interface to transfer file:........\n");
    int snd_len = pUsb->send_image_data(img_buf, buf_len);
    if(snd_len < 0) {
        KnPrint(KN_LOG_ERR, "receving transfer ack msg failed.\n");
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "file transfer completed :%d.\n", snd_len);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for sfid img rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "sfid send img response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "sfid img rsp from device received.\n");

    //extract msg content
    KnSFIDSendImageRsp* sfid_msg = (KnSFIDSendImageRsp*)p_rsp;
    *rsp_code = sfid_msg->_rsp_code;
    *user_id = sfid_msg->_uid;
    *mode = sfid_msg->_mode;
    _msk = sfid_msg->_out_flag;

    if(mask != NULL){
        *mask = _msk;
    }
    int off = 0;
    if(_msk & 0x01) {
        memcpy((void*)(res+off), sfid_msg->_fd_res, FD_RES_LENGTH);
        KnTsPrint(KN_LOG_DBG, "fd result is available.\n");
        off += FD_RES_LENGTH;
    }

    if(_msk & 0x02) {
        memcpy((void*)(res+off), sfid_msg->_lm_res, LM_RES_LENGTH);
        KnTsPrint(KN_LOG_DBG, "fr result is available.\n");
        off += LM_RES_LENGTH;
    }
    
    if(_msk & 0x04) {
        memcpy((void*)(res+off), sfid_msg->_fr_res, FR_RES_LENGTH);
        KnTsPrint(KN_LOG_DBG, "lm result is available.\n");
        off += FR_RES_LENGTH;
    }

    if(_msk & 0x08) {
        memcpy((void*)(res+off), &(sfid_msg->_live_res), sizeof(uint16_t));
        KnTsPrint(KN_LOG_DBG, "liveness result is available.\n");
    }

    if(*rsp_code != 0) {
        KnTsPrint(KN_LOG_DBG, "sfid img rsp with response code:%d.\n", *rsp_code);
        delete p_rsp;
        return *rsp_code;
    }

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, user_id:%d, mode:%d, mask:%d from sfid img rsp.\n", 
            sfid_msg->_rsp_code, sfid_msg->_uid, sfid_msg->_mode, sfid_msg->_out_flag);
    //free resposne memory
    delete p_rsp;
    return 0;
}

int SyncMsgHandler::kdp_sfid_send_img(int dev_idx, char* img_buf, int buf_len, \
                uint32_t* rsp_code, uint16_t* user_id, uint16_t* mode, \
                bool* fd_flag, char* fd_res, bool* fr_flag, char* fr_res,\
                bool* lm_flag, char* lm_res, bool* live_flag, char* live_res) {
    uint16_t msg_type = KN_APP_SFID_SEND_IMG;
    uint16_t msg_len = sizeof(uint32_t) + sizeof(uint16_t)*2;

    BaseDev* pDev = _p_mgr->getDevFromIdx(dev_idx);
    if(!pDev) {
        KnPrint(KN_LOG_ERR, "could not get device for idx:%d.\n", dev_idx);
        return -1;
    }

    USBDev* pUsb = NULL;
    if(pDev->dev_type == KDP_UART_DEV) {
        UARTDev* pUart = (UARTDev*)pDev;
        pUsb = pUart->get_helper_dev();
        if(!pUsb) {
            KnPrint(KN_LOG_ERR, "could not get helper device for idx:%d.\n", dev_idx);
            return -1;
        }
    } else {
        pUsb = (USBDev*)pDev;
    }

    uint16_t mask = 0;
    if(fd_flag && (*fd_flag == true)) mask |= 0x01;
    if(lm_flag && (*lm_flag == true)) mask |= 0x02;
    if(fr_flag && (*fr_flag == true)) mask |= 0x04;
    if(live_flag && (*live_flag == true)) mask |= 0x08;

    //create req message
    KnSFIDSendImageReq* p_req = new KnSFIDSendImageReq(msg_type, msg_len, buf_len, mask);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create sfid send img req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending sfid img to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending sfid send img request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving sfid img req to local map.\n");
    save_req_to_map(msg_type, dev_idx);

    //to call usb interface to transfer image file.
    KnTsPrint(KN_LOG_DBG, "calling usb interface to transfer file:........\n");
    int snd_len = pUsb->send_image_data(img_buf, buf_len);
    if(snd_len < 0) {
        KnPrint(KN_LOG_ERR, "receving transfer ack msg failed.\n");
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "file transfer completed :%d.\n", snd_len);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for sfid img rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "sfid send img response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "sfid img rsp from device received.\n");

    //extract msg content
    KnSFIDSendImageRsp* sfid_msg = (KnSFIDSendImageRsp*)p_rsp;
    *rsp_code = sfid_msg->_rsp_code;
    *user_id = sfid_msg->_uid;
    *mode = sfid_msg->_mode;

    if(fd_flag && (*fd_flag == true)) {
        if(sfid_msg->_out_flag & 0x01) { //fd res
            KnTsPrint(KN_LOG_DBG, "fd result is available.\n");
            memcpy((void*)fd_res, (void*)(sfid_msg->_fd_res), FD_RES_LENGTH);
        } else {
            *fd_flag = false;
        }
    }

    if(lm_flag && (*lm_flag == true)) {
        if(sfid_msg->_out_flag & 0x02) { //lm res
            KnTsPrint(KN_LOG_DBG, "lm result is available.\n");
            memcpy((void*)lm_res, (void*)(sfid_msg->_lm_res), LM_RES_LENGTH);
        } else {
            *lm_flag = false;
        }
    }

    if(fr_flag && (*fr_flag == true)) {
        if(sfid_msg->_out_flag & 0x04) { //fr res
            KnTsPrint(KN_LOG_DBG, "fr result is available.\n");
            memcpy((void*)fr_res, (void*)(sfid_msg->_fr_res), FR_RES_LENGTH);
        } else {
            *fr_flag = false;
        }
    }

    if(live_flag && (*live_flag == true)) {
        if(sfid_msg->_out_flag & 0x08) { //fr res
            KnTsPrint(KN_LOG_DBG, "liveness result is available.\n");
            memcpy((void*)live_res, (void*)&(sfid_msg->_live_res), sizeof(uint16_t));
        } else {
            *live_flag = false;
        }
    }

    if(*rsp_code != 0) {
        KnTsPrint(KN_LOG_DBG, "sfid img rsp with response code:%d.\n", *rsp_code);
        delete p_rsp;
        return *rsp_code;
    }

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, user_id:%d, mode:%d, mask:%d from sfid img rsp.\n", 
            sfid_msg->_rsp_code, sfid_msg->_uid, sfid_msg->_mode, sfid_msg->_out_flag);
    //free resposne memory
    delete p_rsp;
    return 0;
}

void SyncMsgHandler::save_req_to_map(uint16_t req_type, int idx) {
    _map_mtx[idx].lock();

    map<uint16_t, KnBaseMsg*>::iterator iter = _runing_map[idx].find(req_type);
    if(iter != _runing_map[idx].end()) {
        KnBaseMsg* p = iter->second;
        iter->second = NULL;
        _map_mtx[idx].unlock();

        KnPrint(KN_LOG_DBG, "req msg %d exist in map for dev:%d.\n", req_type, idx);
        if(p) delete p; //free the old one
    } else {
        _runing_map[idx][req_type] = NULL;
        _map_mtx[idx].unlock();
    }
}

int SyncMsgHandler::msg_received_rx_queue(KnBaseMsg* p_rsp, uint16_t req_type, int idx) {
    if(idx < 0 || idx >= MAX_COM_DEV) {
        KnPrint(KN_LOG_ERR, "invalid dev idx received from rx queue:%d.\n", idx);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "msg 0x%x received from rx queue for dev:%d.\n", req_type, idx);
    map<uint16_t, KnBaseMsg*>::iterator iter;

    _map_mtx[idx].lock();
    iter = _runing_map[idx].find(req_type);

    if(iter == _runing_map[idx].end()) {
        _map_mtx[idx].unlock();

        if(req_type == KN_MSG_GEN_ACK) { //generic ack
            //_uart_mtx.lock();
            //_uart_wait.notify_all();
            //_uart_mtx.unlock();
            KnTsPrint(KN_LOG_DBG, "uart ack msg received.\n");
        } else {
            KnTsPrint(KN_LOG_DBG, "no req msg is waiting for this rsp:0x%x.\n", req_type);
        }
        delete p_rsp; //ignore it;
        return 0;
    } else {
        KnBaseMsg* p_old = iter->second;
        iter->second = p_rsp;
        _map_mtx[idx].unlock();

        if(p_old) {
            KnPrint(KN_LOG_DBG, "msg %d exist in map for dev:%d.\n", req_type, idx);
            delete p_old; //free the msg
        }
    }

    KnTsPrint(KN_LOG_DBG, "sync msg handler notifying wait thread ........\n");
    //notify
    _cond_mtx.lock();
    _cond_wait.notify_all();
    _cond_mtx.unlock();

    return 0;
}

#if 0
int SyncMsgHandler::wait_uart_ack() {
    _uart_mtx.lock();
    //wait for msg
    #if defined(THREAD_USE_CPP11)
    cv_status ret = _uart_wait.wait_for(_uart_mtx, std::chrono::milliseconds(KDP_UART_ACK_TIME));
    #else
    cv_status ret = _uart_wait.wait_for(_uart_mtx, KDP_UART_ACK_TIME);
    #endif
    _uart_mtx.unlock();

    if(ret == cv_status::timeout) { 
        return -1;
    } else {
        return 0;
    }
}

int SyncMsgHandler::send_image_data_local(char *buf, int data_len, USBDev* pUsb) {
    if(data_len <= 0 || !buf) return -1;
    if(!pUsb) return -1;

    KnTsPrint(KN_LOG_DBG, "sync mgr is starting to send image :%d...\n", data_len);

    int leftover = data_len;
    int i = 0;
    int off = 0;
    int status = -1;

    while(leftover > 0) {

        status = wait_uart_ack();
        if(status < 0) {
            KnPrint(KN_LOG_ERR, "ack msg timeout:%d.\n", status);
            return status;
        }

        int len = static_cast<int> (USB_MAX_BYTES);
        if(len > leftover) len = leftover;

        KnTsPrint(KN_LOG_DBG, "sync mgr is sending segment :%d, with len:%d...\n", i, len);
        status = pUsb->send_bulk(buf + off, len);
        if(status < 0) {
            KnPrint(KN_LOG_ERR, "sending %d time for img, off:%d.\n", i, off);
            return status;
        }

        KnTsPrint(KN_LOG_DBG, "sync mgr sent segment :%d, with len:%d...\n", i, status);
        
        off += status;
        leftover -= status;
        i++;
    }


    KnTsPrint(KN_LOG_DBG, "sync mgr finished sending image :%d...\n", off);
    return off;
}
#endif

int SyncMsgHandler::kdp_reset_sys_sync(int dev_idx, uint32_t reset_mode) {
    uint16_t msg_type = KN_OPR_CMD_RESET_REQ;
    uint16_t msg_len = sizeof(uint32_t) * 2;

    BaseDev* pDev = _p_mgr->getDevFromIdx(dev_idx);
    if(!pDev) {
        KnPrint(KN_LOG_ERR, "could not get device for idx:%d.\n", dev_idx);
        return -1;
    }

    if(reset_mode == 0x22) {
        pDev->restart_dev(); //just for testing. no meaning in real environment.
        return 0;
    }

    if(pDev->dev_type == KDP_USB_DEV) {
        if(reset_mode == 3 || reset_mode == 4) { // for usb dev, 3 and 4 are not supported
            KnPrint(KN_LOG_ERR, "reset mode :%d is not supported by usb dev:%d.\n", 
                    reset_mode, dev_idx);
            return 1; //undefined reset control mode
        }
    }

    //create req message
    KnCmdResetReq* p_req = new KnCmdResetReq(msg_type, msg_len, reset_mode, ~reset_mode);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create reset sys req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending reset sys req to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending reset sys request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving reset sys req to local map.\n");
    save_req_to_map(msg_type);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for sys reset rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "system reset response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "system reset rsp from device received.\n");

    //extract msg content
    KnCmdResetRsp* p_msg = (KnCmdResetRsp*)p_rsp;
    uint32_t rsp_code = p_msg->_err_code;

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d in reset system rsp.\n", rsp_code);
    //free resposne memory
    delete p_rsp;

    if(rsp_code == 0) {
        if(reset_mode == 255) { //system reset
            KnTsPrint(KN_LOG_MSG, "system is doing reset...\n");

            //need wait and re-connect usb
            wait_dev_restart(dev_idx);
        } else if(reset_mode == 256) { //shutdown
            KnTsPrint(KN_LOG_MSG, "shutting down dev...\n");
            pDev->shutdown_dev();
        }
    }

    return rsp_code;
}

int SyncMsgHandler::kdp_report_sys_status_sync(int dev_idx, uint32_t* device, uint32_t* firmware_id,
        uint16_t* sys_status, uint16_t* app_status) {
    uint16_t msg_type = KN_OPR_SYS_STATUS_REQ;
    uint16_t msg_len = 0;

    //create req message
    KnReportSysStsReq* p_req = new KnReportSysStsReq(msg_type, msg_len);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create report sys req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending report sys req to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending report sys request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving report sys req to local map.\n");
    save_req_to_map(msg_type);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for report sys rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "report sys response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "report system rsp from device received.\n");

    //extract msg content
    KnCmdSysStsRsp* p_msg = (KnCmdSysStsRsp*)p_rsp;

    *device = p_msg->_dev_id;
    *firmware_id = p_msg->_fw_id;
    *sys_status = p_msg->_sys_sts;
    *app_status = p_msg->_app_sts;

    KnTsPrint(KN_LOG_DBG, "report system response:%d,%d,%d,%d.\n", *device, *firmware_id, *sys_status, *app_status);
    //free resposne memory
    delete p_rsp;

    return 0;
}


int SyncMsgHandler::kdp_update_fw_sync(int dev_idx, uint32_t* module_id, char* img_buf, int buf_len) {
    uint16_t msg_type = KN_UDT_FW_REQ;
    uint16_t msg_len = sizeof(uint32_t) + sizeof(uint32_t);

    BaseDev* pDev = _p_mgr->getDevFromIdx(dev_idx);
    if(!pDev) {
        KnPrint(KN_LOG_ERR, "could not get device for idx:%d.\n", dev_idx);
        return -1;
    }

    USBDev* pUsb = NULL;
    if(pDev->dev_type == KDP_UART_DEV) {
        UARTDev* pUart = (UARTDev*)pDev;
        pUsb = pUart->get_helper_dev();
        if(!pUsb) {
            KnPrint(KN_LOG_ERR, "could not get helper device for idx:%d.\n", dev_idx);
            return -1;
        }
    } else {
        pUsb = (USBDev*)pDev;
    }

    if(*module_id == 0) { //if no operation
        return 0;
    }

    //create req message
    KnCmdUdtFmReq* p_req = new KnCmdUdtFmReq(msg_type, msg_len, *module_id);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create update firmware req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending update firmware request to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending update firmware request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving update firmware req to local map.\n");
    save_req_to_map(msg_type, dev_idx);

    //to call usb interface to transfer image file.
    KnTsPrint(KN_LOG_DBG, "calling usb interface to transfer file:........\n");
    int snd_len = pUsb->send_image_data(img_buf, buf_len);
    if(snd_len < 0) {
        KnPrint(KN_LOG_ERR, "receving transfer ack msg failed.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "file transfer completed :%d.\n", snd_len);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for update firmware rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type, 250);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "update firmware response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "update firmware rsp from device received.\n");

    //extract msg content
    KnCmdUdtFmRsp* rsp_msg = (KnCmdUdtFmRsp*)p_rsp;
    uint32_t rsp_code = rsp_msg->_rsp_code;
    uint32_t ret_id  = rsp_msg->_module_id;

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, ret_id:%d.\n", rsp_code, ret_id);

    //free resposne memory
    delete p_rsp;

    if(rsp_code != 0) { //error
        KnTsPrint(KN_LOG_ERR, "udt firmware response code:%d.\n", rsp_code);
        return rsp_code;
    }

    //need wait and re-connect usb
    wait_dev_restart(dev_idx);

    if(ret_id != *module_id) {
        KnTsPrint(KN_LOG_DBG, "confirmation of module id is incorrect:%d.\n", ret_id);
        return -1;
    }

    return rsp_code;
}

int SyncMsgHandler::kdp_update_model_sync(int dev_idx, uint32_t* model_id, uint32_t model_size, \
        char* img_buf, int buf_len) {
    uint16_t msg_type = KN_UDT_MODEL_REQ;
    uint16_t msg_len = sizeof(uint32_t) + sizeof(uint32_t);

    BaseDev* pDev = _p_mgr->getDevFromIdx(dev_idx);
    if(!pDev) {
        KnPrint(KN_LOG_ERR, "could not get device for idx:%d.\n", dev_idx);
        return -1;
    }

    USBDev* pUsb = NULL;
    if(pDev->dev_type == KDP_UART_DEV) {
        UARTDev* pUart = (UARTDev*)pDev;
        pUsb = pUart->get_helper_dev();
        if(!pUsb) {
            KnPrint(KN_LOG_ERR, "could not get helper device for idx:%d.\n", dev_idx);
            return -1;
        }
    } else {
        pUsb = (USBDev*)pDev;
    }

    if(*model_id == 0) { //if no operation
        return 0;
    }

    //create req message
    KnCmdUdtModelReq* p_req = new KnCmdUdtModelReq(msg_type, msg_len, *model_id, model_size);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create update model req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending update model request to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending update model request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving update model req to local map.\n");
    save_req_to_map(msg_type, dev_idx);
   
    //to call usb interface to transfer image file.
    KnTsPrint(KN_LOG_DBG, "calling usb interface to transfer file:........\n");
    int snd_len = pUsb->send_image_data(img_buf, buf_len);
    if(snd_len < 0) {
        KnPrint(KN_LOG_ERR, "receving transfer ack msg failed.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "file transfer completed :%d.\n", snd_len);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for update model rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type, 750);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "update model response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "update model rsp from device received.\n");

    //extract msg content
    KnCmdUdtModelRsp* rsp_msg = (KnCmdUdtModelRsp*)p_rsp;
    uint32_t rsp_code = rsp_msg->_rsp_code;
    uint32_t ret_id  = rsp_msg->_model_id;

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, ret_id:%d.\n", rsp_code, ret_id);

    //free resposne memory
    delete p_rsp;

    if(rsp_code != 0) { //error
        KnTsPrint(KN_LOG_ERR, "udt model response code:%d.\n", rsp_code);
        return rsp_code;
    }

    //need wait and re-connect usb
    wait_dev_restart(dev_idx);

    if(ret_id != *model_id) {
        KnTsPrint(KN_LOG_DBG, "confirmation of model id is incorrect:%d.\n", ret_id);
        return -1;
    }

    return rsp_code;
}


int SyncMsgHandler::kdp_start_dme(int dev_idx, uint32_t model_size, char* data, int dat_size, \
    uint32_t* ret_size, char* img_buf, int buf_len) {

    uint16_t msg_type = KN_APP_START_DME_REQ;
    uint16_t msg_len = sizeof(uint32_t) + dat_size;

    BaseDev* pDev = _p_mgr->getDevFromIdx(dev_idx);
    if(!pDev) {
        KnPrint(KN_LOG_ERR, "could not get device for idx:%d.\n", dev_idx);
        return -1;
    }

    USBDev* pUsb = NULL;
    if(pDev->dev_type == KDP_UART_DEV) {
        UARTDev* pUart = (UARTDev*)pDev;
        pUsb = pUart->get_helper_dev();
        if(!pUsb) {
            KnPrint(KN_LOG_ERR, "could not get helper device for idx:%d.\n", dev_idx);
            return -1;
        }
    } else {
        pUsb = (USBDev*)pDev;
    }

    //create req message
    KnStartDmeReq* p_req = new KnStartDmeReq(msg_type, msg_len, model_size, data, dat_size);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create start dme req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending start dme req to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending start dme request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving start dme req to local map.\n");
    save_req_to_map(msg_type, dev_idx);

    //to call usb interface to transfer image file.
    KnTsPrint(KN_LOG_DBG, "calling usb interface to transfer file:........\n");
    int snd_len = pUsb->send_image_data(img_buf, buf_len);
    if(snd_len < 0) {
        KnPrint(KN_LOG_ERR, "receving transfer ack msg failed.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "file transfer completed :%d.\n", snd_len);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for start dme rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "start dme response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "start dme rsp from device received.\n");

    //extract msg content
    KnStartDmeRsp* dme_msg = (KnStartDmeRsp*)p_rsp;
    uint32_t rsp_code = dme_msg->_rsp_code;
    *ret_size = dme_msg->_inf_size;

    //free resposne memory
    delete p_rsp;

    if(rsp_code != 0) {
        KnTsPrint(KN_LOG_DBG, "start dme rsp with response code:%d.\n", rsp_code);
        return rsp_code;
    }

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, infer size:%d from start dme rsp.\n", 
            dme_msg->_rsp_code, dme_msg->_inf_size);
    return 0;
}

int SyncMsgHandler::kdp_dme_configure(int dev_idx, uint32_t setup_conf, char* data, int dat_size) {
    uint16_t msg_type = KN_APP_CONFIG_DME_REQ;
    uint16_t msg_len = dat_size + sizeof(uint32_t);

    //create req message
    KnDmeConfigReq* p_req = new KnDmeConfigReq(msg_type, msg_len, setup_conf, data, dat_size);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create dme config req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending dme config req to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending dme config request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving dme config req to local map.\n");
    save_req_to_map(msg_type, dev_idx);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for dme config rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "dme config response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "dme config rsp from device received.\n");

    //extract msg content
    KnDmeConfigRsp* dme_msg = (KnDmeConfigRsp*)p_rsp;
    uint32_t rsp_code = dme_msg->_rsp_code;

    //free resposne memory
    delete p_rsp;

    if(rsp_code != 0) {
        KnTsPrint(KN_LOG_DBG, "dme config rsp with response code:%d.\n", rsp_code);
        return rsp_code;
    }

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d from dme config rsp.\n", dme_msg->_rsp_code);
    return 0;
}

int SyncMsgHandler::kdp_dme_inference(int dev_idx, char* img_buf, int buf_len, uint32_t* inf_size,\
                bool* res_flag, char* inf_res) {
    uint16_t msg_type = KN_APP_DME_IMG_REQ;
    uint16_t msg_len = sizeof(uint32_t) + sizeof(uint32_t);

    BaseDev* pDev = _p_mgr->getDevFromIdx(dev_idx);
    if(!pDev) {
        KnPrint(KN_LOG_ERR, "could not get device for idx:%d.\n", dev_idx);
        return -1;
    }

    USBDev* pUsb = NULL;
    if(pDev->dev_type == KDP_UART_DEV) {
        UARTDev* pUart = (UARTDev*)pDev;
        pUsb = pUart->get_helper_dev();
        if(!pUsb) {
            KnPrint(KN_LOG_ERR, "could not get helper device for idx:%d.\n", dev_idx);
            return -1;
        }
    } else {
        pUsb = (USBDev*)pDev;
    }

    if(!res_flag) {
        KnPrint(KN_LOG_ERR, "inference result flag need to be specified.\n");
        return -1;
    }

    if(*res_flag == true && !inf_res) {
        KnPrint(KN_LOG_ERR, "inference result is requested, but buffer is null.\n");
        return -1;
    }

    //create req message
    KnDmeImageReq* p_req = new KnDmeImageReq(msg_type, msg_len, buf_len);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create dme img req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending dme img req to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending dme img request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving dme img req to local map.\n");
    save_req_to_map(msg_type, dev_idx);

    //to call usb interface to transfer image file.
    KnTsPrint(KN_LOG_DBG, "calling usb interface to transfer file:........\n");
    int snd_len = pUsb->send_image_data(img_buf, buf_len);
    if(snd_len < 0) {
        KnPrint(KN_LOG_ERR, "receving transfer ack msg failed.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "file transfer completed :%d.\n", snd_len);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for dme img rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "dme img response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "dme img rsp from device received.\n");

    //extract msg content
    KnDmeImageRsp* dme_msg = (KnDmeImageRsp*)p_rsp;
    uint32_t rsp_code = dme_msg->_rsp_code;
    *inf_size = dme_msg->_inf_size;

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, infer size:%d from dme img rsp.\n", 
            dme_msg->_rsp_code, dme_msg->_inf_size);
    //free resposne memory
    delete p_rsp;

    if(rsp_code != 0) {
        KnTsPrint(KN_LOG_DBG, "dme img rsp with response code:%d.\n", rsp_code);
        return rsp_code;
    }

    //retrieve the result
    if(*res_flag == false) {
        return 0;
    }

    snd_len = pUsb->send_data_ack(0, 0);
    if(snd_len < 0) {
        KnPrint(KN_LOG_ERR, "sending ack to device failed.\n");
        *res_flag = false;
        return -1;
    }

    int r_len = pUsb->receive_bulk(inf_res, *inf_size);
    if(r_len < 0 || (uint32_t)r_len != *inf_size) {
        KnPrint(KN_LOG_ERR, "receiving inference result failed: %d.\n", r_len);
        *res_flag = false;
        return -1;
    }
    return 0;
}

int SyncMsgHandler::kdp_get_dme_res(int dev_idx, uint32_t* inf_size, char* inf_res) {
    uint16_t msg_type = KN_APP_DME_GET_REQ;
    uint16_t msg_len = 8;

    //create req message
    KnDmeGetReq* p_req = new KnDmeGetReq(msg_type, msg_len);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create dme get req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending dme get req to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending dme get request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving dme get req to local map.\n");
    save_req_to_map(msg_type, dev_idx);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for dme get rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "dme get response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "dme get rsp from device received.\n");

    //extract msg content
    KnDmeGetRsp* dme_msg = (KnDmeGetRsp*)p_rsp;
    uint32_t rsp_code = dme_msg->_rsp_code;
    *inf_size = dme_msg->_inf_size;

    if(rsp_code != 0) {
        KnTsPrint(KN_LOG_DBG, "dme get rsp with response code:%d.\n", rsp_code);
        return rsp_code;
    }

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, inf size:%d from dme get rsp.\n", 
            dme_msg->_rsp_code, dme_msg->_inf_size);
    //free resposne memory
    delete p_rsp;

    //call usb to get result
    BaseDev* pDev = _p_mgr->getDevFromIdx(dev_idx);
    if(!pDev) {
        KnPrint(KN_LOG_ERR, "could not get device for idx:%d.\n", dev_idx);
        return -1;
    }

    USBDev* pUsb = NULL;
    if(pDev->dev_type == KDP_UART_DEV) {
        UARTDev* pUart = (UARTDev*)pDev;
        pUsb = pUart->get_helper_dev();
        if(!pUsb) {
            KnPrint(KN_LOG_ERR, "could not get helper device for idx:%d.\n", dev_idx);
            return -1;
        }
    } else {
        pUsb = (USBDev*)pDev;
    }

    KnTsPrint(KN_LOG_DBG, "inference result is available.\n");
    uint32_t rcv_len = pUsb->receive_bulk(inf_res, *inf_size);
    if(rcv_len != *inf_size) {
        KnTsPrint(KN_LOG_DBG, "calling usb to get inf result not completed:%d.\n", rcv_len);
        return -1;
    }

    return 0;
}


int SyncMsgHandler::kdp_sfid_send_lw3d_gen(int dev_idx, char* img_buf, int buf_len, \
        char* img_buf_nir, int buf_len_nir, uint32_t* rsp_code, \
        uint16_t* user_id, uint16_t* mode, uint16_t* mask, char* res) {
    uint16_t msg_type = KN_APP_LW3D_IMAGE;
    uint16_t msg_len = sizeof(uint32_t)*2 + sizeof(uint16_t)*2;

    BaseDev* pDev = _p_mgr->getDevFromIdx(dev_idx);
    if(!pDev) {
        KnPrint(KN_LOG_ERR, "could not get device for idx:%d.\n", dev_idx);
        return -1;
    }

    USBDev* pUsb = NULL;
    if(pDev->dev_type == KDP_UART_DEV) {
        UARTDev* pUart = (UARTDev*)pDev;
        pUsb = pUart->get_helper_dev();
        if(!pUsb) {
            KnPrint(KN_LOG_ERR, "could not get helper device for idx:%d.\n", dev_idx);
            return -1;
        }
    } else {
        pUsb = (USBDev*)pDev;
    }

    //create req message
    uint16_t _msk = 0;
    if (mask != NULL) {
        _msk = *mask;
    }
    
    KnSendLw3dImageReq* p_req = new KnSendLw3dImageReq(msg_type, msg_len, buf_len, buf_len_nir, _msk);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create sfid lw3d send img req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending sfid lw3d img to dev ........\n");
    int ret = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "sending sfid send lw3d img request to device failed:%d.\n", ret);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving sfid lw3d img req to local map.\n");
    save_req_to_map(msg_type, dev_idx);

    //to call usb interface to transfer image file.
    KnTsPrint(KN_LOG_DBG, "calling usb interface to transfer rgb file:........\n");
    int snd_len = pUsb->send_image_data(img_buf, buf_len);
    if(snd_len < 0) {
        KnPrint(KN_LOG_ERR, "receving transfer ack msg failed.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "rgb file transfer completed :%d.\n", snd_len);

    KnTsPrint(KN_LOG_DBG, "calling usb interface to transfer nir file:........\n");
    snd_len = pUsb->send_image_data(img_buf_nir, buf_len_nir);
    if(snd_len < 0) {
        KnPrint(KN_LOG_ERR, "receving transfer ack msg failed.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "nir file transfer completed :%d.\n", snd_len);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for sfid lw3d img rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "sfid send lw3d img response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "sfid img lw3d rsp from device received.\n");

    //extract msg content
    KnSendLw3dImageRsp* sfid_msg = (KnSendLw3dImageRsp*)p_rsp;
    *rsp_code = sfid_msg->_rsp_code;
    *user_id = sfid_msg->_uid;
    *mode = sfid_msg->_mode;
    _msk = sfid_msg->_out_flag;

    if(mask != NULL){
        *mask = _msk;
    }
    int off = 0;
    if(_msk & 0x01) {
        memcpy((void*)(res+off), sfid_msg->_fd_rgb_res, FD_RES_LENGTH);
        KnTsPrint(KN_LOG_DBG, "rgb fd result is available.\n");
        off += FD_RES_LENGTH;
    }

    if(_msk & 0x02) {
        memcpy((void*)(res+off), sfid_msg->_lm_rgb_res, LM_RES_LENGTH);
        KnTsPrint(KN_LOG_DBG, "rgb lm result is available.\n");
        off += LM_RES_LENGTH;
    }
    
    if(_msk & 0x04) {
        memcpy((void*)(res+off), sfid_msg->_fr_rgb_res, FR_RES_LENGTH);
        KnTsPrint(KN_LOG_DBG, "rgb fr result is available.\n");
        off += FR_RES_LENGTH;
    }

    if(_msk & 0x10) {
        memcpy((void*)(res+off), sfid_msg->_fd_nir_res, FD_RES_LENGTH);
        KnTsPrint(KN_LOG_DBG, "nir fd result is available.\n");
        off += FD_RES_LENGTH;
    }

    if(_msk & 0x20) {
        memcpy((void*)(res+off), sfid_msg->_lm_nir_res, LM_RES_LENGTH);
        KnTsPrint(KN_LOG_DBG, "nir lm result is available.\n");
        off += LM_RES_LENGTH;
    }
    
    if(_msk & 0x40) {
        memcpy((void*)(res+off), sfid_msg->_fr_nir_res, FR_RES_LENGTH);
        KnTsPrint(KN_LOG_DBG, "nir fr result is available.\n");
        off += FR_RES_LENGTH;
    }

    if(_msk & 0x100) {
        memcpy((void*)(res+off), &(sfid_msg->_live_res), sizeof(uint16_t));
        KnTsPrint(KN_LOG_DBG, "liveness result is available.\n");
    }

    if(*rsp_code != 0) {
        KnTsPrint(KN_LOG_DBG, "sfid img lw3d rsp with response code:%d.\n", *rsp_code);
        delete p_rsp;
        return *rsp_code;
    }

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, user_id:%d, mode:%d, mask:%d from sfid lw3d img rsp.\n", 
            sfid_msg->_rsp_code, sfid_msg->_uid, sfid_msg->_mode, sfid_msg->_out_flag);
    //free resposne memory
    delete p_rsp;
    return 0;
}

int SyncMsgHandler::kdp_sfid_send_lw3d_img(int dev_idx, char* img_buf, int buf_len, \
        char* img_buf_nir, int buf_len_nir, uint32_t* rsp_code, \
        uint16_t* user_id, uint16_t* mode, bool* fd_flag, char* fd_res, \
        bool* fr_flag, char* fr_res, bool* lm_flag, char* lm_res, \
        bool* fd_flag_nir, char* fd_res_nir, bool* fr_flag_nir, \
        char* fr_res_nir, bool* lm_flag_nir, char* lm_res_nir, \
        bool* live_flag, char* live_res) {
    uint16_t msg_type = KN_APP_LW3D_IMAGE;
    uint16_t msg_len = sizeof(uint32_t)*2 + sizeof(uint16_t)*2;

    BaseDev* pDev = _p_mgr->getDevFromIdx(dev_idx);
    if(!pDev) {
        KnPrint(KN_LOG_ERR, "could not get device for idx:%d.\n", dev_idx);
        return -1;
    }

    USBDev* pUsb = NULL;
    if(pDev->dev_type == KDP_UART_DEV) {
        UARTDev* pUart = (UARTDev*)pDev;
        pUsb = pUart->get_helper_dev();
        if(!pUsb) {
            KnPrint(KN_LOG_ERR, "could not get helper device for idx:%d.\n", dev_idx);
            return -1;
        }
    } else {
        pUsb = (USBDev*)pDev;
    }

    uint16_t mask = 0;
    if(fd_flag && (*fd_flag == true)) mask |= 0x01;
    if(lm_flag && (*lm_flag == true)) mask |= 0x02;
    if(fr_flag && (*fr_flag == true)) mask |= 0x04;

    if(fd_flag_nir && (*fd_flag_nir == true)) mask |= 0x10;
    if(lm_flag_nir && (*lm_flag_nir == true)) mask |= 0x20;
    if(fr_flag_nir && (*fr_flag_nir == true)) mask |= 0x40;

    if(live_flag && (*live_flag == true)) mask |= 0x100;

    //create req message
    KnSendLw3dImageReq* p_req = new KnSendLw3dImageReq(msg_type, msg_len, buf_len, buf_len_nir, mask);
    if(!p_req->is_msg_valid()) {
        KnPrint(KN_LOG_ERR, "create sfid send lw3d img req failed.\n");
        delete p_req;
        return -1;
    }

    //sending msg to queue
    KnTsPrint(KN_LOG_DBG, "sending sfid lw3d img to dev ........\n");
    int res = _p_mgr->send_cmd_msg_sync(dev_idx, p_req);
    if(res < 0) {
        KnPrint(KN_LOG_ERR, "sending sfid send lw3d img request to device failed:%d.\n", res);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "saving sfid lw3d img req to local map.\n");
    save_req_to_map(msg_type, dev_idx);

    //to call usb interface to transfer image file.
    KnTsPrint(KN_LOG_DBG, "calling usb interface to transfer rgb file:........\n");
    int snd_len = pUsb->send_image_data(img_buf, buf_len);
    if(snd_len < 0) {
        KnPrint(KN_LOG_ERR, "receving transfer ack msg failed.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "rgb file transfer completed :%d.\n", snd_len);

    KnTsPrint(KN_LOG_DBG, "calling usb interface to transfer nir file:........\n");
    snd_len = pUsb->send_image_data(img_buf_nir, buf_len_nir);
    if(snd_len < 0) {
        KnPrint(KN_LOG_ERR, "receving transfer ack msg failed.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "nir file transfer completed :%d.\n", snd_len);

    //wait for rsp
    KnTsPrint(KN_LOG_DBG, "waiting for sfid lw3d img rsp from device ........\n");
    KnBaseMsg* p_rsp = wait_rsp_msg(dev_idx, msg_type);
    if(!p_rsp) {
        KnPrint(KN_LOG_ERR, "sfid send lw3d img response message was not received.\n");
        return -1;
    }
    KnTsPrint(KN_LOG_DBG, "sfid lw3d img rsp from device received.\n");

    //extract msg content
    KnSendLw3dImageRsp* sfid_msg = (KnSendLw3dImageRsp*)p_rsp;
    *rsp_code = sfid_msg->_rsp_code;
    *user_id = sfid_msg->_uid;
    *mode = sfid_msg->_mode;

    if(fd_flag && (*fd_flag == true)) {
        if(sfid_msg->_out_flag & 0x01) { //fd res
            KnTsPrint(KN_LOG_DBG, "rgb fd result is available.\n");
            memcpy((void*)fd_res, (void*)(sfid_msg->_fd_rgb_res), FD_RES_LENGTH);
        } else {
            *fd_flag = false;
        }
    }

    if(lm_flag && (*lm_flag == true)) {
        if(sfid_msg->_out_flag & 0x02) { //lm res
            KnTsPrint(KN_LOG_DBG, "rgb lm result is available.\n");
            memcpy((void*)lm_res, (void*)(sfid_msg->_lm_rgb_res), LM_RES_LENGTH);
        } else {
            *lm_flag = false;
        }
    }

    if(fr_flag && (*fr_flag == true)) {
        if(sfid_msg->_out_flag & 0x04) { //fr res
            KnTsPrint(KN_LOG_DBG, "rgb fr result is available.\n");
            memcpy((void*)fr_res, (void*)(sfid_msg->_fr_rgb_res), FR_RES_LENGTH);
        } else {
            *fr_flag = false;
        }
    }

    if(fd_flag_nir && (*fd_flag_nir == true)) {
        if(sfid_msg->_out_flag & 0x10) { //fd res
            KnTsPrint(KN_LOG_DBG, "nir fd result is available.\n");
            memcpy((void*)fd_res_nir, (void*)(sfid_msg->_fd_nir_res), FD_RES_LENGTH);
        } else {
            *fd_flag_nir = false;
        }
    }

    if(lm_flag_nir && (*lm_flag_nir == true)) {
        if(sfid_msg->_out_flag & 0x20) { //lm res
            KnTsPrint(KN_LOG_DBG, "nir lm result is available.\n");
            memcpy((void*)lm_res_nir, (void*)(sfid_msg->_lm_nir_res), LM_RES_LENGTH);
        } else {
            *lm_flag_nir = false;
        }
    }

    if(fr_flag_nir && (*fr_flag_nir == true)) {
        if(sfid_msg->_out_flag & 0x40) { //fr res
            KnTsPrint(KN_LOG_DBG, "nir fr result is available.\n");
            memcpy((void*)fr_res_nir, (void*)(sfid_msg->_fr_nir_res), FR_RES_LENGTH);
        } else {
            *fr_flag_nir = false;
        }
    }

    if(live_flag && (*live_flag == true)) {
        if(sfid_msg->_out_flag & 0x100) { //fr res
            KnTsPrint(KN_LOG_DBG, "liveness result is available.\n");
            memcpy((void*)live_res, (void*)&(sfid_msg->_live_res), sizeof(uint16_t));
        } else {
            *live_flag = false;
        }
    }

    if(*rsp_code != 0) {
        KnTsPrint(KN_LOG_DBG, "sfid lw3d img rsp with response code:%d.\n", *rsp_code);
        delete p_rsp;
        return *rsp_code;
    }

    KnTsPrint(KN_LOG_DBG, "rsp_code:%d, user_id:%d, mode:%d, mask:%d from sfid lw3d img rsp.\n", 
            sfid_msg->_rsp_code, sfid_msg->_uid, sfid_msg->_mode, sfid_msg->_out_flag);
    //free resposne memory
    delete p_rsp;
    return 0;
}

