/**
 * @file      USBDev.cpp
 * @brief     implementation file for USB Dev class
 * @version   0.1 - 2019-04-28
 * @copyright (c) 2019 Kneron Inc. All right reserved.
 */

#include "USBDev.h"
#include "KnLog.h"
#include "KdpHostApi.h"
#include "UARTPkt.h"
#include "KnUtil.h"
#include "KnMsg.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include <unistd.h>

#define MAX_RETRY_NUM 2
#define BULK_SIZE_USB 512

int USBDev::_inst_cnt = 0;

USBDev::USBDev(int vid, int pid) {
    m_pkneron_device_h = NULL;
    p_kneron_dev = NULL;
    vendor_id = vid;
    product_id = pid;
    dev_type = KDP_USB_DEV;
    p_dev_list = NULL;

    _inst_cnt++;
    dev_fd = USB_FAKE_FD_START + _inst_cnt;
}


USBDev::~USBDev() {
    if (NULL != m_pkneron_device_h) {

        KnPrint(KN_LOG_DBG, "reseting usb device before exiting.\n");
        libusb_reset_device(m_pkneron_device_h);
        KnPrint(KN_LOG_DBG, "usb device has been reset.\n");

        libusb_release_interface(m_pkneron_device_h, 0);
        libusb_close(m_pkneron_device_h);
        m_pkneron_device_h = NULL;
    }
    if(p_dev_list) {
        libusb_free_device_list(p_dev_list, 1);
    }

    _inst_cnt--;
}

int USBDev::u_init_port()
{
    libusb_device** devs;
    libusb_device* dev;

    ssize_t cnt = libusb_get_device_list(NULL, &devs);
    if(cnt < 0) {
        KnPrint(KN_LOG_ERR, "get usb device list failed, error code:%d.\n", cnt);
        return (int) cnt;
    }

    p_dev_list = devs;

    int i = 0;
    struct libusb_device_descriptor desc;

    int rv = -1;
    while( (dev = devs[i++]) != NULL) {
        rv = libusb_get_device_descriptor(dev, &desc);
        if(rv != 0) {
            KnPrint(KN_LOG_ERR, "get usb device descriptor failed, error code:%d.\n", rv);
            continue;
        }

        if(desc.idProduct == product_id && desc.idVendor == vendor_id) {
            KnPrint(KN_LOG_DBG, "found correct usb dev vid:%d, pid:%d.\n", 
                vendor_id, product_id);
            p_kneron_dev = dev;
            break; 
        }
    }

    if(!p_kneron_dev) {
        KnPrint(KN_LOG_ERR, "no usb device is found, error code:%d.\n", rv);
        return -1;
    } else {
        return 0;
    }
}

int USBDev::u_open() {
    if (NULL == p_kneron_dev) return -1;

    int status = libusb_open(p_kneron_dev, &m_pkneron_device_h);
    if( status != 0 || !m_pkneron_device_h ) {
        KnPrint(KN_LOG_ERR, "usb device vid:%d, pid:%d, open failed.\n", vendor_id, product_id);
        return -1;
    }

    if (0 != (status =  libusb_set_configuration(m_pkneron_device_h, 1))) {
        KnPrint(KN_LOG_ERR, "error for set usb config:%d.\n", status);
        libusb_close(m_pkneron_device_h);
        m_pkneron_device_h = NULL;
        return -1;
    }

    if (0 != (status = libusb_claim_interface(m_pkneron_device_h, 0))) {
        KnPrint(KN_LOG_ERR, "claim interface failed:%d.\n", status);
        libusb_close(m_pkneron_device_h);
        m_pkneron_device_h = NULL;
        return -1;
    }
    KnPrint(KN_LOG_DBG, "usb dev opened successfully.\n");
    return 0;
}

int USBDev::send(int8_t* msgbuf, int msglen) {
    return send((char*) msgbuf, msglen);
}

int USBDev::send(char* msgbuf, int msglen) {
    int status = 0;

    if (NULL == m_pkneron_device_h)
        return -1;

    //need to pack the data
    KnTsPrint(KN_LOG_DBG, "packing payload with uart pkt:%d.\n", msglen);
    UARTPkt pkt((int8_t*)msgbuf, msglen, true, MSG_CRC_FLAG);

    //calling base senf func to send the packed data
    if(!pkt.is_pkt_valid()) {
        KnPrint(KN_LOG_ERR, "packing to UART packet falied.\n");
        return -1;
    }

    kn_dump_msg(pkt.get_pkt_buf(), pkt.get_pkt_len());

    KnTsPrint(KN_LOG_DBG, "kdp msg is packed with uart pkt format:%d.\n", pkt.get_pkt_len());

    int transfered = 0;
    KnTsPrint(KN_LOG_DBG, "calling usb interrupt send to send data :%d...\n", msglen);
    //status = libusb_interrupt_transfer(m_pkneron_device_h, EP_OUT_INT, (unsigned char*)msgbuf, msglen, &transfered, 1000);
    status = libusb_bulk_transfer(m_pkneron_device_h, EP_OUT_INT, 
            (unsigned char*)(pkt.get_pkt_buf()), pkt.get_pkt_len(), &transfered, 1000);
    if(status < 0) {
        KnPrint(KN_LOG_ERR, "write usb failed: %d, %s.\n", status, libusb_error_name(status));
        return status;
    }

    if( (transfered > 0) && (transfered % BULK_SIZE_USB == 0) ) {
        //need a 0 byte transfer
        unsigned char tt[1];
        int tmp = 0;
        KnTsPrint(KN_LOG_DBG, "zero byte sending...\n");
        //libusb_interrupt_transfer(m_pkneron_device_h, EP_OUT_INT, tt, 0, &tmp, 100);
        libusb_bulk_transfer(m_pkneron_device_h, EP_OUT_INT, tt, 0, &tmp, 100);
    }

    KnTsPrint(KN_LOG_DBG, "usb write sent data :%d.\n", transfered);
    return transfered;
}

int USBDev::send_bulk(char* msgbuf, int msglen) {
    int status = 0;

    if (NULL == m_pkneron_device_h)
        return -1;

    int transfered = 0;
    KnTsPrint(KN_LOG_DBG, "calling usb bulk send to send data :%d...\n", msglen);
    status = libusb_bulk_transfer(m_pkneron_device_h, EP_OUT_BULK, (unsigned char*)msgbuf, msglen, &transfered, 1000);
    if(status < 0) {
        KnPrint(KN_LOG_ERR, "write usb failed: %d, %s.\n", status, libusb_error_name(status));
        return status;
    }

    if( (transfered > 0) && (transfered % BULK_SIZE_USB == 0) ) {
        //need a 0 byte transfer
        unsigned char tt[1];
        int tmp = 0;
        KnTsPrint(KN_LOG_DBG, "zero byte sending...\n");
        libusb_bulk_transfer(m_pkneron_device_h, EP_OUT_BULK, tt, 0, &tmp, 100);
    }

    KnTsPrint(KN_LOG_DBG, "usb write sent data :%d.\n", transfered);
    return transfered;
}

int USBDev::receive(int8_t* msgbuf, int msglen) {
    return receive((char*)msgbuf, msglen);
}

void USBDev::shutdown_dev() {
    KnTsPrint(KN_LOG_DBG, "shutdown usb device...\n");

    if (NULL != m_pkneron_device_h) {
        libusb_release_interface(m_pkneron_device_h, 0);
        libusb_close(m_pkneron_device_h);
        m_pkneron_device_h = NULL;
    }
    if(p_dev_list) {
        libusb_free_device_list(p_dev_list, 1);
        p_dev_list = NULL;
        p_kneron_dev = NULL;
    }
}

void USBDev::restart_dev() {
    KnTsPrint(KN_LOG_DBG, "restarting usb device...\n");

    KnTsPrint(KN_LOG_DBG, "calling init port...\n");
    if(u_init_port() < 0) {
        KnPrint(KN_LOG_ERR, "init port failed for restarting usb.\n");
        return;
    }

    KnTsPrint(KN_LOG_DBG, "calling usb open...\n");
    if(u_open() < 0) {
        KnPrint(KN_LOG_ERR, "usb open failed for restarting usb.\n");
        return;
    }
}


int USBDev::reconnect_usb() {
    //close usb
    KnTsPrint(KN_LOG_DBG, "re connecting usb for no such device error...\n");

    if (NULL != m_pkneron_device_h) {
        libusb_release_interface(m_pkneron_device_h, 0);
        libusb_close(m_pkneron_device_h);
        m_pkneron_device_h = NULL;
    }
    if(p_dev_list) {
        libusb_free_device_list(p_dev_list, 1);
        p_dev_list = NULL;
        p_kneron_dev = NULL;
    }

    usleep(10000);

    KnTsPrint(KN_LOG_DBG, "calling init port...\n");
    if(u_init_port() < 0) {
        KnPrint(KN_LOG_ERR, "init port failed for reconnecting usb.\n");
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "calling usb open...\n");
    if(u_open() < 0) {
        KnPrint(KN_LOG_ERR, "usb open failed for reconnecting usb.\n");
        return -1;
    }
    return 0;
}

int USBDev::receive(char* msgbuf, int msglen) {
    int ret_val;
    if (NULL == m_pkneron_device_h)
        return -1;

    int transfered = 0;
    KnTsPrint(KN_LOG_DBG, "calling usb interrupt read to receive data :%d...\n", msglen);

    for(int i = 0; i < MAX_RETRY_NUM; i++) {
        //ret_val = libusb_interrupt_transfer(m_pkneron_device_h, EP_IN_INT , (unsigned char*)msgbuf, msglen, &transfered, 2000);
        ret_val = libusb_bulk_transfer(m_pkneron_device_h, EP_IN_INT , (unsigned char*)msgbuf, msglen, &transfered, 2000);
        if(ret_val != 0) {
            KnTsPrint(KN_LOG_ERR, "USB EP read failed:%d, %s.\n", ret_val, libusb_error_name(ret_val));
            if(ret_val != LIBUSB_ERROR_NO_DEVICE) {
                return -1;
            } else { //no such device
                int s = reconnect_usb();
                if(s < 0) return -1;
                continue; //retry
            }
        } else {
            if( (transfered > 0) && (transfered % BULK_SIZE_USB == 0) ) {
                //need a 0 byte transfer
                unsigned char tt[1];
                int tmp = 0;
                KnTsPrint(KN_LOG_DBG, "zero byte receiving...\n");
                //libusb_interrupt_transfer(m_pkneron_device_h, EP_IN_INT, tt, 0, &tmp, 100);
                libusb_bulk_transfer(m_pkneron_device_h, EP_IN_INT, tt, 0, &tmp, 100);
            }
            break;
        }
    }

    KnTsPrint(KN_LOG_DBG, "usb read received data :%d.\n", transfered);
    return transfered;
}

int USBDev::receive_bulk(char* msgbuf, int msglen) {
    int ret_val;
    if (NULL == m_pkneron_device_h)
        return -1;

    int transfered = 0;
    KnTsPrint(KN_LOG_DBG, "calling usb bulk read to receive data :%d...\n", msglen);

    for(int i = 0; i < MAX_RETRY_NUM; i++) {
        ret_val = libusb_bulk_transfer(m_pkneron_device_h, EP_IN_BULK , (unsigned char*)msgbuf, msglen, &transfered, 2000);
        if(ret_val != 0) {
            KnTsPrint(KN_LOG_ERR, "USB EP read failed:%d, %s.\n", ret_val, libusb_error_name(ret_val));
            if(ret_val != LIBUSB_ERROR_NO_DEVICE) {
                return -1;
            } else { //no such device
                int s = reconnect_usb();
                if(s < 0) return -1;
                continue; //retry
            }
        } else {
            if( (transfered > 0) && (transfered % BULK_SIZE_USB == 0) ) {
                //need a 0 byte transfer
                unsigned char tt[1];
                int tmp = 0;
                KnTsPrint(KN_LOG_DBG, "zero byte receiving...\n");
                libusb_bulk_transfer(m_pkneron_device_h, EP_IN_BULK, tt, 0, &tmp, 100);
            }
            break;
        }
    }

    KnTsPrint(KN_LOG_DBG, "usb read received data :%d.\n", transfered);
    return transfered;
}

int USBDev::usb_recv(int8_t* msgbuf, int msglen) {
    if (NULL == m_pkneron_device_h) return -1;
    int transfered = 0;

    //int ret_val =  libusb_interrupt_transfer(m_pkneron_device_h, EP_IN_INT , (unsigned char*)msgbuf, msglen, &transfered, 2000);
    int ret_val =  libusb_bulk_transfer(m_pkneron_device_h, EP_IN_INT , (unsigned char*)msgbuf, msglen, &transfered, 100);
    if(ret_val != 0) {
        if(ret_val == LIBUSB_ERROR_NO_DEVICE) {
            KnTsPrint(KN_LOG_ERR, "USB EP read failed:%d, %s.\n", ret_val, libusb_error_name(ret_val));
            reconnect_usb(); //reconnect to usb in case of no device
        }
        return ret_val;
    }

    if( (transfered > 0) && (transfered % BULK_SIZE_USB == 0) ) {
        //need a 0 byte transfer
        unsigned char tt[1];
        int tmp = 0;
        KnTsPrint(KN_LOG_DBG, "zero byte sending...\n");
        //libusb_interrupt_transfer(m_pkneron_device_h, EP_IN_INT, tt, 0, &tmp, 100);
        libusb_bulk_transfer(m_pkneron_device_h, EP_IN_INT, tt, 0, &tmp, 100);
    }

    kn_dump_msg(msgbuf, transfered, false);

    //unpack the data
    KnPrint(KN_LOG_DBG, "unpacking msg payload...\n");
    UARTPkt pkt(msgbuf, transfered);
    if(!pkt.is_pkt_valid()) {
        KnPrint(KN_LOG_ERR, "unpacking UART packet falied.\n");
        return -1;
    }

    int bytes_received = pkt.get_payload_len();
    memset(msgbuf, 0, msglen);
    memcpy(msgbuf, pkt.get_payload_buf(), bytes_received);

    KnPrint(KN_LOG_DBG, "get msg payload from uart pkt:%d.\n", bytes_received);

    //payload
    if(bytes_received < (int)sizeof(KnMsgHdr)) {  //less than header
        KnPrint(KN_LOG_ERR, "receive msg len less than hdr.\n");
    } else {
        KnMsgHdr hdr;
        KnBaseMsg::decode_msg_hdr(&hdr, msgbuf, bytes_received);
        int mlen = hdr.msg_len;
        if(bytes_received < mlen + (int)sizeof(KnMsgHdr)) { //only a part msg
            KnPrint(KN_LOG_ERR, "only a part of msg received:%d,%d.\n", bytes_received, mlen);
        }
        //whole msg is ready
    }

    KnPrint(KN_LOG_DBG, "from com dev, got %d bytes msg.\n", bytes_received);

    //kn_dump_msg(msgbuf, bytes_received, false);
    return bytes_received;
}


int USBDev::recv_data_ack() {
    int8_t ack_buffer[ACK_PACKET_SIZE];
    memset(ack_buffer, 0, sizeof(ack_buffer));
    //uint8_t usb_ack[8] = { 0x35, 0x8A, 0x04, 0x00, 0x04, 0x00, 0x08, 0x00 };

    KnTsPrint(KN_LOG_DBG, "usb dev is receiving ack...\n");
    int status = receive_bulk((char*)ack_buffer, ACK_PACKET_SIZE);
    if(status < 4) {
        KnPrint(KN_LOG_ERR, "receiving ack failed:%d.\n", status);
        return -1;
    }

    //get the ack msg
    KnImgTransAck ack_msg(ack_buffer+4, status-4);
    if(ack_msg.is_msg_valid() == false) {
        KnPrint(KN_LOG_ERR, "ack msg parsing failed:%d.\n", status);
        return -1;
    }

    if(ack_msg._action != 0) { //not ack
        KnPrint(KN_LOG_ERR, "got nack msg from device %d.\n", ack_msg._action);
        return -1;
    }

    return 0;
}

int USBDev::send_data_ack(uint32_t action, uint32_t src_id) {
    uint8_t ack_buffer[ACK_PACKET_SIZE] = { 0x83, 0xA5, 0x04, 0x00, }; //pack header

    KnImgTransAck ack_msg(KN_MSG_GEN_ACK, 8, action, src_id);
    if(ack_msg.is_msg_valid() == false) {
        KnPrint(KN_LOG_ERR, "creating ack msg failed.\n");
        return -1;
    }

    //packing msg
    memcpy(ack_buffer+4, ack_msg.get_msg_payload(), ack_msg.get_msg_len());

    //send it
    int status = send_bulk((char*)ack_buffer, ACK_PACKET_SIZE);
    if(status != ACK_PACKET_SIZE) {
        KnPrint(KN_LOG_ERR, "sending ack msg failed:%d.\n", status);
        return -1;
    }

    KnTsPrint(KN_LOG_DBG, "sent ack msg to dev.\n");
    return 0;
}

int USBDev::send_image_data(char *buf, int data_len) {
    if(data_len <= 0 || !buf) return -1;

    KnTsPrint(KN_LOG_DBG, "usb dev is starting to send image :%d...\n", data_len);

    int status = recv_data_ack();
    if(status != 0) {
        KnPrint(KN_LOG_ERR, "ack msg wrong:%d.\n", status);
        return -1;
    }

    int leftover = data_len;
    int i = 0;
    int off = 0;

    while(leftover > 0) {
        int len = static_cast<int> (USB_MAX_BYTES);
        if(len > leftover) len = leftover;

        KnTsPrint(KN_LOG_DBG, "usb dev is sending segment :%d, with len:%d...\n", i, len);
        status = send_bulk(buf + off, len);
        if(status < 0) {
            KnPrint(KN_LOG_ERR, "sending %d time for img, off:%d.\n", i, off);
            return status;
        }

        KnTsPrint(KN_LOG_DBG, "usb dev sent segment :%d, with len:%d...\n", i, status);
        
        off += status;
        leftover -= status;
        i++;
    }

    KnTsPrint(KN_LOG_DBG, "usb dev finished sending image :%d...\n", off);
    return off;
}

