/* -------------------------------------------------------------------------- 
 * Copyright (c) 2018-2019 Kneron Inc. All rights reserved.
 *
 *      Name:    main.c
 *      Purpose: Kneron NCPU
 *
 *---------------------------------------------------------------------------*/


#include <string.h>
#include "kneron_mozart.h"
#include "kneron_mozart_ext.h"
#include "cmsis_os2.h"
#include "os_tick.h"
#include "DrvUART010.h"
#include "ipc.h"
#include "io.h"
#include "base.h"
#include "kdpio.h"
#include "scu_ipc.h"
#include "version.h"
#include "bootloader.h"
#include "power_mgr.h"
#include "dbg.h"
#include "cpu_ops_ex.h"
#include "post_processing_ex.h"
#include "pre_processing_ex.h"

#define FLAG_SCPU_COMM_ISR				BIT0
#define FLAG_SCPU_RESET     			BIT1
#define FLAG_SCPU_ERROR                 BIT2

#define FLAG_SCPU_ALL           (FLAG_SCPU_COMM_ISR | FLAG_SCPU_RESET | FLAG_SCPU_ERROR)

#define FLAG_NPU_COMM_ISR               BIT0

#define SCPU_IRQ                SYSC_SGI_S_STATUS_IRQ

osThreadId_t tid_scpu_comm;
osThreadId_t tid_npu_comm;

struct scpu_to_ncpu_s *in_comm_p;
struct ncpu_to_scpu_s *out_comm_p;

/* IPC_IMAGE_ACTIVE_MAX=2 for 2 parallel input images:
       one for npu, another for post-process */
struct kdp_image_s image_s[IPC_IMAGE_ACTIVE_MAX];
int npu_img_idx;
struct kdp_image_s *npu_img_p;
struct kdp_image_s *pp_img_p;

extern void kdp_ddr_dma_copy(unsigned int *src_ptr, unsigned int *dst_ptr, int size);
extern void mpu_config(void);

/*----------------------------------------------------------------------------
 *      SCPU Communication and Main Thread
 *---------------------------------------------------------------------------*/
void scpu_comm_thread(void *argument)
{
    int32_t flags;
    struct kdp_img_raw_s *raw_img_p;
    struct kdp_model_s *model_p;
    int model_slot_index, image_index;

    /* Init kdpio lib and interrupt/status */
    kdpio_init();

    /* Register pre_processing function dynamically */
    kdpio_pre_processing_register(TINY_YOLO_V3, kdp_preproc_inproc);
    /* An example to register preprocess function in pre_processing_ex.c */
    //kdpio_pre_processing_register(TINY_YOLO_V3, preprocess_rgba_right_shift_yolo);
    /* Register cpu_op function dynamically */
    kdpio_cpu_op_register(NEAREST_UPSAMPLE, nearest_upsample_cpu);
    /* Register post_processing function dynamically */
    kdpio_post_processing_register(TINY_YOLO_V3, post_yolo_v3);

    scu_ipc_enable_to_scpu_int();

    NVIC_EnableIRQ((IRQn_Type)SCPU_IRQ);
    NVIC_EnableIRQ((IRQn_Type)NPU_NPU_IRQ);

    out_comm_p->id = NCPU2SCPU_ID;
    out_comm_p->status = STATUS_OK;

    critical_msg("ncpu: Ready!\n");

    /* Let's run tickless */
    OS_Tick_Disable();

    /* Waiting for all events */
    for (;;) {
        flags = osThreadFlagsWait(FLAG_SCPU_ALL, osFlagsWaitAny, osWaitForever);
        osThreadFlagsClear(flags);

        if (flags & FLAG_SCPU_COMM_ISR) {
            dbg_msg("ncpu: got IPC interrupt\n");

            if (in_comm_p->cmd == CMD_RUN_NPU) {
                // Enable os tick for profiling or disable for saving power
                if (log_get_level_ncpu() >= LOG_PROFILE)
                    OS_Tick_Enable();
                else
                    OS_Tick_Disable();

                model_slot_index = in_comm_p->model_slot_index;
                model_p = &in_comm_p->models[model_slot_index];

                dbg_msg("NPU: model_slot_index=%d, NPU output mem addr=0x%x, len=%d\n", model_slot_index, (uint32_t)model_p->output_mem_addr, model_p->output_mem_len);
                dbg_msg("NPU: input_mem_addr = 0x%x, len=0x%x, cmd_mem_addr=0x%x, len=0x%x, wt_addr=0x%x, wt_len=0x%x, setup_addr=0x%x, setup_len=0x%x\n",
                   (uint32_t)model_p->input_mem_addr, (uint32_t)model_p->input_mem_len,
                   (uint32_t)model_p->cmd_mem_addr, (uint32_t)model_p->cmd_mem_len,
                   (uint32_t)model_p->weight_mem_addr, (uint32_t)model_p->weight_mem_len,
                   (uint32_t)model_p->setup_mem_addr, (uint32_t)model_p->setup_mem_len);

                image_index = in_comm_p->active_img_index;
                raw_img_p = &in_comm_p->raw_images[image_index];
                npu_img_p = &image_s[npu_img_idx];
                npu_img_idx = !npu_img_idx;

                npu_img_p->raw_img_p = raw_img_p;
                npu_img_p->model_id = in_comm_p->models_type[model_slot_index];
                PREPROC_INPUT_MEM_ADDR(npu_img_p) = model_p->input_mem_addr;
                PREPROC_INPUT_MEM_LEN(npu_img_p) = model_p->input_mem_len;
                POSTPROC_RESULT_MEM_ADDR(npu_img_p) = raw_img_p->results[model_slot_index].result_mem_addr;
                POSTPROC_RESULT_MEM_LEN(npu_img_p) = raw_img_p->results[model_slot_index].result_mem_len;

                dbg_msg("NPU: model_type = %d\n",npu_img_p->model_id);

                out_comm_p->img_results[image_index].seq_num = raw_img_p->seq_num;
                out_comm_p->img_results[image_index].status = IMAGE_STATE_ACTIVE;

                PREPROC_INPUT_MEM_ADDR2(npu_img_p) = in_comm_p->input_mem_addr2;
                PREPROC_INPUT_MEM_LEN2(npu_img_p) = in_comm_p->input_mem_len2;
                POSTPROC_OUTPUT_MEM_ADDR3(npu_img_p) = in_comm_p->output_mem_addr3;

                kdpio_set_model(npu_img_p, model_p);

                kdpio_run_preprocess(npu_img_p);
                kdpio_run_npu_op(npu_img_p);
            }

            scu_ipc_clear_from_scpu_int();
            NVIC_ClearPendingIRQ((IRQn_Type)SCPU_IRQ);
            NVIC_EnableIRQ((IRQn_Type)SCPU_IRQ);
        }
    }
}

void scpu_isr(void)
{
    NVIC_DisableIRQ((IRQn_Type)SCPU_IRQ);
    dbg_msg("NCPU: got SCPU interrupt\n");
    osThreadFlagsSet(tid_scpu_comm, FLAG_SCPU_COMM_ISR);
}

static void npu_out_data_move(void *dst_p, void *src_p, int size)
{
    /* Move data out from npu output buffer for parallel handling */
    memcpy(dst_p, src_p, size);

    /* Notify scpu now */
    scu_ipc_trigger_to_scpu_int();
}

/*----------------------------------------------------------------------------
 *      NPU Communication Thread
 *---------------------------------------------------------------------------*/
void npu_comm_thread(void *argument)
{
    int32_t flags;
    int rc, i;

    for (;;) {
        flags = osThreadFlagsWait(FLAG_NPU_COMM_ISR, osFlagsWaitAny, osWaitForever);
        osThreadFlagsClear(flags);

        if (flags & FLAG_NPU_COMM_ISR) {

            rc = kdpio_run_cpu_op(npu_img_p);
            if (rc == RET_NO_ERROR) {
                i = npu_img_p->raw_img_p->ref_idx;

                dbg_msg("npu out: img[%d] state: %d -> %d\n", i, out_comm_p->img_results[i].status, IMAGE_STATE_NPU_DONE);

                out_comm_p->img_index_done = i;
                out_comm_p->img_results[i].status = IMAGE_STATE_NPU_DONE;
                POSTPROC_OUTPUT_MEM_ADDR(npu_img_p) = in_comm_p->output_mem_addr2;
                pp_img_p = npu_img_p;

                rc = kdpio_run_postprocess(pp_img_p, npu_out_data_move);

                if (rc == RET_NO_ERROR) {
                    i = pp_img_p->raw_img_p->ref_idx;

                    dbg_msg("npu out: img[%d] state: %d -> %d\n", i, out_comm_p->img_results[i].status, IMAGE_STATE_DONE);

                    out_comm_p->img_index_done = i;
                    out_comm_p->img_results[i].status = IMAGE_STATE_DONE;

                    scu_ipc_trigger_to_scpu_int();
                }
            }

            NVIC_ClearPendingIRQ((IRQn_Type)NPU_NPU_IRQ);
            NVIC_EnableIRQ((IRQn_Type)NPU_NPU_IRQ);
        }
    }
}

void npu_isr(void)
{
    NVIC_DisableIRQ((IRQn_Type)NPU_NPU_IRQ);
    osThreadFlagsSet(tid_npu_comm, FLAG_NPU_COMM_ISR);
}

/*----------------------------------------------------------------------------
 *      Init
 *---------------------------------------------------------------------------*/
static void ncpu_init()
{
    osThreadAttr_t attr;

    memset(&attr, 0, sizeof(attr));
    attr.priority = osPriorityHigh;
    tid_scpu_comm = osThreadNew(scpu_comm_thread, NULL, &attr);
    attr.priority = osPriorityNormal;
    tid_npu_comm = osThreadNew(npu_comm_thread, NULL, &attr);

    NVIC_SetVector((IRQn_Type)SCPU_IRQ, (uint32_t)scpu_isr);
    NVIC_SetVector((IRQn_Type)NPU_NPU_IRQ, (uint32_t)npu_isr);

    in_comm_p = (struct scpu_to_ncpu_s *)IPC_MEM_ADDR;
    out_comm_p = (struct ncpu_to_scpu_s *)IPC_MEM_ADDR2;

    power_mgr_init();
}

int cpu_type = 0;
void info(uint32_t boot_loader_flag) {
#define WHOAMI_KEY 0x12345678
    char infobuf[100];
    osVersion_t osv;
    osStatus_t status;
    uint32_t rom_table;
	
    if (WHOAMI_KEY == boot_loader_flag) {
        dbg_msg("\n=== From Bootloader ===\n");
    }
    else {
        dbg_msg("\n=== From JTag debug ===\n");
    }	
    
    rom_table = inw(CPU_ID_STS_ADDR);
    if (CPU_ID_SCPU == rom_table) {
        dbg_msg("=== [I am SCPU] firmware version: %x ===\n", g_fw_misc.version);
        cpu_type = 0;
    }
    else if (CPU_ID_NCPU == rom_table) {
        dbg_msg("=== [I am NCPU] firmware version: %x ===\n", g_fw_misc.version);
        cpu_type = 1;
    }  
    
    status = osKernelGetInfo(&osv, infobuf, sizeof(infobuf));
    if(status == osOK) {	
        dbg_msg("+-------------------Keil RTX5----------------------+\n");
        dbg_msg("Kernel Information: %s\r\n", infobuf);
        dbg_msg("Kernel Version    : %d\r\n", osv.kernel);
        dbg_msg("Kernel API Version: %d\r\n", osv.api);      
    }
}

/*----------------------------------------------------------------------------
 *      Main: Initialize and start RTX Kernel
 *---------------------------------------------------------------------------*/
int main(void)
{
#if 0
    /* Conflict with new UART driver on scpu at boot time
        - mat result in hard fault */
    unsigned int boot_loader_flag = 0x0;
    boot_loader_flag = bootloader_handler();
    info(boot_loader_flag);
#endif

    // System Initialization
    SystemCoreClockUpdate();

    osKernelInitialize();

    ncpu_init();
    mpu_config();

    if (osKernelGetState() == osKernelReady) {
        osKernelStart();                    // Start thread execution
    }

    while(1);
}

