/*
 * Kneron Example Post-Processing driver
 *
 * Copyright (C) 2018-2019 Kneron, Inc. All rights reserved.
 *
 */
 
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "base.h"
#include "kdpio.h"
#include "ipc.h"
#include "post_processing_ex.h"
#include "dbg.h"

#define YOLO_V3_O1_GRID_W       7
#define YOLO_V3_O1_GRID_H       7
#define YOLO_V3_O1_GRID_MAX     (YOLO_V3_O1_GRID_W * YOLO_V3_O1_GRID_H)
#define YOLO_V3_O2_GRID_W       14
#define YOLO_V3_O2_GRID_H       14
#define YOLO_V3_O2_GRID_MAX     (YOLO_V3_O2_GRID_W * YOLO_V3_O2_GRID_H)
#define YOLO_V3_CELL_BOX_NUM    3

#define YOLO_CELL_BOX_NUM       5
#define YOLO_CLASS_MAX          80
#define YOLO_GOOD_BOX_MAX       30
#define YOLO_BOX_FIX_CH         5   /* x, y, w, h, confidence score */

#define KDP_COL_MIN             16 /* Bytes, i.e. 128 bits */

const float prob_thresh_yolov3 = 0.5;     // probability threshold for yolo v3
const float nms_thresh_yolov3 = 0.45;     // non max suppression threshold for yolo v3
const float anchers_v0[3][2] = {{81,82}, {135,169}, {344,319}};
const float anchers_v1[3][2] = {{23,27}, {37,58}, {81,82}};

/* Shared global variable area among models */
struct yolo_v3_post_globals_s {
    float box_class_probs[YOLO_CLASS_MAX];
    struct bounding_box_s bboxes_v3[YOLO_V3_O2_GRID_MAX * YOLO_V3_CELL_BOX_NUM];
    struct bounding_box_s result_tmp_s[YOLO_GOOD_BOX_MAX];
};

union post_globals_u_s {
    struct yolo_v3_post_globals_s     yolov3;
} u_globals;

static float do_div_scale(float v, int div, float scale)
{
    return ((v / div) / scale);
}

static uint32_t round_up(uint32_t num){
    return ((num + (KDP_COL_MIN - 1)) & ~(KDP_COL_MIN - 1));
}

static float sigmoid(float x)
{
    float exp_value;
    float return_value;

    exp_value = expf(-x);

    return_value = 1 / (1 + exp_value);

    return return_value;
}

static int float_comparator(float a, float b)
{
    float diff = a - b;

    if (diff < 0)
        return 1;
    else if (diff > 0)
        return -1;
    return 0;
}

static int box_score_comparator(const void *pa, const void *pb)
{
    float a, b;

    a = ((struct bounding_box_s *)pa)->score;
    b = ((struct bounding_box_s *)pb)->score;

    return float_comparator(a, b);
}

static float overlap(float l1, float r1, float l2, float r2)
{
    float left = l1 > l2 ? l1 : l2;
    float right = r1 < r2 ? r1 : r2;
    return right - left;
}

static float box_intersection(struct bounding_box_s *a, struct bounding_box_s *b)
{
    float w, h, area;

    w = overlap(a->x1, a->x2, b->x1, b->x2);
    h = overlap(a->y1, a->y2, b->y1, b->y2);

    if (w < 0 || h < 0)
        return 0;

    area = w * h;
    return area;
}

static float box_union(struct bounding_box_s *a, struct bounding_box_s *b)
{
    float i, u;

    i = box_intersection(a, b);
    u = (a->y2 - a->y1) * (a->x2 - a->x1) + (b->y2 - b->y1) * (b->x2 - b->x1) - i;

    return u;
}

static float box_iou(struct bounding_box_s *a, struct bounding_box_s *b)
{
    float c;

    if (box_intersection(a, b) / box_intersection(a, a) > box_intersection(a, b) / box_intersection(b, b))
        c = box_intersection(a, b) / box_intersection(a, a);
    else
        c = box_intersection(a, b) / box_intersection(b, b);

    if (c < box_intersection(a, b) / box_union(a, b))
        c = box_intersection(a, b) / box_union(a, b);

    return c;
}

int post_yolo_v3(int model_id, struct kdp_image_s *image_p)
{
    struct yolo_v3_post_globals_s *gp = &u_globals.yolov3;
    int i, j, k, div, ch, row, col, max_score_class, good_box_count, class_good_box_count, good_result_count, len;
    float box_x, box_y, box_w, box_h, box_confidence, max_score;
    int8_t *src_p, *x_p, *y_p, *width_p, *height_p, *score_p, *class_p;
    struct bounding_box_s *bbox;
    struct yolo_result_s *result;
    struct bounding_box_s *result_box_p, *result_tmp_p, *r_tmp_p;

    int data_size, grid_w, grid_h, class_num, grid_w_bytes_aligned, w_bytes_to_skip;
    int incr_off;

    data_size = (POSTPROC_OUTPUT_FORMAT(image_p) & BIT(0)) + 1;     /* 1 or 2 in bytes */

    result = (struct yolo_result_s *)(POSTPROC_RESULT_MEM_ADDR(image_p));
    result_box_p = result->boxes;

    result_tmp_p = gp->result_tmp_s;

    class_num = POSTPROC_OUT_NODE_CH(image_p) / YOLO_V3_CELL_BOX_NUM - YOLO_BOX_FIX_CH;

    result->class_count = class_num;
    bbox = gp->bboxes_v3;
    good_box_count = 0;

    float anchers_v[3][2];
    int idx;
    int offset = sizeof(struct out_node_s);
    struct out_node_s *out_p;

    for (idx = 0; idx < POSTPROC_OUTPUT_NUM(image_p); idx++) {
        out_p = (struct out_node_s *)((uint32_t)POSTPROC_OUT_NODE(image_p) + idx * offset);

        src_p = (int8_t *)OUT_NODE_ADDR(out_p);
        grid_w = OUT_NODE_COL(out_p);
        grid_h = OUT_NODE_ROW(out_p);
        len = grid_w * data_size;
        grid_w_bytes_aligned = round_up(len);
        w_bytes_to_skip = grid_w_bytes_aligned - len;
        len = grid_w_bytes_aligned;

        x_p = src_p;
        y_p = x_p + len;
        width_p = y_p + len;
        height_p = width_p + len;
        score_p = height_p + len;
        class_p = score_p + len;

        div = 1 << OUT_NODE_RADIX(out_p);

        float fScale = *(float *)&OUT_NODE_SCALE(out_p);

        //For tiny yolo v3
        if (0 == idx) {
            memcpy(anchers_v, anchers_v0, 6*sizeof(float));
        } else {
            memcpy(anchers_v, anchers_v1, 6*sizeof(float));
        }

        for (row = 0; row < grid_h; row++) {
            for (ch = 0; ch < YOLO_V3_CELL_BOX_NUM; ch++) {
                for (col = 0; col < grid_w; col++) {
                    if (data_size == 1) {
                        box_x = (float)*x_p;
                        box_y = (float)*y_p;
                        box_w = (float)*width_p;
                        box_h = (float)*height_p;
                        box_confidence = (float)*score_p;
                    } else {
                        box_x = (float)*(uint16_t *)x_p;
                        box_y = (float)*(uint16_t *)y_p;
                        box_w = (float)*(uint16_t *)width_p;
                        box_h = (float)*(uint16_t *)height_p;
                        box_confidence = (float)*(uint16_t *)score_p;
                    }
					//fLib_printf("box %f %f %f %f %f\n", box_x, box_y, box_w, box_h, box_confidence);
                    box_confidence = sigmoid(do_div_scale(box_confidence, div, fScale));

                    /* Get scores of all class */
                    for (i = 0; i < class_num; i++) {
                        if (data_size == 1)
                            gp->box_class_probs[i] = (float)*(class_p + i * len);
                        else
                            gp->box_class_probs[i] = (float)*(uint16_t *)(class_p + i * len );

                        gp->box_class_probs[i] = sigmoid(do_div_scale(gp->box_class_probs[i], div, fScale));
                    }

                    //increase pointer to next position
                    x_p += data_size;
                    y_p += data_size;
                    width_p += data_size;
                    height_p += data_size;
                    score_p += data_size;
                    class_p += data_size;

                    /* Find highest score class */
                    max_score = 0;
                    max_score_class = -1;
                    for (i = 0; i < class_num; i++) {
                        gp->box_class_probs[i] = gp->box_class_probs[i] * box_confidence;
                        if (gp->box_class_probs[i] >= prob_thresh_yolov3) {
                            if (gp->box_class_probs[i] >= prob_thresh_yolov3 && gp->box_class_probs[i] > max_score) {
                                max_score = gp->box_class_probs[i];
                                max_score_class = i;
                            }
                        }
                    }

                    if (max_score_class != -1) {
                        box_x = do_div_scale(box_x, div, fScale);
                        box_y = do_div_scale(box_y, div, fScale);
                        box_w = do_div_scale(box_w, div, fScale);
                        box_h = do_div_scale(box_h, div, fScale);

                        box_x = (sigmoid(box_x) + col) / grid_w;
                        box_y = (sigmoid(box_y) + row) / grid_h;
                        box_w = expf(box_w) * anchers_v[ch][0] / DIM_INPUT_COL(image_p);
                        box_h = expf(box_h) * anchers_v[ch][1] / DIM_INPUT_ROW(image_p);

                        bbox->x1 = (box_x - (box_w / 2)) * DIM_INPUT_COL(image_p);
                        bbox->y1 = (box_y - (box_h / 2)) * DIM_INPUT_ROW(image_p);
                        bbox->x2 = (box_x + (box_w / 2)) * DIM_INPUT_COL(image_p);
                        bbox->y2 = (box_y + (box_h / 2)) * DIM_INPUT_ROW(image_p);
                        bbox->score = gp->box_class_probs[max_score_class];
                        bbox->class_num = max_score_class;

                        bbox++;
                        good_box_count++;
                    }
                }

                /* go to next 7*85 values for next anch*/
                incr_off = w_bytes_to_skip + grid_w_bytes_aligned*84; //84 = 85 -1
                x_p += incr_off;
                y_p += incr_off;
                width_p += incr_off;
                height_p += incr_off;
                score_p += incr_off;
                class_p += incr_off;
            }
        }

    }

    good_result_count = 0;

    for (i = 0; i < class_num; i++) {
        bbox = gp->bboxes_v3;
        class_good_box_count = 0;
        r_tmp_p = result_tmp_p;

        for (j = 0; j < good_box_count; j++) {
            if (bbox->class_num == i) {
                memcpy(r_tmp_p, bbox, sizeof(struct bounding_box_s));
                r_tmp_p++;
                class_good_box_count++;
            }
            bbox++;
        }

        if (class_good_box_count == 1) {
            memcpy(&result_box_p[good_result_count], &result_tmp_p[0], sizeof(struct bounding_box_s));
            good_result_count++;
        } else if (class_good_box_count >= 2) {
            qsort(result_tmp_p, class_good_box_count, sizeof(struct bounding_box_s), box_score_comparator);
            for (j = 0; j < class_good_box_count; j++) {
                if (result_tmp_p[j].score == 0)
                    continue;
                for (k = j + 1; k < class_good_box_count; k++) {
                    if (box_iou(&result_tmp_p[j], &result_tmp_p[k]) > nms_thresh_yolov3) {
                        result_tmp_p[k].score = 0;
                    }
                }
            }

            for (j = 0; j < class_good_box_count; j++) {
                if (result_tmp_p[j].score > 0) {
                    memcpy(&result_box_p[good_result_count], &result_tmp_p[j], sizeof(struct bounding_box_s));
                    good_result_count++;
                }
            }
        }
    }

    dbg_msg("good_result_count: %d\n", good_result_count);
    result->box_count = good_result_count;
    for (unsigned int i = 0; i < good_result_count; i++) {
        dbg_msg("post_yolo3 %f %f %f %f %f %d\n", result->boxes[i].x1, result->boxes[i].y1, result->boxes[i].x2, result->boxes[i].y2, result->boxes[i].score, result->boxes[i].class_num);
    }
    len = good_result_count * sizeof(struct bounding_box_s);

    return len;
}
