/*
 * @name : kl520_api_camera.c
 * @brief : camera interface
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */
#include "config/board_kdp520.h"
#include <string.h> 
#include <stdio.h> 
#include <stdlib.h>
#include <cmsis_os2.h>
#include "framework/v2k.h"
#include "framework/v2k_image.h"
#include "framework/event.h"
#include "kl520_api_camera.h"
#include "media/kdp_inference.h"
#include "dbg.h"

static int defaultcam = CAMERA_RGB_IDX;

/*-----------------------------------------------------------------------
 *                             Camera API
 *-----------------------------------------------------------------------*/
/**
* Initialize parameter of camera driver
*
* @param  cam_idx : camera index ref to defgroup@Camera_setting
*         *cam_ctx: ref to @struct camera_ctx
* 
* @return -1: Invalid camera index
*         -2: Camera is already initialized
*          0: Setting success
*/ 
int kl520_api_camera_setting(uint8_t cam_idx, unsigned int width, unsigned int height, unsigned int pixelformat)
{
    int ret = -1;
    if (cam_idx >= IMGSRC_NUM) {
        DSG("[%s] index %d invalid", __func__, cam_idx);
        return ret;
    } 
    ret = kdp_camera_setting(cam_idx, width, height, pixelformat);
    info_msg("[%s] m_camera_ctx[%d].width:%d, height:%d, ret:%d\n", __func__, cam_idx, width, height, ret);
    return 0;
}

/**
* @brief  Open the specific camera
*
* @param  cam_idx
* 
* @return 0: Succeed to open the specific camera
*        <0: Fail to open the specific camera
*/
int kl520_api_camera_open(uint8_t cam_idx)
{
    int ret = -1;
    struct kdp_camera_comm_params *cam_params = &m_kdp_all_comm_params.cam_params;
    osThreadId_t target_tid;

    if (cam_idx >= IMGSRC_NUM) {
        DSG("camera index %d invalid", cam_idx);
        return -1;    
    }

    if(NULL == (target_tid = kdp_get_inference_thread_id()))
        return -1;

    cam_params->caller_tid = osThreadGetId();
    cam_params->cam_idx = cam_idx;
    if(CAMERA_RGB_IDX == cam_params->cam_idx)
        set_thread_event(target_tid, FLAGS_KDP520_OPEN_CAMERA0_EVT);
    else
        set_thread_event(target_tid, FLAGS_KDP520_OPEN_CAMERA1_EVT);
    wait_thread_event(FLAGS_CAMERA_ACK);
    ret = cam_params->ret;
    info_msg("[%s %d] Get FLAGS_CAMERA_ACK, cam_idx:%d, ret:%d\n", __func__, __LINE__, cam_idx, ret);
    return ret;
}

/**
* @brief  Start to capture image stream
*
* @param  cam_idx
* 
* @return 0: Succeed to start capturing image stream on the specific camera
*        <0: Fail to start capturing image stream on the specific camera
*/
int kl520_api_camera_start(uint8_t cam_idx)
{
    int ret = -1;
    struct kdp_camera_comm_params *cam_params = &m_kdp_all_comm_params.cam_params;
    osThreadId_t target_tid;

    if(CAMERA_STATE_RUNNING == kdp_get_camera_ctx_state(cam_idx))
    {
       DSG("camera[%d] is running!", cam_idx); 
       return -1;
    }

    if(NULL == (target_tid = kdp_get_inference_thread_id()))
        return -1;

    cam_params->caller_tid = osThreadGetId();
    cam_params->cam_idx = cam_idx;
    if(CAMERA_RGB_IDX == cam_params->cam_idx)
        set_thread_event(target_tid, FLAGS_KDP520_START_CAMERA0_EVT);
    else
        set_thread_event(target_tid, FLAGS_KDP520_START_CAMERA1_EVT);
    wait_thread_event(FLAGS_CAMERA_ACK);
    ret = cam_params->ret;
    info_msg("[%s %d] Get FLAGS_CAMERA_ACK, cam_idx:%d, ret:%d\n", __func__, __LINE__, cam_idx, ret);
    return ret;
}

/**
* @brief  Close the specific camera
*
* @param  cam_idx
* 
* @return 0: Succeed to close the specific camera
*        <0: Fail to close the specific camera
*/
int kl520_api_camera_close(uint8_t cam_idx)
{
    int ret = -1;
    osThreadId_t target_tid;
    struct kdp_camera_comm_params *cam_params = &m_kdp_all_comm_params.cam_params;

    if(CAMERA_STATE_IDLE != kdp_get_camera_ctx_state(cam_idx) &&
       CAMERA_STATE_INITED != kdp_get_camera_ctx_state(cam_idx) )
    {
        critical_msg("Please stop camera[%d] before close it!\n", cam_idx); 
        return -1;
    }

    if(NULL == (target_tid = kdp_get_inference_thread_id()))
        return -1;

    cam_params->caller_tid= osThreadGetId();
    cam_params->cam_idx = cam_idx;
    if(CAMERA_RGB_IDX == cam_params->cam_idx)
        set_thread_event(target_tid, FLAGS_KDP520_CLOSE_CAMERA0_EVT);
    else
        set_thread_event(target_tid, FLAGS_KDP520_CLOSE_CAMERA1_EVT);
    wait_thread_event(FLAGS_CAMERA_ACK);
    ret = cam_params->ret;
    info_msg("[%s %d] Get FLAGS_CAMERA_ACK, cam_idx:%d, ret:%d\n", __func__, __LINE__, cam_idx, ret);

    return ret;
}

/**
* @brief  Stop to capture image stream
*
* @param  cam_idx
* 
* @return 0: Succeed to stop capturing image stream on the specific camera
*        <0: Fail to stop capturing image stream on the specific camera
*/
int kl520_api_camera_stop(uint8_t cam_idx)
{
    int ret = -1;
    osThreadId_t target_tid;    
    struct kdp_camera_comm_params *cam_params = &m_kdp_all_comm_params.cam_params;

    if(CAMERA_STATE_RUNNING != kdp_get_camera_ctx_state(cam_idx))
    {
       err_msg("camera[%d] is not running!", cam_idx); 
       return -1;
    }

    if(NULL == (target_tid = kdp_get_inference_thread_id()))
        return -1;

    cam_params->caller_tid= osThreadGetId();
    cam_params->cam_idx = cam_idx;
    if(CAMERA_RGB_IDX == cam_params->cam_idx)
        set_thread_event(target_tid, FLAGS_KDP520_STOP_CAMERA0_EVT);
    else
        set_thread_event(target_tid, FLAGS_KDP520_STOP_CAMERA1_EVT);
    wait_thread_event(FLAGS_CAMERA_ACK);
    ret = cam_params->ret;
    info_msg("[%s %d] Get FLAGS_CAMERA_ACK, cam_idx:%d, ret:%d\n", __func__, __LINE__, cam_idx, ret);
    return ret;
}

/**
* @brief  Check the status if the specific camera is running
*
* @param  cam_idx
* 
* @return TRUE: The specific camera is running
*         FALSE: The specific camera is NOT running
*/
bool kl520_api_camera_is_running(uint8_t cam_idx)
{
    return (CAMERA_STATE_RUNNING == kdp_get_camera_ctx_state(cam_idx));
}

/**
* @brief  Check the status if both cameras are running
*
* @param  N/A
* 
* @return TRUE: Both cameras are running
*         FALSE: Both cameras are NOT running
*/
bool kl520_api_all_cameras_are_running(void)
{
    return (kl520_api_camera_is_running(CAMERA_RGB_IDX)
            && kl520_api_camera_is_running(CAMERA_NIR_IDX));
}

/**
* @brief  Switch camera source to display
*
* @param  N/A
* 
* @return N/A
*/
extern void kl520_api_dp_close(void);
extern bool kl520_api_dp_open(uint8_t, unsigned int, unsigned int, unsigned int);
void kl520_api_camera_switch_rbg_nir(void)
{
    defaultcam = !defaultcam;
    if(CAMERA_RGB_IDX == defaultcam) {
        kl520_api_dp_close();
        kl520_api_dp_open(defaultcam, CAMERA_RGB_IMG_WIDTH, CAMERA_RGB_IMG_HEIGHT, V2K_PIX_FMT_RGB565);
    }
    else if(CAMERA_NIR_IDX == defaultcam) {
        kl520_api_dp_close();
        kl520_api_dp_open(defaultcam, CAMERA_NIR_IMG_WIDTH, CAMERA_NIR_IMG_HEIGHT, V2K_PIX_FMT_RAW8);
    }
}

