#include "config/board_kdp520.h"
#include <string.h> 
#include <stdio.h> 
#include <stdlib.h>
#include <cmsis_os2.h>
#include "framework/v2k.h"
#include "framework/v2k_image.h"
#include "framework/event.h"
#include "framework/result.h"
#include "kdp_memory.h"
#include "kdp_app.h"
#include "kdp_app_db.h"
#include "kdp_app_fid.h"
#include "dbg.h"
#include "media/kdp_camera.h"
#include "media/display/lcdc.h"
#include "kl520_api.h"
#include "kl520_api_fdfr.h"
#include "kl520_api_camera.h"
#include "media/kdp_inference.h"


static bool m_fdfr_is_inited = FALSE;

static osThreadId_t _kl520_api_face_preexecute(void)
{
    osThreadId_t target_tid = NULL;

    kl520_api_fdfr_model_init();

    if(FALSE == kl520_api_all_cameras_are_running()) {
        if(FALSE == kl520_api_camera_is_running(CAMERA_RGB_IDX)) {
            kl520_api_camera_open(CAMERA_RGB_IDX);
            kl520_api_camera_start(CAMERA_RGB_IDX);
        }
        kl520_api_dp_open(CAMERA_RGB_IDX, LCD_WIDTH, LCD_HEIGHT, V2K_PIX_FMT_RGB565);
        if(FALSE == kl520_api_camera_is_running(CAMERA_NIR_IDX)) {
            kl520_api_camera_open(CAMERA_NIR_IDX);
            kl520_api_camera_start(CAMERA_NIR_IDX);
        }
    }

    if ((m_fdfr_is_inited) && kl520_api_all_cameras_are_running()) {
        target_tid = kdp_get_inference_thread_id();
    }
    else {
        critical_msg("[%s] m_fdfr_is_inited:%d and both camera aren't opened", __func__, m_fdfr_is_inited);
    }
    return target_tid;    
}

int _kl520_api_face_register(int rgst_type, uint8_t user_idx, int rgst_idx, int num_rgst)
{
    int ret = -1;
    osThreadId_t target_tid;
    struct kdp_fdfr_comm_params *fdfr_params = &m_kdp_all_comm_params.fdfr_params;
    
    if(TRUE == kdp_fdfr_inference_is_running()) {
        err_msg(" Invalid command. Please Stop inference first\n");
        return ret;
    }

    if(NULL != (target_tid = _kl520_api_face_preexecute())) {
        fdfr_params->rgst_ctx.caller_tid = osThreadGetId();
        fdfr_params->rgst_ctx.num_rgst = num_rgst;
        fdfr_params->rgst_ctx.user_idx = user_idx;
        fdfr_params->rgst_ctx.rgst_idx = rgst_idx;
        fdfr_params->rgst_ctx.rgst_type = rgst_type;
        dbg_msg("[%s %d] rgst_type:%d, user_idx:%d, rgst_idx:%d, num_rgst:%d\n", __func__, __LINE__, rgst_type, user_idx, rgst_idx, num_rgst);
        set_thread_event(target_tid, rgst_type);
        wait_thread_event(FLAGS_FDFR_ACK);
        ret = fdfr_params->rgst_ctx.ret;
        dbg_msg("[%s %d] Get FLAGS_FDFR_ACK, user_idx:%d, ret:0x%x\n", __func__, __LINE__, user_idx, ret);
    }
    return ret;
}


int _kl520_api_5_face_register(uint8_t user_idx, int pose_type)
{
    int ret = -1;
    osThreadId_t target_tid;
    struct kdp_fdfr_comm_params *fdfr_params = &m_kdp_all_comm_params.fdfr_params;
    
    if(TRUE == kdp_fdfr_inference_is_running()) {
        err_msg(" Invalid command. Please Stop inference first\n");
        return ret;
    }

    if(NULL != (target_tid = _kl520_api_face_preexecute())) {
        fdfr_params->rgst_5_ctx.caller_tid = osThreadGetId();
        fdfr_params->rgst_5_ctx.user_idx = user_idx;
        fdfr_params->rgst_5_ctx.pose_type = pose_type;
        dbg_msg("[%s %d] user_idx:%d, pose_type:%d\n", __func__, __LINE__, user_idx, pose_type);
        set_thread_event(target_tid, FLAGS_KDP520_AUTO_REGISTER_5_FACE_EVT);
        wait_thread_event(FLAGS_FDFR_ACK);
        ret = fdfr_params->rgst_5_ctx.ret;
        dbg_msg("[%s %d] Get FLAGS_FDFR_ACK, user_idx:%d, pose_type:%d, ret:0x%x\n", __func__, __LINE__, user_idx, pose_type, ret);
    }
    return ret;
}

/*-----------------------------------------------------------------------
 *                          Face FDFR API 
 *-----------------------------------------------------------------------*/
/**
* @brief  Initialize FDFR models
*
* @param  N/A
* 
* @return TRUE:  Succeed to Initialize
*         FALSE: Fail to initialize
*/
bool kl520_api_fdfr_model_init(void)
{
    if (m_fdfr_is_inited)
        return TRUE;

    m_fdfr_is_inited = kdp_fdfr_model_init();
    return m_fdfr_is_inited;
}

/**
* @brief  Run FDFR recognition in loop until received stop FDFR inference loop command 
*         or call kl520_api_face_recognition_loop_stop()
*
* @param  N/A
* 
* @return 0: Get started to run FDFR recognition\n
*        -1: Both of RGB and NIR cameras aren't opened yet
*/
bool kl520_api_face_recognition_loop(void)
{
    osThreadId_t target_tid;    

    if(NULL != (target_tid = _kl520_api_face_preexecute())) {
        kdp_fdfr_inference_start();
        set_thread_event(target_tid, FLAGS_KDP520_START_INFERENCE_LOOP_EVT);
        return TRUE;
    }
    else {
        return FALSE;
    }
}

/**
* @brief  Stop loop running FDFR recognition
*
* @param  N/A
* 
* @return 0: Stop running FDFR recognition
*        -1: Both of RGB and NIR cameras aren't opened yet
*/
bool kl520_api_face_recognition_loop_stop(void)
{
    osThreadId_t target_tid;    

    if(NULL != (target_tid = _kl520_api_face_preexecute())) {
        kdp_fdfr_inference_stop();
        set_thread_event(target_tid, FLAGS_KDP520_STOP_INFERENCE_LOOP_EVT);
        return TRUE;
    }
    else {
        return FALSE;
    }
}

/**
* @brief  Run FDFR recognition in loop until FDFR result compare DB ok
*
* @param  N/A
* 
* @return 0: Get started to run FDFR recognition
*        -1: Both of RGB and NIR cameras aren't opened yet
*/
int kl520_api_face_recognition(void)
{
    int ret = -1;
    osThreadId_t target_tid;    
    struct kdp_fdfr_comm_params *fdfr_params = &m_kdp_all_comm_params.fdfr_params;

    if(NULL != (target_tid = _kl520_api_face_preexecute())) {
        kl520_api_timer_init(PWMTIMER3, PWMTMR_1000MSEC_PERIOD);
        m_fdfr_tick = kdp_current_t3_tick();
        kdp_fdfr_inference_start();
        fdfr_params->fdfr_ctx.caller_tid = osThreadGetId();
        set_thread_event(target_tid, FLAGS_KDP520_START_INFERENCE_EVT);
        wait_thread_event(FLAGS_FDFR_ACK);
        kdp_fdfr_inference_stop();
        ret = fdfr_params->fdfr_ctx.ret;
        dbg_msg("[%s %d] Get FLAGS_FDFR_ACK, ret:0x%x\n", __func__, __LINE__, ret);
    }
    return ret;
}

/**
* @brief  Register user face into DDR DB
*
* @param  user_idx: index of user to be registered into DDR DB
* 
* @return 0: Stop running FDFR recognition
*        -1: Both of RGB and NIR cameras aren't opened yet
*        -2: kl520 is running FDFR inference
*         1: No space of database
*        >1: enum kdp_app_fid_status_code_e
*/
int kl520_api_face_register(uint8_t user_idx)
{                     
    return _kl520_api_face_register(FLAGS_KDP520_REGISTER_FDR_EVT, user_idx, -1, 1/*num_reg*/);
}

/**
* @brief  Register user face into DDR DB
*
* @param  user_idx: index of user to be registered into DDR DB
* 
* @return 0: Stop running FDFR recognition
*        -1: Both of RGB and NIR cameras aren't opened yet
*        -2: kl520 is running FDFR inference
*         1: No space of database
*        >1: enum kdp_app_fid_status_code_e
*/
int kl520_api_face_auto_register(uint8_t user_idx, int reg_index)
{
   return _kl520_api_face_register(FLAGS_KDP520_AUTO_REGISTER_FDR_EVT, user_idx, reg_index, 1/*num_reg*/);
}

/**
* @brief  Register user face into DDR DB
*
* @param  user_idx: index of user to be registered into DDR DB
* 
* @return 0: Stop running FDFR recognition
*        -1: Both of RGB and NIR cameras aren't opened yet
*        -2: kl520 is running FDFR inference
*         1: No space of database
*        >1: enum kdp_app_fid_status_code_e
*/
int kl520_api_face_register_cfm(uint8_t user_idx)
{
    return _kl520_api_face_register(FLAGS_KDP520_REGISTER_CFM_FDR_EVT, user_idx, -1, 1/*num_reg*/);
}


int kl520_api_face_auto_register_both(uint8_t user_idx, int reg_index)
{
    return _kl520_api_face_register(FLAGS_KDP520_AUTO_REGISTER_BOTH_FDR_EVT, user_idx, reg_index, 2/*num_reg*/);
}

int kl520_api_face_register_cfm_both(uint8_t user_idx)
{
    return _kl520_api_face_register(FLAGS_KDP520_REGISTER_CFM_BOTH_FDR_EVT, user_idx, -1, 2/*num_reg*/);
}

int kl520_api_face_auto_register_5_face(uint8_t user_idx, int reg_index)
{
    return _kl520_api_5_face_register(user_idx, reg_index);
}


/**
* @brief  Clear the register user face record in DDR DB
*
* @param  N/A
* 
* @return 0: Succeed to cancel register\n
*        -1: Fail to cancel register
*/
int kl520_api_face_cancel_register(void)
{
    int ret = -1;
    osThreadId_t target_tid;
    struct kdp_fdfr_comm_params *fdfr_params = &m_kdp_all_comm_params.fdfr_params;

    if(NULL != (target_tid = _kl520_api_face_preexecute())) {
        fdfr_params->rgst_cancel_ctx.caller_tid = osThreadGetId();
        set_thread_event(target_tid, FLAGS_KDP520_CANCEL_REGISTER_FDR_EVT);
        wait_thread_event(FLAGS_FDFR_ACK);
        ret = fdfr_params->rgst_cancel_ctx.ret;
        dbg_msg("[%s %d] Get FLAGS_FDFR_ACK, ret:0x%x\n", __func__, __LINE__, ret);
    }
    return ret;
}

/**
* @brief  Add user face into Flash DB
*
* @param  user_idx: index of user to be added into Flash DB
* 
* @return 0: Stop running FDFR recognition
*        -1: Both of RGB and NIR cameras aren't opened yet
*         1: No space of database
*        >1: enum kdp_app_fid_status_code_e
*/
int kl520_api_face_add(uint8_t user_idx)
{
    int ret = -1;
    osThreadId_t target_tid;
    struct kdp_fdfr_comm_params *fdfr_params = &m_kdp_all_comm_params.fdfr_params;

    if(NULL != (target_tid = _kl520_api_face_preexecute())) {
        fdfr_params->add_ctx.caller_tid = osThreadGetId();
        fdfr_params->add_ctx.user_idx = user_idx;
        set_thread_event(target_tid, FLAGS_KDP520_ADD_FDR_FACE_EVT);
        wait_thread_event(FLAGS_FDFR_ACK);
        ret = fdfr_params->add_ctx.ret;
        dbg_msg("[%s %d] Get FLAGS_FDFR_ACK, user_idx:%d, ret:0x%x", __func__, __LINE__, user_idx, ret);
    }
   
    return ret;
}

/**
* @brief  Delete user face into Flash DB
*
* @param  user_idx: index of user to be delete from Flash DB
* 
* @return 0: Stop running FDFR recognition
*        -1: Both of RGB and NIR cameras aren't opened yet
*         1: No space of database
*        >1: enum kdp_app_fid_status_code_e
*/
int kl520_api_face_del(bool all, u8 face_id)
{
    int ret = -1;  
    osThreadId_t target_tid;
    struct kdp_fdfr_comm_params *fdfr_params = &m_kdp_all_comm_params.fdfr_params;

    if((target_tid =_kl520_api_face_preexecute()) != NULL) {
        fdfr_params->del_ctx.caller_tid = osThreadGetId();
        fdfr_params->del_ctx.user_idx = face_id;
        fdfr_params->del_ctx.all = all;

        dbg_msg("[%s %d] Get FLAGS_FDFR_ACK, all:%d, face_id:%d\n", __func__, __LINE__, all, face_id);
        set_thread_event(target_tid, FLAGS_KDP520_DEL_FDR_FACE_EVT);
        wait_thread_event(FLAGS_FDFR_ACK);
        ret = fdfr_params->del_ctx.ret;
        dbg_msg("[%s %d] Get FLAGS_FDFR_ACK, all:%d, face_id:%d, ret:0x%x\n", __func__, __LINE__, all, face_id, ret);
    }
    return ret;
}

/**
* @brief  Stop/Close both cameras and display preview
*
* @param  N/A
* 
* @return N/A
*/
void kl520_api_face_close(void)
{    
    kdp_fdfr_recognition_variable_reset();
    kl520_api_camera_stop(CAMERA_RGB_IDX);
    kl520_api_camera_close(CAMERA_RGB_IDX);
    kl520_api_camera_stop(CAMERA_NIR_IDX);
    kl520_api_camera_close(CAMERA_NIR_IDX);
    dbg_msg("kl520_api_face_close done");

    kl520_api_dp_close();

    kdp_fb_mgr_free_write_buf(CAMERA_NIR_IDX);
    kdp_fb_mgr_free_write_buf(CAMERA_RGB_IDX);
    kdp_fb_mgr_free_inf_buf(CAMERA_NIR_IDX);
    kdp_fb_mgr_free_inf_buf(CAMERA_RGB_IDX);
	m_fdfr_is_inited = FALSE;
}

/**
* @brief  Use to set timeout of fdfr inference
*
* @param  timeout: second unit
* 
* @return N/A
*/
void kl520_api_face_recognition_set_timeout(int timeout)
{
    m_face_recognition_timeout = timeout;
}

bool kl520_api_timer_init(pwmtimer timer, u32 tick)
{
    return kdp_timer_init(timer, tick);
}

void kl520_api_timer_tick_reset(pwmtimer timer)
{
    kdp_timer_tick_reset(timer);
}

int kl520_api_timer_close(pwmtimer timer)
{
    return kdp_timer_close(timer);
}


