#ifdef V2K_ENABLE
/*
 * @name : kdp520_i2c.c
 * @brief : i2c driver for kdp520 i2c controller
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include "config/board_kdp520.h"
#include "kdp520_i2c.h"
#include <string.h> 
#include <stdlib.h>
#include "kneron_mozart.h"
#include "types.h"
#include "base.h"
#include "io.h"
#include "framework/ioport.h"
#include "framework/framework_driver.h"
#include "framework/framework_errno.h"
#include "media/camera/sensor.h"
#include "dbg.h"


#define I2C_GSR_Value       5
#define I2C_TSR_Value       5
#define I2C_MAX_FREQ        (400 * 1000)//(400 * 1024)

#define I2C_RD(dev)         ((((dev) << 1) & 0xfe) | 1)
#define I2C_WR(dev)         (((dev) << 1) & 0xfe)


/* I2C Register */
#define I2C_REG_CONTROL     0x0
#define I2C_REG_STATUS      0x4
#define I2C_REG_CLOCKDIV    0x8
#define I2C_REG_DATA        0xC
#define I2C_REG_TGSR        0x14

/* I2C control register */
#define I2C_CTRL_ALIRQ      0x2000  /* arbitration lost interrupt (master) */
#define I2C_CTRL_NAKRIRQ    0x400   /* NACK response interrupt (master) */
#define I2C_CTRL_DRIRQ      0x200   /* rx interrupt (both) */
#define I2C_CTRL_DTIRQ      0x100   /* tx interrupt (both) */
#define I2C_CTRL_TBEN       0x80    /* tx enable (both) */
#define I2C_CTRL_NAK        0x40    /* NACK (both) */
#define I2C_CTRL_STOP       0x20    /* stop (master) */
#define I2C_CTRL_START      0x10    /* start (master) */
#define I2C_CTRL_SCLEN      0x4     /* enable clock out (master) */
#define I2C_CTRL_I2CEN      0x2     /* enable I2C (both) */
#define I2C_CTRL_ENABLE     \
    (I2C_CTRL_ALIRQ | I2C_CTRL_NAKRIRQ | I2C_CTRL_DRIRQ | I2C_CTRL_DTIRQ | I2C_CTRL_SCLEN | I2C_CTRL_I2CEN)

/* I2C Status register */
#define I2C_STATUS_TD       0x20    /* tx/rx finish */
#define I2C_STATUS_BUSY     0x4     /* chip busy */	

struct kdp520_i2c_context {
    u32 base;
    int retries;
    struct kdp520_i2c_params params;
    int (*master_xfer)(struct kdp520_i2c_context *i2c_ctx, struct i2c_msg *msgs, int num);    
};

u32 _kdp520_i2c_getgsr(struct kdp520_i2c_context *i2c_ctx)
{
    u32 tmp;

    tmp = inw(((u8*)i2c_ctx->base + I2C_REG_TGSR));

    return ((tmp >> 10) & 0x7);
}

static BOOL _kdp_i2c_check_busy(u32 base) {
    if((inw(base + I2C_REG_STATUS) & I2C_STATUS_BUSY)== I2C_STATUS_BUSY)
     return TRUE;

    return FALSE;
}

static int _kdp_i2c_wait(int base, uint32_t mask)
{
    int ret = -KDP_FRAMEWORK_ERRNO_IO;
    uint32_t stat, ts;

    //for (ts = 0; ts < 55000; ts++) {
    for (ts = 0; ts < 100000; ts++) {
        stat = inw(base + I2C_REG_STATUS);
        outw(base + I2C_REG_STATUS, stat);
        if ((stat & mask) == mask) {
            ret = 0;//it means ok
            break;
        }
    }

    return ret;
}

static int _kdp520_i2c_xfer_msg(
        struct kdp520_i2c_context *i2c_ctx, struct i2c_msg *msg, BOOL is_stop)
{
    int ret, i;

    u32 base = i2c_ctx->base;    
    u16 slave_addr = msg->addr;    
    u16 flags = msg->flags;
    u16 len = msg->len;
    u8 *buf = msg->buf;
    
    BOOL is_read = (flags & I2C_MASTER_READ);

    //DSG("kdp520_i2c_xfer_msg base=%x slave_addr=%x len=%d is_stop=%d", base, slave_addr, len, is_stop);
    if (is_read)
        outw(base + I2C_REG_DATA, I2C_RD(slave_addr));
    else
        outw(base + I2C_REG_DATA, I2C_WR(slave_addr));

    outw(base + I2C_REG_CONTROL, I2C_CTRL_ENABLE | I2C_CTRL_TBEN | I2C_CTRL_START);
    ret = _kdp_i2c_wait(base, I2C_STATUS_TD);
    if (0 == ret)
    {
        u32 ctrl = I2C_CTRL_ENABLE | I2C_CTRL_TBEN;
        for (i = 0; i < len; ++i) {
            if ((is_stop) && (i == len - 1)) {
                if (is_read)
                    ctrl |= I2C_CTRL_NAK | I2C_CTRL_STOP;
                else
                    ctrl |= I2C_CTRL_STOP;
            }
                
            if (!is_read)
                outw(base + I2C_REG_DATA, buf[i]);
            
            outw(base + I2C_REG_CONTROL, ctrl);
            
            ret = _kdp_i2c_wait(base, I2C_STATUS_TD);
            if (ret)
                break;
            
            if (is_read)
                buf[i] = (u8)(inw(base + I2C_REG_DATA) & 0xFF);
        }
    }

    //kdp520_i2c_reset(i2c_ctx);

    return ret;
}

static int _kdp520_i2c_xfer(
        struct kdp520_i2c_context *i2c_ctx, struct i2c_msg *msgs, int num)
{
    int ret = 0;
    int i;

    //enable i2c clock

    for (i = 0; i < num; ++i) {
        ret = _kdp520_i2c_xfer_msg(i2c_ctx, &msgs[i], (i == (num - 1)));
        if (0 != ret)
            break;
    }

    while (_kdp_i2c_check_busy(i2c_ctx->base));

    //disable i2c clock

    return ret;
}

static int _i2c_transfer(struct kdp520_i2c_context *i2c_ctx, struct i2c_msg *msgs, int num)
{
    int ret = -KDP_FRAMEWORK_ERRNO_IO;
    int try;

    for (ret = 0, try = 0; try <= i2c_ctx->retries; try++) {
        ret = i2c_ctx->master_xfer(i2c_ctx, msgs, num);
        if (ret != -KDP_FRAMEWORK_ERRNO_AGAIN)
            break;
    }
    
    return ret;
}

static int _kdp520_i2c_probe(struct core_device *core_d)
{
    struct kdp520_i2c_context *i2c_ctx;
        
    i2c_ctx = (struct kdp520_i2c_context*)calloc(1, sizeof(*i2c_ctx));
    if (!i2c_ctx)
        return -KDP_FRAMEWORK_ERRNO_NOMEM;

    i2c_ctx->base = driver_get_ioport_setting(core_d, IORESOURCE_MEM)->start;
    i2c_ctx->retries = 5;
    i2c_ctx->master_xfer = _kdp520_i2c_xfer;

    framework_drv_set_drvdata(&core_d->pin_ctx, (void*)i2c_ctx);    
    
    return 0;
}

static int _kdp520_i2c_remove(struct core_device *core_d)
{
    return 0;
}

static int _kdp520_i2c_set_params(struct core_device *core_d, struct kdp520_i2c_params *params)
{
    struct kdp520_i2c_context *ctx = framework_drv_get_drvdata(&core_d->pin_ctx);
        
    memcpy((void*)(&ctx->params), params, sizeof(ctx->params));
    ctx->params.public_mode = TRUE;
    
    return 0;
}

static int _kdp520_i2c_init(struct core_device *core_d)
{
    struct kdp520_i2c_context *ctx = framework_drv_get_drvdata(&core_d->pin_ctx);
    int bus_speed = 0;
    
    outw(((u8*)ctx->base + I2C_REG_TGSR), (I2C_GSR_Value << 10)|I2C_TSR_Value);

    DSG("ctx->params.public_mode=%d",ctx->params.public_mode);
    DSG("ctx->params.bus_speed=%d",ctx->params.bus_speed);
    if (ctx->params.public_mode)
        bus_speed = ctx->params.bus_speed;
    
    if (bus_speed > 0) {
        int bits = ((APB_CLOCK - (bus_speed * _kdp520_i2c_getgsr(ctx))) / ( 2 * bus_speed)) - 2;
        if (bits < BIT10) 
            outw(((u8*)ctx->base + I2C_REG_CLOCKDIV), bits);
    }
    
    return 0;
}

static int _kdp520_i2c_reset(struct core_device *core_d)
{
    struct kdp520_i2c_context *ctx = framework_drv_get_drvdata(&core_d->pin_ctx);
    
    u32 ts;
    outw((u8*)ctx->base + I2C_REG_CONTROL, 1);
    for (ts = 0; ts < 50000; ts++) {
        if (!(inw(((u8*)ctx->base + I2C_REG_CONTROL)) & 0x1)) {
            break;
        }
    }
    
    return 0;    
}

static int _kdp520_i2c_write(
        struct core_device *core_d, u8 slave_addr, u16 reg, u16 reg_len, u8 data)
{
    int ret;
    struct kdp520_i2c_context *ctx = framework_drv_get_drvdata(&core_d->pin_ctx);
    
    u8 paddr[3];
    if (2 == reg_len) {
        paddr[0] = ((reg >> 8) & 0xff);
        paddr[1] = reg & 0XFF;
        paddr[2] = data;
    }
    else {
        paddr[0] = (reg & 0XFF);
        paddr[1] = data;        
    }
    
    struct i2c_msg msgs[] = {
        {
            .addr = slave_addr,
            .flags = I2C_MASTER_WRITE,
            .len = reg_len + 1,
            .buf = paddr,
        }
    };
    
    ret = _i2c_transfer(ctx, msgs, 1);

    return ret;
}

static int _kdp520_i2c_read(
        struct core_device *core_d, u8 slave_addr, u16 reg, u16 reg_len, u8 *data)
{
	int ret;
    struct kdp520_i2c_context *ctx = framework_drv_get_drvdata(&core_d->pin_ctx);
    
    u8 paddr[2];
    if (2 == reg_len) {
        paddr[0] = ((reg >> 8) & 0xff);
        paddr[1] = reg & 0XFF;
    }
    else {
        paddr[0] = (reg & 0XFF);
    }

	struct i2c_msg msgs[] = {
		{
			.addr = slave_addr,
			.flags = I2C_MASTER_WRITE,
			.len = reg_len,
			.buf = paddr,
		},
		{
			.addr = slave_addr,
			.flags = I2C_MASTER_READ,
			.len = 1,
			.buf = data,
		},
	};

	ret = _i2c_transfer(ctx, msgs, 2);

	return ret;
}
static int _kdp520_i2c_read_bytes(
        struct core_device *core_d, u8 slave_addr, u16 reg, u16 reg_len, u8 *data, u8 data_len)
{
	int ret;
    struct kdp520_i2c_context *ctx = framework_drv_get_drvdata(&core_d->pin_ctx);

    u8 paddr[2];
    if (2 == reg_len) {
        paddr[0] = ((reg >> 8) & 0xff);
        paddr[1] = reg & 0XFF;
    }
    else {
        paddr[0] = (reg & 0XFF);
    }

	struct i2c_msg msgs[] = {
		{
			.addr = slave_addr,
			.flags = I2C_MASTER_WRITE,
			.len = reg_len,
			.buf = paddr,
		},
		{
			.addr = slave_addr,
			.flags = I2C_MASTER_READ,
			.len = data_len,
			.buf = data,
		},
	};

	ret = _i2c_transfer(ctx, msgs, 2);

	return ret;
}

#if I2C_0_ENABLE == YES
extern struct core_device kdp520_i2c_0;
struct core_i2c_driver kdp520_i2c_0_driver = {
    .driver         = {
        .name       = "kdp520_i2c_0",
    },
    .probe          = _kdp520_i2c_probe,
    .remove         = _kdp520_i2c_remove,
    .core_dev       = &kdp520_i2c_0,
    
    .power_mgr      = {
        .power      = NULL,
        .suspend    = NULL,
        .resume     = NULL,
    }, 

    //.cfg_func     = (void*)&_i2c_set_params,
    .set_params     = _kdp520_i2c_set_params,    
    .init           = _kdp520_i2c_init,
    .reset          = _kdp520_i2c_reset,
    .write          = _kdp520_i2c_write,
    .read           = _kdp520_i2c_read,
	.readbytes		= _kdp520_i2c_read_bytes,	    
    .inited         = FALSE,
    .params         = { .bus_speed = 50000 },
};
#endif

#if I2C_1_ENABLE == YES
extern struct core_device kdp520_i2c_1;
struct core_i2c_driver kdp520_i2c_1_driver = {
    .driver         = {
        .name       = "kdp520_i2c_1",
    },
    .probe          = _kdp520_i2c_probe,
    .remove         = _kdp520_i2c_remove,
    .core_dev       = &kdp520_i2c_1,
    
    .power_mgr      = {
        .power      = NULL,
        .suspend    = NULL,
        .resume     = NULL,
    }, 

    //.cfg_func     = (void*)&_i2c_set_params,
    .set_params     = _kdp520_i2c_set_params,    
    .init           = _kdp520_i2c_init,
    .reset          = _kdp520_i2c_reset,
    .write          = _kdp520_i2c_write,
    .read           = _kdp520_i2c_read,
	.readbytes		= _kdp520_i2c_read_bytes,
    .inited         = FALSE,
    .params         = { .bus_speed = 50000 },
};
#endif
#if I2C_2_ENABLE == YES
extern struct core_device kdp520_i2c_2;
struct core_i2c_driver kdp520_i2c_2_driver = {
    .driver         = {
        .name       = "kdp520_i2c_2",
    },
    .probe          = _kdp520_i2c_probe,
    .remove         = _kdp520_i2c_remove,
    .core_dev       = &kdp520_i2c_2,
    
    .power_mgr      = {
        .power      = NULL,
        .suspend    = NULL,
        .resume     = NULL,
    }, 

    //.cfg_func     = (void*)&_i2c_set_params,
    .set_params     = _kdp520_i2c_set_params,    
    .init           = _kdp520_i2c_init,
    .reset          = _kdp520_i2c_reset,
    .write          = _kdp520_i2c_write,
    .read           = _kdp520_i2c_read,
	.readbytes		= _kdp520_i2c_read_bytes,
    .inited         = FALSE,
    .params         = { .bus_speed = 50000 },
};
#endif
#if I2C_3_ENABLE == YES
extern struct core_device kdp520_i2c_3;
struct core_i2c_driver kdp520_i2c_3_driver = {
    .driver         = {
        .name       = "kdp520_i2c_3",
    },
    .probe          = _kdp520_i2c_probe,
    .remove         = _kdp520_i2c_remove,
    .core_dev       = &kdp520_i2c_3,
    
    .power_mgr      = {
        .power      = NULL,
        .suspend    = NULL,
        .resume     = NULL,
    }, 

    .set_params     = _kdp520_i2c_set_params,    
    .init           = _kdp520_i2c_init,
    .reset          = _kdp520_i2c_reset,
    .write          = _kdp520_i2c_write,
    .read           = _kdp520_i2c_read,
	.readbytes		= _kdp520_i2c_read_bytes,	
    .inited         = FALSE,
    .params         = { .bus_speed = 50000 },
};
#endif

static inline void _get_i2c_dev_drv(
        enum i2c_adap_id id, struct core_device **lpdev, struct core_i2c_driver **lpdrv)
{
    switch (id) {
#if I2C_0_ENABLE == YES        
    case I2C_ADAP_0:
        *lpdev = &kdp520_i2c_0;
        *lpdrv = &kdp520_i2c_0_driver;
        break;
#endif
#if I2C_1_ENABLE == YES    
    case I2C_ADAP_1:
        *lpdev = &kdp520_i2c_1;
        *lpdrv = &kdp520_i2c_1_driver;        
        break;
#endif
#if I2C_2_ENABLE == YES    
    case I2C_ADAP_2:
        *lpdev = &kdp520_i2c_2;
        *lpdrv = &kdp520_i2c_2_driver;        
        break;
#endif
#if I2C_3_ENABLE == YES    
    case I2C_ADAP_3:
        *lpdev = &kdp520_i2c_3;
        *lpdrv = &kdp520_i2c_3_driver;        
        break;
#endif    
    default:;
    }    
}

int kdp_drv_i2c_init(enum i2c_adap_id id, unsigned long bus_speed, BOOL force)
{
    int ret = -1;
    struct core_device *dev = NULL;
    struct core_i2c_driver *drv = NULL;
    _get_i2c_dev_drv(id, &dev, &drv);
    
    if (dev && drv && ((force) || (!drv->inited)))
    {
        struct kdp520_i2c_params params;
        params.bus_speed = bus_speed;    
        drv->probe(dev);
        drv->set_params(dev, &params);
        drv->init(dev);
        drv->reset(dev); 
        drv->inited = TRUE;
        
        ret = 0;
    }
    
    return ret;
}

int kdp_drv_i2c_write(enum i2c_adap_id id, u8 subaddr, u16 reg, u8 reg_len, u8 data)
{
    int ret = -1;
    
    struct core_device *dev = NULL;
    struct core_i2c_driver *drv = NULL;
    _get_i2c_dev_drv(id, &dev, &drv);
    
    ret = drv->write(dev, subaddr, reg, reg_len, data);    
    
    return ret;
}

int kdp_drv_i2c_read(enum i2c_adap_id id, u8 subaddr, u16 reg, u8 reg_len, u8 *lpdata)
{
    int ret = -1;
    
    struct core_device *dev = NULL;
    struct core_i2c_driver *drv = NULL;
    _get_i2c_dev_drv(id, &dev, &drv);
    
    ret = drv->read(dev, subaddr, reg, reg_len, lpdata);    
    
    return ret;
}

int kdp_drv_i2c_read_bytes(enum i2c_adap_id id, u8 subaddr, u16 reg, u8 reg_len, u8 *lpdata, u8 data_len)
{
    int ret = -1;
    
    struct core_device *dev = NULL;
    struct core_i2c_driver *drv = NULL;
    _get_i2c_dev_drv(id, &dev, &drv);
    
    ret = drv->readbytes(dev, subaddr, reg, reg_len, lpdata, data_len);    
    
    return ret;
}

int kdp_drv_init(void)
{
#if I2C_0_ENABLE == YES
    kdp_drv_i2c_init(I2C_ADAP_0, 400000, FALSE);
#endif    
#if I2C_1_ENABLE == YES
    kdp_drv_i2c_init(I2C_ADAP_1, 400000, FALSE);
#endif
#if I2C_2_ENABLE == YES    
    kdp_drv_i2c_init(I2C_ADAP_2, 400000, FALSE);
#endif
#if I2C_3_ENABLE == YES     
    kdp_drv_i2c_init(I2C_ADAP_3, 400000, FALSE);
#endif
    
    return 0;
}

#endif // V2K_ENABLE
