/*
 *  Video for Linux Two header file
 *
 *  Copyright (C) 1999-2012 the contributors
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  Alternatively you can redistribute this file under the terms of the
 *  BSD license as stated below:
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *  3. The names of its contributors may not be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	Header file for  and applications
 *  with public API.
 * All kernel-specific stuff were moved to media/K_UVC-dev.h, so
 *
 *
 *	See https://linuxtv.org for more info
 *
 *	Author:  <bill@thedirks.org>
 *
 *
 *
 */
#ifndef _KAPI_UVC_H
#define _KAPI_UVC_H

#include <sys/time.h> 
#include <stdint.h>
#include <ioctl.h>


#define VIDEO_MAX_FRAME               32
#define VIDEO_MAX_PLANES               8


//struct k_uvc_queryctrl;
/*
 *	M I S C E L L A N E O U S
 */

/*  Four-character-code (FOURCC) */
#define k_uvc_fourcc(a, b, c, d)\
	((uint32_t)(a) | ((uint32_t)(b) << 8) | ((uint32_t)(c) << 16) | ((uint32_t)(d) << 24))
#define k_uvc_fourcc_be(a, b, c, d)	(k_uvc_fourcc(a, b, c, d) | (1 << 31))

/*
 *	E N U M S
 */
enum k_uvc_field {
    K_UVC_FIELD_ANY           = 0, /* driver can choose from none,
					 top, bottom, interlaced
					 depending on whatever it thinks
					 is approximate ... */
    K_UVC_FIELD_NONE          = 1, /* this device has no fields ... */
    K_UVC_FIELD_TOP           = 2, /* top field only */
    K_UVC_FIELD_BOTTOM        = 3, /* bottom field only */
    K_UVC_FIELD_INTERLACED    = 4, /* both fields interlaced */
    K_UVC_FIELD_SEQ_TB        = 5, /* both fields sequential into one
					 buffer, top-bottom order */
    K_UVC_FIELD_SEQ_BT        = 6, /* same as above + bottom-top order */
    K_UVC_FIELD_ALTERNATE     = 7, /* both fields alternating into
					 separate buffers */
    K_UVC_FIELD_INTERLACED_TB = 8, /* both fields interlaced, top field
					 first and the top field is
					 transmitted first */
    K_UVC_FIELD_INTERLACED_BT = 9, /* both fields interlaced, top field
					 first and the bottom field is
					 transmitted first */
};

#define K_UVC_FIELD_HAS_TOP(field)	\
    ((field) == K_UVC_FIELD_TOP 	||\
     (field) == K_UVC_FIELD_INTERLACED ||\
     (field) == K_UVC_FIELD_INTERLACED_TB ||\
     (field) == K_UVC_FIELD_INTERLACED_BT ||\
     (field) == K_UVC_FIELD_SEQ_TB	||\
     (field) == K_UVC_FIELD_SEQ_BT)
#define K_UVC_FIELD_HAS_BOTTOM(field)	\
    ((field) == K_UVC_FIELD_BOTTOM 	||\
     (field) == K_UVC_FIELD_INTERLACED ||\
     (field) == K_UVC_FIELD_INTERLACED_TB ||\
     (field) == K_UVC_FIELD_INTERLACED_BT ||\
     (field) == K_UVC_FIELD_SEQ_TB	||\
     (field) == K_UVC_FIELD_SEQ_BT)
#define K_UVC_FIELD_HAS_BOTH(field)	\
    ((field) == K_UVC_FIELD_INTERLACED ||\
     (field) == K_UVC_FIELD_INTERLACED_TB ||\
     (field) == K_UVC_FIELD_INTERLACED_BT ||\
     (field) == K_UVC_FIELD_SEQ_TB ||\
     (field) == K_UVC_FIELD_SEQ_BT)
#define K_UVC_FIELD_HAS_T_OR_B(field)	\
    ((field) == K_UVC_FIELD_BOTTOM ||\
     (field) == K_UVC_FIELD_TOP ||\
     (field) == K_UVC_FIELD_ALTERNATE)

enum k_uvc_buf_type {
    K_UVC_BUF_TYPE_VIDEO_CAPTURE        = 1,
    K_UVC_BUF_TYPE_VIDEO_OUTPUT         = 2,
    K_UVC_BUF_TYPE_VIDEO_OVERLAY        = 3,
    K_UVC_BUF_TYPE_VIDEO_OUTPUT_OVERLAY = 8,
    K_UVC_BUF_TYPE_VIDEO_CAPTURE_MPLANE = 9,
    K_UVC_BUF_TYPE_VIDEO_OUTPUT_MPLANE  = 10,
};

#define K_UVC_TYPE_IS_MULTIPLANAR(type)			\
    ((type) == K_UVC_BUF_TYPE_VIDEO_CAPTURE_MPLANE	\
     || (type) == K_UVC_BUF_TYPE_VIDEO_OUTPUT_MPLANE)

#define K_UVC_TYPE_IS_OUTPUT(type)				\
    ((type) == K_UVC_BUF_TYPE_VIDEO_OUTPUT)

enum k_uvc_memory {
    K_UVC_MEMORY_MMAP             = 1,
    K_UVC_MEMORY_USERPTR          = 2,
    K_UVC_MEMORY_OVERLAY          = 3,
};

/* see also http://vektor.theorem.ca/graphics/ycbcr/ */
enum k_uvc_colorspace {
	/*
	 * Default colorspace, i.e. let the driver figure it out.
	 * Can only be used with video capture.
	 */
    K_UVC_COLORSPACE_DEFAULT       = 0,

	/* SMPTE 170M: used for broadcast NTSC/PAL SDTV */
    K_UVC_COLORSPACE_SMPTE170M     = 1,

	/* Obsolete pre-1998 SMPTE 240M HDTV standard, superseded by Rec 709 */
    K_UVC_COLORSPACE_SMPTE240M     = 2,

	/* Rec.709: used for HDTV */
    K_UVC_COLORSPACE_REC709        = 3,

	/*
	 * Deprecated, do not use. No driver will ever return this. This was
	 * based on a misunderstanding of the bt878 datasheet.
	 */
    K_UVC_COLORSPACE_BT878         = 4,

	/*
	 * NTSC 1953 colorspace. This only makes sense when dealing with
	 * really, really old NTSC recordings. Superseded by SMPTE 170M.
	 */
    K_UVC_COLORSPACE_470_SYSTEM_M  = 5,

	/*
	 * EBU Tech 3213 PAL/SECAM colorspace. This only makes sense when
	 * dealing with really old PAL/SECAM recordings. Superseded by
	 * SMPTE 170M.
	 */
    K_UVC_COLORSPACE_470_SYSTEM_BG = 6,

	/*
     * Effectively shorthand for K_UVC_COLORSPACE_SRGB, K_UVC_YCBCR_ENC_601
     * and K_UVC_QUANTIZATION_FULL_RANGE. To be used for (Motion-)JPEG.
	 */
    K_UVC_COLORSPACE_JPEG          = 7,

	/* For RGB colorspaces such as produces by most webcams. */
    K_UVC_COLORSPACE_SRGB          = 8,

	/* AdobeRGB colorspace */
    K_UVC_COLORSPACE_ADOBERGB      = 9,

	/* BT.2020 colorspace, used for UHDTV. */
    K_UVC_COLORSPACE_BT2020        = 10,

	/* Raw colorspace: for RAW unprocessed images */
    K_UVC_COLORSPACE_RAW           = 11,

	/* DCI-P3 colorspace, used by cinema projectors */
    K_UVC_COLORSPACE_DCI_P3        = 12,
};

/*
 * Determine how COLORSPACE_DEFAULT should map to a proper colorspace.
 * This depends on whether this is a SDTV image (use SMPTE 170M), an
 * HDTV image (use Rec. 709), or something else (use sRGB).
 */
#define K_UVC_MAP_COLORSPACE_DEFAULT(is_sdtv, is_hdtv) \
    ((is_sdtv) ? K_UVC_COLORSPACE_SMPTE170M : \
     ((is_hdtv) ? K_UVC_COLORSPACE_REC709 : K_UVC_COLORSPACE_SRGB))

enum k_uvc_xfer_func {
	/*
     * Mapping of K_UVC_XFER_FUNC_DEFAULT to actual transfer functions
	 * for the various colorspaces:
	 *
     * K_UVC_COLORSPACE_SMPTE170M, K_UVC_COLORSPACE_470_SYSTEM_M,
     * K_UVC_COLORSPACE_470_SYSTEM_BG, K_UVC_COLORSPACE_REC709 and
     * K_UVC_COLORSPACE_BT2020: K_UVC_XFER_FUNC_709
	 *
     * K_UVC_COLORSPACE_SRGB, K_UVC_COLORSPACE_JPEG: K_UVC_XFER_FUNC_SRGB
	 *
     * K_UVC_COLORSPACE_ADOBERGB: K_UVC_XFER_FUNC_ADOBERGB
	 *
     * K_UVC_COLORSPACE_SMPTE240M: K_UVC_XFER_FUNC_SMPTE240M
	 *
     * K_UVC_COLORSPACE_RAW: K_UVC_XFER_FUNC_NONE
	 *
     * K_UVC_COLORSPACE_DCI_P3: K_UVC_XFER_FUNC_DCI_P3
	 */
    K_UVC_XFER_FUNC_DEFAULT     = 0,
    K_UVC_XFER_FUNC_709         = 1,
    K_UVC_XFER_FUNC_SRGB        = 2,
    K_UVC_XFER_FUNC_ADOBERGB    = 3,
    K_UVC_XFER_FUNC_SMPTE240M   = 4,
    K_UVC_XFER_FUNC_NONE        = 5,
    K_UVC_XFER_FUNC_DCI_P3      = 6,
    K_UVC_XFER_FUNC_SMPTE2084   = 7,
};

/*
 * Determine how XFER_FUNC_DEFAULT should map to a proper transfer function.
 * This depends on the colorspace.
 */
#define K_UVC_MAP_XFER_FUNC_DEFAULT(colsp) \
    ((colsp) == K_UVC_COLORSPACE_ADOBERGB ? K_UVC_XFER_FUNC_ADOBERGB : \
     ((colsp) == K_UVC_COLORSPACE_SMPTE240M ? K_UVC_XFER_FUNC_SMPTE240M : \
      ((colsp) == K_UVC_COLORSPACE_DCI_P3 ? K_UVC_XFER_FUNC_DCI_P3 : \
       ((colsp) == K_UVC_COLORSPACE_RAW ? K_UVC_XFER_FUNC_NONE : \
        ((colsp) == K_UVC_COLORSPACE_SRGB || (colsp) == K_UVC_COLORSPACE_JPEG ? \
         K_UVC_XFER_FUNC_SRGB : K_UVC_XFER_FUNC_709)))))

enum k_uvc_ycbcr_encoding {
	/*
     * Mapping of K_UVC_YCBCR_ENC_DEFAULT to actual encodings for the
	 * various colorspaces:
	 *
     * K_UVC_COLORSPACE_SMPTE170M, K_UVC_COLORSPACE_470_SYSTEM_M,
     * K_UVC_COLORSPACE_470_SYSTEM_BG, K_UVC_COLORSPACE_SRGB,
     * K_UVC_COLORSPACE_ADOBERGB and K_UVC_COLORSPACE_JPEG: K_UVC_YCBCR_ENC_601
	 *
     * K_UVC_COLORSPACE_REC709 and K_UVC_COLORSPACE_DCI_P3: K_UVC_YCBCR_ENC_709
	 *
     * K_UVC_COLORSPACE_BT2020: K_UVC_YCBCR_ENC_BT2020
	 *
     * K_UVC_COLORSPACE_SMPTE240M: K_UVC_YCBCR_ENC_SMPTE240M
	 */
    K_UVC_YCBCR_ENC_DEFAULT        = 0,

	/* ITU-R 601 -- SDTV */
    K_UVC_YCBCR_ENC_601            = 1,

	/* Rec. 709 -- HDTV */
    K_UVC_YCBCR_ENC_709            = 2,

	/* ITU-R 601/EN 61966-2-4 Extended Gamut -- SDTV */
    K_UVC_YCBCR_ENC_XV601          = 3,

	/* Rec. 709/EN 61966-2-4 Extended Gamut -- HDTV */
    K_UVC_YCBCR_ENC_XV709          = 4,

	/*
	 * sYCC (Y'CbCr encoding of sRGB), identical to ENC_601. It was added
	 * originally due to a misunderstanding of the sYCC standard. It should
     * not be used, instead use K_UVC_YCBCR_ENC_601.
	 */
    K_UVC_YCBCR_ENC_SYCC           = 5,

	/* BT.2020 Non-constant Luminance Y'CbCr */
    K_UVC_YCBCR_ENC_BT2020         = 6,

	/* BT.2020 Constant Luminance Y'CbcCrc */
    K_UVC_YCBCR_ENC_BT2020_CONST_LUM = 7,

	/* SMPTE 240M -- Obsolete HDTV */
    K_UVC_YCBCR_ENC_SMPTE240M      = 8,
};

/*
 * Determine how YCBCR_ENC_DEFAULT should map to a proper Y'CbCr encoding.
 * This depends on the colorspace.
 */
#define K_UVC_MAP_YCBCR_ENC_DEFAULT(colsp) \
    (((colsp) == K_UVC_COLORSPACE_REC709 || \
      (colsp) == K_UVC_COLORSPACE_DCI_P3) ? K_UVC_YCBCR_ENC_709 : \
     ((colsp) == K_UVC_COLORSPACE_BT2020 ? K_UVC_YCBCR_ENC_BT2020 : \
      ((colsp) == K_UVC_COLORSPACE_SMPTE240M ? K_UVC_YCBCR_ENC_SMPTE240M : \
       K_UVC_YCBCR_ENC_601)))

enum k_uvc_quantization {
	/*
	 * The default for R'G'B' quantization is always full range, except
	 * for the BT2020 colorspace. For Y'CbCr the quantization is always
	 * limited range, except for COLORSPACE_JPEG, SRGB, ADOBERGB,
	 * XV601 or XV709: those are full range.
	 */
    K_UVC_QUANTIZATION_DEFAULT     = 0,
    K_UVC_QUANTIZATION_FULL_RANGE  = 1,
    K_UVC_QUANTIZATION_LIM_RANGE   = 2,
};

/*
 * Determine how QUANTIZATION_DEFAULT should map to a proper quantization.
 * This depends on whether the image is RGB or not, the colorspace and the
 * Y'CbCr encoding.
 */
#define K_UVC_MAP_QUANTIZATION_DEFAULT(is_rgb, colsp, ycbcr_enc) \
    (((is_rgb) && (colsp) == K_UVC_COLORSPACE_BT2020) ? K_UVC_QUANTIZATION_LIM_RANGE : \
     (((is_rgb) || (ycbcr_enc) == K_UVC_YCBCR_ENC_XV601 || \
      (ycbcr_enc) == K_UVC_YCBCR_ENC_XV709 || (colsp) == K_UVC_COLORSPACE_JPEG) || \
      (colsp) == K_UVC_COLORSPACE_ADOBERGB || (colsp) == K_UVC_COLORSPACE_SRGB ? \
     K_UVC_QUANTIZATION_FULL_RANGE : K_UVC_QUANTIZATION_LIM_RANGE))

enum k_uvc_priority {
    K_UVC_PRIORITY_UNSET       = 0,  /* not initialized */
    K_UVC_PRIORITY_BACKGROUND  = 1,
    K_UVC_PRIORITY_INTERACTIVE = 2,
    K_UVC_PRIORITY_RECORD      = 3,
    K_UVC_PRIORITY_DEFAULT     = K_UVC_PRIORITY_INTERACTIVE,
};

struct k_uvc_rect {
	int32_t   left;
	int32_t   top;
	uint32_t   width;
	uint32_t   height;
};

struct k_uvc_fract {
	uint32_t   numerator;
	uint32_t   denominator;
};

/**
  * struct k_uvc_capability - Describes K_UVC device caps returned by VIDIOC_QUERYCAP
  *
  * @driver:	   name of the driver module (e.g. "bttv")
  * @card:	   name of the card (e.g. "Hauppauge WinTV")
  * @bus_info:	   name of the bus (e.g. "PCI:" + pci_name(pci_dev) )
  * @version:	   KERNEL_VERSION
  * @capabilities: capabilities of the physical device as a whole
  * @device_caps:  capabilities accessed via this particular device (node)
  * @reserved:	   reserved fields for future extensions
  */
struct k_uvc_capability {
    uint8_t	driver[16];
    uint8_t	card[32];
	uint8_t	bus_info[32];
	uint32_t version;
	uint32_t capabilities;
	uint32_t device_caps;
	uint32_t reserved[3];
};

/* Values for 'capabilities' field */
#define K_UVC_CAP_VIDEO_CAPTURE		0x00000001  /* Is a video capture device */
#define K_UVC_CAP_VIDEO_OUTPUT		0x00000002  /* Is a video output device */
#define K_UVC_CAP_VIDEO_OVERLAY		0x00000004  /* Can do video overlay */
#define K_UVC_CAP_VIDEO_OUTPUT_OVERLAY	0x00000200  /* Can do video output overlay */
#define K_UVC_CAP_HW_FREQ_SEEK		0x00000400  /* Can do hardware frequency seek  */

/* Is a video capture device that supports multiplanar formats */
#define K_UVC_CAP_VIDEO_CAPTURE_MPLANE	0x00001000
/* Is a video output device that supports multiplanar formats */
#define K_UVC_CAP_VIDEO_OUTPUT_MPLANE	0x00002000
/* Is a video mem-to-mem device that supports multiplanar formats */
#define K_UVC_CAP_VIDEO_M2M_MPLANE	0x00004000
/* Is a video mem-to-mem device */
#define K_UVC_CAP_VIDEO_M2M		0x00008000

//#define K_UVC_CAP_SDR_CAPTURE		0x00100000  /* Is a SDR capture device */
#define K_UVC_CAP_EXT_PIX_FORMAT		0x00200000  /* Supports the extended pixel format */
//#define K_UVC_CAP_SDR_OUTPUT		0x00400000  /* Is a SDR output device */

#define K_UVC_CAP_READWRITE              0x01000000  /* read/write systemcalls */
#define K_UVC_CAP_ASYNCIO                0x02000000  /* async I/O */
#define K_UVC_CAP_STREAMING              0x04000000  /* streaming I/O ioctls */

#define K_UVC_CAP_TOUCH                  0x10000000  /* Is a touch device */

#define K_UVC_CAP_DEVICE_CAPS            0x80000000  /* sets device capabilities field */

/*
 *	V I D E O   I M A G E   F O R M A T
 */
struct k_uvc_pix_format {
	uint32_t  width;
	uint32_t  height;
	uint32_t  pixelformat;
    uint32_t  field;		/* enum k_uvc_field */
	uint32_t  bytesperline;	/* for padding, zero if unused */
	uint32_t  sizeimage;
    uint32_t  colorspace;	/* enum K_UVC_colorspace */
	uint32_t  priv;		/* private data, depends on pixelformat */
    uint32_t  flags;		/* format flags (K_UVC_PIX_FMT_FLAG_*) */
    uint32_t  ycbcr_enc;	/* enum K_UVC_ycbcr_encoding */
    uint32_t  quantization;	/* enum K_UVC_quantization */
    uint32_t  xfer_func;	/* enum k_uvc_xfer_func */
};

/*      Pixel format         FOURCC                          depth  Description  */

/* RGB formats */
#define K_UVC_PIX_FMT_RGB332  k_uvc_fourcc('R', 'G', 'B', '1') /*  8  RGB-3-3-2     */
#define K_UVC_PIX_FMT_RGB444  k_uvc_fourcc('R', '4', '4', '4') /* 16  xxxxrrrr ggggbbbb */
#define K_UVC_PIX_FMT_ARGB444 k_uvc_fourcc('A', 'R', '1', '2') /* 16  aaaarrrr ggggbbbb */
#define K_UVC_PIX_FMT_XRGB444 k_uvc_fourcc('X', 'R', '1', '2') /* 16  xxxxrrrr ggggbbbb */
#define K_UVC_PIX_FMT_RGB555  k_uvc_fourcc('R', 'G', 'B', 'O') /* 16  RGB-5-5-5     */
#define K_UVC_PIX_FMT_ARGB555 k_uvc_fourcc('A', 'R', '1', '5') /* 16  ARGB-1-5-5-5  */
#define K_UVC_PIX_FMT_XRGB555 k_uvc_fourcc('X', 'R', '1', '5') /* 16  XRGB-1-5-5-5  */
#define K_UVC_PIX_FMT_RGB565  k_uvc_fourcc('R', 'G', 'B', 'P') /* 16  RGB-5-6-5     */
#define K_UVC_PIX_FMT_RGB555X k_uvc_fourcc('R', 'G', 'B', 'Q') /* 16  RGB-5-5-5 BE  */
#define K_UVC_PIX_FMT_ARGB555X k_uvc_fourcc_be('A', 'R', '1', '5') /* 16  ARGB-5-5-5 BE */
#define K_UVC_PIX_FMT_XRGB555X k_uvc_fourcc_be('X', 'R', '1', '5') /* 16  XRGB-5-5-5 BE */
#define K_UVC_PIX_FMT_RGB565X k_uvc_fourcc('R', 'G', 'B', 'R') /* 16  RGB-5-6-5 BE  */
#define K_UVC_PIX_FMT_BGR666  k_uvc_fourcc('B', 'G', 'R', 'H') /* 18  BGR-6-6-6	  */
#define K_UVC_PIX_FMT_BGR24   k_uvc_fourcc('B', 'G', 'R', '3') /* 24  BGR-8-8-8     */
#define K_UVC_PIX_FMT_RGB24   k_uvc_fourcc('R', 'G', 'B', '3') /* 24  RGB-8-8-8     */
#define K_UVC_PIX_FMT_BGR32   k_uvc_fourcc('B', 'G', 'R', '4') /* 32  BGR-8-8-8-8   */
#define K_UVC_PIX_FMT_ABGR32  k_uvc_fourcc('A', 'R', '2', '4') /* 32  BGRA-8-8-8-8  */
#define K_UVC_PIX_FMT_XBGR32  k_uvc_fourcc('X', 'R', '2', '4') /* 32  BGRX-8-8-8-8  */
#define K_UVC_PIX_FMT_RGB32   k_uvc_fourcc('R', 'G', 'B', '4') /* 32  RGB-8-8-8-8   */
#define K_UVC_PIX_FMT_ARGB32  k_uvc_fourcc('B', 'A', '2', '4') /* 32  ARGB-8-8-8-8  */
#define K_UVC_PIX_FMT_XRGB32  k_uvc_fourcc('B', 'X', '2', '4') /* 32  XRGB-8-8-8-8  */

/* Grey formats */
#define K_UVC_PIX_FMT_GREY    k_uvc_fourcc('G', 'R', 'E', 'Y') /*  8  Greyscale     */
#define K_UVC_PIX_FMT_Y4      k_uvc_fourcc('Y', '0', '4', ' ') /*  4  Greyscale     */
#define K_UVC_PIX_FMT_Y6      k_uvc_fourcc('Y', '0', '6', ' ') /*  6  Greyscale     */
#define K_UVC_PIX_FMT_Y10     k_uvc_fourcc('Y', '1', '0', ' ') /* 10  Greyscale     */
#define K_UVC_PIX_FMT_Y12     k_uvc_fourcc('Y', '1', '2', ' ') /* 12  Greyscale     */
#define K_UVC_PIX_FMT_Y16     k_uvc_fourcc('Y', '1', '6', ' ') /* 16  Greyscale     */
#define K_UVC_PIX_FMT_Y16_BE  k_uvc_fourcc_be('Y', '1', '6', ' ') /* 16  Greyscale BE  */

/* Grey bit-packed formats */
#define K_UVC_PIX_FMT_Y10BPACK    k_uvc_fourcc('Y', '1', '0', 'B') /* 10  Greyscale bit-packed */

/* Palette formats */
#define K_UVC_PIX_FMT_PAL8    k_uvc_fourcc('P', 'A', 'L', '8') /*  8  8-bit palette */

/* Chrominance formats */
#define K_UVC_PIX_FMT_UV8     k_uvc_fourcc('U', 'V', '8', ' ') /*  8  UV 4:4 */

/* Luminance+Chrominance formats */
#define K_UVC_PIX_FMT_YUYV    k_uvc_fourcc('Y', 'U', 'Y', 'V') /* 16  YUV 4:2:2     */
#define K_UVC_PIX_FMT_YYUV    k_uvc_fourcc('Y', 'Y', 'U', 'V') /* 16  YUV 4:2:2     */
#define K_UVC_PIX_FMT_YVYU    k_uvc_fourcc('Y', 'V', 'Y', 'U') /* 16 YVU 4:2:2 */
#define K_UVC_PIX_FMT_UYVY    k_uvc_fourcc('U', 'Y', 'V', 'Y') /* 16  YUV 4:2:2     */
#define K_UVC_PIX_FMT_VYUY    k_uvc_fourcc('V', 'Y', 'U', 'Y') /* 16  YUV 4:2:2     */
#define K_UVC_PIX_FMT_Y41P    k_uvc_fourcc('Y', '4', '1', 'P') /* 12  YUV 4:1:1     */
#define K_UVC_PIX_FMT_YUV444  k_uvc_fourcc('Y', '4', '4', '4') /* 16  xxxxyyyy uuuuvvvv */
#define K_UVC_PIX_FMT_YUV555  k_uvc_fourcc('Y', 'U', 'V', 'O') /* 16  YUV-5-5-5     */
#define K_UVC_PIX_FMT_YUV565  k_uvc_fourcc('Y', 'U', 'V', 'P') /* 16  YUV-5-6-5     */
#define K_UVC_PIX_FMT_YUV32   k_uvc_fourcc('Y', 'U', 'V', '4') /* 32  YUV-8-8-8-8   */
#define K_UVC_PIX_FMT_HI240   k_uvc_fourcc('H', 'I', '2', '4') /*  8  8-bit color   */
#define K_UVC_PIX_FMT_HM12    k_uvc_fourcc('H', 'M', '1', '2') /*  8  YUV 4:2:0 16x16 macroblocks */
#define K_UVC_PIX_FMT_M420    k_uvc_fourcc('M', '4', '2', '0') /* 12  YUV 4:2:0 2 lines y, 1 line uv interleaved */

/* two planes -- one Y, one Cr + Cb interleaved  */
#define K_UVC_PIX_FMT_NV12    k_uvc_fourcc('N', 'V', '1', '2') /* 12  Y/CbCr 4:2:0  */
#define K_UVC_PIX_FMT_NV21    k_uvc_fourcc('N', 'V', '2', '1') /* 12  Y/CrCb 4:2:0  */
#define K_UVC_PIX_FMT_NV16    k_uvc_fourcc('N', 'V', '1', '6') /* 16  Y/CbCr 4:2:2  */
#define K_UVC_PIX_FMT_NV61    k_uvc_fourcc('N', 'V', '6', '1') /* 16  Y/CrCb 4:2:2  */
#define K_UVC_PIX_FMT_NV24    k_uvc_fourcc('N', 'V', '2', '4') /* 24  Y/CbCr 4:4:4  */
#define K_UVC_PIX_FMT_NV42    k_uvc_fourcc('N', 'V', '4', '2') /* 24  Y/CrCb 4:4:4  */

/* two non contiguous planes - one Y, one Cr + Cb interleaved  */
#define K_UVC_PIX_FMT_NV12M   k_uvc_fourcc('N', 'M', '1', '2') /* 12  Y/CbCr 4:2:0  */
#define K_UVC_PIX_FMT_NV21M   k_uvc_fourcc('N', 'M', '2', '1') /* 21  Y/CrCb 4:2:0  */
#define K_UVC_PIX_FMT_NV16M   k_uvc_fourcc('N', 'M', '1', '6') /* 16  Y/CbCr 4:2:2  */
#define K_UVC_PIX_FMT_NV61M   k_uvc_fourcc('N', 'M', '6', '1') /* 16  Y/CrCb 4:2:2  */
#define K_UVC_PIX_FMT_NV12MT  k_uvc_fourcc('T', 'M', '1', '2') /* 12  Y/CbCr 4:2:0 64x32 macroblocks */
#define K_UVC_PIX_FMT_NV12MT_16X16 k_uvc_fourcc('V', 'M', '1', '2') /* 12  Y/CbCr 4:2:0 16x16 macroblocks */

/* three planes - Y Cb, Cr */
#define K_UVC_PIX_FMT_YUV410  k_uvc_fourcc('Y', 'U', 'V', '9') /*  9  YUV 4:1:0     */
#define K_UVC_PIX_FMT_YVU410  k_uvc_fourcc('Y', 'V', 'U', '9') /*  9  YVU 4:1:0     */
#define K_UVC_PIX_FMT_YUV411P k_uvc_fourcc('4', '1', '1', 'P') /* 12  YVU411 planar */
#define K_UVC_PIX_FMT_YUV420  k_uvc_fourcc('Y', 'U', '1', '2') /* 12  YUV 4:2:0     */
#define K_UVC_PIX_FMT_YVU420  k_uvc_fourcc('Y', 'V', '1', '2') /* 12  YVU 4:2:0     */
#define K_UVC_PIX_FMT_YUV422P k_uvc_fourcc('4', '2', '2', 'P') /* 16  YVU422 planar */

/* three non contiguous planes - Y, Cb, Cr */
#define K_UVC_PIX_FMT_YUV420M k_uvc_fourcc('Y', 'M', '1', '2') /* 12  YUV420 planar */
#define K_UVC_PIX_FMT_YVU420M k_uvc_fourcc('Y', 'M', '2', '1') /* 12  YVU420 planar */
#define K_UVC_PIX_FMT_YUV422M k_uvc_fourcc('Y', 'M', '1', '6') /* 16  YUV422 planar */
#define K_UVC_PIX_FMT_YVU422M k_uvc_fourcc('Y', 'M', '6', '1') /* 16  YVU422 planar */
#define K_UVC_PIX_FMT_YUV444M k_uvc_fourcc('Y', 'M', '2', '4') /* 24  YUV444 planar */
#define K_UVC_PIX_FMT_YVU444M k_uvc_fourcc('Y', 'M', '4', '2') /* 24  YVU444 planar */

/* Bayer formats - see http://www.siliconimaging.com/RGB%20Bayer.htm */
#define K_UVC_PIX_FMT_SBGGR8  k_uvc_fourcc('B', 'A', '8', '1') /*  8  BGBG.. GRGR.. */
#define K_UVC_PIX_FMT_SGBRG8  k_uvc_fourcc('G', 'B', 'R', 'G') /*  8  GBGB.. RGRG.. */
#define K_UVC_PIX_FMT_SGRBG8  k_uvc_fourcc('G', 'R', 'B', 'G') /*  8  GRGR.. BGBG.. */
#define K_UVC_PIX_FMT_SRGGB8  k_uvc_fourcc('R', 'G', 'G', 'B') /*  8  RGRG.. GBGB.. */
#define K_UVC_PIX_FMT_SBGGR10 k_uvc_fourcc('B', 'G', '1', '0') /* 10  BGBG.. GRGR.. */
#define K_UVC_PIX_FMT_SGBRG10 k_uvc_fourcc('G', 'B', '1', '0') /* 10  GBGB.. RGRG.. */
#define K_UVC_PIX_FMT_SGRBG10 k_uvc_fourcc('B', 'A', '1', '0') /* 10  GRGR.. BGBG.. */
#define K_UVC_PIX_FMT_SRGGB10 k_uvc_fourcc('R', 'G', '1', '0') /* 10  RGRG.. GBGB.. */
	/* 10bit raw bayer packed, 5 bytes for every 4 pixels */
#define K_UVC_PIX_FMT_SBGGR10P k_uvc_fourcc('p', 'B', 'A', 'A')
#define K_UVC_PIX_FMT_SGBRG10P k_uvc_fourcc('p', 'G', 'A', 'A')
#define K_UVC_PIX_FMT_SGRBG10P k_uvc_fourcc('p', 'g', 'A', 'A')
#define K_UVC_PIX_FMT_SRGGB10P k_uvc_fourcc('p', 'R', 'A', 'A')
	/* 10bit raw bayer a-law compressed to 8 bits */
#define K_UVC_PIX_FMT_SBGGR10ALAW8 k_uvc_fourcc('a', 'B', 'A', '8')
#define K_UVC_PIX_FMT_SGBRG10ALAW8 k_uvc_fourcc('a', 'G', 'A', '8')
#define K_UVC_PIX_FMT_SGRBG10ALAW8 k_uvc_fourcc('a', 'g', 'A', '8')
#define K_UVC_PIX_FMT_SRGGB10ALAW8 k_uvc_fourcc('a', 'R', 'A', '8')
	/* 10bit raw bayer DPCM compressed to 8 bits */
#define K_UVC_PIX_FMT_SBGGR10DPCM8 k_uvc_fourcc('b', 'B', 'A', '8')
#define K_UVC_PIX_FMT_SGBRG10DPCM8 k_uvc_fourcc('b', 'G', 'A', '8')
#define K_UVC_PIX_FMT_SGRBG10DPCM8 k_uvc_fourcc('B', 'D', '1', '0')
#define K_UVC_PIX_FMT_SRGGB10DPCM8 k_uvc_fourcc('b', 'R', 'A', '8')
#define K_UVC_PIX_FMT_SBGGR12 k_uvc_fourcc('B', 'G', '1', '2') /* 12  BGBG.. GRGR.. */
#define K_UVC_PIX_FMT_SGBRG12 k_uvc_fourcc('G', 'B', '1', '2') /* 12  GBGB.. RGRG.. */
#define K_UVC_PIX_FMT_SGRBG12 k_uvc_fourcc('B', 'A', '1', '2') /* 12  GRGR.. BGBG.. */
#define K_UVC_PIX_FMT_SRGGB12 k_uvc_fourcc('R', 'G', '1', '2') /* 12  RGRG.. GBGB.. */
#define K_UVC_PIX_FMT_SBGGR16 k_uvc_fourcc('B', 'Y', 'R', '2') /* 16  BGBG.. GRGR.. */

/* compressed formats */
#define K_UVC_PIX_FMT_MJPEG    k_uvc_fourcc('M', 'J', 'P', 'G') /* Motion-JPEG   */
#define K_UVC_PIX_FMT_JPEG     k_uvc_fourcc('J', 'P', 'E', 'G') /* JFIF JPEG     */
#define K_UVC_PIX_FMT_DV       k_uvc_fourcc('d', 'v', 's', 'd') /* 1394          */
#define K_UVC_PIX_FMT_MPEG     k_uvc_fourcc('M', 'P', 'E', 'G') /* MPEG-1/2/4 Multiplexed */
#define K_UVC_PIX_FMT_H264     k_uvc_fourcc('H', '2', '6', '4') /* H264 with start codes */
#define K_UVC_PIX_FMT_H264_NO_SC k_uvc_fourcc('A', 'V', 'C', '1') /* H264 without start codes */
#define K_UVC_PIX_FMT_H264_MVC k_uvc_fourcc('M', '2', '6', '4') /* H264 MVC */
#define K_UVC_PIX_FMT_H263     k_uvc_fourcc('H', '2', '6', '3') /* H263          */
#define K_UVC_PIX_FMT_MPEG1    k_uvc_fourcc('M', 'P', 'G', '1') /* MPEG-1 ES     */
#define K_UVC_PIX_FMT_MPEG2    k_uvc_fourcc('M', 'P', 'G', '2') /* MPEG-2 ES     */
#define K_UVC_PIX_FMT_MPEG4    k_uvc_fourcc('M', 'P', 'G', '4') /* MPEG-4 part 2 ES */
#define K_UVC_PIX_FMT_XVID     k_uvc_fourcc('X', 'V', 'I', 'D') /* Xvid           */
#define K_UVC_PIX_FMT_VC1_ANNEX_G k_uvc_fourcc('V', 'C', '1', 'G') /* SMPTE 421M Annex G compliant stream */
#define K_UVC_PIX_FMT_VC1_ANNEX_L k_uvc_fourcc('V', 'C', '1', 'L') /* SMPTE 421M Annex L compliant stream */
#define K_UVC_PIX_FMT_VP8      k_uvc_fourcc('V', 'P', '8', '0') /* VP8 */

/*  Vendor-specific formats   */
#define K_UVC_PIX_FMT_CPIA1    k_uvc_fourcc('C', 'P', 'I', 'A') /* cpia1 YUV */
#define K_UVC_PIX_FMT_WNVA     k_uvc_fourcc('W', 'N', 'V', 'A') /* Winnov hw compress */
#define K_UVC_PIX_FMT_SN9C10X  k_uvc_fourcc('S', '9', '1', '0') /* SN9C10x compression */
#define K_UVC_PIX_FMT_SN9C20X_I420 k_uvc_fourcc('S', '9', '2', '0') /* SN9C20x YUV 4:2:0 */
#define K_UVC_PIX_FMT_PWC1     k_uvc_fourcc('P', 'W', 'C', '1') /* pwc older webcam */
#define K_UVC_PIX_FMT_PWC2     k_uvc_fourcc('P', 'W', 'C', '2') /* pwc newer webcam */
#define K_UVC_PIX_FMT_ET61X251 k_uvc_fourcc('E', '6', '2', '5') /* ET61X251 compression */
#define K_UVC_PIX_FMT_SPCA501  k_uvc_fourcc('S', '5', '0', '1') /* YUYV per line */
#define K_UVC_PIX_FMT_SPCA505  k_uvc_fourcc('S', '5', '0', '5') /* YYUV per line */
#define K_UVC_PIX_FMT_SPCA508  k_uvc_fourcc('S', '5', '0', '8') /* YUVY per line */
#define K_UVC_PIX_FMT_SPCA561  k_uvc_fourcc('S', '5', '6', '1') /* compressed GBRG bayer */
#define K_UVC_PIX_FMT_PAC207   k_uvc_fourcc('P', '2', '0', '7') /* compressed BGGR bayer */
#define K_UVC_PIX_FMT_MR97310A k_uvc_fourcc('M', '3', '1', '0') /* compressed BGGR bayer */
#define K_UVC_PIX_FMT_JL2005BCD k_uvc_fourcc('J', 'L', '2', '0') /* compressed RGGB bayer */
#define K_UVC_PIX_FMT_SN9C2028 k_uvc_fourcc('S', 'O', 'N', 'X') /* compressed GBRG bayer */
#define K_UVC_PIX_FMT_SQ905C   k_uvc_fourcc('9', '0', '5', 'C') /* compressed RGGB bayer */
#define K_UVC_PIX_FMT_PJPG     k_uvc_fourcc('P', 'J', 'P', 'G') /* Pixart 73xx JPEG */
#define K_UVC_PIX_FMT_OV511    k_uvc_fourcc('O', '5', '1', '1') /* ov511 JPEG */
#define K_UVC_PIX_FMT_OV518    k_uvc_fourcc('O', '5', '1', '8') /* ov518 JPEG */
#define K_UVC_PIX_FMT_STV0680  k_uvc_fourcc('S', '6', '8', '0') /* stv0680 bayer */
#define K_UVC_PIX_FMT_TM6000   k_uvc_fourcc('T', 'M', '6', '0') /* tm5600/tm60x0 */
#define K_UVC_PIX_FMT_CIT_YYVYUY k_uvc_fourcc('C', 'I', 'T', 'V') /* one line of Y then 1 line of VYUY */
#define K_UVC_PIX_FMT_KONICA420  k_uvc_fourcc('K', 'O', 'N', 'I') /* YUV420 planar in blocks of 256 pixels */
#define K_UVC_PIX_FMT_JPGL	k_uvc_fourcc('J', 'P', 'G', 'L') /* JPEG-Lite */
#define K_UVC_PIX_FMT_SE401      k_uvc_fourcc('S', '4', '0', '1') /* se401 janggu compressed rgb */
#define K_UVC_PIX_FMT_S5C_UYVY_JPG k_uvc_fourcc('S', '5', 'C', 'I') /* S5C73M3 interleaved UYVY/JPEG */
#define K_UVC_PIX_FMT_Y8I      k_uvc_fourcc('Y', '8', 'I', ' ') /* Greyscale 8-bit L/R interleaved */
#define K_UVC_PIX_FMT_Y12I     k_uvc_fourcc('Y', '1', '2', 'I') /* Greyscale 12-bit L/R interleaved */
#define K_UVC_PIX_FMT_Z16      k_uvc_fourcc('Z', '1', '6', ' ') /* Depth data 16-bit */

/* SDR formats - used only for Software Defined Radio devices */
#define K_UVC_SDR_FMT_CU8          k_uvc_fourcc('C', 'U', '0', '8') /* IQ u8 */
#define K_UVC_SDR_FMT_CU16LE       k_uvc_fourcc('C', 'U', '1', '6') /* IQ u16le */
#define K_UVC_SDR_FMT_CS8          k_uvc_fourcc('C', 'S', '0', '8') /* complex s8 */
#define K_UVC_SDR_FMT_CS14LE       k_uvc_fourcc('C', 'S', '1', '4') /* complex s14le */
#define K_UVC_SDR_FMT_RU12LE       k_uvc_fourcc('R', 'U', '1', '2') /* real u12le */

/* Touch formats - used for Touch devices */
#define K_UVC_TCH_FMT_DELTA_TD16	k_uvc_fourcc('T', 'D', '1', '6') /* 16-bit signed deltas */
#define K_UVC_TCH_FMT_DELTA_TD08	k_uvc_fourcc('T', 'D', '0', '8') /* 8-bit signed deltas */
#define K_UVC_TCH_FMT_TU16	k_uvc_fourcc('T', 'U', '1', '6') /* 16-bit unsigned touch data */
#define K_UVC_TCH_FMT_TU08	k_uvc_fourcc('T', 'U', '0', '8') /* 8-bit unsigned touch data */

/* priv field value to indicates that subsequent fields are valid. */
#define K_UVC_PIX_FMT_PRIV_MAGIC		0xfeedcafe

/* Flags */
#define K_UVC_PIX_FMT_FLAG_PREMUL_ALPHA	0x00000001

/*
 *	F O R M A T   E N U M E R A T I O N
 */
struct k_uvc_fmtdesc {
	uint32_t  index;  /* Format number      */
    uint32_t  k_uvc_buf_type;   /* enum k_uvc_buf_type */
	uint32_t  flags;
	uint8_t  description[32];   /* Description string */
	uint32_t  pixelformat;       /* Format fourcc      */
	uint32_t  reserved[4];
};

#define K_UVC_FMT_FLAG_COMPRESSED 0x0001
#define K_UVC_FMT_FLAG_EMULATED   0x0002

	/* Frame Size and frame rate enumeration */
/*
 *	F R A M E   S I Z E   E N U M E R A T I O N
 */
enum k_uvc_frmsizetypes {
    K_UVC_FRMSIZE_TYPE_DISCRETE	= 1,
    K_UVC_FRMSIZE_TYPE_CONTINUOUS	= 2,
    K_UVC_FRMSIZE_TYPE_STEPWISE	= 3,
};

struct k_uvc_frmsize_discrete {
	uint32_t  width;  /* Frame width [pixel] */
	uint32_t  height;  /* Frame height [pixel] */
};

struct k_uvc_frmsize_stepwise {
	uint32_t  min_width;	/* Minimum frame width [pixel] */
	uint32_t  max_width;	/* Maximum frame width [pixel] */
	uint32_t  step_width;	/* Frame width step size [pixel] */
	uint32_t  min_height;	/* Minimum frame height [pixel] */
	uint32_t  max_height;	/* Maximum frame height [pixel] */
	uint32_t  step_height;	/* Frame height step size [pixel] */
};

struct k_uvc_frmsizeenum {
	uint32_t  index;		/* Frame size number */
	uint32_t  pixel_format;	/* Pixel format */
	uint32_t  type;		/* Frame size type the device supports. */

	union {					/* Frame size */
        struct k_uvc_frmsize_discrete	discrete;
        struct k_uvc_frmsize_stepwise	stepwise;
	};

	uint32_t   reserved[2];			/* Reserved space for future use */
};

/*
 *	F R A M E   R A T E   E N U M E R A T I O N
 */
enum k_uvc_frmivaltypes {
    K_UVC_FRMIVAL_TYPE_DISCRETE	= 1,
    K_UVC_FRMIVAL_TYPE_CONTINUOUS = 2,
    K_UVC_FRMIVAL_TYPE_STEPWISE	= 3,
};

struct k_uvc_frmival_stepwise {
    struct k_uvc_fract  min;		/* Minimum frame interval [s] */
    struct k_uvc_fract  max;		/* Maximum frame interval [s] */
    struct k_uvc_fract  step;		/* Frame interval step size [s] */
};

struct k_uvc_frmivalenum {
	uint32_t  index;		/* Frame format index */
	uint32_t  pixel_format;	/* Pixel format */
	uint32_t  width;		/* Frame width */
	uint32_t  height;		/* Frame height */
	uint32_t  type;		/* Frame interval type the device supports. */

	union {					/* Frame interval */
        struct k_uvc_fract		discrete;
        struct k_uvc_frmival_stepwise	stepwise;
	};

	uint32_t	reserved[2];			/* Reserved space for future use */
};

/*
 *	T I M E C O D E
 */
struct k_uvc_timecode {
	uint32_t  type;
	uint32_t  flags;
	uint8_t	frames;
	uint8_t	seconds;
	uint8_t	minutes;
	uint8_t	hours;
	uint8_t	userbits[4];
};

/*  Type  */
#define K_UVC_TC_TYPE_24FPS		1
#define K_UVC_TC_TYPE_25FPS		2
#define K_UVC_TC_TYPE_30FPS		3
#define K_UVC_TC_TYPE_50FPS		4
#define K_UVC_TC_TYPE_60FPS		5

/*  Flags  */
#define K_UVC_TC_FLAG_DROPFRAME		0x0001 /* "drop-frame" mode */
#define K_UVC_TC_FLAG_COLORFRAME		0x0002
#define K_UVC_TC_USERBITS_field		0x000C
#define K_UVC_TC_USERBITS_USERDEFINED	0x0000
#define K_UVC_TC_USERBITS_8BITCHARS	0x0008
/* The above is based on SMPTE timecodes */

struct k_uvc_jpegcompression {
	int quality;

	int  APPn;  /* Number of APP segment to be written,
				 * must be 0..15 */
	int  APP_len;        /* Length of data in JPEG APPn segment */
	char APP_data[60];   /* Data in the JPEG APPn segment. */

	int  COM_len;        /* Length of data in JPEG COM segment */
	char COM_data[60];   /* Data in JPEG COM segment */

	uint32_t jpeg_markers;  /* Which markers should go into the JPEG
				 * output. Unless you exactly know what
				 * you do, leave them untouched.
				 * Including less markers will make the
				 * resulting code smaller, but there will
				 * be fewer applications which can read it.
				 * The presence of the APP and COM marker
				 * is influenced by APP_len and COM_len
				 * ONLY, not by this property! */

#define K_UVC_JPEG_MARKER_DHT (1<<3)    /* Define Huffman Tables */
#define K_UVC_JPEG_MARKER_DQT (1<<4)    /* Define Quantization Tables */
#define K_UVC_JPEG_MARKER_DRI (1<<5)    /* Define Restart Interval */
#define K_UVC_JPEG_MARKER_COM (1<<6)    /* Comment segment */
#define K_UVC_JPEG_MARKER_APP (1<<7)    /* App segment, driver will
					* always use APP0 */
};

/*
 *	M E M O R Y - M A P P I N G   B U F F E R S
 */
struct k_uvc_requestbuffers {
	uint32_t  count;
    uint32_t  k_uvc_buf_type;		/* enum k_uvc_buf_type */
    uint32_t  memory_type;		/* enum k_uvc_memory */
	uint32_t  reserved[2];
};

/**
 * struct k_uvc_plane - plane info for multi-planar buffers
 * @bytesused:		number of bytes occupied by data in the plane (payload)
 * @length:		size of this plane (NOT the payload) in bytes
 * @mem_offset:		when memory in the associated struct k_uvc_buffer is
 *			K_UVC_MEMORY_MMAP, equals the offset from the start of
 *			the device memory for this plane (or is a "cookie" that
 *			should be passed to mmap() called on the video node)
 * @userptr:		when memory is K_UVC_MEMORY_USERPTR, a userspace pointer
 *			pointing to this plane
 * @fd:			when memory is K_UVC_MEMORY_DMABUF, a userspace file
 *			descriptor associated with this plane
 * @data_offset:	offset in the plane to the start of data; usually 0,
 *			unless there is a header in front of the data
 *
 * Multi-planar buffers consist of one or more planes, e.g. an YCbCr buffer
 * with two planes can have one plane for Y, and another for interleaved CbCr
 * components. Each plane can reside in a separate memory buffer, or even in
 * a completely separate memory node (e.g. in embedded devices).
 */
struct k_uvc_plane {
	uint32_t  bytesused;
	uint32_t  length;
	union {
		uint32_t       mem_offset;
		unsigned long  userptr;
		int32_t        fd;
	} m;
	uint32_t  data_offset;
	uint32_t  reserved[11];
};

/**
 * struct k_uvc_buffer - video buffer info
 * @index:	id number of the buffer
 * @type:	enum k_uvc_buf_type; buffer type (type == *_MPLANE for
 *		multiplanar buffers);
 * @bytesused:	number of bytes occupied by data in the buffer (payload);
 *		unused (set to 0) for multiplanar buffers
 * @flags:	buffer informational flags
 * @field:	enum k_uvc_field; field order of the image in the buffer
 * @timestamp:	frame timestamp
 * @timecode:	frame timecode
 * @sequence:	sequence count of this frame
 * @memory:	enum k_uvc_memory; the method, in which the actual video data is
 *		passed
 * @offset:	for non-multiplanar buffers with memory == K_UVC_MEMORY_MMAP;
 *		offset from the start of the device memory for this plane,
 *		(or a "cookie" that should be passed to mmap() as offset)
 * @userptr:	for non-multiplanar buffers with memory == K_UVC_MEMORY_USERPTR;
 *		a userspace pointer pointing to this buffer
 * @fd:		for non-multiplanar buffers with memory == K_UVC_MEMORY_DMABUF;
 *		a userspace file descriptor associated with this buffer
 * @planes:	for multiplanar buffers; userspace pointer to the array of plane
 *		info structs for this buffer
 * @length:	size in bytes of the buffer (NOT its payload) for single-plane
 *		buffers (when type != *_MPLANE); number of elements in the
 *		planes array for multi-plane buffers
 *
 * Contains data exchanged by application and driver using one of the Streaming
 * I/O methods.
 */
//struct timeval {
//        long         tv_sec;         /* seconds */
//        long    tv_usec;        /* microseconds */
//};
 
 
struct k_uvc_buffer {
	uint32_t  index;
	uint32_t  type;
	uint32_t  bytesused;
	uint32_t  flags;
	uint32_t  field;
	struct timeval  timestamp;
    struct k_uvc_timecode  timecode;
	uint32_t  sequence;

	/* memory location */
	uint32_t  memory_type;
	union {
		uint32_t  offset;
//		unsigned long   userptr;
        struct k_uvc_plane *planes;
//		int32_t  fd;
	} m;
	uint32_t  length;
	uint32_t  reserved2;
	uint32_t  reserved;
};

/*  Flags for 'flags' field */
/* Buffer is mapped (flag) */
#define K_UVC_BUF_FLAG_MAPPED			0x00000001
/* Buffer is queued for processing */
#define K_UVC_BUF_FLAG_QUEUED			0x00000002
/* Buffer is ready */
#define K_UVC_BUF_FLAG_DONE			0x00000004
/* Image is a keyframe (I-frame) */
#define K_UVC_BUF_FLAG_KEYFRAME			0x00000008
/* Image is a P-frame */
#define K_UVC_BUF_FLAG_PFRAME			0x00000010
/* Image is a B-frame */
#define K_UVC_BUF_FLAG_BFRAME			0x00000020
/* Buffer is ready, but the data contained within is corrupted. */
#define K_UVC_BUF_FLAG_ERROR			0x00000040
/* timecode field is valid */
#define K_UVC_BUF_FLAG_TIMECODE			0x00000100
/* Buffer is prepared for queuing */
#define K_UVC_BUF_FLAG_PREPARED			0x00000400
/* Cache handling flags */
#define K_UVC_BUF_FLAG_NO_CACHE_INVALIDATE	0x00000800
#define K_UVC_BUF_FLAG_NO_CACHE_CLEAN		0x00001000
/* Timestamp type */
#define K_UVC_BUF_FLAG_TIMESTAMP_MASK		0x0000e000
#define K_UVC_BUF_FLAG_TIMESTAMP_UNKNOWN		0x00000000
#define K_UVC_BUF_FLAG_TIMESTAMP_MONOTONIC	0x00002000
#define K_UVC_BUF_FLAG_TIMESTAMP_COPY		0x00004000
/* Timestamp sources. */
#define K_UVC_BUF_FLAG_TSTAMP_SRC_MASK		0x00070000
#define K_UVC_BUF_FLAG_TSTAMP_SRC_EOF		0x00000000
#define K_UVC_BUF_FLAG_TSTAMP_SRC_SOE		0x00010000
/* mem2mem encoder/decoder */
#define K_UVC_BUF_FLAG_LAST			0x00100000

/**
 * struct k_uvc_exportbuffer - export of video buffer as DMABUF file descriptor
 *
 * @index:	id number of the buffer
 * @type:	enum k_uvc_buf_type; buffer type (type == *_MPLANE for
 *		multiplanar buffers);
 * @plane:	index of the plane to be exported, 0 for single plane queues
 * @flags:	flags for newly created file, currently only O_CLOEXEC is
 *		supported, refer to manual of open syscall for more details
 * @fd:		file descriptor associated with DMABUF (set by driver)
 *
 * Contains data used for exporting a video buffer as DMABUF file descriptor.
 * The buffer is identified by a 'cookie' returned by VIDIOC_QUERYBUF
 * (identical to the cookie used to mmap() the buffer to userspace). All
 * reserved fields must be set to zero. The field reserved0 is expected to
 * become a structure 'type' allowing an alternative layout of the structure
 * content. Therefore this field should not be used for any other extensions.
 */
struct k_uvc_exportbuffer {
    uint32_t  k_uvc_buf_type; /* enum k_uvc_buf_type */
	uint32_t  index;
	uint32_t  plane;
	uint32_t  flags;
	int32_t  fd;
	uint32_t  reserved[11];
};

/*
 *	O V E R L A Y   P R E V I E W
 */
struct k_uvc_framebuffer {
	uint32_t  capability;
	uint32_t  flags;
/* FIXME: in theory we should pass something like PCI device + memory
 * region + offset instead of some physical address */
	void  *base;
	struct {
		uint32_t  width;
		uint32_t  height;
		uint32_t  pixelformat;
        uint32_t  field;		/* enum k_uvc_field */
		uint32_t  bytesperline;	/* for padding, zero if unused */
		uint32_t  sizeimage;
        uint32_t  colorspace;	/* enum k_uvc_colorspace */
		uint32_t  priv;		/* reserved field, set to 0 */
	} fmt;
};
/*  Flags for the 'capability' field. Read only */
#define K_UVC_FBUF_CAP_EXTERNOVERLAY	0x0001
#define K_UVC_FBUF_CAP_CHROMAKEY		0x0002
#define K_UVC_FBUF_CAP_LIST_CLIPPING     0x0004
#define K_UVC_FBUF_CAP_BITMAP_CLIPPING	0x0008
#define K_UVC_FBUF_CAP_LOCAL_ALPHA	0x0010
#define K_UVC_FBUF_CAP_GLOBAL_ALPHA	0x0020
#define K_UVC_FBUF_CAP_LOCAL_INV_ALPHA	0x0040
#define K_UVC_FBUF_CAP_SRC_CHROMAKEY	0x0080
/*  Flags for the 'flags' field. */
#define K_UVC_FBUF_FLAG_PRIMARY		0x0001
#define K_UVC_FBUF_FLAG_OVERLAY		0x0002
#define K_UVC_FBUF_FLAG_CHROMAKEY	0x0004
#define K_UVC_FBUF_FLAG_LOCAL_ALPHA	0x0008
#define K_UVC_FBUF_FLAG_GLOBAL_ALPHA	0x0010
#define K_UVC_FBUF_FLAG_LOCAL_INV_ALPHA	0x0020
#define K_UVC_FBUF_FLAG_SRC_CHROMAKEY	0x0040

struct k_uvc_clip {
    struct k_uvc_rect  c;
    struct k_uvc_clip   *next;
};

struct k_uvc_window {
    struct k_uvc_rect  w;
    uint32_t  field;	 /* enum k_uvc_field */
	uint32_t  chromakey;
    struct k_uvc_clip   *clips;
	uint32_t  clipcount;
	void   *bitmap;
	uint8_t  global_alpha;
};

/*
 *	C A P T U R E   P A R A M E T E R S
 */
struct k_uvc_captureparm {
	uint32_t  capability;	  /*  Supported modes */
	uint32_t  capturemode;	  /*  Current mode */
    struct k_uvc_fract  timeperframe;  /*  Time per frame in seconds */
	uint32_t  extendedmode;  /*  Driver-specific extensions */
	uint32_t  readbuffers;   /*  # of buffers for read */
	uint32_t  reserved[4];
};

/*  Flags for 'capability' and 'capturemode' fields */
#define K_UVC_MODE_HIGHQUALITY	0x0001	/*  High quality imaging mode */
#define K_UVC_CAP_TIMEPERFRAME	0x1000	/*  timeperframe field is supported */

struct k_uvc_outputparm {
	uint32_t  capability;	 /*  Supported modes */
	uint32_t  outputmode;	 /*  Current mode */
    struct k_uvc_fract  timeperframe; /*  Time per frame in seconds */
	uint32_t  extendedmode; /*  Driver-specific extensions */
	uint32_t  writebuffers; /*  # of buffers for write */
	uint32_t  reserved[4];
};

/*
 *	I N P U T   I M A G E   C R O P P I N G
 */
struct k_uvc_cropcap {
    uint32_t  k_uvc_buf_type;	/* enum k_uvc_buf_type */
    struct k_uvc_rect  bounds;
    struct k_uvc_rect  defrect;
    struct k_uvc_fract  pixelaspect;
};

struct k_uvc_crop {
    uint32_t  k_uvc_buf_type;	/* enum k_uvc_buf_type */
    struct k_uvc_rect c;
};

/**
 * struct k_uvc_selection - selection info
 * @type:	buffer type (do not use *_MPLANE types)
 * @target:	Selection target, used to choose one of possible rectangles;
 *		defined in K_UVC-common.h; K_UVC_SEL_TGT_* .
 * @flags:	constraints flags, defined in K_UVC-common.h; K_UVC_SEL_FLAG_*.
 * @r:		coordinates of selection window
 * @reserved:	for future use, rounds structure size to 64 bytes, set to zero
 *
 * Hardware may use multiple helper windows to process a video stream.
 * The structure is used to exchange this selection areas between
 * an application and a driver.
 */
struct k_uvc_selection {
	uint32_t  buf_type;
	uint32_t  target;
	uint32_t  flags;
    struct k_uvc_rect  r;
	uint32_t reserved[9];
};


/*
 *      A N A L O G   V I D E O   S T A N D A R D
 */

typedef uint64_t k_uvc_std_id;

/* one bit for each */
//#define K_UVC_STD_PAL_B    ((k_uvc_std_id)0x00000001)
//#define K_UVC_STD_PAL_B1   ((k_uvc_std_id)0x00000002)
//#define K_UVC_STD_PAL_G    ((k_uvc_std_id)0x00000004)
//#define K_UVC_STD_PAL_H    ((k_uvc_std_id)0x00000008)
//#define K_UVC_STD_PAL_I    ((k_uvc_std_id)0x00000010)
//#define K_UVC_STD_PAL_D    ((k_uvc_std_id)0x00000020)
//#define K_UVC_STD_PAL_D1   ((k_uvc_std_id)0x00000040)
//#define K_UVC_STD_PAL_K    ((k_uvc_std_id)0x00000080)

//#define K_UVC_STD_PAL_M    ((k_uvc_std_id)0x00000100)
//#define K_UVC_STD_PAL_N    ((k_uvc_std_id)0x00000200)
//#define K_UVC_STD_PAL_Nc   ((k_uvc_std_id)0x00000400)
#define K_UVC_STD_PAL_60   ((k_uvc_std_id)0x00000800)

//#define K_UVC_STD_NTSC_M     ((k_uvc_std_id)0x00001000)	/* BTSC */
//#define K_UVC_STD_NTSC_M_JP  ((k_uvc_std_id)0x00002000)	/* EIA-J */
//#define K_UVC_STD_NTSC_443   ((k_uvc_std_id)0x00004000)
//#define K_UVC_STD_NTSC_M_KR  ((k_uvc_std_id)0x00008000)	/* FM A2 */

//#define K_UVC_STD_SECAM_B    ((k_uvc_std_id)0x00010000)
//#define K_UVC_STD_SECAM_D    ((k_uvc_std_id)0x00020000)
//#define K_UVC_STD_SECAM_G    ((k_uvc_std_id)0x00040000)
//#define K_UVC_STD_SECAM_H    ((k_uvc_std_id)0x00080000)
//#define K_UVC_STD_SECAM_K    ((k_uvc_std_id)0x00100000)
//#define K_UVC_STD_SECAM_K1   ((k_uvc_std_id)0x00200000)
//#define K_UVC_STD_SECAM_L    ((k_uvc_std_id)0x00400000)
//#define K_UVC_STD_SECAM_LC   ((k_uvc_std_id)0x00800000)

/* ATSC/HDTV */
//#define K_UVC_STD_ATSC_8_VSB   ((k_uvc_std_id)0x01000000)
//#define K_UVC_STD_ATSC_16_VSB  ((k_uvc_std_id)0x02000000)

/* FIXME:
   Although std_id is 64 bits, there is an issue on PPC32 architecture that
   makes switch(__u64) to break. So, there's a hack on K_UVC-common.c rounding
   this value to 32 bits.
   As, currently, the max value is for K_UVC_STD_ATSC_16_VSB (30 bits wide),
   it should work fine. However, if needed to add more than two standards,
   K_UVC-common.c should be fixed.
 */

/*
 * Some macros to merge video standards in order to make live easier for the
 * drivers and K_UVC applications
 */

/*
 * "Common" NTSC/M - It should be noticed that K_UVC_STD_NTSC_443 is
 * Missing here.
 */
#define K_UVC_STD_NTSC           (K_UVC_STD_NTSC_M	|\
                 K_UVC_STD_NTSC_M_JP     |\
                 K_UVC_STD_NTSC_M_KR)
///* Secam macros */
#define K_UVC_STD_SECAM_DK      	(K_UVC_STD_SECAM_D	|\
                 K_UVC_STD_SECAM_K	|\
                 K_UVC_STD_SECAM_K1)
///* All Secam Standards */
#define K_UVC_STD_SECAM		(K_UVC_STD_SECAM_B	|\
                 K_UVC_STD_SECAM_G	|\
                 K_UVC_STD_SECAM_H	|\
                 K_UVC_STD_SECAM_DK	|\
                 K_UVC_STD_SECAM_L       |\
                 K_UVC_STD_SECAM_LC)
///* PAL macros */
#define K_UVC_STD_PAL_BG		(K_UVC_STD_PAL_B		|\
                 K_UVC_STD_PAL_B1	|\
                 K_UVC_STD_PAL_G)
#define K_UVC_STD_PAL_DK		(K_UVC_STD_PAL_D		|\
                 K_UVC_STD_PAL_D1	|\
                 K_UVC_STD_PAL_K)
/*
 * "Common" PAL - This macro is there to be compatible with the old
 * V4L1 concept of "PAL": /BGDKHI.
 * Several PAL standards are missing here: /M, /N and /Nc
 */
//#define K_UVC_STD_PAL		(K_UVC_STD_PAL_BG	|\
//				 K_UVC_STD_PAL_DK	|\
//				 K_UVC_STD_PAL_H		|\
//				 K_UVC_STD_PAL_I)
///* Chroma "agnostic" standards */
#define K_UVC_STD_B		(K_UVC_STD_PAL_B		|\
                 K_UVC_STD_PAL_B1	|\
                 K_UVC_STD_SECAM_B)
#define K_UVC_STD_G		(K_UVC_STD_PAL_G		|\
                 K_UVC_STD_SECAM_G)
//#define K_UVC_STD_H		(K_UVC_STD_PAL_H		|\
//				 K_UVC_STD_SECAM_H)
//#define K_UVC_STD_L		(K_UVC_STD_SECAM_L	|\
//				 K_UVC_STD_SECAM_LC)
//#define K_UVC_STD_GH		(K_UVC_STD_G		|\
//				 K_UVC_STD_H)
//#define K_UVC_STD_DK		(K_UVC_STD_PAL_DK	|\
//				 K_UVC_STD_SECAM_DK)
//#define K_UVC_STD_BG		(K_UVC_STD_B		|\
//				 K_UVC_STD_G)
#define K_UVC_STD_MN		(K_UVC_STD_PAL_M		|\
                 K_UVC_STD_PAL_N		|\
                 K_UVC_STD_PAL_Nc	|\
                 K_UVC_STD_NTSC)

///* Standards where MTS/BTSC stereo could be found */
#define K_UVC_STD_MTS		(K_UVC_STD_NTSC_M	|\
                 K_UVC_STD_PAL_M		|\
                 K_UVC_STD_PAL_N		|\
                 K_UVC_STD_PAL_Nc)

///* Standards for Countries with 60Hz Line frequency */
#define K_UVC_STD_525_60		(K_UVC_STD_PAL_M		|\
                 K_UVC_STD_PAL_60	|\
                 K_UVC_STD_NTSC		|\
                 K_UVC_STD_NTSC_443)
/* Standards for Countries with 50Hz Line frequency */
#define K_UVC_STD_625_50		(K_UVC_STD_PAL		|\
                 K_UVC_STD_PAL_N		|\
                 K_UVC_STD_PAL_Nc	|\
                 K_UVC_STD_SECAM)

//#define K_UVC_STD_ATSC           (K_UVC_STD_ATSC_8_VSB    |\
//				 K_UVC_STD_ATSC_16_VSB)
///* Macros with none and all analog standards */
//#define K_UVC_STD_UNKNOWN        0
//#define K_UVC_STD_ALL            (K_UVC_STD_525_60	|\
//				 K_UVC_STD_625_50)

//struct k_uvc_standard {
//	uint32_t            index;
//	k_uvc_std_id         id;
//	uint8_t             name[24];
//	struct k_uvc_fract   frameperiod; /* Frames, not fields */
//	uint32_t		    framelines;
//	uint32_t		    reserved[4];
//};



/*
 *	V I D E O   I N P U T S
 */
struct k_uvc_input {
	uint32_t  index;    /*  Which input */
	uint8_t	  name[32];	/*  Label */
	uint32_t  type;		/*  Type of input */
    //uint32_t  audioset;	/*  Associated audios (bitfield) */
//	uint32_t  tuner;    /*  enum K_UVC_tuner_type */
//	k_uvc_std_id  std;
	uint32_t  status;
	uint32_t  capabilities;
	uint32_t  reserved[3];
};

/*  Values for the 'type' field */
#define k_uvc_input_TYPE_TUNER		1
#define k_uvc_input_TYPE_CAMERA		2
#define k_uvc_input_TYPE_TOUCH		3

/* field 'status' - general */
#define K_UVC_IN_ST_NO_POWER    0x00000001  /* Attached device is off */
#define K_UVC_IN_ST_NO_SIGNAL   0x00000002
#define K_UVC_IN_ST_NO_COLOR    0x00000004

/* field 'status' - sensor orientation */
/* If sensor is mounted upside down set both bits */
#define K_UVC_IN_ST_HFLIP       0x00000010 /* Frames are flipped horizontally */
#define K_UVC_IN_ST_VFLIP       0x00000020 /* Frames are flipped vertically */

/* field 'status' - analog */
#define K_UVC_IN_ST_NO_H_LOCK   0x00000100  /* No horizontal sync lock */
#define K_UVC_IN_ST_COLOR_KILL  0x00000200  /* Color killer is active */
#define K_UVC_IN_ST_NO_V_LOCK   0x00000400  /* No vertical sync lock */
#define K_UVC_IN_ST_NO_STD_LOCK 0x00000800  /* No standard format lock */

/* field 'status' - digital */
#define K_UVC_IN_ST_NO_SYNC     0x00010000  /* No synchronization lock */
#define K_UVC_IN_ST_NO_EQU      0x00020000  /* No equalizer lock */
#define K_UVC_IN_ST_NO_CARRIER  0x00040000  /* Carrier recovery failed */

/* field 'status' - VCR and set-top box */
#define K_UVC_IN_ST_MACROVISION 0x01000000  /* Macrovision detected */
#define K_UVC_IN_ST_NO_ACCESS   0x02000000  /* Conditional access denied */
#define K_UVC_IN_ST_VTR         0x04000000  /* VTR time constant */

/* capabilities flags */
#define K_UVC_IN_CAP_DV_TIMINGS		0x00000002 /* Supports S_DV_TIMINGS */
#define K_UVC_IN_CAP_CUSTOM_TIMINGS	K_UVC_IN_CAP_DV_TIMINGS /* For compatibility */
#define K_UVC_IN_CAP_STD			0x00000004 /* Supports S_STD */
#define K_UVC_IN_CAP_NATIVE_SIZE		0x00000008 /* Supports setting native size */

/*
 *	V I D E O   O U T P U T S
 */
struct k_uvc_output {
	uint32_t    index;		/*  Which output */
	uint8_t     name[32];	/*  Label */
	uint32_t    type;		/*  Type of output */
    //uint32_t    audioset;	/*  Associated audios (bitfield) */
//	uint32_t    modulator;  /*  Associated modulator */
//	k_uvc_std_id  std;
	uint32_t    capabilities;
	uint32_t    reserved[3];
};
/*  Values for the 'type' field */
#define K_UVC_OUTPUT_TYPE_MODULATOR		1
#define K_UVC_OUTPUT_TYPE_ANALOG			2
#define K_UVC_OUTPUT_TYPE_ANALOGVGAOVERLAY	3

/* capabilities flags */
#define K_UVC_OUT_CAP_DV_TIMINGS		0x00000002 /* Supports S_DV_TIMINGS */
#define K_UVC_OUT_CAP_CUSTOM_TIMINGS	K_UVC_OUT_CAP_DV_TIMINGS /* For compatibility */
#define K_UVC_OUT_CAP_STD		0x00000004 /* Supports S_STD */
#define K_UVC_OUT_CAP_NATIVE_SIZE	0x00000008 /* Supports setting native size */

/*
 *	C O N T R O L S
 */
struct k_uvc_control {
	uint32_t  id;
	int32_t	  value;
};

struct k_uvc_ext_control {
	uint32_t id;
	uint32_t size;
	uint32_t reserved2[1];
	union {
		int32_t value;
		int64_t value64;
		char  *string;
		uint8_t  *p_u8;
		uint16_t  *p_u16;
		uint32_t  *p_u32;
		void  *ptr;
	};
} __attribute__ ((packed));

struct k_uvc_ext_controls {
	union {
//#ifndef __KERNEL__
		uint32_t ctrl_class;
//#endif
		uint32_t which;
	};
	uint32_t count;
	uint32_t error_idx;
	uint32_t reserved[2];
    struct k_uvc_ext_control *controls;
};

#define K_UVC_CTRL_ID_MASK      	  (0x0fffffff)
//#ifndef __KERNEL__
#define K_UVC_CTRL_ID2CLASS(id)    ((id) & 0x0fff0000UL)
//#endif
#define K_UVC_CTRL_ID2WHICH(id)    ((id) & 0x0fff0000UL)
#define K_UVC_CTRL_DRIVER_PRIV(id) (((id) & 0xffff) >= 0x1000)
#define K_UVC_CTRL_MAX_DIMS	  (4)
#define K_UVC_CTRL_WHICH_CUR_VAL   0
#define K_UVC_CTRL_WHICH_DEF_VAL   0x0f000000

enum k_uvc_ctrl_type {
    K_UVC_CTRL_TYPE_INTEGER	     = 1,
    K_UVC_CTRL_TYPE_BOOLEAN	     = 2,
    K_UVC_CTRL_TYPE_MENU	     = 3,
    K_UVC_CTRL_TYPE_BUTTON	     = 4,
    K_UVC_CTRL_TYPE_INTEGER64     = 5,
    K_UVC_CTRL_TYPE_CTRL_CLASS    = 6,
    K_UVC_CTRL_TYPE_STRING        = 7,
    K_UVC_CTRL_TYPE_BITMASK       = 8,
    K_UVC_CTRL_TYPE_INTEGER_MENU  = 9,

	/* Compound types are >= 0x0100 */
    K_UVC_CTRL_COMPOUND_TYPES     = 0x0100,
    K_UVC_CTRL_TYPE_U8	     = 0x0100,
    K_UVC_CTRL_TYPE_U16	     = 0x0101,
    K_UVC_CTRL_TYPE_U32	     = 0x0102,
};

/*  Used in the VIDIOC_QUERYCTRL ioctl for querying controls */
struct k_uvc_queryctrl {
	uint32_t  id;
    uint32_t  type;	/* enum K_UVC_ctrl_type */
	uint8_t  name[32];	/* Whatever */
	int32_t  minimum;	/* Note signedness */
	int32_t  maximum;
	int32_t  step;
	int32_t  default_value;
	uint32_t  flags;
	uint32_t  reserved[2];
};

/*  Used in the VIDIOC_QUERY_EXT_CTRL ioctl for querying extended controls */
struct k_uvc_query_ext_ctrl {
	uint32_t  id;
	uint32_t  type;
	char      name[32];
	int64_t	  minimum;
	int64_t   maximum;
	uint64_t  step;
	int64_t   default_value;
	uint32_t  flags;
	uint32_t  elem_size;
	uint32_t  elems;
	uint32_t  nr_of_dims;
    uint32_t  dims[K_UVC_CTRL_MAX_DIMS];
	uint32_t  reserved[32];
};

/*  Used in the VIDIOC_QUERYMENU ioctl for querying menu items */
struct k_uvc_querymenu {
	uint32_t  id;
	uint32_t  index;
	union {
		uint8_t	name[32];	/* Whatever */
		int64_t	value;
	};
	uint32_t  reserved;
} __attribute__ ((packed));

/*  Control flags  */
#define K_UVC_CTRL_FLAG_DISABLED		0x0001
#define K_UVC_CTRL_FLAG_GRABBED		0x0002
#define K_UVC_CTRL_FLAG_READ_ONLY 	0x0004
#define K_UVC_CTRL_FLAG_UPDATE 		0x0008
#define K_UVC_CTRL_FLAG_INACTIVE 	0x0010
#define K_UVC_CTRL_FLAG_SLIDER 		0x0020
#define K_UVC_CTRL_FLAG_WRITE_ONLY 	0x0040
#define K_UVC_CTRL_FLAG_VOLATILE		0x0080
#define K_UVC_CTRL_FLAG_HAS_PAYLOAD	0x0100
#define K_UVC_CTRL_FLAG_EXECUTE_ON_WRITE	0x0200

/*  Query flags, to be ORed with the control ID */
#define K_UVC_CTRL_FLAG_NEXT_CTRL	0x80000000
#define K_UVC_CTRL_FLAG_NEXT_COMPOUND	0x40000000

/*  User-class control IDs defined by K_UVC */
#define K_UVC_CID_MAX_CTRLS		1024
/*  IDs reserved for driver specific controls */
#define K_UVC_CID_PRIVATE_BASE		0x08000000

/**
 * struct k_uvc_plane_pix_format - additional, per-plane format definition
 * @sizeimage:          maximum size in bytes required for data, for which
 *                      this plane will be used
 * @bytesperline:       distance in bytes between the leftmost pixels in two
 *                      adjacent lines
 */
struct k_uvc_plane_pix_format {
    uint32_t    sizeimage;
    uint32_t    bytesperline;
    uint16_t    reserved[6];
} __attribute__ ((packed));



/**
 * struct k_uvc_pix_format_mplane - multiplanar format definition
 * @width:              image width in pixels
 * @height:             image height in pixels
 * @pixelformat:        little endian four character code (fourcc)
 * @field:              enum k_uvc_field; field order (for interlaced video)
 * @colorspace:         enum k_uvc_colorspace; supplemental to pixelformat
 * @plane_fmt:          per-plane information
 * @num_planes:         number of planes for this format
 * @flags:              format flags (K_UVC_PIX_FMT_FLAG_*)
 * @ycbcr_enc:          enum k_uvc_ycbcr_encoding, Y'CbCr encoding
 * @quantization:       enum k_uvc_quantization, colorspace quantization
 * @xfer_func:          enum k_uvc_xfer_func, colorspace transfer function
 */
struct k_uvc_pix_format_mplane {
    uint32_t    width;
    uint32_t    height;
    uint32_t    pixelformat;
    uint32_t    field;
    uint32_t    colorspace;

    struct k_uvc_plane_pix_format    plane_fmt[VIDEO_MAX_PLANES];
    uint8_t    num_planes;
    uint8_t    flags;
    uint8_t    ycbcr_enc;
    uint8_t    quantization;
    uint8_t    xfer_func;
    uint8_t    reserved[7];
} __attribute__ ((packed));

/**
 * struct k_uvc_format - stream data format
 * @type:	enum k_uvc_buf_type; type of the data stream
 * @pix:	definition of an image format
 * @pix_mp:	definition of a multiplanar image format
 * @win:	definition of an overlaid image
 * @vbi:	raw VBI capture or output parameters
 * @sliced:	sliced VBI capture or output parameters
 * @raw_data:	placeholder for future extensions and custom formats
 */
struct k_uvc_format {
    uint32_t	 k_uvc_buf_type;
    union {
        struct k_uvc_pix_format		pix;     /* K_UVC_BUF_TYPE_VIDEO_CAPTURE */
        struct k_uvc_pix_format_mplane	pix_mp;  /* K_UVC_BUF_TYPE_VIDEO_CAPTURE_MPLANE */
        struct k_uvc_window		win;     /* K_UVC_BUF_TYPE_VIDEO_OVERLAY */
        uint8_t	raw_data[200];                   /* user-defined */
    } fmt;
};
/*
 *      A G G R E G A T E   S T R U C T U R E S
 */


/*	Stream type-dependent parameters
 */
struct k_uvc_streamparm {
    uint32_t	 k_uvc_buf_type;			/* enum k_uvc_buf_type */
	union {
        struct k_uvc_captureparm	capture;
        struct k_uvc_outputparm	output;
		uint8_t	raw_data[200];  /* user-defined */
	} parm;
};

/*
 *	E V E N T S
 */

#define K_UVC_EVENT_ALL				0
#define K_UVC_EVENT_VSYNC			1
#define K_UVC_EVENT_EOS				2
#define K_UVC_EVENT_CTRL				3
#define K_UVC_EVENT_FRAME_SYNC			4
#define K_UVC_EVENT_SOURCE_CHANGE		5
#define K_UVC_EVENT_MOTION_DET			6
#define K_UVC_EVENT_PRIVATE_START		0x08000000

/* Payload for K_UVC_EVENT_VSYNC */
struct k_uvc_event_vsync {
    /* Can be K_UVC_FIELD_ANY, _NONE, _TOP or _BOTTOM */
	uint8_t field;
} __attribute__ ((packed));

/* Payload for K_UVC_EVENT_CTRL */
#define K_UVC_EVENT_CTRL_CH_VALUE		(1 << 0)
#define K_UVC_EVENT_CTRL_CH_FLAGS		(1 << 1)
#define K_UVC_EVENT_CTRL_CH_RANGE		(1 << 2)

struct k_uvc_event_ctrl {
	uint32_t changes;
	uint32_t type;
	union {
		int32_t value;
		int64_t value64;
	};
	uint32_t flags;
	int32_t minimum;
	int32_t maximum;
	int32_t step;
	int32_t default_value;
};

struct k_uvc_event_frame_sync {
	uint32_t frame_sequence;
};

#define K_UVC_EVENT_SRC_CH_RESOLUTION		(1 << 0)

struct k_uvc_event_src_change {
	uint32_t changes;
};

#define K_UVC_EVENT_MD_FL_HAVE_FRAME_SEQ	(1 << 0)

/**
 * struct k_uvc_event_motion_det - motion detection event
 * @flags:             if K_UVC_EVENT_MD_FL_HAVE_FRAME_SEQ is set, then the
 *                     frame_sequence field is valid.
 * @frame_sequence:    the frame sequence number associated with this event.
 * @region_mask:       which regions detected motion.
 */
struct k_uvc_event_motion_det {
	uint32_t flags;
	uint32_t frame_sequence;
	uint32_t region_mask;
};

struct k_uvc_event {
	uint32_t				type;
	union {
        struct k_uvc_event_vsync		vsync;
        struct k_uvc_event_ctrl		ctrl;
        struct k_uvc_event_frame_sync	frame_sync;
        struct k_uvc_event_src_change	src_change;
        struct k_uvc_event_motion_det	motion_det;
		uint8_t	data[64];
	} u;
	uint32_t  pending;
	uint32_t  sequence;
	struct timespec  timestamp;
	uint32_t  id;
	uint32_t  reserved[8];
};

#define K_UVC_EVENT_SUB_FL_SEND_INITIAL		(1 << 0)
#define K_UVC_EVENT_SUB_FL_ALLOW_FEEDBACK	(1 << 1)

struct k_uvc_event_subscription {
	uint32_t				type;
	uint32_t				id;
	uint32_t				flags;
	uint32_t				reserved[5];
};


/**
 * struct k_uvc_create_buffers - VIDIOC_CREATE_BUFS argument
 * @index:	on return, index of the first created buffer
 * @count:	entry: number of requested buffers,
 *		return: number of created buffers
 * @memory:	enum k_uvc_memory; buffer memory type
 * @format:	frame format, for which buffers are requested
 * @reserved:	future extensions
 */
struct k_uvc_create_buffers {
	uint32_t			index;
	uint32_t			count;
	uint32_t			memory_type;
    struct k_uvc_format	format;
	uint32_t			reserved[8];
};

/*
 *	I O C T L   C O D E S   F O R   V I D E O   D E V I C E S
 *
 */
#define VIDIOC_QUERYCAP		 _IOR('V',  0, struct k_uvc_capability)
#define VIDIOC_ENUM_FMT      _IOWR('V',  1, struct k_uvc_fmtdesc)
#define VIDIOC_G_FMT		_IOWR('V',  2, struct k_uvc_format)
#define VIDIOC_S_FMT		_IOWR('V',  3, struct k_uvc_format)
#define VIDIOC_REQBUFS		_IOWR('V',  4, struct k_uvc_requestbuffers)
#define VIDIOC_QUERYBUF		_IOWR('V',  5, struct k_uvc_buffer)
#define VIDIOC_QBUF	    	_IOWR('V', 6, struct k_uvc_buffer)
#define VIDIOC_EXPBUF		_IOWR('V', 7, struct k_uvc_exportbuffer)
#define VIDIOC_DQBUF		_IOWR('V', 8, struct k_uvc_buffer)
#define VIDIOC_STREAMON		 _IOW('V', 9, int)
#define VIDIOC_STREAMOFF	 _IOW('V', 10, int)
#define VIDIOC_G_PARM		_IOWR('V', 11, struct k_uvc_streamparm)
#define VIDIOC_S_PARM		_IOWR('V', 12, struct k_uvc_streamparm)
#define VIDIOC_ENUMINPUT	_IOWR('V', 13, struct k_uvc_input)
#define VIDIOC_G_CTRL		_IOWR('V', 14, struct k_uvc_control)
#define VIDIOC_S_CTRL		_IOWR('V', 15, struct k_uvc_control)
#define VIDIOC_QUERYCTRL	_IOWR('V', 16, struct k_uvc_queryctrl)
#define VIDIOC_QUERYMENU	_IOWR('V', 17, struct k_uvc_querymenu)
#define VIDIOC_G_INPUT		 _IOR('V', 18, int)
#define VIDIOC_S_INPUT		_IOWR('V', 19, int)
#define VIDIOC_G_CROP		_IOWR('V', 20, struct k_uvc_crop)
#define VIDIOC_QUERYSTD      _IOR('V', 21, k_uvc_std_id)
#define VIDIOC_TRY_FMT      _IOWR('V', 22, struct k_uvc_format)

#define VIDIOC_G_PRIORITY	 _IOR('V', 23, uint32_t) /* enum K_UVC_priority */
#define VIDIOC_S_PRIORITY	 _IOW('V', 24, uint32_t) /* enum K_UVC_priority */
#define VIDIOC_G_EXT_CTRLS	_IOWR('V', 25, struct k_uvc_ext_controls)
#define VIDIOC_S_EXT_CTRLS	_IOWR('V', 26, struct k_uvc_ext_controls)
#define VIDIOC_TRY_EXT_CTRLS	_IOWR('V', 27, struct k_uvc_ext_controls)
#define VIDIOC_ENUM_FRAMESIZES	_IOWR('V', 28, struct k_uvc_frmsizeenum)
#define VIDIOC_ENUM_FRAMEINTERVALS _IOWR('V', 29, struct k_uvc_frmivalenum)
#define VIDIOC_QUERY_EXT_CTRL   _IOWR('V', 30, struct k_uvc_query_ext_ctrl)
#define	VIDIOC_DQEVENT		 _IOR('V', 31, struct K_UVC_event)
#define	VIDIOC_SUBSCRIBE_EVENT	 _IOW('V', 32, struct K_UVC_event_subscription)
#define	VIDIOC_UNSUBSCRIBE_EVENT _IOW('V', 33, struct K_UVC_event_subscription)
#define VIDIOC_CREATE_BUFS	_IOWR('V', 34, struct k_uvc_create_buffers)
#define VIDIOC_G_SELECTION	_IOWR('V', 35, struct k_uvc_selection)


/* Current cropping area */
#define K_UVC_SEL_TGT_CROP               0x0000
/* Default cropping area */
#define K_UVC_SEL_TGT_CROP_DEFAULT       0x0001
/* Cropping bounds */
#define K_UVC_SEL_TGT_CROP_BOUNDS        0x0002
/* Native frame size */
#define K_UVC_SEL_TGT_NATIVE_SIZE        0x0003
/* Current composing area */
#define K_UVC_SEL_TGT_COMPOSE            0x0100
/* Default composing area */
#define K_UVC_SEL_TGT_COMPOSE_DEFAULT    0x0101
/* Composing bounds */
#define K_UVC_SEL_TGT_COMPOSE_BOUNDS     0x0102
/* Current composing area plus all padding pixels */
#define K_UVC_SEL_TGT_COMPOSE_PADDED     0x0103

#define K_UVC_SEL_TGT_CROP_ACTIVE        K_UVC_SEL_TGT_CROP
#define K_UVC_SEL_TGT_COMPOSE_ACTIVE     K_UVC_SEL_TGT_COMPOSE
#define K_UVC_SUBDEV_SEL_TGT_CROP_ACTUAL K_UVC_SEL_TGT_CROP
#define K_UVC_SUBDEV_SEL_TGT_COMPOSE_ACTUAL K_UVC_SEL_TGT_COMPOSE
#define K_UVC_SUBDEV_SEL_TGT_CROP_BOUNDS K_UVC_SEL_TGT_CROP_BOUNDS
#define K_UVC_SUBDEV_SEL_TGT_COMPOSE_BOUNDS K_UVC_SEL_TGT_COMPOSE_BOUNDS


/* Control classes */
#define K_UVC_CTRL_CLASS_USER		0x00980000	/* Old-style 'user' controls */
#define K_UVC_CTRL_CLASS_MPEG		0x00990000	/* MPEG-compression controls */
#define K_UVC_CTRL_CLASS_CAMERA		0x009a0000	/* Camera class controls */
#define K_UVC_CTRL_CLASS_FM_TX		0x009b0000	/* FM Modulator controls */
#define K_UVC_CTRL_CLASS_FLASH		0x009c0000	/* Camera flash controls */
#define K_UVC_CTRL_CLASS_JPEG		0x009d0000	/* JPEG-compression controls */
#define K_UVC_CTRL_CLASS_IMAGE_SOURCE	0x009e0000	/* Image source controls */
#define K_UVC_CTRL_CLASS_IMAGE_PROC	0x009f0000	/* Image processing controls */
#define K_UVC_CTRL_CLASS_DV		0x00a00000	/* Digital Video controls */
#define K_UVC_CTRL_CLASS_FM_RX		0x00a10000	/* FM Receiver controls */
#define K_UVC_CTRL_CLASS_RF_TUNER	0x00a20000	/* RF tuner controls */
#define K_UVC_CTRL_CLASS_DETECT		0x00a30000	/* Detection controls */

/* User-class control IDs */

#define K_UVC_CID_BASE			(K_UVC_CTRL_CLASS_USER | 0x900)
#define K_UVC_CID_USER_BASE 		K_UVC_CID_BASE
#define K_UVC_CID_USER_CLASS 		(K_UVC_CTRL_CLASS_USER | 1)
#define K_UVC_CID_BRIGHTNESS		(K_UVC_CID_BASE+0)
#define K_UVC_CID_CONTRAST		(K_UVC_CID_BASE+1)
#define K_UVC_CID_SATURATION		(K_UVC_CID_BASE+2)
#define K_UVC_CID_HUE			(K_UVC_CID_BASE+3)
#define K_UVC_CID_AUTO_WHITE_BALANCE	(K_UVC_CID_BASE+12)
#define K_UVC_CID_RED_BALANCE         (K_UVC_CID_BASE+14)
#define K_UVC_CID_BLUE_BALANCE                (K_UVC_CID_BASE+15)
#define K_UVC_CID_GAMMA			(K_UVC_CID_BASE+16)
//#define K_UVC_CID_WHITENESS		(K_UVC_CID_GAMMA) /* Deprecated */
//#define K_UVC_CID_EXPOSURE		(K_UVC_CID_BASE+17)
//#define K_UVC_CID_AUTOGAIN		(K_UVC_CID_BASE+18)
#define K_UVC_CID_GAIN			(K_UVC_CID_BASE+19)
//#define K_UVC_CID_HFLIP			(K_UVC_CID_BASE+20)
//#define K_UVC_CID_VFLIP			(K_UVC_CID_BASE+21)

#define K_UVC_CID_POWER_LINE_FREQUENCY	(K_UVC_CID_BASE+24)
enum k_uvc_power_line_frequency {
    K_UVC_CID_POWER_LINE_FREQUENCY_DISABLED	= 0,
    K_UVC_CID_POWER_LINE_FREQUENCY_50HZ	= 1,
    K_UVC_CID_POWER_LINE_FREQUENCY_60HZ	= 2,
    K_UVC_CID_POWER_LINE_FREQUENCY_AUTO	= 3,
};
#define K_UVC_CID_HUE_AUTO			(K_UVC_CID_BASE+25)
#define K_UVC_CID_WHITE_BALANCE_TEMPERATURE	(K_UVC_CID_BASE+26)
#define K_UVC_CID_SHARPNESS			(K_UVC_CID_BASE+27)
#define K_UVC_CID_BACKLIGHT_COMPENSATION 	(K_UVC_CID_BASE+28)
#define K_UVC_CID_CHROMA_AGC                     (K_UVC_CID_BASE+29)
#define K_UVC_CID_COLOR_KILLER                   (K_UVC_CID_BASE+30)
#define K_UVC_CID_COLORFX			(K_UVC_CID_BASE+31)
enum k_uvc_colorfx {
    K_UVC_COLORFX_NONE			= 0,
    K_UVC_COLORFX_BW				= 1,
    K_UVC_COLORFX_SEPIA			= 2,
    K_UVC_COLORFX_NEGATIVE			= 3,
    K_UVC_COLORFX_EMBOSS			= 4,
    K_UVC_COLORFX_SKETCH			= 5,
    K_UVC_COLORFX_SKY_BLUE			= 6,
    K_UVC_COLORFX_GRASS_GREEN		= 7,
    K_UVC_COLORFX_SKIN_WHITEN		= 8,
    K_UVC_COLORFX_VIVID			= 9,
    K_UVC_COLORFX_AQUA			= 10,
    K_UVC_COLORFX_ART_FREEZE			= 11,
    K_UVC_COLORFX_SILHOUETTE			= 12,
    K_UVC_COLORFX_SOLARIZATION		= 13,
    K_UVC_COLORFX_ANTIQUE			= 14,
    K_UVC_COLORFX_SET_CBCR			= 15,
};
#define K_UVC_CID_AUTOBRIGHTNESS			(K_UVC_CID_BASE+32)
#define K_UVC_CID_BAND_STOP_FILTER		(K_UVC_CID_BASE+33)

#define K_UVC_CID_ROTATE				(K_UVC_CID_BASE+34)
#define K_UVC_CID_BG_COLOR			(K_UVC_CID_BASE+35)

#define K_UVC_CID_CHROMA_GAIN                    (K_UVC_CID_BASE+36)

#define K_UVC_CID_ILLUMINATORS_1			(K_UVC_CID_BASE+37)
#define K_UVC_CID_ILLUMINATORS_2			(K_UVC_CID_BASE+38)

#define K_UVC_CID_MIN_BUFFERS_FOR_CAPTURE	(K_UVC_CID_BASE+39)
#define K_UVC_CID_MIN_BUFFERS_FOR_OUTPUT		(K_UVC_CID_BASE+40)

#define K_UVC_CID_ALPHA_COMPONENT		(K_UVC_CID_BASE+41)
#define K_UVC_CID_COLORFX_CBCR			(K_UVC_CID_BASE+42)

/* last CID + 1 */
#define K_UVC_CID_LASTP1                         (K_UVC_CID_BASE+43)

/* USER-class private control IDs */

/* The base for the meye driver controls. See linux/meye.h for the list
 * of controls. We reserve 16 controls for this driver. */
#define K_UVC_CID_USER_MEYE_BASE			(K_UVC_CID_USER_BASE + 0x1000)

/* The base for the bttv driver controls.
 * We reserve 32 controls for this driver. */
#define K_UVC_CID_USER_BTTV_BASE			(K_UVC_CID_USER_BASE + 0x1010)


/* The base for the s2255 driver controls.
 * We reserve 16 controls for this driver. */
#define K_UVC_CID_USER_S2255_BASE		(K_UVC_CID_USER_BASE + 0x1030)

/*
 * The base for the si476x driver controls. See include/media/drv-intf/si476x.h
 * for the list of controls. Total of 16 controls is reserved for this driver
 */
#define K_UVC_CID_USER_SI476X_BASE		(K_UVC_CID_USER_BASE + 0x1040)

/* The base for the TI VPE driver controls. Total of 16 controls is reserved for
 * this driver */
#define K_UVC_CID_USER_TI_VPE_BASE		(K_UVC_CID_USER_BASE + 0x1050)

/* The base for the saa7134 driver controls.
 * We reserve 16 controls for this driver. */
#define K_UVC_CID_USER_SAA7134_BASE		(K_UVC_CID_USER_BASE + 0x1060)

/* The base for the adv7180 driver controls.
 * We reserve 16 controls for this driver. */
#define K_UVC_CID_USER_ADV7180_BASE		(K_UVC_CID_USER_BASE + 0x1070)

/* The base for the tc358743 driver controls.
 * We reserve 16 controls for this driver. */
#define K_UVC_CID_USER_TC358743_BASE		(K_UVC_CID_USER_BASE + 0x1080)


/*  Camera class control IDs */

#define K_UVC_CID_CAMERA_CLASS_BASE 	(K_UVC_CTRL_CLASS_CAMERA | 0x900)
#define K_UVC_CID_CAMERA_CLASS 		(K_UVC_CTRL_CLASS_CAMERA | 1)

#define K_UVC_CID_EXPOSURE_AUTO			(K_UVC_CID_CAMERA_CLASS_BASE+1)
enum  k_uvc_exposure_auto_type {
    K_UVC_EXPOSURE_AUTO = 0,
    K_UVC_EXPOSURE_MANUAL = 1,
    K_UVC_EXPOSURE_SHUTTER_PRIORITY = 2,
    K_UVC_EXPOSURE_APERTURE_PRIORITY = 3
};
#define K_UVC_CID_EXPOSURE_ABSOLUTE		(K_UVC_CID_CAMERA_CLASS_BASE+2)
#define K_UVC_CID_EXPOSURE_AUTO_PRIORITY		(K_UVC_CID_CAMERA_CLASS_BASE+3)

#define K_UVC_CID_PAN_RELATIVE			(K_UVC_CID_CAMERA_CLASS_BASE+4)
#define K_UVC_CID_TILT_RELATIVE			(K_UVC_CID_CAMERA_CLASS_BASE+5)
#define K_UVC_CID_PAN_RESET			(K_UVC_CID_CAMERA_CLASS_BASE+6)
#define K_UVC_CID_TILT_RESET			(K_UVC_CID_CAMERA_CLASS_BASE+7)

#define K_UVC_CID_PAN_ABSOLUTE			(K_UVC_CID_CAMERA_CLASS_BASE+8)
#define K_UVC_CID_TILT_ABSOLUTE			(K_UVC_CID_CAMERA_CLASS_BASE+9)

#define K_UVC_CID_FOCUS_ABSOLUTE			(K_UVC_CID_CAMERA_CLASS_BASE+10)
#define K_UVC_CID_FOCUS_RELATIVE			(K_UVC_CID_CAMERA_CLASS_BASE+11)
#define K_UVC_CID_FOCUS_AUTO			(K_UVC_CID_CAMERA_CLASS_BASE+12)

#define K_UVC_CID_ZOOM_ABSOLUTE			(K_UVC_CID_CAMERA_CLASS_BASE+13)
#define K_UVC_CID_ZOOM_RELATIVE			(K_UVC_CID_CAMERA_CLASS_BASE+14)
#define K_UVC_CID_ZOOM_CONTINUOUS		(K_UVC_CID_CAMERA_CLASS_BASE+15)

#define K_UVC_CID_PRIVACY			(K_UVC_CID_CAMERA_CLASS_BASE+16)

#define K_UVC_CID_IRIS_ABSOLUTE			(K_UVC_CID_CAMERA_CLASS_BASE+17)
#define K_UVC_CID_IRIS_RELATIVE			(K_UVC_CID_CAMERA_CLASS_BASE+18)

//#define K_UVC_CID_AUTO_EXPOSURE_BIAS		(K_UVC_CID_CAMERA_CLASS_BASE+19)

//#define K_UVC_CID_AUTO_N_PRESET_WHITE_BALANCE	(K_UVC_CID_CAMERA_CLASS_BASE+20)
enum k_uvc_auto_n_preset_white_balance {
    K_UVC_WHITE_BALANCE_MANUAL		= 0,
    K_UVC_WHITE_BALANCE_AUTO			= 1,
    K_UVC_WHITE_BALANCE_INCANDESCENT		= 2,
    K_UVC_WHITE_BALANCE_FLUORESCENT		= 3,
    K_UVC_WHITE_BALANCE_FLUORESCENT_H	= 4,
    K_UVC_WHITE_BALANCE_HORIZON		= 5,
    K_UVC_WHITE_BALANCE_DAYLIGHT		= 6,
    K_UVC_WHITE_BALANCE_FLASH		= 7,
    K_UVC_WHITE_BALANCE_CLOUDY		= 8,
    K_UVC_WHITE_BALANCE_SHADE		= 9,
};

#define K_UVC_CID_WIDE_DYNAMIC_RANGE		(K_UVC_CID_CAMERA_CLASS_BASE+21)
#define K_UVC_CID_IMAGE_STABILIZATION		(K_UVC_CID_CAMERA_CLASS_BASE+22)

#define K_UVC_CID_ISO_SENSITIVITY		(K_UVC_CID_CAMERA_CLASS_BASE+23)
#define K_UVC_CID_ISO_SENSITIVITY_AUTO		(K_UVC_CID_CAMERA_CLASS_BASE+24)
enum k_uvc_iso_sensitivity_auto_type {
    K_UVC_ISO_SENSITIVITY_MANUAL		= 0,
    K_UVC_ISO_SENSITIVITY_AUTO		= 1,
};

#define K_UVC_CID_EXPOSURE_METERING		(K_UVC_CID_CAMERA_CLASS_BASE+25)
enum k_uvc_exposure_metering {
    K_UVC_EXPOSURE_METERING_AVERAGE		= 0,
    K_UVC_EXPOSURE_METERING_CENTER_WEIGHTED	= 1,
    K_UVC_EXPOSURE_METERING_SPOT		= 2,
    K_UVC_EXPOSURE_METERING_MATRIX		= 3,
};

#define K_UVC_CID_SCENE_MODE			(K_UVC_CID_CAMERA_CLASS_BASE+26)
enum k_uvc_scene_mode {
    K_UVC_SCENE_MODE_NONE			= 0,
    K_UVC_SCENE_MODE_BACKLIGHT		= 1,
    K_UVC_SCENE_MODE_BEACH_SNOW		= 2,
    K_UVC_SCENE_MODE_CANDLE_LIGHT		= 3,
    K_UVC_SCENE_MODE_DAWN_DUSK		= 4,
    K_UVC_SCENE_MODE_FALL_COLORS		= 5,
    K_UVC_SCENE_MODE_FIREWORKS		= 6,
    K_UVC_SCENE_MODE_LANDSCAPE		= 7,
    K_UVC_SCENE_MODE_NIGHT			= 8,
    K_UVC_SCENE_MODE_PARTY_INDOOR		= 9,
    K_UVC_SCENE_MODE_PORTRAIT		= 10,
    K_UVC_SCENE_MODE_SPORTS			= 11,
    K_UVC_SCENE_MODE_SUNSET			= 12,
    K_UVC_SCENE_MODE_TEXT			= 13,
};

#define K_UVC_CID_3A_LOCK			(K_UVC_CID_CAMERA_CLASS_BASE+27)
#define K_UVC_LOCK_EXPOSURE			(1 << 0)
#define K_UVC_LOCK_WHITE_BALANCE			(1 << 1)
#define K_UVC_LOCK_FOCUS				(1 << 2)

#define K_UVC_CID_AUTO_FOCUS_START		(K_UVC_CID_CAMERA_CLASS_BASE+28)
#define K_UVC_CID_AUTO_FOCUS_STOP		(K_UVC_CID_CAMERA_CLASS_BASE+29)
#define K_UVC_CID_AUTO_FOCUS_STATUS		(K_UVC_CID_CAMERA_CLASS_BASE+30)
#define K_UVC_AUTO_FOCUS_STATUS_IDLE		(0 << 0)
#define K_UVC_AUTO_FOCUS_STATUS_BUSY		(1 << 0)
#define K_UVC_AUTO_FOCUS_STATUS_REACHED		(1 << 1)
#define K_UVC_AUTO_FOCUS_STATUS_FAILED		(1 << 2)

#define K_UVC_CID_AUTO_FOCUS_RANGE		(K_UVC_CID_CAMERA_CLASS_BASE+31)
enum k_uvc_auto_focus_range {
    K_UVC_AUTO_FOCUS_RANGE_AUTO		= 0,
    K_UVC_AUTO_FOCUS_RANGE_NORMAL		= 1,
    K_UVC_AUTO_FOCUS_RANGE_MACRO		= 2,
    K_UVC_AUTO_FOCUS_RANGE_INFINITY		= 3,
};

#define K_UVC_CID_PAN_SPEED			(K_UVC_CID_CAMERA_CLASS_BASE+32)
#define K_UVC_CID_TILT_SPEED		(K_UVC_CID_CAMERA_CLASS_BASE+33)

struct file* k_video_open (const char *name);
unsigned long k_video_get_unmapped_area(struct file *, unsigned long, unsigned long, unsigned long, unsigned long);
long k_video_ioctl (struct file *, unsigned int cmd, unsigned long arg );
//long int k_video_ioctl2(struct file *pfile,  unsigned int cmd, unsigned long int arg);

#endif /* _UAPI_K_UVC_H */
