/*
 * Kneron Peripheral API
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#ifndef __KDP_USBD_H__
#define __KDP_USBD_H__

#pragma anon_unions
#include "kdp_peripheral.h"

/*
    This is USB device mode driver API
    Some limitations:
    1) support only control and bulk transfer
    2) for bulk transfer, maximum endpoint number is 4
 */

#define MAX_USBD_CONFIG 1    // FIXME: need more ?
#define MAX_USBD_INTERFACE 1 // FIMXE: need more ?
#define MAX_USBD_ENDPOINT 4  // it is a limitation

/* 8-byte setup packet struct */
typedef struct __attribute__((__packed__))
{
    uint8_t bmRequestType;
    uint8_t bRequest;
    uint16_t wValue;
    uint16_t wIndex;
    uint16_t wLength;
} kdp_usbd_setup_packet_t;

/* Endpoint descriptor */
typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint8_t bEndpointAddress;
    uint8_t bmAttributes;
    uint16_t wMaxPacketSize;
    uint8_t bInterval;
} kdp_usbd_endpoint_descriptor_t;

/* Interface descriptor */
typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint8_t bInterfaceNumber;
    uint8_t bAlternateSetting;
    uint8_t bNumEndpoints;
    uint8_t bInterfaceClass;
    uint8_t bInterfaceSubClass;
    uint8_t bInterfaceProtocol;
    uint8_t iInterface;

    // support maximum number of endpoint is 4, bNumEndpoints should be <= 4
    kdp_usbd_endpoint_descriptor_t *endpoint[MAX_USBD_ENDPOINT];

} kdp_usbd_interface_descriptor_t;

/* Configuration descriptor */
typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint16_t wTotalLength;
    uint8_t bNumInterfaces;
    uint8_t bConfigurationValue;
    uint8_t iConfiguration;
    uint8_t bmAttributes;
    uint8_t MaxPower;

    // support maximum number of interface is 2, bNumInterfaces should be <= 2
    kdp_usbd_interface_descriptor_t *interface[MAX_USBD_INTERFACE];

} kdp_usbd_config_descriptor_t;

/* Device descriptor */
typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint16_t bcdUSB;
    uint8_t bDeviceClass;
    uint8_t bDeviceSubClass;
    uint8_t bDeviceProtocol;
    uint8_t bMaxPacketSize0;
    uint16_t idVendor;
    uint16_t idProduct;
    uint16_t bcdDevice;
    uint8_t iManufacturer;
    uint8_t iProduct;
    uint8_t iSerialNumber;
    uint8_t bNumConfigurations;

    // support only 1 configuration, so bNumConfigurations should be <= 1
    kdp_usbd_config_descriptor_t *config[MAX_USBD_CONFIG];

} kdp_usbd_device_descriptor_t;

typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint16_t bcdUSB;
    uint8_t bDeviceClass;
    uint8_t bDeviceSubClass;
    uint8_t bDeviceProtocol;
    uint8_t bMaxPacketSize0;
    uint8_t bNumConfigurations;
    uint8_t bReserved;
} kdp_usbd_device_qualifier_descriptor_t;

typedef enum
{
    KDP_USBD_HIGH_SPEED,
    KDP_USBD_FULL_SPEED // FIXME: not support yet
} kdp_usbd_speed_e;

typedef enum
{
    KDP_USBD_DEVICE_DESCRIPTORS = 0, /* include Configuration, Interface and Endpoint descriptors */
    KDP_USBD_STRING_DESCRIPTOR
} kdp_usbd_descriptor_type_e;

typedef enum
{
    KDP_USBD_EVENT_BUS_RESET = 1,
    KDP_USBD_EVENT_BUS_SUSPEND,
    KDP_USBD_EVENT_BUS_RESUME,
    KDP_USBD_EVENT_SETUP_PACKET,
    KDP_USBD_EVENT_DEV_CONFIGURED,
    KDP_USBD_EVENT_TRANSFER_BUF_FULL,
    KDP_USBD_EVENT_TRANSFER_DONE,
    KDP_USBD_EVENT_TRANSFER_OUT,
    KDP_USBD_EVENT_TRANSFER_TERMINATED,
    KDP_USBD_EVENT_DMA_ERROR,
} kdp_usbd_event_name_e;

// define a usbd event to notify user
// explain 'edata' as below:
// # 'configuration number' for KDP_USBD_EVENT_DEV_CONFIGURED
// # 'endpoint address' for KDP_USBD_EVENT_TRANSFER_BUF_FULL
// # 'endpoint address' for KDP_USBD_EVENT_TRANSFER_OUT
// # 'endpoint address' for KDP_USBD_EVENT_TRANSFER_DONE
typedef struct
{
    kdp_usbd_event_name_e ename;
    union {
        kdp_usbd_setup_packet_t setup;
        struct
        {
            uint32_t data1;
            uint32_t data2;
        };
    };

} kdp_usbd_event_t;

typedef enum
{
    KDP_USBD_RESPOND_OK,   // send ACK in the status stage
    KDP_USBD_RESPOND_ERROR // send STALL in the status stage
} kdp_usbd_status_report_e;

// init USB device mode
extern kdp_status_e kdp_usbd_init(void);

// de-init USB device mode
extern kdp_status_e kdp_usbd_deinit(void);

// reset the device and make it be re-enumerated
extern kdp_status_e kdp_usbd_reset_device(void);

// set up device, configuration, interface, endpoint descriptors descriptor
extern kdp_status_e kdp_usbd_set_device_descriptor(
    kdp_usbd_speed_e speed,
    kdp_usbd_device_descriptor_t *dev_desc);

// set device qualifier descriptor
extern kdp_status_e kdp_usbd_set_device_qualifier_descriptor(
    kdp_usbd_speed_e speed,
    kdp_usbd_device_qualifier_descriptor_t *dev_qual_desc);

// set thread flag for usb event notifications
extern kdp_status_e kdp_usbd_set_notify(osThreadId_t tid, uint32_t tflag);

// enable/disable USB device mode
// while enabled, the device can be enumerated by a host
// otherwise cannot be enumerated
extern kdp_status_e kdp_usbd_set_enable(kdp_bool_e enable);

extern kdp_bool_e kdp_usbd_is_dev_configured(void);

// get an usb event from internal event queue
extern kdp_status_e kdp_usbd_get_event(kdp_usbd_event_t *uevent);

// for a user-defined vendor reqeust & control IN & wLength > 0
// user should use this function to send data to host
// or respond an error via kdp_usbd_control_respond(KDP_USBD_RESPOND_ERROR) to claim STALL
extern kdp_status_e kdp_usbd_control_send_sync(
    uint8_t *buf,
    uint32_t size,
    uint32_t timeout_ms);

// for a user-defined vendor reqeust & control OUT & wLength > 0
// user should use this function to receive data from host
// or respond an error via kdp_usbd_control_respond(KDP_USBD_RESPOND_ERROR) to claim STALL
extern kdp_status_e kdp_usbd_control_receive_sync(
    uint8_t *buf,
    uint32_t *size,
    uint32_t timeout_ms);

// to report status for a user-defined vendor request
extern kdp_status_e kdp_usbd_control_respond(kdp_usbd_status_report_e status);

// reset endponit
extern kdp_status_e kdp_usbd_reset_endpoint(uint32_t endpoint);

// Bulk-In transfer
// send data to the host through bulk-in transfer
// blocking-wait API with timeout in millisecond
extern kdp_status_e kdp_usbd_bulk_send_sync(
    uint32_t endpoint, // should be the value from bEndpointAddress
    uint32_t *buf,     // memory address to be write to
    uint32_t txLen,    // transfer length
    uint32_t timeout_ms);

// Bulk-In transfer
// user can commit a buffer for Bulk In transfer
// and then wait for KDP_USBD_EVENT_TRANSFER_DONE
// to be notified that the transfer is done or some error code if failed
// txLen : trnasfer size, should be smaller or equal to buffer size
extern kdp_status_e kdp_usbd_bulk_send_async(
    uint32_t endpoint, // should be the value from bEndpointAddress
    uint32_t *buf,     // memory addres to be read from
    uint32_t txLen);   // transfer length

// Bulk-Out transfer
// receive data from the host through bulk-out transfer
// blocking-wait API with timeout in millisecond
extern kdp_status_e kdp_usbd_bulk_receive_sync(
    uint32_t endpoint, // should be the value from bEndpointAddress
    uint32_t *buf,     // memory address to be write to
    uint32_t *blen,    // input: buf len, output: actual received len
    uint32_t timeout_ms);

// Bulk-Out transfer
// when receiving a 'KDP_USBD_EVENT_TRANSFER_OUT' event,
// user should commit a buffer for Bulk Out transfer through this function
// when transfer is done by usbd, firmware will receive
// eihter a 'KDP_USBD_EVENT_TRANSFER_DONE' or 'KDP_USBD_EVENT_TRANSFER_BUF_FULL' event
// it is implemented in DMA interrupted-based way, no-wait
extern kdp_status_e kdp_usbd_bulk_receive_async(
    uint32_t endpoint, // should be the value from bEndpointAddress
    uint32_t *buf,     // memory address to be write to
    uint32_t blen      // buf len
);

// Interrupt-In transfer
// write data to the FIFO buffer for periodic interrupt-in transfer
// this is a blocking function
extern kdp_status_e kdp_usbd_interrupt_send_sync(
    uint32_t endpoint, // should be the value from bEndpointAddress
    uint32_t *buf,     // memory addres to be read from
    uint32_t txLen,    // transfer length, shoudl be less then MaxPacketSize
    uint32_t timeout_ms);

// Interrupt-Out transfer
// receive interupt out data with timeout
// this is a blocking function
extern kdp_status_e kdp_usbd_interrupt_receive_sync(
    uint32_t endpoint, // should be the value from bEndpointAddress
    uint32_t *buf,     // memory addres to be read from
    uint32_t *rxLen,   // input: buf len, output: actual received len, should be less than MaxPacketSize
    uint32_t timeout_ms);

#endif
