/*
 * KDP Camera driver for KL520
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include "config/board_kdp520.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cmsis_os2.h>
#include "kneron_mozart.h"
#include "base.h"
#include "utility.h"
#include "io.h"
#include "scu_extreg.h"
#include "framework/v2k.h"

#include "media/kdp_camera.h"
#include "media/kdp_fb_mgr.h"
#include "media/kdp_sensor.h"
#include "dbg.h"

//#define CAM_DEBUG

#ifdef CAM_DEBUG
#define cam_msg(fmt, ...) MSG(LOG_ERROR, "[%s] " fmt, __func__, ##__VA_ARGS__)
#else
#define cam_msg(fmt, ...)
#endif

#define CSI2RX_REG_VIDR         0x00
#define CSI2RX_REG_DIDR         0x01
#define CSI2RX_REG_CR           0x04 //Control Register
#define CSI2RX_REG_VSCR         0x05 //DPI V Sync Control Register
#define CSI2RX_REG_ECR          0x06
#define CSI2RX_REG_TCNR         0x08
#define CSI2RX_REG_HRTVR        0x0A //HS RX Timeout Value Register
#define CSI2RX_REG_FIUR         0x0B 
#define CSI2RX_REG_ITR          0x12 //Initialization Timer Register
#define CSI2RX_REG_VSTR0        0x14 //DPI VC0 V Sync Timing Register
#define CSI2RX_REG_HSTR0        0x15 //DPI VC0 H Sync Timing Register
#define CSI2RX_REG_VSTR1        0x16
#define CSI2RX_REG_HSTR1        0x17
#define CSI2RX_REG_VSTR2        0x18
#define CSI2RX_REG_HSTR2        0x19
#define CSI2RX_REG_VSTR3        0x1A
#define CSI2RX_REG_HSTR3        0x1B
#define CSI2RX_REG_MCR          0x1C //DPI Mapping Control Register
#define CSI2RX_REG_VSTER        0x1E
#define CSI2RX_REG_HPNR         0x20 //DPI Horizontal Pixel Number
#define CSI2RX_REG_PECR         0x28 //PPI Enable Control Register
#define CSI2RX_REG_DLMR         0x2A
#define CSI2RX_REG_CSIERR       0x30
#define CSI2RX_REG_INTSTS       0x33
#define CSI2RX_REG_ESR          0x34
#define CSI2RX_REG_DPISR        0x38
#define CSI2RX_REG_INTER        0x3C
#define CSI2RX_REG_FFR          0x3D
#define CSI2RX_REG_DPCMR        0x48
#define CSI2RX_REG_FRR          0x4C
#define CSI2RX_REG_PFTR         0x50
#define CSI2RX_REG_PUDTR        0x58
#define CSI2RX_REG_FRCR         0x80
#define CSI2RX_REG_FNR          0x88
#define CSI2RX_REG_BPGLR        0x90
#define CSI2RX_REG_BPGHR        0x91

#define MAX_VIDEO_MEM           8 << 20 

#define CSIRX_0_BASE            CSIRX_FTCSIRX100_PA_BASE
#define CSIRX_1_BASE            CSIRX_FTCSIRX100_1_PA_BASE

/* DPI V Sync Control Register (VSCR, Address = 0x05) */
#define CSIRX_REG_VSCR_SET_VSTU0(addr, val) SET_MASKED_BIT(addr, val, 0)

/* PPI Enable Control Register (PECR, Address = 0x28) */
#define CSIRX_REG_PECR_GET_PEC(addr)        GET_BITS(addr, 0, 7)
#define CSIRX_REG_PECR_SET_PEC(addr, val)   SET_MASKED_BITS(addr, val, 0, 7)

/* Data Lane Mapping Register 0 */
#define CSIRX_REG_0_DLMR0                   (CSIRX_0_BASE + 0x2A)
#define CSIRX_REG_0_DLMR0_GET_L1()          GET_BITS(CSIRX_REG_0_DLMR0, 4, 6)
#define CSIRX_REG_0_DLMR0_GET_L0()          GET_BITS(CSIRX_REG_0_DLMR0, 0, 2)
#define CSIRX_REG_1_DLMR0                   (CSIRX_1_BASE + 0x2A)
#define CSIRX_REG_1_DLMR0_GET_L1()          GET_BITS(CSIRX_REG_1_DLMR0, 4, 6)
#define CSIRX_REG_1_DLMR0_GET_L0()          GET_BITS(CSIRX_REG_1_DLMR0, 0, 2)


#define CSIRX_REG_0_FEATURE0                (CSIRX_0_BASE + 0x40)
#define CSIRX_REG_1_FEATURE0                (CSIRX_1_BASE + 0x40)
/* DPI FIFO Address Depth */
#define CSIRX_REG_0_FEATURE0_GET_DFAD()     GET_BITS(CSIRX_REG_0_FEATURE0, 5, 7)
#define CSIRX_REG_1_FEATURE0_GET_DFAD()     GET_BITS(CSIRX_REG_1_FEATURE0, 5, 7)
/* Lane Number */
#define CSIRX_REG_0_FEATURE0_GET_LN()       GET_BITS(CSIRX_REG_0_FEATURE0, 2, 4)
#define CSIRX_REG_1_FEATURE0_GET_LN()       GET_BITS(CSIRX_REG_1_FEATURE0, 2, 4)
/* Virtual channel number */
#define CSIRX_REG_0_FEATURE0_GET_VCN()      GET_BITS(CSIRX_REG_0_FEATURE0, 0, 1)
#define CSIRX_REG_1_FEATURE0_GET_VCN()      GET_BITS(CSIRX_REG_1_FEATURE0, 0, 1)

#define CSIRX_REG_0_FEATURE6                (CSIRX_0_BASE + 0x46)
#define CSIRX_REG_1_FEATURE6                (CSIRX_1_BASE + 0x46)
/* ASC Indicator */
#define CSIRX_REG_0_FEATURE6_GET_ASC()      GET_BIT(CSIRX_REG_0_FEATURE6, 7)
#define CSIRX_REG_1_FEATURE6_GET_ASC()      GET_BIT(CSIRX_REG_1_FEATURE6, 7)
/* APB Indicator */
#define CSIRX_REG_0_FEATURE6_GET_APB()      GET_BIT(CSIRX_REG_0_FEATURE6, 6)
#define CSIRX_REG_1_FEATURE6_GET_APB()      GET_BIT(CSIRX_REG_1_FEATURE6, 6)
/* I2C Indicator */
#define CSIRX_REG_0_FEATURE6_GET_I2C()      GET_BIT(CSIRX_REG_0_FEATURE6, 5)
#define CSIRX_REG_1_FEATURE6_GET_I2C()      GET_BIT(CSIRX_REG_1_FEATURE6, 5)
/* DPCM_EN Indicator */
#define CSIRX_REG_0_FEATURE6_GET_DPCM()     GET_BIT(CSIRX_REG_0_FEATURE6, 4)
#define CSIRX_REG_1_FEATURE6_GET_DPCM()     GET_BIT(CSIRX_REG_1_FEATURE6, 4)
/* CHK_DATA Indicator */
#define CSIRX_REG_0_FEATURE6_GET_CD()       GET_BIT(CSIRX_REG_0_FEATURE6, 3)
#define CSIRX_REG_1_FEATURE6_GET_CD()       GET_BIT(CSIRX_REG_1_FEATURE6, 3)
/* DUAL_PIXEL Indicator */
#define CSIRX_REG_0_FEATURE6_GET_DP()       GET_BIT(CSIRX_REG_0_FEATURE6, 2)
#define CSIRX_REG_1_FEATURE6_GET_DP()       GET_BIT(CSIRX_REG_1_FEATURE6, 2)
/* FRC Indicator */
#define CSIRX_REG_0_FEATURE6_GET_FRC()      GET_BIT(CSIRX_REG_0_FEATURE6, 1)
#define CSIRX_REG_1_FEATURE6_GET_FRC()      GET_BIT(CSIRX_REG_1_FEATURE6, 1)
/* DPI_PG Indicator */
#define CSIRX_REG_0_FEATURE6_GET_PG()       GET_BIT(CSIRX_REG_0_FEATURE6, 0)
#define CSIRX_REG_1_FEATURE6_GET_PG()       GET_BIT(CSIRX_REG_1_FEATURE6, 0)

#define CSIRX_REG_0_FEATURE7                (CSIRX_0_BASE + 0x47)
#define CSIRX_REG_1_FEATURE7                (CSIRX_1_BASE + 0x47)
/* OCTA_PIXEL Indicator */
#define CSIRX_REG_0_FEATURE7_GET_OP()       GET_BIT(CSIRX_REG_0_FEATURE7, 4)
#define CSIRX_REG_1_FEATURE7_GET_OP()       GET_BIT(CSIRX_REG_1_FEATURE7, 4)
/* MONO_RGB_EN Indicator */
#define CSIRX_REG_0_FEATURE7_GET_MR()       GET_BIT(CSIRX_REG_0_FEATURE7, 3)
#define CSIRX_REG_1_FEATURE7_GET_MR()       GET_BIT(CSIRX_REG_1_FEATURE7, 3)
/* CRC_CHK Indicator */
#define CSIRX_REG_0_FEATURE7_GET_CRC()      GET_BIT(CSIRX_REG_0_FEATURE7, 2)
#define CSIRX_REG_1_FEATURE7_GET_CRC()      GET_BIT(CSIRX_REG_1_FEATURE7, 2)
/* LANE_SWAP Indicator */
#define CSIRX_REG_0_FEATURE7_GET_SWAP()     GET_BIT(CSIRX_REG_0_FEATURE7, 1)
#define CSIRX_REG_1_FEATURE7_GET_SWAP()     GET_BIT(CSIRX_REG_1_FEATURE7, 1)
/* QUAD_PIXEL Indicator */
#define CSIRX_REG_0_FEATURE7_GET_QP()       GET_BIT(CSIRX_REG_0_FEATURE7, 0)
#define CSIRX_REG_1_FEATURE7_GET_QP()       GET_BIT(CSIRX_REG_1_FEATURE7, 0)


#define CSIRX_REG_0_BPGLR0                  (CSIRX_0_BASE + 0x90)
#define CSIRX_REG_1_BPGLR0                  (CSIRX_1_BASE + 0x90)
#define CSIRX_REG_0_BPGLR0_SET_VLN(val)     SET_MASKED_BITS(CSIRX_REG_0_BPGLR0, val, 0, 7)
#define CSIRX_REG_1_BPGLR0_SET_VLN(val)     SET_MASKED_BITS(CSIRX_REG_1_BPGLR0, val, 0, 7)

#define CSIRX_REG_0_BPGHR0                  (CSIRX_0_BASE + 0x91)
#define CSIRX_REG_1_BPGHR0                  (CSIRX_1_BASE + 0x91)
#define CSIRX_REG_0_BPGHR0_SET_PT(val)      SET_MASKED_BITS(CSIRX_REG_0_BPGHR0, val, 6, 7)
#define CSIRX_REG_1_BPGHR0_SET_PT(val)      SET_MASKED_BITS(CSIRX_REG_1_BPGHR0, val, 6, 7)
#define CSIRX_REG_0_BPGHR0_SET_PS(val)      SET_MASKED_BIT(CSIRX_REG_0_BPGHR0, val, 5)
#define CSIRX_REG_1_BPGHR0_SET_PS(val)      SET_MASKED_BIT(CSIRX_REG_1_BPGHR0, val, 5)
#define CSIRX_REG_0_BPGHR0_SET_GE(val)      SET_MASKED_BIT(CSIRX_REG_0_BPGHR0, val, 4)
#define CSIRX_REG_1_BPGHR0_SET_GE(val)      SET_MASKED_BIT(CSIRX_REG_1_BPGHR0, val, 4)
#define CSIRX_REG_0_BPGHR0_SET_VLN(val)     SET_MASKED_BITS(CSIRX_REG_0_BPGHR0, val, 0, 3)
#define CSIRX_REG_1_BPGHR0_SET_VLN(val)     SET_MASKED_BITS(CSIRX_REG_1_BPGHR0, val, 0, 3)

#define DPI2AHB_PAGE_NUM            2           // # of pages by one controller

#define D2A_REG_CTRL            0x00        // Control
#define D2A_REG_FNC             0x04        // Frame Number Control
#define D2A_REG_P0ADDR          0x08        // Page 0 Address
#define D2A_REG_P1ADDR          0x0C        // Page 1 Address
#define D2A_REG_ICT             0x10        // Interrupt Control
#define D2A_REG_IS              0x14        // Interrupt Status
#define D2A_REG_ST              0x18        // Status
#define D2A_REG_PT              0x1C        // Packet Type
#define D2A_REG_FIU0            0x20        // FI Use 0
#define D2A_REG_FIU1            0x24        // FI Use 1
#define D2A_REG_TAVR            0x28        // Tile Average Result (n)

/* D2A interrupt control & status register */
#define D2A_INT_TILE_AVG_D       BIT5
#define D2A_INT_AHB_TX_ERR       BIT4
#define D2A_INT_FIFO_UF          BIT3
#define D2A_INT_FIFO_OF          BIT2
#define D2A_INT_FN_OL            BIT1
#define D2A_INT_WRD              BIT0

#define D2A_INT_ALL              0x3F

/* D2A status register */
#define D2A_ST_PG               0x3

/* D2A Packet type register */
#define D2A_PT_YUV422           0x1E
#define D2A_PT_RGB565           0x22
#define D2A_PT_RGB888           0x24
#define D2A_PT_RAW8             0x2A
#define D2A_PT_RAW10            0x2B
#define D2A_PT_RAW12            0x2C
#define D2A_PT_RAW14            0x2D
#define D2A_PT_RAW16            0x2E

#define TILE_AVG_SIZE_128       0x00000000
#define TILE_AVG_SIZE_64        BIT16
#define TILE_AVG_SIZE_32        BIT17
#define TILE_AVG_SIZE_VAL       TILE_AVG_SIZE_128
#define TILE_AVG_SIZE_PIXELS    128

#define TILE_BLOCK_MAX_W        10
#define TILE_BLOCK_MAX_H        6
#define TILE_BLOCKS_MAX         (TILE_BLOCK_MAX_W * TILE_BLOCK_MAX_H)
#define TILE_REGS_MAX           (TILE_BLOCKS_MAX / 4)

struct kdp520_cam_context {
    int             id;
    int             irq;

    u32 dpi2ahb_base;
    u32 csi_rx_base;
    int page_done_num;
    int page_num;

    u32 sensor_id;

    struct cam_format fmt;
    u32 capabilities;

    int mipi_lane_num;

    int tile_avg_en;
    int tile_n_w;
    int tile_n_h;
    u32 tile_avgrs[TILE_REGS_MAX];
};

struct kdp520_cam_context cam_ctx[KDP_CAM_NUM];
static int buf0_idx[2] = {-1,-1};
static int buf1_idx[2] = {-1,-1};

#if defined(BOARD_96) || defined(BOARD_CUSTOMER1)
extern void gc2145_sensor_init(int cam_idx);
extern void sc132gs_sensor_init(int cam_idx);
#endif

static void camera_isr(struct kdp520_cam_context *ctx)
{   
    u32 sta_is, sta_st;
    u32 buf_addr[2];
    int cam_idx = ctx->id;

    sta_is = inw(ctx->dpi2ahb_base + D2A_REG_IS);

    if (sta_is & D2A_INT_TILE_AVG_D) {
        for (int i = 0; i < TILE_REGS_MAX; i++) {
            u32 tmp = inw(ctx->dpi2ahb_base + D2A_REG_TAVR + i * 4);
            ctx->tile_avgrs[i] = tmp;
        }
        // TODO: convert and pass to app provided buffer
    }

    if (sta_is & D2A_INT_WRD) {
        sta_st = inw(ctx->dpi2ahb_base + D2A_REG_ST);
        if (ctx->page_done_num != sta_st) // 1: page 1 done 2: page 0 done
        {
            if (sta_st == BIT0) {
                if (buf0_idx[cam_idx] != -1)
                    kdp_fb_mgr_write_done(cam_idx, buf0_idx[cam_idx]);
                buf_addr[cam_idx] =  kdp_fb_mgr_next_write(cam_idx, &buf0_idx[cam_idx]);
                outw(ctx->dpi2ahb_base + D2A_REG_P0ADDR, buf_addr[cam_idx]);
            } else {
                if (buf1_idx[cam_idx] != -1)
                    kdp_fb_mgr_write_done(cam_idx, buf1_idx[cam_idx]);
                buf_addr[cam_idx] =  kdp_fb_mgr_next_write(cam_idx, &buf1_idx[cam_idx]);
                outw(ctx->dpi2ahb_base + D2A_REG_P1ADDR, buf_addr[cam_idx]);
            }
            ctx->page_done_num = sta_st;
        }
    }   

    outw(ctx->dpi2ahb_base + D2A_REG_IS, sta_is);
}

void camera_isr_0(void)
{   
    struct kdp520_cam_context *ctx = &cam_ctx[KDP_CAM_0];

    camera_isr(ctx);
}

void camera_isr_1(void)
{
    struct kdp520_cam_context *ctx = &cam_ctx[KDP_CAM_1];

    camera_isr(ctx);
}

void dpi2ahb_enable(struct kdp520_cam_context *ctx)
{
    u32 val = 0;
    switch (ctx->sensor_id) {
    case SENSOR_TYPE_HMX2056:
        val |= 0x3000;
        break;
    case SENSOR_TYPE_OV9286:
        val |= 0x3008;
        break;
    case SENSOR_TYPE_HMXRICA:
        val |= 0x3008;
        break;
    case SENSOR_TYPE_GC2145:
        val |= 0x2800;
        break;
    case SENSOR_TYPE_SC132GS:
        val |= 0x3000;
        break;
    }

    if (ctx->tile_avg_en) {
        val |= TILE_AVG_SIZE_VAL;
        outw(ctx->dpi2ahb_base + D2A_REG_FNC, 0x2);
        outw(ctx->dpi2ahb_base + D2A_REG_CTRL, val);
        outw(ctx->dpi2ahb_base + D2A_REG_ICT, D2A_INT_WRD | D2A_INT_TILE_AVG_D);

        ctx->tile_n_w = (ctx->fmt.width + TILE_AVG_SIZE_PIXELS - 1) / TILE_AVG_SIZE_PIXELS;
        ctx->tile_n_h = (ctx->fmt.height + TILE_AVG_SIZE_PIXELS - 1) / TILE_AVG_SIZE_PIXELS;
        if (ctx->tile_n_w > TILE_BLOCK_MAX_W || ctx->tile_n_h >= TILE_BLOCK_MAX_H) {
            err_msg("[%s][%d] can't support %d x %d tiles\n", __func__, ctx->id, ctx->tile_n_w, ctx->tile_n_h);
            ctx->tile_avg_en = 0;
        }
    } else {
        outw(ctx->dpi2ahb_base + D2A_REG_FNC, 0x0);
        outw(ctx->dpi2ahb_base + D2A_REG_CTRL, val);
        outw(ctx->dpi2ahb_base + D2A_REG_ICT, D2A_INT_WRD);
    }

#ifdef SOURCE_FROM_PATTERN
    if (MIPI_CSI2RX_0 == ctx->id) {
    #if IMGSRC_0_FORMAT == IMAGE_FORMAT_YCBCR
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_YUV422);
    #elif IMGSRC_0_FORMAT == IMAGE_FORMAT_RGB565
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_RGB565);
    #elif IMGSRC_0_FORMAT == IMAGE_FORMAT_RAW10
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_RAW10);
    #elif IMGSRC_0_FORMAT == IMAGE_FORMAT_RAW8
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_RAW8);
    #endif
    }
    else if (MIPI_CSI2RX_1 == ctx->id) {
    #if IMGSRC_1_FORMAT == IMAGE_FORMAT_YCBCR
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_YUV422);
    #elif IMGSRC_1_FORMAT == IMAGE_FORMAT_RGB565
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_RGB565);
    #elif IMGSRC_1_FORMAT == IMAGE_FORMAT_RAW10
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_RAW10);
    #elif IMGSRC_1_FORMAT == IMAGE_FORMAT_RAW8
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_RAW8);
    #endif
    }
#else
    switch (ctx->fmt.pixelformat) {
    case V2K_PIX_FMT_YCBCR:
        cam_msg("V2K_PIX_FMT_YCBCR\n");
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_YUV422);
        break;        
    case V2K_PIX_FMT_RGB565:
        cam_msg("V2K_PIX_FMT_RGB565\n");
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_RGB565);
        break;
    case V2K_PIX_FMT_RAW10:	
        cam_msg("V2K_PIX_FMT_RAW10\n");
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_RAW10);
        break;
    case V2K_PIX_FMT_RAW8:
        cam_msg("V2K_PIX_FMT_RAW8\n");
        outw(ctx->dpi2ahb_base + D2A_REG_PT, D2A_PT_RAW8);
        break;
    default:;
    }
#endif

    outw(ctx->dpi2ahb_base + D2A_REG_FIU0, 0x1);
}

#define ROUND_UP(x, y) ((((x) + (y - 1)) / y) * y)

static void csi2rx_init(struct kdp520_cam_context *ctx)
{ 
    int val, width;

    width = ROUND_UP(ctx->fmt.width, 4);
    outw(ctx->csi_rx_base + CSI2RX_REG_HPNR , width);

    val = inw(ctx->csi_rx_base + CSI2RX_REG_CR);
    //val = (val&(~0xffff)) | 0x010d; 
    val = (val&(~0xff)) | 0x0d; 
    outw(ctx->csi_rx_base + CSI2RX_REG_CR , val);
    
    val = inw(ctx->csi_rx_base + CSI2RX_REG_MCR);
    val = (val&(~0xff)) | 0x00;//0x22;
    outw(ctx->csi_rx_base + CSI2RX_REG_MCR , val);

#ifdef SOURCE_FROM_PATTERN    
    CSIRX_REG_PECR_SET_PEC(ctx->csi_rx_base + CSI2RX_REG_PECR, 1);
    
    //outw(ctx->csi_rx_base + CSI2RX_REG_BPGLR, 0x01); 
    val = (inw(ctx->csi_rx_base + CSI2RX_REG_BPGLR) & 0xFF);
    cam_msg("CSI2RX_REG_BPGLR val=%x\n", val);
    
    val = (inw(ctx->csi_rx_base + CSI2RX_REG_BPGHR) & 0x0F);
        //val = 0;

    if (MIPI_CSI2RX_0 == ctx->id) {
    #if IMGSRC_0_FORMAT == IMAGE_FORMAT_RGB565
        val |= 0x40;         //RAW10
    #elif IMGSRC_0_FORMAT == IMAGE_FORMAT_RAW10
        val |= 0x40;         //RAW10
    #elif IMGSRC_0_FORMAT == IMAGE_FORMAT_RAW8
        val |= 0x00;         //RAW8
    #endif
    }
    else if (MIPI_CSI2RX_1 == ctx->id) {
    #if IMGSRC_1_FORMAT == IMAGE_FORMAT_RGB565
        val |= 0x40;         //RAW10
    #elif IMGSRC_1_FORMAT == IMAGE_FORMAT_RAW10
        val |= 0x40;         //RAW10
    #elif IMGSRC_1_FORMAT == IMAGE_FORMAT_RAW8
        val |= 0x00;         //RAW8
    #endif        
    }
    //val |= 0x20; //swap          
    outw(ctx->csi_rx_base + CSI2RX_REG_BPGHR, val); 
    val = inw(ctx->csi_rx_base + CSI2RX_REG_BPGHR);
#endif    

    switch (ctx->sensor_id) {
    case SENSOR_TYPE_OV9286:
    case SENSOR_TYPE_HMXRICA:
    case SENSOR_TYPE_SC132GS:
        //val = inw(ctx->csi_rx_base + CSI2RX_REG_PFTR);
        val = 160;//100; // * 4 = 400 pixels
        outw(ctx->csi_rx_base + CSI2RX_REG_PFTR , val);
        break;
    case SENSOR_TYPE_GC2145:
        CSIRX_REG_VSCR_SET_VSTU0(ctx->csi_rx_base, 1);
        break;
    default:;
    }
    
    
}

static void csi2rx_power(struct kdp520_cam_context *ctx, int on)
{
    uint32_t mask, val = 0;

    if (ctx->id == 0) {
        mask = (SCU_EXTREG_CSIRX_CTRL0_ClkLnEn | SCU_EXTREG_CSIRX_CTRL0_Enable);
        if (on)
            val = mask;

        masked_outw(SCU_EXTREG_CSIRX_CTRL0, val, mask);
    } else if (ctx->id == 1) {
        mask = (SCU_EXTREG_CSIRX_CTRL1_ClkLnEn | SCU_EXTREG_CSIRX_CTRL1_Enable);
        if (on)
            val = mask;

        masked_outw(SCU_EXTREG_CSIRX_CTRL1, val, mask);
    }
}

static void csi2rx_reset(struct kdp520_cam_context *ctx)
{
    if (ctx->id == 0) {
        masked_outw( SCU_EXTREG_CSIRX_CTRL0,
                    ((SCU_EXTREG_CSIRX_CTRL0_ClkLnEn | SCU_EXTREG_CSIRX_CTRL0_Enable) |
                     (SCU_EXTREG_CSIRX_CTRL0_apb_rst_n | SCU_EXTREG_CSIRX_CTRL0_pwr_rst_n |
                      SCU_EXTREG_CSIRX_CTRL0_sys_rst_n)),
                    ((SCU_EXTREG_CSIRX_CTRL0_ClkLnEn | SCU_EXTREG_CSIRX_CTRL0_Enable) |
                     (SCU_EXTREG_CSIRX_CTRL0_apb_rst_n | SCU_EXTREG_CSIRX_CTRL0_pwr_rst_n |
                      SCU_EXTREG_CSIRX_CTRL0_sys_rst_n)));

        masked_outw( SCU_EXTREG_CSIRX_CTRL1,
                    ((SCU_EXTREG_CSIRX_CTRL1_ClkLnEn | SCU_EXTREG_CSIRX_CTRL1_Enable) |
                     (SCU_EXTREG_CSIRX_CTRL1_apb_rst_n | SCU_EXTREG_CSIRX_CTRL1_pwr_rst_n |
                      SCU_EXTREG_CSIRX_CTRL1_sys_rst_n)),
                    ((SCU_EXTREG_CSIRX_CTRL1_ClkLnEn | SCU_EXTREG_CSIRX_CTRL1_Enable) |
                     (SCU_EXTREG_CSIRX_CTRL1_apb_rst_n | SCU_EXTREG_CSIRX_CTRL1_pwr_rst_n |
                      SCU_EXTREG_CSIRX_CTRL1_sys_rst_n)));

        masked_outw( SCU_EXTREG_DPI2AHB_CTRL,
                    (SCU_EXTREG_DPI2AHB_CTRL_rst_n | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n |
                     SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n),
                    (SCU_EXTREG_DPI2AHB_CTRL_rst_n | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n |
                     SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n));

        masked_outw( SCU_EXTREG_DPI2AHB_CTRL,
                    (SCU_EXTREG_DPI2AHB_CTRL_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n_1 |
                     SCU_EXTREG_DPI2AHB_CTRL_sys_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n_1),
                    (SCU_EXTREG_DPI2AHB_CTRL_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n_1 |
                     SCU_EXTREG_DPI2AHB_CTRL_sys_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n_1));

#if V2K_ENABLE_TYPE == V2K_ENABLE_GC2145 || V2K_ENABLE_TYPE == V2K_ENABLE_GC2145_SC132GS
        outb(MIPIRX_PHY_CSR_PA_BASE + 0x11, 0x7);
#else
        outb(MIPIRX_PHY_CSR_PA_BASE + 0x11, 0x3);
#endif
    } else if (ctx->id == 1) {
        masked_outw( SCU_EXTREG_CSIRX_CTRL0,
                    ((SCU_EXTREG_CSIRX_CTRL0_ClkLnEn | SCU_EXTREG_CSIRX_CTRL0_Enable) |
                     (SCU_EXTREG_CSIRX_CTRL0_apb_rst_n | SCU_EXTREG_CSIRX_CTRL0_pwr_rst_n |
                      SCU_EXTREG_CSIRX_CTRL0_sys_rst_n)),
                    ((SCU_EXTREG_CSIRX_CTRL0_ClkLnEn | SCU_EXTREG_CSIRX_CTRL0_Enable) |
                     (SCU_EXTREG_CSIRX_CTRL0_apb_rst_n | SCU_EXTREG_CSIRX_CTRL0_pwr_rst_n |
                      SCU_EXTREG_CSIRX_CTRL0_sys_rst_n)));

        masked_outw( SCU_EXTREG_CSIRX_CTRL1,
                    ((SCU_EXTREG_CSIRX_CTRL1_ClkLnEn | SCU_EXTREG_CSIRX_CTRL1_Enable) |
                     (SCU_EXTREG_CSIRX_CTRL1_apb_rst_n | SCU_EXTREG_CSIRX_CTRL1_pwr_rst_n |
                      SCU_EXTREG_CSIRX_CTRL1_sys_rst_n)),
                    ((SCU_EXTREG_CSIRX_CTRL1_ClkLnEn | SCU_EXTREG_CSIRX_CTRL1_Enable) |
                     (SCU_EXTREG_CSIRX_CTRL1_apb_rst_n | SCU_EXTREG_CSIRX_CTRL1_pwr_rst_n |
                      SCU_EXTREG_CSIRX_CTRL1_sys_rst_n)));

        masked_outw( SCU_EXTREG_DPI2AHB_CTRL,
                    (SCU_EXTREG_DPI2AHB_CTRL_rst_n | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n |
                     SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n),
                    (SCU_EXTREG_DPI2AHB_CTRL_rst_n | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n |
                     SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n));

        masked_outw( SCU_EXTREG_DPI2AHB_CTRL,
                    (SCU_EXTREG_DPI2AHB_CTRL_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n_1 |
                     SCU_EXTREG_DPI2AHB_CTRL_sys_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n_1),
                    (SCU_EXTREG_DPI2AHB_CTRL_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n_1 |
                     SCU_EXTREG_DPI2AHB_CTRL_sys_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n_1));

#if V2K_ENABLE_TYPE == V2K_ENABLE_SC132GS_GC2145
        outb(MIPIRX_PHY_CSR_1_PA_BASE + 0x11, 0x7);
#else
        outb(MIPIRX_PHY_CSR_1_PA_BASE + 0x11, 0x3);
#endif

        masked_outw( SCU_EXTREG_CSIRX_CTRL0,
                    (SCU_EXTREG_CSIRX_CTRL0_apb_rst_n | SCU_EXTREG_CSIRX_CTRL0_pwr_rst_n |
                     SCU_EXTREG_CSIRX_CTRL0_sys_rst_n),
                    (SCU_EXTREG_CSIRX_CTRL0_apb_rst_n | SCU_EXTREG_CSIRX_CTRL0_pwr_rst_n |
                     SCU_EXTREG_CSIRX_CTRL0_sys_rst_n));

        masked_outw( SCU_EXTREG_CSIRX_CTRL1,
                    (SCU_EXTREG_CSIRX_CTRL1_apb_rst_n | SCU_EXTREG_CSIRX_CTRL1_pwr_rst_n |
                     SCU_EXTREG_CSIRX_CTRL1_sys_rst_n ),
                    (SCU_EXTREG_CSIRX_CTRL1_apb_rst_n | SCU_EXTREG_CSIRX_CTRL1_pwr_rst_n |
                     SCU_EXTREG_CSIRX_CTRL1_sys_rst_n ));

        //dpi2ahb_1 reset
        masked_outw( SCU_EXTREG_DPI2AHB_CTRL,
                    (SCU_EXTREG_DPI2AHB_CTRL_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n_1 |
                     SCU_EXTREG_DPI2AHB_CTRL_sys_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n_1),

                    (SCU_EXTREG_DPI2AHB_CTRL_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n_1 |
                     SCU_EXTREG_DPI2AHB_CTRL_sys_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n_1));

#if V2K_ENABLE_TYPE == V2K_ENABLE_GC2145
        outb(MIPIRX_PHY_CSR_1_PA_BASE + 0x11, 0x7);
#else
        outb(MIPIRX_PHY_CSR_1_PA_BASE + 0x11, 0x3);
#endif
    }
}

/* API */

static int kdp520_cam_open(int cam_idx)
{
    struct kdp520_cam_context *ctx = &cam_ctx[cam_idx];

    cam_msg("cam: %d\n", cam_idx);

    csi2rx_power(ctx, 1);
    csi2rx_reset(ctx);

    return 0;
}

static void kdp520_cam_close(int cam_idx)
{
    struct kdp520_cam_context *ctx = &cam_ctx[cam_idx];

    cam_msg("cam: %d\n", cam_idx);
    csi2rx_power(ctx, 0);
}

static int kdp520_cam_query_capability(int cam_idx, struct cam_capability *cap)
{
    struct kdp520_cam_context *ctx = &cam_ctx[cam_idx];

    cam_msg("cam: %d\n", cam_idx);

    ctx->capabilities = V2K_CAP_VIDEO_CAPTURE | V2K_CAP_STREAMING | V2K_CAP_DEVICE_CAPS;

    strcpy(cap->driver, "kl520_camera");
    strcpy(cap->desc, "kl520_camera");
    cap->version = 0x00010001;
    cap->capabilities = ctx->capabilities;
    return 0;
}

static int kdp520_cam_set_format(int cam_idx, struct cam_format *format)
{
    struct kdp520_cam_context *ctx = &cam_ctx[cam_idx];
    int bpp;

    ctx->fmt = *format;

    if (format->pixelformat == V2K_PIX_FMT_RGB565)
        bpp = 2;
    else if (format->pixelformat == V2K_PIX_FMT_RAW8)
        bpp = 1;

    ctx->fmt.sizeimage = format->width * format->height * bpp;

    cam_msg("cam %d: w=%d h=%d p=0x%x f=%d b=%d s=%d c=%d\n", cam_idx,
            ctx->fmt.width, ctx->fmt.height, ctx->fmt.pixelformat, ctx->fmt.field,
            ctx->fmt.bytesperline, ctx->fmt.sizeimage, ctx->fmt.colorspace);

    dpi2ahb_enable(ctx);
    csi2rx_init(ctx);

    kdp_sensor_set_fmt(ctx->sensor_id, &ctx->fmt);

    NVIC_EnableIRQ((IRQn_Type)ctx->irq);
    return 0;
}

static int kdp520_cam_get_format(int cam_idx, struct cam_format *format)
{
    struct kdp520_cam_context *ctx = &cam_ctx[cam_idx];

    cam_msg("cam: %d\n", cam_idx);

    *format = ctx->fmt;
    return 0;
}

static int kdp520_cam_buffer_init(int cam_idx)
{
    struct kdp520_cam_context *ctx = &cam_ctx[cam_idx];

    cam_msg("cam %d: size=%d\n", cam_idx, ctx->fmt.sizeimage);

    kdp_fb_mgr_init(cam_idx, ctx->fmt.sizeimage, MAX_FRAME_BUFFER);

    outw(ctx->dpi2ahb_base + D2A_REG_P0ADDR, kdp_fb_mgr_next_write(cam_idx, &buf0_idx[cam_idx]));
    outw(ctx->dpi2ahb_base + D2A_REG_P1ADDR, kdp_fb_mgr_next_write(cam_idx, &buf1_idx[cam_idx]));
    outw(ctx->dpi2ahb_base + D2A_REG_FNC, 0);

    return 0;
}

static int kdp520_cam_start_capture(int cam_idx)
{
    struct kdp520_cam_context *ctx = &cam_ctx[cam_idx];
    int i;
    uint32_t val;

    cam_msg("cam: %d\n", cam_idx);

    val = 0;
    for (i = 0; i < ctx->mipi_lane_num; i++)
        val |= BIT0 << i;

    CSIRX_REG_PECR_SET_PEC(ctx->csi_rx_base + CSI2RX_REG_PECR, val);

    NVIC_EnableIRQ((IRQn_Type)ctx->irq);

    return 0;
}

static int kdp520_cam_stop_capture(int cam_idx)
{
    struct kdp520_cam_context *ctx = &cam_ctx[cam_idx];

    cam_msg("cam: %d\n", cam_idx);

    NVIC_DisableIRQ((IRQn_Type)ctx->irq);

    return 0;
}

static int kdp520_cam_buffer_prepare(int cam_idx)
{
    cam_msg("cam: %d\n", cam_idx);
    return 0;
}

static int kdp520_cam_buffer_capture(int cam_idx, int *addr, int *size)
{
    cam_msg("cam: %d\n", cam_idx);
    return 0;
}

static int kdp520_cam_stream_on(int cam_idx)
{
    cam_msg("cam: %d\n", cam_idx);
//    kdp_csi_rx_stream_on(cam_idx);
    return 0;
}

static int kdp520_cam_stream_off(int cam_idx)
{
    cam_msg("cam: %d\n", cam_idx);
//    kdp_csi_rx_stream_off(cam_idx);
    return 0;
}

static struct cam_ops kdp520_camera_ops = {
    .open               = kdp520_cam_open,
    .close              = kdp520_cam_close,
    .query_capability   = kdp520_cam_query_capability,
    .set_format         = kdp520_cam_set_format,
    .get_format         = kdp520_cam_get_format,
    .buffer_init        = kdp520_cam_buffer_init,
    .start_capture      = kdp520_cam_start_capture,
    .stop_capture       = kdp520_cam_stop_capture,
    .buffer_prepare     = kdp520_cam_buffer_prepare,
    .buffer_capture     = kdp520_cam_buffer_capture,
    .stream_on          = kdp520_cam_stream_on,
    .stream_off         = kdp520_cam_stream_off,
};

void sys_camera_global_init(void)
{
    SCU_EXTREG_CLK_EN1_SET_csirx0_TxEscClk(1);
    SCU_EXTREG_CLK_EN1_SET_csirx0_vc0(1);
    SCU_EXTREG_CLK_EN1_SET_csirx0_csi(1);

    SCU_EXTREG_CLK_EN1_SET_csirx1_TxEscClk(1);
    SCU_EXTREG_CLK_EN1_SET_csirx1_vc0(1);
    SCU_EXTREG_CLK_EN1_SET_csirx1_csi(1);

    masked_outw( SCU_EXTREG_CSIRX_CTRL0,
               ((SCU_EXTREG_CSIRX_CTRL0_ClkLnEn | SCU_EXTREG_CSIRX_CTRL0_Enable) |
                (SCU_EXTREG_CSIRX_CTRL0_apb_rst_n | SCU_EXTREG_CSIRX_CTRL0_pwr_rst_n |
                 SCU_EXTREG_CSIRX_CTRL0_sys_rst_n)),
               ((SCU_EXTREG_CSIRX_CTRL0_ClkLnEn | SCU_EXTREG_CSIRX_CTRL0_Enable) |
                (SCU_EXTREG_CSIRX_CTRL0_apb_rst_n | SCU_EXTREG_CSIRX_CTRL0_pwr_rst_n |
                 SCU_EXTREG_CSIRX_CTRL0_sys_rst_n)));

    masked_outw( SCU_EXTREG_CSIRX_CTRL1,
               ((SCU_EXTREG_CSIRX_CTRL1_ClkLnEn | SCU_EXTREG_CSIRX_CTRL1_Enable) |
                (SCU_EXTREG_CSIRX_CTRL1_apb_rst_n | SCU_EXTREG_CSIRX_CTRL1_pwr_rst_n |
                 SCU_EXTREG_CSIRX_CTRL1_sys_rst_n)),
               ((SCU_EXTREG_CSIRX_CTRL1_ClkLnEn | SCU_EXTREG_CSIRX_CTRL1_Enable) |
                (SCU_EXTREG_CSIRX_CTRL1_apb_rst_n | SCU_EXTREG_CSIRX_CTRL1_pwr_rst_n |
                 SCU_EXTREG_CSIRX_CTRL1_sys_rst_n)));

    masked_outw( SCU_EXTREG_DPI2AHB_CTRL,
               (SCU_EXTREG_DPI2AHB_CTRL_rst_n | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n |
                SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n),
               (SCU_EXTREG_DPI2AHB_CTRL_rst_n | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n |
                SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n));

    masked_outw( SCU_EXTREG_DPI2AHB_CTRL,
               (SCU_EXTREG_DPI2AHB_CTRL_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n_1 |
                SCU_EXTREG_DPI2AHB_CTRL_sys_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n_1),
               (SCU_EXTREG_DPI2AHB_CTRL_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_apb_rst_n_1 |
                SCU_EXTREG_DPI2AHB_CTRL_sys_rst_n_1 | SCU_EXTREG_DPI2AHB_CTRL_pwr_rst_n_1));
}

int kdp_camera_kl520_init(void)
{
    cam_msg("init\n");

    cam_ctx[KDP_CAM_0].irq = D2A_FTDPI2AHB_IRQ;
    cam_ctx[KDP_CAM_1].irq = D2A_FTDPI2AHB_1_IRQ;
    
    NVIC_SetVector((IRQn_Type)cam_ctx[KDP_CAM_0].irq, (u32)camera_isr_0);
    NVIC_SetVector((IRQn_Type)cam_ctx[KDP_CAM_1].irq, (u32)camera_isr_1);

    cam_ctx[KDP_CAM_0].dpi2ahb_base = DPI2AHB_CSR_PA_BASE;
    cam_ctx[KDP_CAM_1].dpi2ahb_base = DPI2AHB_CSR_1_PA_BASE;

    cam_ctx[KDP_CAM_0].csi_rx_base = CSIRX_FTCSIRX100_PA_BASE;
    cam_ctx[KDP_CAM_1].csi_rx_base = CSIRX_FTCSIRX100_1_PA_BASE;

    cam_ctx[KDP_CAM_0].id = KDP_CAM_0;
    cam_ctx[KDP_CAM_1].id = KDP_CAM_1;

    cam_ctx[KDP_CAM_0].mipi_lane_num = 2;
    cam_ctx[KDP_CAM_1].mipi_lane_num = 2;

    kdp_camera_controller_register(KDP_CAM_0, &kdp520_camera_ops);
    kdp_camera_controller_register(KDP_CAM_1, &kdp520_camera_ops);

    sys_camera_global_init();

#if defined(BOARD_96)

    cam_ctx[KDP_CAM_0].sensor_id = SENSOR_TYPE_GC2145;
    cam_ctx[KDP_CAM_1].sensor_id = SENSOR_TYPE_SC132GS;

    gc2145_sensor_init(KDP_CAM_0);
    sc132gs_sensor_init(KDP_CAM_1);

    cam_ctx[KDP_CAM_0].tile_avg_en = 0;
    cam_ctx[KDP_CAM_1].tile_avg_en = 1;

#elif defined(BOARD_CUSTOMER1)

    cam_ctx[KDP_CAM_0].sensor_id = SENSOR_TYPE_SC132GS;
    cam_ctx[KDP_CAM_1].sensor_id = SENSOR_TYPE_GC2145;

    sc132gs_sensor_init(KDP_CAM_0);
    gc2145_sensor_init(KDP_CAM_1);

    cam_ctx[KDP_CAM_0].tile_avg_en = 1;
    cam_ctx[KDP_CAM_1].tile_avg_en = 0;

#endif

    return 0;
}
