/*
 * KDP Frame Buffer driver
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include <string.h>
#include "cmsis_os2.h"
#include "config/board_kdp520.h"
#include "kdp_memory.h"
#include "media/kdp_fb_mgr.h"
#include "dbg.h"
#include "framework/irqflags.h"
#include "RTX_Config.h"

#define USE_GUARDBAND           4   // 1/n on both sides

#define INF_WAIT_MAX_TIME_MS    500
#define INF_WAIT_OS_DELAY_MS    10
#define INF_WAIT_OS_DELAY_TICKS (OS_TICK_FREQ * INF_WAIT_OS_DELAY_MS / 1000)

#define FOLLOW_INFERENCE_FIRST
#define READ_FROM_ISR
//#define FB_DEBUG
//#define FB_DEBUG2

#ifdef FB_DEBUG
#define fb_msg(fmt, ...) MSG(LOG_ERROR, "[%8.8d] " fmt, osKernelGetTickCount(), ##__VA_ARGS__)
#else
#define fb_msg(fmt, ...)
#endif

static struct kdp_fb_mgr_s {
    int         buf_size;
    int         buf_num;

    uint32_t    buf_addr[MAX_FRAME_BUFFER];
    uint32_t    buf_seq_num[MAX_FRAME_BUFFER];
    
    int         write_idx;      // camera
    int         write_next_idx; // camera
    int         write_done_idx; // camera

    int         inf_idx;        // inference
    int         inf_done_idx;   // inference

    int         read_idx;       // display

    uint32_t    write_count;
    uint32_t    read_count;
    uint32_t    inf_count;

    fb_write_done_notify    write_done_cb;
    bool        init_done;

} kdp_frame_buffer[IMGSRC_NUM];

static struct kdp_fb_mgr_s *fb_p[IMGSRC_NUM] = {
    &kdp_frame_buffer[0],
    &kdp_frame_buffer[1],
};

static inline void inc_write_next_idx(struct kdp_fb_mgr_s *fbp)
{
    fbp->write_next_idx = (fbp->write_next_idx + 1) % fbp->buf_num;
}

int kdp_fb_mgr_init(int cam_idx, uint32_t buf_size, int buf_num)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];
    int i, size;

    if(TRUE == fbp->init_done)
        return 0;

#if USE_GUARDBAND
    size = buf_size + (2 * buf_size / USE_GUARDBAND);
#else
    size = buf_size;
#endif

    if ((buf_num > MAX_FRAME_BUFFER) || (0 == buf_num))
        buf_num = MAX_FRAME_BUFFER;

    for (i = 0; i < buf_num; i++) {
        fbp->buf_addr[i] = kdp_ddr_reserve(size);
#if USE_GUARDBAND
        fbp->buf_addr[i] += (buf_size / USE_GUARDBAND);
#endif
        err_msg("  cam %d: frame buf[%d] : 0x%x\n", cam_idx, i, fbp->buf_addr[i]);
    }

    fbp->write_idx = fbp->read_idx = fbp->inf_idx = -1;
    fbp->write_done_idx = fbp->write_next_idx = fbp->inf_done_idx = -1;
    fbp->buf_size = buf_size;
    fbp->buf_num = buf_num;
    fbp->init_done = TRUE;
    return 0;
}

uint32_t kdp_fb_mgr_next_write(int cam_idx, int *write_idx)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    inc_write_next_idx(fbp);

    if (fbp->write_idx == -1)
        fbp->write_idx = fbp->write_next_idx;

    fb_msg("[%d]%s: w-next:%d (read_idx %d, inf-done:%d, inf:%d, w-done:%d)\n",
        cam_idx, __func__, fbp->write_next_idx, fbp->read_idx,
        fbp->inf_done_idx, fbp->inf_idx, fbp->write_done_idx);

    // skip the buffers being or could be used for inference/read
    if (fbp->write_next_idx == fbp->read_idx)
        inc_write_next_idx(fbp);
    if (fbp->write_next_idx == fbp->inf_done_idx)
        inc_write_next_idx(fbp);
    if (fbp->write_next_idx == fbp->inf_idx)
        inc_write_next_idx(fbp);
    if (fbp->write_next_idx == fbp->write_done_idx)
        inc_write_next_idx(fbp);

    fb_msg("[%d]%s: w-next:%d (final)\n",
        cam_idx, __func__, fbp->write_next_idx);

    if (write_idx)
        *write_idx = fbp->write_next_idx;

    local_irq_enable();

    return fbp->buf_addr[fbp->write_next_idx];
}

int kdp_fb_mgr_write_done(int cam_idx, int write_idx)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    fb_msg("[%d]%s: %d\n", cam_idx, __func__, write_idx);

    if (fbp->inf_idx != -1 ) {
        if (fbp->write_next_idx == fbp->read_idx) {
            err_msg("\n [%d]%s: [%d done] w-next:%d == read_idx (prev w-done:%d, inf:%d, inf-done:%d) !\n",
                cam_idx, __func__, write_idx, fbp->write_next_idx, fbp->write_done_idx,
                fbp->inf_idx, fbp->inf_done_idx);
        } else if (fbp->write_next_idx == fbp->inf_idx) {
            err_msg("\n [%d]%s: [%d done] w-next:%d == inf_idx %d !!\n",
                cam_idx, __func__, write_idx, fbp->write_next_idx, fbp->inf_idx);
        }
    }

    fbp->write_done_idx = write_idx;
    fbp->write_idx = fbp->write_next_idx;

#ifdef FB_DEBUG2
    if (fbp->write_count < 3) {
        uint32_t *addr0, *addr1, *addr2;

        addr0 = (uint32_t *)fbp->buf_addr[0] + 0x100;
        addr1 = (uint32_t *)fbp->buf_addr[1] + 0x100;
        addr2 = (uint32_t *)fbp->buf_addr[2] + 0x100;
        err_msg("[%d] write-done: %d: [0]=0x%x, [1]=0x%x, [2]=0x%x\n",
            cam_idx, write_idx, *addr0, *addr1, *addr2);
        *addr0 = 1;
        *addr1 = 2;
        *addr2 = 4;
    }
#endif

    /* Mark with sequential number */
    fbp->write_count++;
    fbp->buf_seq_num[write_idx] = fbp->write_count;

    if (fbp->write_done_cb)
        fbp->write_done_cb(cam_idx, write_idx);

    local_irq_enable();

    return 0;
}

void kdp_fb_mgr_free_write_buf(int cam_idx)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];
    int i;

    local_irq_disable();

    fbp->write_idx = -1;
    fbp->write_next_idx = -1;
    fbp->write_done_idx = -1;

#ifdef FB_DEBUG2
    fbp->write_count = 0;   // for test purposes

    memset((void *)fbp->buf_addr[0], 0, fbp->buf_size);
    memset((void *)fbp->buf_addr[1], 0, fbp->buf_size);
    memset((void *)fbp->buf_addr[2], 0, fbp->buf_size);
#else
    memset((void *)fbp->buf_addr[0], 0, fbp->buf_size);
    memset((void *)fbp->buf_addr[1], 0, fbp->buf_size);
#endif

    for (i = 0; i < fbp->buf_num; i++) {
        fbp->buf_seq_num[i] = 0;
    }

    fb_msg("[%d]%s\n", cam_idx, __func__);

    local_irq_enable();
}

uint32_t kdp_fb_mgr_get_buf_seq_num(int cam_idx, int idx)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];

    return fbp->buf_seq_num[idx];
}

uint32_t kdp_fb_mgr_get_buf(int cam_idx, int idx)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    idx %= fbp->buf_num;
    uint32_t buf_addr = fbp->buf_addr[idx];
    memset((void *)buf_addr, 0, fbp->buf_size);

    fb_msg("[%d]%s: %d\n", cam_idx, __func__, idx);

    local_irq_enable();

    return buf_addr;
}

uint32_t kdp_fb_mgr_next_read(int cam_idx, int *read_idx)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];
#ifndef READ_FROM_ISR
    int try_max_count = INF_WAIT_MAX_TIME_MS / INF_WAIT_OS_DELAY_MS;

next_read_buf:
#endif
    local_irq_disable();

#ifdef FOLLOW_INFERENCE_FIRST
    if (fbp->inf_done_idx != -1) {
        // keep displaying the inference done buffer
        fbp->read_idx = fbp->inf_done_idx;
        fb_msg("[%d]%s: %d (inf-done)\n", cam_idx, __func__, fbp->read_idx);
    } else if (fbp->write_done_idx != -1) {
#else // Just follow write-done
    if (fbp->write_done_idx != -1) {
#endif
        fbp->read_idx = fbp->write_done_idx;
        fb_msg("[%d]%s: %d (w-done)\n", cam_idx, __func__, fbp->read_idx);
#ifdef FB_DEBUG2
        if (fbp->write_count < 3) {
            err_msg("[%d]%s: %d (w-done)\n", cam_idx, __func__, fbp->read_idx);
        }
#endif
    } else {
#ifndef READ_FROM_ISR
        // let's wait for next available one
        if (try_max_count-- >= 0) {
            local_irq_enable();
            osDelay(INF_WAIT_OS_DELAY_TICKS);
            goto next_read_buf;
        }

        if (fbp->read_idx == -1)
            fbp->read_idx = 0;
        fb_msg("[%d]%s: reuse read_idx:%d after waiting %d ms\n",
                cam_idx, __func__, fbp->read_idx, INF_WAIT_MAX_TIME_MS);
#else
        if (fbp->read_idx == -1)
            fbp->read_idx = 0;
#endif

#ifdef FB_DEBUG2
        if (fbp->write_count < 3) {
            err_msg("[%d]%s: reuse read_idx:%d after waiting %d ms\n",
                cam_idx, __func__, fbp->read_idx, INF_WAIT_MAX_TIME_MS);
        }
#endif
    }

    if (read_idx)
        *read_idx = fbp->read_idx;

    fbp->read_count++;

    local_irq_enable();

    return fbp->buf_addr[fbp->read_idx];
}

void kdp_fb_mgr_free_read_buf(int cam_idx)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    fbp->read_idx = -1;

    fb_msg("[%d]%s\n", cam_idx, __func__);

    local_irq_enable();
}

uint32_t kdp_fb_mgr_next_inf(int cam_idx, int *inf_idx)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];
    int try_max_count = INF_WAIT_MAX_TIME_MS / INF_WAIT_OS_DELAY_MS;

next_inf_buf:
    local_irq_disable();

    // inference uses the latest write_done_idx
    if (fbp->write_done_idx == fbp->inf_done_idx) {
        // latest write_done_idx is already used. let's wait for new one
        if (try_max_count-- >= 0) {
            local_irq_enable();
            osDelay(INF_WAIT_OS_DELAY_TICKS);
            goto next_inf_buf;
        }
        fb_msg("[%d]%s: reuse w-done:%d after waiting %d ms\n",
                cam_idx, __func__, fbp->write_done_idx, INF_WAIT_MAX_TIME_MS);
    } else {
        fb_msg("[%d]%s: %d\n", cam_idx, __func__, fbp->write_done_idx);
    }

    if (fbp->inf_done_idx != -1) {
        int new_buffers = fbp->write_done_idx - fbp->inf_done_idx;
        if (new_buffers < 0)
            new_buffers += fbp->buf_num;

        if (new_buffers > 1) {
            fb_msg("[%d]%s: skip %d buffers: from %d to %d\n",
                cam_idx, __func__, new_buffers - 1,
                fbp->inf_done_idx, fbp->write_done_idx);
        }
    }

    // inference uses the latest write_done_idx
    fbp->inf_idx = fbp->write_done_idx;

    if (inf_idx)
        *inf_idx = fbp->inf_idx;

    local_irq_enable();

    return fbp->buf_addr[fbp->inf_idx];
}

int kdp_fb_mgr_inf_done(int cam_idx, int inf_idx)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    fbp->inf_done_idx = inf_idx;
    fbp->inf_count++;

    fb_msg("[%d]%s: %d\n", cam_idx, __func__, inf_idx);

    local_irq_enable();

    return 0;
}

void kdp_fb_mgr_free_inf_buf(int cam_idx)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];

    local_irq_disable();

    fbp->inf_idx = -1;
    fbp->inf_done_idx = -1;

    fb_msg("[%d]%s\n", cam_idx, __func__);

    local_irq_enable();
}

int kdp_fb_mgr_notifier_register(int cam_idx, fb_write_done_notify callback)
{
    struct kdp_fb_mgr_s *fbp = fb_p[cam_idx];

    fbp->write_done_cb = callback;
    return 0;
}
