/*
 * KDP Inference support driver
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include <string.h>
#include "framework/event.h"
#include "media/display/lcdc.h"
#include "media/display/kdp_display.h"
#include "media/display/kdp_2d.h"
#include "framework/result.h"
#include "framework/v2k_color.h"
#include "kdp_app.h"
#include "kdp_app_db.h"
#include "kdp_app_fid.h"
#include "kdp_app_lw3d.h"
#include "kdp_memory.h"
#include "media/kdp_fb_mgr.h"
#include "media/kdp_camera.h"
#include "media/kdp_inference.h"
#include "dbg.h"
#include "kl520_include.h"
#include "power_manager.h"
#include "media/camera/sensor.h"
#include "kdp520_pwm_timer.h"

static int g_buf_idx[2] = {-1, -1};

#define NUM_REG_INFERENCE 				    (20)
#define FM_SIZE                             (512)
#define FM_SIZE_BYTE                        (FM_SIZE * 4)
#define LM_SIZE								(10)
#define LM_SIZE_BYTE					    (LM_SIZE * 4)	
#define REG_TEMP_FR_SIZE					(FM_SIZE_BYTE * NUM_REG_INFERENCE)
#define REG_TEMP_LM_SIZE					(LM_SIZE_BYTE * NUM_REG_INFERENCE)

#define KL520_THRESH_NORMAL_MAX             (30)
#define KL520_THRESH_LEFT_MIN               (55)
#define KL520_THRESH_LEFT_MAX               (80)
#define KL520_THRESH_RIGHT_MIN              (55)
#define KL520_THRESH_RIGHT_MAX              (80)
#define KL520_THRESH_UP_MIN                 (50)
#define KL520_THRESH_UP_MAX                 (65)
#define KL520_THRESH_DOWN_MIN               (50)
#define KL520_THRESH_DOWN_MAX               (65)

#define KL520_MIDDLE_LINE					(60)

#define KDP_INF_WITHOUT_RESULT_ACK          (0)
#define KDP_INF_WITH_RESULT_ACK             (1)

static uint8_t face_reg_thresh[] = {0, KL520_THRESH_NORMAL_MAX, KL520_THRESH_LEFT_MIN,  KL520_THRESH_LEFT_MAX,
                             KL520_THRESH_RIGHT_MIN, KL520_THRESH_RIGHT_MAX, KL520_THRESH_UP_MIN, KL520_THRESH_UP_MAX,
                             KL520_THRESH_DOWN_MIN, KL520_THRESH_DOWN_MAX
                            };

static kdp_app_fid_inout_t     *m_fdr_inout = NULL;
static kdp_app_face_det_res_t  *m_fd_out  = NULL;
static kdp_app_face_fmap_t     *m_fr_out  = NULL;
static kdp_app_db_user_data_t  *m_db_data = NULL;
static kdp_app_face_fmap_t     *m_fr_1_out  = NULL;
static uint32_t                 *db_tmp_lm = 0;
static float                    *db_tmp_fr = 0;
static uint32_t                 *db_tmp_1_lm = 0;
static float                    *db_tmp_1_fr = 0;

static bool db_tmp_valid_pose[NUM_REG_INFERENCE];
static bool g_fdfr_run_inference = FALSE;

/* For fdfr inference result */
struct kdp_rect m_boundingbox = { 0 };
struct kdp_rect s_boundingbox = { 0 };
/* For face hint boundingbox */
struct kdp_rect h_boundingbox = { 0 };

osThreadId_t tid_inference = NULL;
int g_draw_boundingbox_mode = ALL_DRAW_BOUNDINGBOX_MODE;
int g_frame_margin_length = LCD_HINT_BOUNDINGBOX_MARGIN_LEN;
int m_face_add_timeout = 0;
int m_face_recognition_timeout = 0;
u32 m_fdfr_tick = 0;

int reg_type = 0;
int reg_count = 0;

kdp_all_comm_params m_kdp_all_comm_params;

kdp_camera_ctx m_kdp_cam_ctx[IMGSRC_NUM + 2] = {
    {FALSE, CAMERA_STATE_IDLE, 0, 0, 0},
    {FALSE, CAMERA_STATE_IDLE, 0, 0, 0},
};


/* Middleware layer */
/*-----------------------------------------------------------------------------------------
 *                                  Private power manager API 
 *-----------------------------------------------------------------------------------------*/
int ncpu_inference_nap(enum pm_device_id dev_id)
{
    if(g_fdfr_run_inference == TRUE)
        return -1;
    else
        return 0;
}

int ncpu_inference_deep_nap(enum pm_device_id dev_id)
{
    if(g_fdfr_run_inference == TRUE)
        return -1;
    else
        return 0;
}

int ncpu_inference_sleep(enum pm_device_id dev_id)
{
    if(g_fdfr_run_inference == TRUE)
        return -1;
    else
        return 0;
}

int ncpu_inference_deep_sleep(enum pm_device_id dev_id)
{
    if(g_fdfr_run_inference == TRUE)
        return -1;
    else
        return 0;
}

struct pm_s pm_ncpu_inference =
{
    .nap = ncpu_inference_nap,
    .wakeup_nap = 0,
    .deep_nap = ncpu_inference_deep_nap,
    .wakeup_deep_nap = 0,
    .sleep = 0,
    .wakeup_sleep = ncpu_inference_sleep,
    .deep_sleep = 0,
    .wakeup_deep_sleep = ncpu_inference_deep_sleep,
};

/*-----------------------------------------------------------------------------------------
 *                                  Private kdp draw API 
 *-----------------------------------------------------------------------------------------*/
static void _kdp_draw_fdfr_result(fdr_result fdr_result, u32 buf_addr)
{
     if (FDR_RET_OK <= fdr_result) {
        if (fdr_result == FDR_RET_OK) {
            kdp_display_set_pen_rgb565(YELLOW, 4);
        } else if (fdr_result == FDR_RET_DB_COMP_OK) {
            kdp_display_set_pen_rgb565(GREEN, 4);
        } else {
            kdp_display_set_pen_rgb565(RED, 4);
        }

        if(fdr_result != FDR_RET_NO_FACE) {
            if(g_draw_boundingbox_mode & DRAW_FDR_RESULT_BOUNDINGBOX) {
                kdp_display_draw_rect(buf_addr, m_boundingbox.start_x, m_boundingbox.start_y, m_boundingbox.end_x - m_boundingbox.start_x, m_boundingbox.end_y - m_boundingbox.start_y, DRAW_FDR_RESULT_BOUNDINGBOX);
                memset((void *)&m_boundingbox, 0 , sizeof(m_boundingbox));
            }
        }

        if(g_draw_boundingbox_mode & DRAW_HINT_BOUNDINGBOX)  {
            if(g_draw_boundingbox_mode == DRAW_HINT_BOUNDINGBOX)
            {
                if (fdr_result == FDR_RET_OK) {
                    kdp_display_set_pen_rgb565(YELLOW, 8);
                } else if (fdr_result == FDR_RET_DB_COMP_OK) {
                    kdp_display_set_pen_rgb565(GREEN, 8);
                } else {
                    kdp_display_set_pen_rgb565(RED, 8);
                }
            }
            else if(g_draw_boundingbox_mode == ALL_DRAW_BOUNDINGBOX_MODE)
                kdp_display_set_pen_rgb565(BLUE, 8);

            kdp_display_draw_rect(buf_addr, s_boundingbox.start_x, s_boundingbox.start_y, s_boundingbox.end_x - s_boundingbox.start_x, s_boundingbox.end_y - s_boundingbox.start_y, DRAW_HINT_BOUNDINGBOX);
        }
    }
}

/*-----------------------------------------------------------------------------------------
 *                                  Private kdp camera API 
 *-----------------------------------------------------------------------------------------*/
static int _kdp_camera_open(uint8_t cam_idx)
{
#ifdef V2K_ENABLE
    int ret;
    unsigned int width = 0;
    unsigned int height = 0;
    unsigned int pixelformat;
    
    struct cam_capability cap;
    struct cam_format fmt;

    char fmtstr[8];
    memset(&cap, 0, sizeof(cap));
    memset(&fmt, 0, sizeof(fmt));

    pixelformat = m_kdp_cam_ctx[cam_idx].pixelformat;
    width = m_kdp_cam_ctx[cam_idx].width;
    height = m_kdp_cam_ctx[cam_idx].height;
    dbg_msg("[%s] cam_idx:%d, width:%d, height:%d", __func__, cam_idx, width, height);

    if (0 != (ret = kdp_camera_open(cam_idx)))
        return ret;

    if (0 != (ret = kdp_camera_get_device_info(cam_idx, &cap)))
        return ret;

    fmt.width = width;
    fmt.height = height;
    fmt.pixelformat = pixelformat;
    if (0 != (ret = kdp_camera_set_frame_format(cam_idx, (struct cam_format *)&fmt)))
        return ret;

    if (0 != (ret = kdp_camera_get_frame_format(cam_idx, &fmt)))
        return ret;    

    memset(fmtstr, 0, 8); 
    memcpy(fmtstr, &fmt.pixelformat, 4); 

    if (0 != (ret = kdp_camera_buffer_init(cam_idx)))
        return ret;

    m_kdp_cam_ctx[cam_idx].state = CAMERA_STATE_INITED;
    return 0;
#endif
}

static int _kdp_camera_start(unsigned int cam_idx)
{
	int ret = -1;

    if (0 != (ret = kdp_camera_start(cam_idx))) {
        return ret;
    }

    m_kdp_cam_ctx[cam_idx].state = CAMERA_STATE_RUNNING;
    return 0;
}

static int _kdp_camera_stop(unsigned int cam_idx)
{
		int ret = -1;
    if(0 != (ret = kdp_camera_stop(cam_idx)))
        return ret;

    m_kdp_cam_ctx[cam_idx].state = CAMERA_STATE_INITED;
    return 0;
}

static int _kdp_camera_close(unsigned int cam_idx)
{
    int ret = -1;

    if(0 != (ret = kdp_camera_close(cam_idx)))
        return ret;

    m_kdp_cam_ctx[cam_idx].state = CAMERA_STATE_IDLE;
    return ret;
}


/*-----------------------------------------------------------------------------------------------------------------------------------
 *                                          Private kdp FDFR API
 * From sample_app_fdr.c
 *
 * sample_kneron_fdr_inference(void)                      -> _kdp_fdfr_inference(void)
 * _sample_kneron_fdr_both_only(struct kdp_rect *lp_rect) -> _kdp_fdfr_both_only(struct kdp_rect *lp_rect)
 * sample_kneron_fdr_both_inference(void)                 -> _kdp_fdfr_both_inference(void)
 *
 * _sample_kneron_reg_fdr_only(struct kdp_rect *lp_rect, int index)     (Combine together, use parameter "num_rgst" to distinguish them)
 * _sample_kneron_reg_fdr_both_only(struct kdp_rect *lp_rect, int index) -> _kdp_rgst_fdfr_only(struct kdp_rect *lp_rect, int index, int num_rgst)
 * sample_kneron_fdr_registration(u32 curr_user_idx)                     -> _kdp_fdfr_registration(uint32_t user_idx)
 * sample_kneron_5face_registration(u32 curr_user_idx, int type)         -> _kdp_5_face_registration(uint32_t user_idx, int type)
 * sample_kneron_fdr_auto_reg(u32 curr_user_idx, int reg_index)         (Combine together, use parameter "num_rgst" to distinguish them)
 * sample_kneron_fdr_both_reg(u32 curr_user_idx, int reg_index)          -> _kdp_fdfr_auto_registeration(uint32_t user_idx, int reg_index, int num_reg)
 * sample_kneron_fdr_registration_confirm(u32 curr_user_idx)            (Combine together, use parameter "num_rgst" to distinguish them)
 * sample_kneron_fdr_both_registration_confirm(u32 curr_user_idx)        ->_kdp_fdfr_registration_confirm(uint32_t user_idx, int num_rgst);
 *
 * sample_kneron_fdr_add_user(u32 curr_user_idx)                    ->  _kdp_fdfr_add_user(uint32_t user_idx)
 * sample_kneron_fdr_del_user(unsigned char all, u32 curr_user_idx) ->  _kdp_fdfr_del_user(bool all, uint32_t user_idx)
 * sample_kneron_fdr_cancel(void)                                   ->  _kdp_fdfr_cancel_registeration(void)
 * sample_kneron_fdr_abort(void)                                    ->  _kdp_fdfr_abort(void)
 *-----------------------------------------------------------------------------------------------------------------------------------*/
#define INIT_TILE2
#define Max_EXPO_TIME 43072
uint32_t nir_frm_cnt = 0;
static int rgb_fd_get_flag = 10;
static u32 m_iCurExpoTime = 43072;	
u8 tile_val[20];
static u8 distance_flag = 10;
//static u8 CurLedPwm = 0;
//static u8 TgtLedPwm = 99;

int _kdp_check_DB_space(uint32_t user_idx)
{
    if ((user_idx + 1) > DB_SLOT_SIZE)/*User id(curr_user_idx) starts from 0*/
    {
        info_msg("[registration] : No space\n");
        return KDP_APP_UNKNOWN_ERR;
    }
    return 0;
}

fdr_result _kdp_fdfr_inference(void)
{
    fdr_result fdr_ret = FDR_RET_OOPS;
    int ret = -1;
    int32_t lw3d_ret;
    struct kdp_rect *lp_rect = &m_boundingbox;

    m_fdr_inout->op = KDP_APP_FID_OP_INFERENCE;
    m_fdr_inout->thresh_fid = 0.425;

    do {
        //todo dummy
        memset(m_fd_out, 0, sizeof(kdp_app_face_det_res_t));
        memset(m_fr_out, 0, sizeof(kdp_app_face_fmap_t));   
        
        // get exp time
        u16 a = 0;
        u8 b, c;
        //todo need rename
        rgb_get_lux( &a , &b , &c );
        //dbg_msg("exp_rgb_time = %d", a);
        
        ret = kdp_app_fdr(m_fdr_inout, m_fd_out, NULL, m_fr_out);
        
        if (ret == KDP_APP_FDR_NO_FACE ){
            info_msg("[inference %d] FD result : NO FACE\n", __LINE__);
            // if no face, aec_roi set as 480x480 rgb right img 
            if (rgb_fd_get_flag == 0){
                //dbg_msg(" rgb_fd_get_flag  = 0 ");
            }else{
                //todo need rename
                rgb_fd_get_flag = 0;
                rgb_set_aec_roi(20, 80, 0, 60, 30, 70, 10, 50);
                //dbg_msg("set rgb_fd_get_flag = 0 ");
            }    
        }else if (ret == KDP_APP_FDR_WRONG_USAGE ||   ret == KDP_APP_FDR_BAD_POSE || ret == KDP_APP_FDR_NO_ALIVE || ret == KDP_APP_FDR_FR_FAIL ||
         ret == KDP_APP_LM_LOW_CONFIDENT || ret == KDP_APP_LM_BLUR ){
         info_msg("other than KDP_APP_FDR_NO_FACE ret : %d\n", ret);  
    
        }else{ // if get face, aec_roi set as fd region.
            lp_rect->start_x = (u16) m_fd_out->xywh[0];
            lp_rect->start_y = (u16) m_fd_out->xywh[1];
            lp_rect->end_x = ((u16) m_fd_out->xywh[0]) + ((u16) m_fd_out->xywh[2]);
            lp_rect->end_y = ((u16) m_fd_out->xywh[1]) + ((u16) m_fd_out->xywh[3]);
          
            u8 xmin_inter = (u8) (m_fd_out->xywh[0] >> 3);
            u8 xmax_inter = (u8) ( (m_fd_out->xywh[0] + m_fd_out->xywh[2]) >> 3 );
            u8 ymin_inter = (u8) (m_fd_out->xywh[1] >> 3);
            u8 ymax_inter = (u8) ( (m_fd_out->xywh[1] + m_fd_out->xywh[3]) >> 3 );
            
            if(xmin_inter >= 20){ // if get face and fd region in rgb right region(480x480)
        
                if (rgb_fd_get_flag == 1){
                    //dbg_msg(" rgb_fd_get_flag = 1 ");
                }else{
                    rgb_fd_get_flag = 1;
                    rgb_set_aec_roi(20, 80, 0, 60, xmin_inter, xmax_inter, ymin_inter, ymax_inter);
                }
            }else{ // if fd region not in rgb right region (480x480)
                
                if (rgb_fd_get_flag == 0){
                    //dbg_msg(" rgb_fd_get_flag  = 0 ");
                }else{
                    rgb_fd_get_flag = 0;
                    rgb_set_aec_roi(20, 80, 0, 60, 30, 70, 10, 50);
                    //dbg_msg("set rgb_fd_get_flag = 0 ");
                }
            }
            int DIST_W = m_fd_out->xywh[2];
            //dbg_msg("DIST_W = %d \n", DIST_W);
            if( DIST_W >= 256 ){  // short distance  
                if( distance_flag == 0 ){
                    dbg_msg(" nir gain as 6 \n");
                    //dbg_msg(" nir gain as 6 \n");
                }else{
                    distance_flag = 0;
                    nir_set_gain(0x27, 0x35); // nir set gain as 6
                    dbg_msg("set nir gain = 6 \n");
                }
            }
            else if ( DIST_W < 256 ){ // long distance
                if( distance_flag == 1 ){
                    dbg_msg(" nir gain as 12 \n");
                    //dbg_msg(" nir gain as 12 \n");
                }else{
                    distance_flag = 1;
                    nir_set_gain(0x2F, 0x35); // nir set gain as 12
                    dbg_msg("set nir gain = 12\n");
                }
            }
        }
        
        lw3d_ret = kdp_app_lw3d_fdr(distance_flag);
        if (KDP_APP_OK == ret ) {
            if ( KDP_APP_OK == lw3d_ret) {
                fdr_ret = FDR_RET_DB_COMP_OK;
                info_msg("[Green] UID : %d\n", m_fdr_inout->user_id);
            }else {
                dbg_msg("[inference] LW3D Failed\n");
                fdr_ret = FDR_RET_OK;
                info_msg("[Yellow] UID : %d\n", m_fdr_inout->user_id);
            }
		}
        else
        {
            fdr_ret = FDR_RET_DB_COMP_FAIL;
        }
    } while (0);

    if(ret == KDP_APP_FDR_NO_FACE)
        fdr_ret = FDR_RET_NO_FACE;
 
    return fdr_ret;
}

static int _kdp_fdfr_both_only(struct kdp_rect *lp_rect)
{
    int ret, lw3d_ret;
    int32_t fdr_ret = FDR_RET_OOPS;
    do {
        m_fdr_inout->op = KDP_APP_FID_OP_EXTRACT;
        //ret = kdp_app_fdr_only(m_fdr_inout, m_fd_out, NULL, m_fr_out);
        ret = kdp_app_fdr(m_fdr_inout, m_fd_out, NULL, m_fr_out);
        dbg_msg("[%s] ret=0x%x\n", __func__, ret);
        if (KDP_APP_FDR_NO_FACE != ret)
        {
            lp_rect->start_x = (u16) m_fd_out->xywh[0];
            lp_rect->start_y = (u16) m_fd_out->xywh[1];
            lp_rect->end_x = ((u16) m_fd_out->xywh[0]) + ((u16) m_fd_out->xywh[2]);
            lp_rect->end_y = ((u16) m_fd_out->xywh[1]) + ((u16) m_fd_out->xywh[3]);
        }
        else {
            info_msg("[inference %d] FD result : NO FACE\n", __LINE__);
            //break;
        }

        if (ret == KDP_APP_FDR_NO_FACE){
            // if no face, aec_roi set as 480x480 rgb right img 
            if (rgb_fd_get_flag == 0){
                //dbg_msg(" rgb_fd_get_flag  = 0 ");
            }else{
                rgb_fd_get_flag = 0;
                rgb_set_aec_roi(20, 80, 0, 60, 30, 70, 10, 50);
                //dbg_msg("set rgb_fd_get_flag = 0 ");
            }        
        }else{ // if get face, aec_roi set as fd region.
            
            u8 xmin_inter = (u8) (m_fd_out->xywh[0] >> 3);
            u8 xmax_inter = (u8) ( (m_fd_out->xywh[0] + m_fd_out->xywh[2]) >> 3 );
            u8 ymin_inter = (u8) (m_fd_out->xywh[1] >> 3);
            u8 ymax_inter = (u8) ( (m_fd_out->xywh[1] + m_fd_out->xywh[3]) >> 3 );
            
            if( (xmin_inter >= 20)){ // if get face and fd region in rgb right region(480x480)
        
                if (rgb_fd_get_flag == 1){
                    //dbg_msg(" rgb_fd_get_flag = 1 ");
                }else{
                    rgb_fd_get_flag = 1;
                    rgb_set_aec_roi(20, 80, 0, 60, xmin_inter, xmax_inter, ymin_inter, ymax_inter);
                }
            }else{ // if fd region not in rgb right region (480x480)
                
                if (rgb_fd_get_flag == 0){
                    //dbg_msg(" rgb_fd_get_flag  = 0 ");
                }else{
                    rgb_fd_get_flag = 0;
                    rgb_set_aec_roi(20, 80, 0, 60, 30, 70, 10, 50);
                    //dbg_msg("set rgb_fd_get_flag = 0 ");
                }
            }
            int DIST_W = m_fd_out->xywh[2];
            //dbg_msg("DIST_W = %d \n", DIST_W);
            if( DIST_W >= 256 ){  // short distance  
                if( distance_flag == 0 ){
                    dbg_msg(" nir gain as 6 \n");
                }else{
                    distance_flag = 0;
                    nir_set_gain(0x27, 0x35); // nir set gain as 6
                    dbg_msg("set nir gain = 6 \n");
                }
            }
            else if ( DIST_W < 256 ){ // long distance
                if( distance_flag == 1 ){
                    dbg_msg(" nir gain as 12 \n");
                }else{
                    distance_flag = 1;
                    nir_set_gain(0x2F, 0x35); // nir set gain as 12
                    dbg_msg("set nir gain = 12\n");
                }
            }
        }
        lw3d_ret = kdp_app_nir_fdr(distance_flag, m_fdr_inout, NULL, NULL, m_fr_1_out);
        info_msg("[inference] finish LW3D");

        m_fdr_inout->thresh_fid = 0.425;
        // fdr_ret = kdp_app_db_lw3d_compare((uint32_t)m_fr_out->feature_map, (uint32_t)m_fr_1_out->feature_map, &(m_fdr_inout->user_id), m_fdr_inout->thresh_fid);
        fdr_ret = kdp_app_db_lw3d_compare_rgb_nir(
                (uint32_t)m_fr_out->feature_map, 0.5 /*rgb_weight*/, 
                (uint32_t)m_fr_1_out->feature_map, 0.5 /*nir_weight*/, 
                &(m_fdr_inout->user_id), m_fdr_inout->thresh_fid);
        
        info_msg("[inference] finish LW3D compare, close is id %d\n", m_fdr_inout->user_id);
        if (fdr_ret == KDP_APP_OK) {
                info_msg("[inference] LW3D Passed, match with id %d\n",  m_fdr_inout->user_id);
        }         
        else
        {
                fdr_ret = FDR_RET_DB_COMP_FAIL;
                info_msg("[inference] Not Match\n");
        }

        if(fdr_ret != FDR_RET_DB_COMP_FAIL){
            if(KDP_APP_FDR_NO_ALIVE != lw3d_ret){
                fdr_ret = FDR_RET_DB_COMP_OK;
            }else{
                info_msg("[inference] LW3D Failed\n");
                fdr_ret = FDR_RET_OK;
            }
        }               
    } while (0);

    return fdr_ret;
}

fdr_result _kdp_fdfr_both_inference(void)
{
    fdr_result fdr_ret = FDR_RET_OOPS;
    struct kdp_rect *lp_rect = &m_boundingbox;

    m_fdr_inout->op = KDP_APP_FID_OP_INFERENCE;
    m_fdr_inout->thresh_fid = 0.425;

    do {
        memset(m_fd_out, 0, sizeof(kdp_app_face_det_res_t));
        memset(m_fr_out, 0, sizeof(kdp_app_face_fmap_t));   
        
        // get exp time
        u16 a = 0;
        u8 b, c;
        rgb_get_lux( &a , &b , &c );
        //dbg_msg("exp_rgb_time = %d", a);
			 //dump NIR image
			  //memcpy((const void *) 0x62000000, (const void *) 0x637E7000, 640*480*sizeof(char));
        //memcpy((const void *) 0x62000000, (const void *) 0x6357C400, 640*480*sizeof(char));
		//memcpy((const void *) 0x62000000, (const void *) 0x62FFE000, 640*480*2*sizeof(char));
			
        fdr_ret = (fdr_result)_kdp_fdfr_both_only(lp_rect);


    } while (0);
    
    return fdr_ret;
}

static int _kdp_rgst_fdfr_only(struct kdp_rect *lp_rect, int index, int num_rgst)
{
    int ret, ret0 = 0, ret1 = 0;

    do {
        m_fdr_inout->op = KDP_APP_FID_OP_EXTRACT;  
        if(1 == num_rgst) {
            ret = kdp_app_fdr(m_fdr_inout, m_fd_out, (kdp_app_face_lm_res_t*)&db_tmp_lm[index*10], (kdp_app_face_fmap_t*)&db_tmp_fr[index*512]);
            dbg_msg("[%s] ret=0x%x\n", __func__, ret);

            if (KDP_APP_FDR_NO_FACE != ret)
            {
                lp_rect->start_x = (u16) m_fd_out->xywh[0];
                lp_rect->start_y = (u16) m_fd_out->xywh[1];
                lp_rect->end_x = ((u16) m_fd_out->xywh[0]) + ((u16) m_fd_out->xywh[2]);
                lp_rect->end_y = ((u16) m_fd_out->xywh[1]) + ((u16) m_fd_out->xywh[3]);
            }
            else {
                info_msg("[inference %d] FD result : NO FACE\n", __LINE__);
                break;
            }

        }
        else if(2 == num_rgst) {
            ret0 = kdp_app_fdr(m_fdr_inout, m_fd_out, (kdp_app_face_lm_res_t*)&db_tmp_lm[index*10], (kdp_app_face_fmap_t*)&db_tmp_fr[index*512]);
            ret1 = kdp_app_nir_fdr_only(NULL, NULL, (kdp_app_face_lm_res_t*)&db_tmp_1_lm[index*10], (kdp_app_face_fmap_t*)&db_tmp_1_fr[index*512]);
            dbg_msg("[%s] ret0=0x%x, ret1=0x%x\n", __func__, ret0, ret1);

            if ((KDP_APP_FDR_NO_FACE != ret0) && (KDP_APP_FDR_NO_FACE != ret1))
            {
                lp_rect->start_x = (u16) m_fd_out->xywh[0];
                lp_rect->start_y = (u16) m_fd_out->xywh[1];
                lp_rect->end_x = ((u16) m_fd_out->xywh[0]) + ((u16) m_fd_out->xywh[2]);
                lp_rect->end_y = ((u16) m_fd_out->xywh[1]) + ((u16) m_fd_out->xywh[3]);
                ret = ret0; //might be some cases left for different status of ret1
            }
            else {
                ret = KDP_APP_FDR_NO_FACE;
                dbg_msg("[inference %d] FD result : NO FACE (ret0:%d, ret1:%d)\n", __LINE__, ret0, ret1);
                break;
            }
        }
    } while (0);

    return ret;
}

int _kdp_fdfr_registration(uint32_t user_idx)
{
    int ret = -1;

    do {
        //todo : should restructure 
        if(0 != (ret = _kdp_check_DB_space(user_idx))) {
            break;
        }
        else if (kdp_app_db_get_fm0_cnt(user_idx) >= MAX_FID) {
            info_msg("[registration] : saved_fmap_cnt > MAX_FID(5)\n");
            ret = KDP_APP_OK;
            break;
        }
        else {
            memset(m_fd_out, 0, sizeof(kdp_app_face_det_res_t));
            memset(m_fr_out, 0, sizeof(kdp_app_face_fmap_t));
 
            m_fdr_inout->op = KDP_APP_FID_OP_REGISTER;
            m_fdr_inout->face_id = kdp_app_db_get_fm0_cnt((int)user_idx) + 1;  //start from 1
            m_fdr_inout->user_id = (uint16_t)user_idx; 
            m_fdr_inout->thresh_fid = 0.425;

            ret = kdp_app_fdr(m_fdr_inout, m_fd_out, NULL, m_fr_out);
            if (KDP_APP_OK == ret) {
                info_msg("[registration] : -- OK!\n\n");
            }
            else if (KDP_APP_DB_ALREADY_SAVED == ret) {
                info_msg("[registration] : -- ERROR!, face has been registered\n\n");
            }
            else {
                info_msg("[registration] : -- ERROR!, err=0x%x\n\n", ret);
            }
            break;
        }
    } while (0);

    return ret;
}

//Datang's 5 face register
int _kdp_5_face_registration(uint32_t user_idx, int type) {
    //DSG_NOLF(" sample_kneron_fdr_auto_reg!\n");
    int ret = -1;
    int pose_ret = -1;
    // db_tmp_valid_pose[type] = false;

    if(0 != (ret = _kdp_check_DB_space(user_idx))) {
        return ret;
    }
    //try different angle 5 faces
    ret = _kdp_rgst_fdfr_only(&m_boundingbox, type, 1);

    if(ret != KDP_APP_OK){
        return ret;
    }
    //left or up

    if (type==1 || type==3) {
        pose_ret = kdp_app_check_head_pose((kdp_app_face_lm_res_t*)&db_tmp_lm[type*10], type, 0, face_reg_thresh);
    } else  {
        pose_ret = kdp_app_check_head_pose((kdp_app_face_lm_res_t*)&db_tmp_lm[type*10], type, 1, face_reg_thresh);
    }

    if (pose_ret == -1) {
        return pose_ret;
    } else {
        //if normal face
        if (type==0) {
            //see if there is any one in the data base match this guy
            ret = kdp_app_db_compare((uint32_t)&db_tmp_fr[0], &m_fdr_inout->user_id, m_fdr_inout->thresh_fid);
            if(ret != KDP_APP_DB_NO_MATCH) {
                return KDP_APP_DB_ALREADY_SAVED;
            } else {
                //register normal face
                uint32_t fr_addr = (uint32_t)db_tmp_fr;
                ret = kdp_app_db_register(fr_addr, user_idx, type);   
            }
        } else {
                uint32_t fr_addr = (uint32_t)db_tmp_fr+(type+1)*512*4;
                ret = kdp_app_db_register(fr_addr, user_idx, type);                
        }
    }

    return ret;
}

static int _kdp_fdfr_auto_registeration(uint32_t user_idx, int reg_index, int num_rgst)
{
	dbg_msg("[%s %d] user_idx:%d, reg_index:%d, num_rgst:%d", __func__, __LINE__, user_idx, reg_index, num_rgst);
    int ret = -1;
	db_tmp_valid_pose[reg_index] = false;

    if(0 != (ret = _kdp_check_DB_space(user_idx))) {
        return ret;
    }

    /*_kdp_rgst_fdr_only() combines _sample_kneron_fdr_both_only() + _sample_kneron_reg_fdr_both_only()
      num_reg = 1 => _sample_kneron_fdr_both_only()
      num_reg = 2 => _sample_kneron_reg_fdr_both_only()*/
    ret = _kdp_rgst_fdfr_only(&m_boundingbox, reg_index, num_rgst);
            
    //for (int j=0; j<10; j++) {
    //    DSG_NOLF(" lm %d is %d!\n", j, db_tmp_lm[reg_index*10+j]);
    //}

    if (ret == KDP_APP_OK) {
        //DSG_NOLF(" have face!\n");
		//fLib_printf("start_x: %hu\n", m_boundingbox.start_x);
		//fLib_printf("start_y: %hu\n", m_boundingbox.start_y);
		//fLib_printf("end_x: %hu\n", m_boundingbox.end_x);
		//fLib_printf("end_y: %hu\n", m_boundingbox.end_y);
        //check if it is bad pose
        if(1 == num_rgst) {
            bool bad_pose = kdp_app_check_bad_pose(&db_tmp_lm[reg_index*10]);
            if (!bad_pose) {
                //DSG_NOLF(" good pose for index %d!\n", reg_index);
                db_tmp_valid_pose[reg_index] = true;
            }
        }
        else if(2 == num_rgst) {
            //check if it is bad pose
            bool rgb_bad_pose = kdp_app_check_bad_pose(&db_tmp_lm[reg_index*10]);
            bool nir_bad_pose = kdp_app_check_bad_pose(&db_tmp_1_lm[reg_index*10]);
            if (!rgb_bad_pose && !nir_bad_pose) {
                //DSG_NOLF(" good pose for index %d!\n", reg_index);
                db_tmp_valid_pose[reg_index] = true;
            }
        }
    }

    return ret;
}


static int _kdp_fdfr_registration_confirm(uint32_t user_idx, int num_rgst) {
	info_msg(" Registration confrim for id %d!\n", user_idx);
    int ret = -1;
    int near_cnt = 0;

    m_fdr_inout->thresh_fid = 0.425;

    //need to check if the user is already in db??
    for (int i = 0; i<NUM_REG_INFERENCE; i++) {
		//DSG_NOLF("check %d FM!\n", i);
        if (db_tmp_valid_pose[i]) {
			//DSG_NOLF("valid pose %d!\n", i);
            ret = kdp_app_db_compare((uint32_t)&db_tmp_fr[i*512], &m_fdr_inout->user_id, m_fdr_inout->thresh_fid);
            if(ret != KDP_APP_DB_NO_MATCH) {
				//DSG_NOLF("%d FM match with %d in database!\n", i, m_fdr_inout->user_id);
				near_cnt++;
				db_tmp_valid_pose[i] = 0;
                //return KDP_APP_DB_ALREADY_SAVED; // person already in DB ?? 
            }
        }
        //KDP_APP_DB_NO_MATCH
    }
		if (near_cnt>5) {
			dbg_msg("more than 5 FM are too close to those in database!\n");
			return KDP_APP_DB_ALREADY_SAVED; // person already in DB ?? 
		}
    //check best pose for registration

    int index[5], index1[5], index2[5], index3[5], index4[5];

    float tmp_lm[5];
    float min_lm[5];
    float min_lm1[5], min_lm2[5], min_lm3[5], min_lm4[5];
    //initialize
    for (int i=0; i<5; i++) {
        min_lm[i]  = 1000;
        min_lm1[i] = 1000;
        min_lm2[i] = 1000;
        min_lm3[i] = 1000;
        min_lm4[i] = 1000;
        index[i]  = 0;
        index1[i] = 0;
        index2[i] = 0;
        index3[i] = 0;
        index4[i] = 0;
    }   
    info_msg("check for best pose!\n");
    for (int i=0;i<NUM_REG_INFERENCE;i++) {
        if (db_tmp_valid_pose[i]) {
			//DSG_NOLF("pose number %d!\n",i);
            kdp_app_get_pose_value(&db_tmp_lm[i*10], tmp_lm);
            for (int j=0; j<5; j++) {
				bool change = FALSE;
                if (tmp_lm[j] < min_lm[j]) {
                    index[j] = i;
                    min_lm[j] = tmp_lm[j];
					change = TRUE;
                }
                if (!change && (tmp_lm[j] < min_lm1[j])) {
                        index1[j] = i;
                        min_lm1[j] = tmp_lm[j];
                        change = TRUE;
                }
                if (!change && (tmp_lm[j] < min_lm2[j])) {
                    index2[j] = i;
                        min_lm2[j] = tmp_lm[j];
                        change = TRUE;
                }
                if (!change && (tmp_lm[j] < min_lm3[j])) {
                        index3[j] = i;
                        min_lm3[j] = tmp_lm[j];
                        change = TRUE;
                }
                if (!change && (tmp_lm[j] < min_lm4[j])) {
                        index4[j] = i;
                        min_lm4[j] = tmp_lm[j];
                        change = TRUE;
                }
            }
        }
    }
    //check if index has repeated 
    //DSG_NOLF("check if any index is repeat!\n");
    int target;
    for (int i= 0; i<5; i++) {
        target = index[i];
        bool repeat = FALSE; 
        for (int j=0; j<5; j++) {
            if ((i!=j)&&(target == index[j])) {
                repeat = TRUE;
                break;
            }
        }
        if (!repeat) {
            //DSG_NOLF("keep index %d for pose %d!\n", target, i);
            index[i] = target;
            continue;
        }	 
        target = index1[i];
        repeat = FALSE;			 
        for (int j=0; j<5; j++) {
            if ((i!=j)&&(target == index[j])) {
                repeat = TRUE;
                break;
            }
        }
        if (!repeat) {
            //DSG_NOLF("replace min 1 index %d for pose %d!\n", target, i);
            index[i] = target;
            continue;
        }	 				 	 
        target = index2[i];
        repeat = FALSE;			 
        for (int j=0; j<5; j++) {
            if ((i!=j)&&(target == index[j])) {
                    repeat = TRUE;
                    break;
            }
        }
        if (!repeat) {
            //DSG_NOLF("replace min 2 index %d for pose %d!\n", target, i);
            index[i] = target;
            continue;
        }					 
        target = index3[i];
        repeat = FALSE;			 
        for (int j=0; j<5; j++) {
            if ((i!=j)&&(target == index[j])) {
                    repeat = TRUE;
                    break;
            }
        }
        if (!repeat) {
            //DSG_NOLF("replace min 3 index %d for pose %d!\n", target, i);
            index[i] = target;
            continue;
        }
        target = index4[i];
        repeat = FALSE;			 
        for (int j=0; j<5; j++) {
            if ((i!=j)&&(target == index[j])) {
                    repeat = TRUE;
                    break;
            }
        }
        if (!repeat) {
            //DSG_NOLF("replace min 4 index %d for pose %d!\n", target, i);
            index[i] = target;
            continue;
        }			 
	}
    info_msg("best pose index %d %d %d %d %d!\n", index[0], index[1], index[2], index[3], index[4]);
    //check if these lm valid
    for (int i=0; i<5; i++) {
        if (!db_tmp_valid_pose[index[i]]) {
            critical_msg("ERROR! not enough valid pose in registration!");
            ret = KDP_APP_FDR_BAD_POSE;
            return ret;
        }
    }	
	//DSG_NOLF("register for user id %d!\n", curr_user_idx);
	if(1 == num_rgst) {
        for (int i=0; i<5; i++) {
            uint32_t fr_addr = (uint32_t)db_tmp_fr+index[i]*512*4;
    		//DSG_NOLF("db_tmp_fr addr is %d, index[i]*512 is %d, fr_addr is %d!\n", db_tmp_fr, index[i]*512, fr_addr);
            ret = kdp_app_db_register(fr_addr, user_idx, i+1);		 
        }
    }
    else if(2 == num_rgst) {
        for (int i=0; i<5; i++) {
            uint32_t rgb_fr_addr = (uint32_t)db_tmp_fr+index[i]*512*4;
            uint32_t nir_fr_addr = (uint32_t)db_tmp_1_fr+index[i]*512*4;        
    		//DSG_NOLF("db_tmp_fr addr is %d, index[i]*512 is %d, fr_addr is %d!\n", db_tmp_fr, index[i]*512, fr_addr);
            ret = kdp_app_db_both_register(rgb_fr_addr, nir_fr_addr, user_idx, i+1);		 
        }   
    }
	return ret;
}

static int _kdp_fdfr_add_user(uint32_t user_idx)
{
    int ret = -1;

    do {
        //todo no need
        if(0 != (ret = _kdp_check_DB_space(user_idx))) {
            break;
        }
        else {
            m_db_data->user_id_in = (uint16_t)user_idx;
            ret = kdp_app_db_add(m_db_data);
        }

        if (KDP_APP_OK == ret) { 
            info_msg("[add_user] : UID:%d, OK!\n", user_idx);
        }
        else {
            info_msg("[add_user] : UID:%d, ERROR!, ret=%x\n", user_idx, ret);
            break;
        }
    } while (0);
    return ret;
}

static int _kdp_fdfr_del_user(bool all, uint32_t user_idx)//int user_id)
{
    int ret = -1;
    do {    
        if (all)
            m_db_data->del_all = TRUE;
        else {
            m_db_data->del_all = FALSE;
            m_db_data->user_id_in = (uint16_t)user_idx;
        }
        ret = kdp_app_db_delete(m_db_data);
        if (KDP_APP_OK == ret) {
            info_msg("[delete user face] delete user face -- OK!\n");
        }
        else {
            info_msg("[delete user face] delete user face -- ERROR, ret=%x\n", ret);
            break;            
        }
        ret = 0;
    } while (0);
    
    return ret;    
}

static int _kdp_fdfr_cancel_registeration(void)
{
    int ret = -1;

    /* under construction */
    do {    
         ret = kdp_app_db_abort_reg();      
         if (KDP_APP_OK == ret) { 
             info_msg("[cancel registration] -- OK!\n");
         } else {
             info_msg("[cancel registration] -- ERROR!, err=%x\n", ret);
         }
         ret = 0;    
    } while (0);
    
    //0808@customer1 should release the memory which alloc in sample_kneron_fdr_model_init();  
    return ret;
}

static unsigned int _kdp_fdfr_fb_capture(void)
{
    int cam_idx;
    unsigned int frame_addr_rgb, frame_addr_nir;

    cam_idx = CAMERA_RGB_IDX;
    frame_addr_rgb = kdp_fb_mgr_next_inf(cam_idx, &g_buf_idx[cam_idx]);
    kdp_app_config_image_memory_address(cam_idx, frame_addr_rgb);
    cam_idx = CAMERA_NIR_IDX;
    frame_addr_nir = kdp_fb_mgr_next_inf(cam_idx, &g_buf_idx[cam_idx]);
    kdp_app_config_image_memory_address(cam_idx, frame_addr_nir);
    return frame_addr_rgb;
}

static void _kdp_frdr_fb_prepare(void)
{
    int cam_idx;

    cam_idx = CAMERA_RGB_IDX;
    kdp_fb_mgr_inf_done(cam_idx, g_buf_idx[cam_idx]);
    cam_idx = CAMERA_NIR_IDX;
    kdp_fb_mgr_inf_done(cam_idx, g_buf_idx[cam_idx]);
}

static void _kdp_fdfr_fb_free(void)
{
    kdp_fb_mgr_free_inf_buf(CAMERA_RGB_IDX);
    kdp_fb_mgr_free_inf_buf(CAMERA_NIR_IDX);
}

static int _kdp_fdfr_registeration_check(struct kdp_fdfr_comm_params *fdfr_params)
{
    if(FLAGS_KDP520_REGISTER_FDR_EVT == fdfr_params->rgst_ctx.rgst_type) {
        return _kdp_fdfr_registration(fdfr_params->rgst_ctx.user_idx);
    }
    else if((FLAGS_KDP520_AUTO_REGISTER_FDR_EVT == fdfr_params->rgst_ctx.rgst_type) ||
            (FLAGS_KDP520_AUTO_REGISTER_BOTH_FDR_EVT == fdfr_params->rgst_ctx.rgst_type)) {
        return _kdp_fdfr_auto_registeration(fdfr_params->rgst_ctx.user_idx, fdfr_params->rgst_ctx.rgst_idx, fdfr_params->rgst_ctx.num_rgst);
    }
    else if((FLAGS_KDP520_REGISTER_CFM_FDR_EVT == fdfr_params->rgst_ctx.rgst_type) || 
            (FLAGS_KDP520_REGISTER_CFM_BOTH_FDR_EVT == fdfr_params->rgst_ctx.rgst_type)){
        return _kdp_fdfr_registration_confirm(fdfr_params->rgst_ctx.user_idx, fdfr_params->rgst_ctx.num_rgst);
    }
		return -1;
}

fdr_result _kdp_inference_check_ack_status(fdr_result fdr_ret)
{
    u32 curr_tick;
    int _nShowimerCur = -1;
    if(FDR_RET_DB_COMP_OK == fdr_ret) {
        memcpy(&h_boundingbox, &m_boundingbox, sizeof(m_boundingbox));
        _kdp_fdfr_fb_free();
    } 
    else {
        curr_tick = kdp_current_t3_tick();
        _nShowimerCur = curr_tick - m_fdfr_tick;
        if ( _nShowimerCur >= m_face_recognition_timeout) {
            fdr_ret = FDR_RET_TIMEOUT;
            critical_msg("fdfr inference timeout =%d/%d",_nShowimerCur, m_face_recognition_timeout);
            _kdp_fdfr_fb_free();
        }
    }
    return fdr_ret;
}

static fdr_result _kdp_fdfr_inference_running(int inf_type)
{
    fdr_result fdr_ret;
    unsigned int frame_addr_rgb;

    while(g_fdfr_run_inference == TRUE) {
        frame_addr_rgb = _kdp_fdfr_fb_capture();
        fdr_ret = _kdp_fdfr_inference();
        if(fdr_ret != FDR_RET_OOPS) {
            _kdp_draw_fdfr_result(fdr_ret, frame_addr_rgb);

            if(KDP_INF_WITH_RESULT_ACK == inf_type) {
                fdr_ret = _kdp_inference_check_ack_status(fdr_ret);
                if((fdr_ret == FDR_RET_DB_COMP_OK) || (fdr_ret == FDR_RET_TIMEOUT))
                    return fdr_ret;
            }
        }
        _kdp_frdr_fb_prepare();
    }
    return FDR_RET_OOPS;
}

static void kdp_inference_thread(void *argument)
{
    unsigned int flags;
    kdp_all_comm_params *all_params = (kdp_all_comm_params *)argument;
    kdp_camera_comm_params *cam_params = &all_params->cam_params;
    kdp_fdfr_comm_params *fdfr_params = &all_params->fdfr_params;

    for (;;) {
        flags = osThreadFlagsWait(FLAGS_KDP520_ALL_EVENTS, osFlagsWaitAny, osWaitForever);
        osThreadFlagsClear(flags);
        /* =========================== KDP camera ===========================  */
        if (flags & (FLAGS_KDP520_OPEN_CAMERA0_EVT | FLAGS_KDP520_OPEN_CAMERA1_EVT)) {
            cam_params->ret = _kdp_camera_open(cam_params->cam_idx);
            set_thread_event(cam_params->caller_tid, FLAGS_CAMERA_ACK);
            flags &= ~(FLAGS_KDP520_OPEN_CAMERA0_EVT | FLAGS_KDP520_OPEN_CAMERA1_EVT);
            continue;
        }

        if (flags & (FLAGS_KDP520_START_CAMERA0_EVT | FLAGS_KDP520_START_CAMERA1_EVT)) {
            cam_params->ret = _kdp_camera_start(cam_params->cam_idx);
            set_thread_event(cam_params->caller_tid, FLAGS_CAMERA_ACK);
            flags &= ~(FLAGS_KDP520_START_CAMERA0_EVT | FLAGS_KDP520_START_CAMERA1_EVT);
            continue;
        }

        if (flags & (FLAGS_KDP520_STOP_CAMERA0_EVT | FLAGS_KDP520_STOP_CAMERA1_EVT)) {
            cam_params->ret = _kdp_camera_stop(cam_params->cam_idx);
            set_thread_event(cam_params->caller_tid, FLAGS_CAMERA_ACK);
            flags &= ~(FLAGS_KDP520_STOP_CAMERA0_EVT | FLAGS_KDP520_STOP_CAMERA1_EVT);
            continue;
        }

        if (flags & (FLAGS_KDP520_CLOSE_CAMERA0_EVT | FLAGS_KDP520_CLOSE_CAMERA1_EVT)) {
            cam_params->ret = _kdp_camera_close(cam_params->cam_idx);
            set_thread_event(cam_params->caller_tid, FLAGS_CAMERA_ACK);
            flags &= ~(FLAGS_KDP520_CLOSE_CAMERA0_EVT | FLAGS_KDP520_CLOSE_CAMERA1_EVT);
            continue;
        }

         /* =========================== KDP fdfr inference =========================== */
        if(flags & FLAGS_KDP520_START_INFERENCE_LOOP_EVT) {
            _kdp_fdfr_inference_running(KDP_INF_WITHOUT_RESULT_ACK);
            flags &= ~(FLAGS_KDP520_START_INFERENCE_LOOP_EVT);
            continue;
        }

        if (flags & FLAGS_KDP520_STOP_INFERENCE_LOOP_EVT) {
            _kdp_fdfr_fb_free();
            flags &= ~(FLAGS_KDP520_STOP_INFERENCE_LOOP_EVT | FLAGS_KDP520_START_INFERENCE_LOOP_EVT);
            continue;
        }  

        if(flags & FLAGS_KDP520_START_INFERENCE_EVT) {
            fdfr_params->fdfr_ctx.ret = _kdp_fdfr_inference_running(KDP_INF_WITH_RESULT_ACK);
            flags &= ~(FLAGS_KDP520_START_INFERENCE_EVT);
            set_thread_event(fdfr_params->fdfr_ctx.caller_tid, FLAGS_FDFR_ACK);
            continue;
        }

        if (flags & (FLAGS_KDP520_REGISTER_FDR_EVT | FLAGS_KDP520_AUTO_REGISTER_FDR_EVT | FLAGS_KDP520_REGISTER_CFM_FDR_EVT | \
                     FLAGS_KDP520_AUTO_REGISTER_BOTH_FDR_EVT | FLAGS_KDP520_REGISTER_CFM_BOTH_FDR_EVT)) {
            _kdp_fdfr_fb_capture();
            fdfr_params->rgst_ctx.ret = _kdp_fdfr_registeration_check(fdfr_params);
            _kdp_fdfr_fb_free();
            set_thread_event(fdfr_params->rgst_ctx.caller_tid, FLAGS_FDFR_ACK);
            flags &= ~(FLAGS_KDP520_REGISTER_FDR_EVT | FLAGS_KDP520_AUTO_REGISTER_FDR_EVT | FLAGS_KDP520_REGISTER_CFM_FDR_EVT |\
                       FLAGS_KDP520_AUTO_REGISTER_BOTH_FDR_EVT | FLAGS_KDP520_REGISTER_CFM_BOTH_FDR_EVT);
            continue;
        }

        if (flags & FLAGS_KDP520_AUTO_REGISTER_5_FACE_EVT) {
            _kdp_fdfr_fb_capture();
            fdfr_params->rgst_5_ctx.ret = _kdp_5_face_registration(fdfr_params->rgst_5_ctx.user_idx, fdfr_params->rgst_5_ctx.pose_type);
            _kdp_fdfr_fb_free();

            if(fdfr_params->rgst_5_ctx.ret == KDP_APP_OK){
               // memcpy((void *) (0x62000000 + 640*480*2*reg_type), (void *) g_fdr_buf_addr[0], 640*480*2*sizeof(char));
                /*Should NOT use global variable to control value, should revise to use API to control later...*/
                reg_type += 1;
                reg_count = 0;
            }
            set_thread_event(fdfr_params->rgst_5_ctx.caller_tid, FLAGS_FDFR_ACK);
            flags &= ~(FLAGS_KDP520_AUTO_REGISTER_5_FACE_EVT);
            continue;
        }   

        if (flags & FLAGS_KDP520_CANCEL_REGISTER_FDR_EVT) {
            fdfr_params->rgst_cancel_ctx.ret = _kdp_fdfr_cancel_registeration();
            set_thread_event(fdfr_params->rgst_cancel_ctx.caller_tid, FLAGS_FDFR_ACK);
            flags &= ~(FLAGS_KDP520_CANCEL_REGISTER_FDR_EVT);
            continue;
        }   

        if (flags & FLAGS_KDP520_ADD_FDR_FACE_EVT) {
            fdfr_params->add_ctx.ret = _kdp_fdfr_add_user(fdfr_params->add_ctx.user_idx);
            set_thread_event(fdfr_params->add_ctx.caller_tid, FLAGS_FDFR_ACK);
            flags &= ~(FLAGS_KDP520_ADD_FDR_FACE_EVT);
            continue;
        }

        if (flags & FLAGS_KDP520_DEL_FDR_FACE_EVT) {
            fdfr_params->del_ctx.ret = _kdp_fdfr_del_user(fdfr_params->del_ctx.all, fdfr_params->del_ctx.user_idx);
            set_thread_event(fdfr_params->del_ctx.caller_tid, FLAGS_FDFR_ACK);
            flags &= ~(FLAGS_KDP520_DEL_FDR_FACE_EVT);
            continue;
        }

        if(flags) {
            err_msg("%s: unknown flags 0x%x\n", flags);
        }
    }
}

/*----------------------------------------------------------------------------------------
 *                                      Puclic FDFR API 
 *----------------------------------------------------------------------------------------*/
bool kdp_fdfr_model_init(void)
{  
    int j;
    int saved_count;

    if(m_fdr_inout == NULL)
        m_fdr_inout = (kdp_app_fid_inout_t*)kdp_ddr_reserve(sizeof(kdp_app_fid_inout_t));
    if(m_fd_out == NULL)
        m_fd_out = (kdp_app_face_det_res_t*)kdp_ddr_reserve(sizeof(kdp_app_face_det_res_t));
    if(m_fr_out == NULL)
        m_fr_out = (kdp_app_face_fmap_t*)kdp_ddr_reserve(sizeof(kdp_app_face_fmap_t));

    if(m_fr_1_out == NULL)
        m_fr_1_out = (kdp_app_face_fmap_t*)kdp_ddr_reserve(sizeof(kdp_app_face_fmap_t));

    if(m_db_data == NULL)
        m_db_data = (kdp_app_db_user_data_t*)kdp_ddr_reserve(sizeof(kdp_app_db_user_data_t));
    //todo
    if (db_tmp_fr == NULL) 
        db_tmp_fr = (float *)kdp_ddr_reserve(sizeof(uint8_t)*REG_TEMP_FR_SIZE); 
    if (db_tmp_lm == NULL)
        db_tmp_lm = (uint32_t *)kdp_ddr_reserve(sizeof(uint8_t)*REG_TEMP_LM_SIZE); 

    if (db_tmp_1_fr == NULL) 
        db_tmp_1_fr = (float *)kdp_ddr_reserve(sizeof(uint8_t)*REG_TEMP_FR_SIZE); 
    if (db_tmp_1_lm == NULL)
            db_tmp_1_lm = (uint32_t *)kdp_ddr_reserve(sizeof(uint8_t)*REG_TEMP_LM_SIZE);  

    saved_count = kdp_app_init();
    kdp_app_config_image(CAMERA_RGB_IDX, RGB_IMG_SOURCE_W, RGB_IMG_SOURCE_H, 3, KDP_APP_IMAGE_FORMAT_RGB565, 0/*default addr*/);
    kdp_app_config_image(CAMERA_NIR_IDX, NIR_IMG_SOURCE_CROP_H, NIR_IMG_SOURCE_CROP_W, 1, KDP_APP_IMAGE_FORMAT_RAW8, 0/*default addr*/);

    kdp_app_db_init(KDP_APP_DB_FLASH_ADDR);     //load db from flash

    //init m_sample_db.add_tag according to existed data in DB
    for (j=0 ; j<MAX_USER;j++) {
        if(saved_count == 0)
            break;
            
        if(kdp_app_db_slot_is_used(j)) {
            saved_count--;
        }
    }
    return TRUE;
}

int kdp_fdfr_inference_thread_create(void)
{
    int ret = -1;
    if (NULL != tid_inference)
        return ret;

    if(NULL != (tid_inference = osThreadNew(kdp_inference_thread, (void*)&m_kdp_all_comm_params, NULL))) {
        osThreadSetPriority (tid_inference, osPriorityNormal);
        power_manager_register(PM_DEVICE_NCPU_INFERENCE, &pm_ncpu_inference);
        ret = 0;
    }
    return ret;
}

void kdp_fdfr_set_boundingbox_mode(int mode)
{
    g_draw_boundingbox_mode = mode;
}

void kdp_fdfr_inference_start(void)
{
    memset(&m_boundingbox, 0, sizeof(m_boundingbox));
    memset(&h_boundingbox, 0, sizeof(h_boundingbox));
    kdp_fdfr_set_hint_boundingbox();
    g_fdfr_run_inference = TRUE;
}

void kdp_fdfr_inference_stop(void)
{
    memset(&s_boundingbox, 0, sizeof(s_boundingbox));
    g_fdfr_run_inference = FALSE;
}

bool kdp_fdfr_inference_is_running(void)
{
    return g_fdfr_run_inference;
}

void kdp_fdfr_set_hint_boundingbox(void)
{
    unsigned short dp_xre = 0, dp_yres = 0;
    kdp520_lcdc_get_dp_size(&dp_xre, &dp_yres);
    s_boundingbox.start_x = g_frame_margin_length;
    s_boundingbox.start_y = g_frame_margin_length;
    s_boundingbox.end_x = dp_xre - s_boundingbox.start_x;
    s_boundingbox.end_y = dp_yres - s_boundingbox.start_y;
}

int kdp_fdfr_get_available_db_idx(void)
{
    return kdp_app_db_get_available_ID_slot();
}

int kdp_fdfr_get_remaining_fmap_cnt(int user_idx)
{
    return (MAX_FID - kdp_app_db_get_fm0_cnt(user_idx));
}

void kdp_fdfr_recognition_variable_reset(void)
{
    m_iCurExpoTime = Max_EXPO_TIME;
    nir_set_exp_time((m_iCurExpoTime&0x0000FF00)>>8, m_iCurExpoTime&0xFF);
    nir_led_close();
    //rgb_led_close();
    distance_flag = 10; 
    nir_set_gain(0x27, 0x24);
    //dbg_msg("close nir led \n");          
    nir_frm_cnt = 0;
//    TgtLedPwm = 0;
//    CurLedPwm = 0;     
}

/*----------------------------------------------------------------------------------------
 *                                  Puclic camera API 
 *----------------------------------------------------------------------------------------*/
int kdp_camera_setting(uint8_t cam_idx, unsigned int width, unsigned int height, unsigned int pixelformat)
{
    if(TRUE == m_kdp_cam_ctx[cam_idx].init) {
        dbg_msg("[%s] camera index %d is already initialized", __func__, cam_idx);
        return -1;
    }
    m_kdp_cam_ctx[cam_idx].init = TRUE;
    m_kdp_cam_ctx[cam_idx].width = width;
    m_kdp_cam_ctx[cam_idx].height = height;
    m_kdp_cam_ctx[cam_idx].pixelformat = pixelformat;
    return 0;
}

int kdp_get_camera_ctx_state(uint8_t cam_idx)
{
    return m_kdp_cam_ctx[cam_idx].state;
}

osThreadId_t kdp_get_inference_thread_id(void)
{
    return tid_inference;
}
