/*
 * Kneron Peripheral API - GPIO 
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include <stdlib.h>

#include "cmsis_os2.h"
#include "kneron_mozart.h"

#include "kdp_io.h"
#include "kdp_gpio.h"

/* GPIO registers offset */
#define REG_GPIO_DOUT_OFFSET            0x0
#define REG_GPIO_DIN_OFFSET             0x4
#define REG_GPIO_PINOUT_OFFSET          0x8
#define REG_GPIO_PIN_BYPASS             0xC
#define REG_GPIO_DATASET                0x10
#define REG_GPIO_DATACLR                0x14
#define REG_GPIO_PULLENABLE             0x18
#define REG_GPIO_PULLTYPE               0x1C
#define REG_GPIO_INT_ENABLE             0x20
#define REG_GPIO_INT_RAWSTATE           0x24
#define REG_GPIO_INT_MASKSTATE          0x28
#define REG_GPIO_INT_MASK               0x2C
#define REG_GPIO_INT_CLEAR              0x30
#define REG_GPIO_INT_TRIGGER            0x34
#define REG_GPIO_INT_BOTH               0x38
#define REG_GPIO_INT_RISENEG            0x3C
#define REG_GPIO_INT_BOUNCEENABLE       0x40
#define REG_GPIO_INT_PRESCALE           0x44

#define GpioRegRead(reg_offset)  inw(GPIO_FTGPIO010_PA_BASE + reg_offset)
#define GpioRegWrite(reg_offset, val)  outw(GPIO_FTGPIO010_PA_BASE + reg_offset, val)
#define GpioRegMaskedSet(reg_offset, val) masked_outw(GPIO_FTGPIO010_PA_BASE + reg_offset, val, val)
#define GpioRegMaskedClr(reg_offset, val) masked_outw(GPIO_FTGPIO010_PA_BASE + reg_offset, 0, val)

// internal struct to record user thread flag notification
typedef struct _GPIO_THREAD_NOTIFY_t{
    uint32_t pin;
    osThreadId_t tid;
    uint32_t tflag;
    struct _GPIO_THREAD_NOTIFY_t *next;
}kdp_i_gpio_notify_t;

// user thread flag notification linked list
// dynamically allocate for saving memory usage
static kdp_i_gpio_notify_t *gpio_notify_list;

// GPIO controller to NVIC ISR
static void kdp_gpio_isr(void)
{
    // read GPIO interrupt pins with masked status
    uint32_t intr_status = GpioRegRead(REG_GPIO_INT_MASKSTATE);

    // clear the interrupt source coming from GPIO peripheral
    GpioRegWrite(REG_GPIO_INT_CLEAR, intr_status);

    // check if needed to notify user thread
    kdp_i_gpio_notify_t *notify_item = gpio_notify_list;
    while (notify_item)
    {
        if (intr_status & (0x1 << notify_item->pin))
            osThreadFlagsSet(notify_item->tid, notify_item->tflag);

        notify_item = notify_item->next;
    }
}

kdp_status_e kdp_gpio_init()
{
    // enable gpio_pclk
    // FIXME: should use internal system clock function (not yet implemented)
    // to manage all peripherals pclk
    read_set_bit(SCU_FTSCU100_PA_BASE + 0x60, 20);

    NVIC_SetVector(GPIO_FTGPIO010_IRQ, (uint32_t)kdp_gpio_isr);

    // Clear and Enable SAI IRQ
    NVIC_ClearPendingIRQ(GPIO_FTGPIO010_IRQ);
    NVIC_EnableIRQ(GPIO_FTGPIO010_IRQ);

    return KDP_STATUS_OK;
}

kdp_status_e kdp_gpio_deinit()
{
    NVIC_DisableIRQ(GPIO_FTGPIO010_IRQ);

    // disable gpio_pclk
    // FIXME: should use system-clock API (not yet implemented) to manage all peripherals pclk
    read_clear_bit(SCU_FTSCU100_PA_BASE + 0x60, 20);

    // free allocated memory for user thread notification
    kdp_i_gpio_notify_t *notify_item = gpio_notify_list;
    kdp_i_gpio_notify_t *tn_temp;
    while (notify_item)
    {
        tn_temp = notify_item;
        notify_item = notify_item->next;
        free(tn_temp);
    }

    return KDP_STATUS_OK;
}

kdp_status_e kdp_gpio_set_pin_attribute(kdp_gpio_pin_e pin, uint32_t attributes)
{
    uint32_t pin_val = (1 << pin);

    // disable interrup in case of wrong conditions
    GpioRegMaskedClr(REG_GPIO_INT_ENABLE, pin_val);
    GpioRegMaskedSet(REG_GPIO_INT_MASK, pin_val);

    // set pin direction
    if (attributes & KDP_GPIO_DIR_INPUT)
        GpioRegMaskedClr(REG_GPIO_PINOUT_OFFSET, pin_val);
    else // KDP_GPIO_DIR_OUTPUT
        GpioRegMaskedSet(REG_GPIO_PINOUT_OFFSET, pin_val);

    uint32_t edge_attributes =
        (KDP_GPIO_INT_EDGE_RISING | KDP_GPIO_INT_EDGE_FALLING | KDP_GPIO_INT_EDGE_BOTH);

    if (attributes & edge_attributes)
    {
        // edge trigger
        GpioRegMaskedClr(REG_GPIO_INT_TRIGGER, pin_val);

        // set both or single edge
        if (attributes & KDP_GPIO_INT_EDGE_BOTH)
            GpioRegMaskedSet(REG_GPIO_INT_BOTH, pin_val);
        else
            GpioRegMaskedClr(REG_GPIO_INT_BOTH, pin_val);

        // set rising or falling edge
        if (attributes & KDP_GPIO_INT_EDGE_RISING)
            GpioRegMaskedClr(REG_GPIO_INT_RISENEG, pin_val);
        else
            GpioRegMaskedSet(REG_GPIO_INT_RISENEG, pin_val);
    }

     uint32_t level_attributes =
        (KDP_GPIO_INT_LEVEL_HIGH | KDP_GPIO_INT_LEVEL_LOW);

    if (attributes & level_attributes)
    {
        // level trigger
        GpioRegMaskedSet(REG_GPIO_INT_TRIGGER, pin_val);

        // set high or low level
        if (attributes & KDP_GPIO_INT_LEVEL_HIGH)
            GpioRegMaskedClr(REG_GPIO_INT_RISENEG, pin_val);
        else
            GpioRegMaskedSet(REG_GPIO_INT_RISENEG, pin_val);
    }

    // clear pull-enable here, should use IO control register to set pull-up or pull-down
    GpioRegMaskedClr(REG_GPIO_PULLENABLE, pin_val);

    return KDP_STATUS_OK;
}

kdp_status_e kdp_gpio_set_notify(kdp_gpio_pin_e pin, osThreadId_t tid, uint32_t tflag)
{
    kdp_i_gpio_notify_t *notify_item = (kdp_i_gpio_notify_t *)malloc(sizeof(kdp_i_gpio_notify_t));

    // insert the new thread notification item into tn list
    notify_item->pin = pin;
    notify_item->tid = tid;
    notify_item->tflag = tflag;
    notify_item->next = gpio_notify_list;

    gpio_notify_list = notify_item;

    return KDP_STATUS_OK;
}

kdp_status_e kdp_gpio_set_interrupt_enable(kdp_gpio_pin_e pin, kdp_bool_e isEnable)
{
    uint32_t pin_val = (1 << pin);

    if (isEnable)
    {
        GpioRegMaskedSet(REG_GPIO_INT_CLEAR, pin_val);
        GpioRegMaskedSet(REG_GPIO_INT_ENABLE, pin_val);
        GpioRegMaskedClr(REG_GPIO_INT_MASK, pin_val);
    }
    else
    {
        GpioRegMaskedClr(REG_GPIO_INT_ENABLE, pin_val);
        GpioRegMaskedSet(REG_GPIO_INT_MASK, pin_val);
    }

    return KDP_STATUS_OK;
}

kdp_status_e kdp_gpio_set_debounce_enable(kdp_gpio_pin_e pin, kdp_bool_e isEnable, uint32_t debounce_clock)
{
    uint32_t pin_val = (1 << pin);

    if (isEnable){
        GpioRegMaskedSet(REG_GPIO_INT_BOUNCEENABLE, pin_val);
        GpioRegWrite(REG_GPIO_INT_PRESCALE, (APB_CLOCK/debounce_clock));
    }
    else
        GpioRegMaskedClr(REG_GPIO_INT_BOUNCEENABLE, pin_val);

    return KDP_STATUS_OK;
}

kdp_status_e kdp_gpio_set_pin_value(kdp_gpio_pin_e pin, kdp_bool_e value)
{
    uint32_t pin_val = (1 << pin);

    if (value == KDP_BOOL_TRUE)
        GpioRegWrite(REG_GPIO_DATASET, pin_val);
    else
        GpioRegWrite(REG_GPIO_DATACLR, pin_val);

    return KDP_STATUS_OK;
}

kdp_status_e kdp_gpio_get_pin_value(kdp_gpio_pin_e pin, kdp_bool_e *pValue)
{
    uint32_t all_bits = GpioRegRead(REG_GPIO_DIN_OFFSET);

    *pValue = (all_bits & (0x1 << pin)) ? KDP_BOOL_TRUE : KDP_BOOL_FALSE;

    return KDP_STATUS_OK;
}
