/*
 * Kneron Peripheral API - I2C
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include <stdlib.h>

#include "cmsis_os2.h"
#include "scu_extreg.h"
#include "kneron_mozart.h"

#include "kdp_io.h"
#include "kdp_i2c.h"
#include "dbg.h"
//#define I2C_DEBUG

#define I2CRegRead(id, reg_offset)  inw(IIC_FTIIC010_0_PA_BASE + id*0x100000 + reg_offset)
#define I2CRegWrite(id, reg_offset, val) outw(IIC_FTIIC010_0_PA_BASE + id*0x100000 + reg_offset, val)
#define I2CRegMaskedSet(id, reg_offset, val) masked_outw(IIC_FTIIC010_0_PA_BASE + id*0x100000 + reg_offset, val, val)
#define I2CRegMaskedClr(id, reg_offset, val) masked_outw(IIC_FTIIC010_0_PA_BASE + id*0x100000 + reg_offset, 0, val)


// I2C Control Register
#define REG_I2C_CR          0x00
#define CR_ALIRQ            BIT13  /* arbitration lost interrupt (master) */
#define CR_SAMIRQ           BIT12  /* slave address match interrupt (slave) */
#define CR_STOPIRQ          BIT11   /* stop condition interrupt (slave) */
#define CR_NAKRIRQ          BIT10   /* NACK response interrupt (master) */
#define CR_DRIRQ            BIT9   /* rx interrupt (both) */
#define CR_DTIRQ            BIT8   /* tx interrupt (both) */
#define CR_TBEN             BIT7    /* tx enable (both) */
#define CR_NAK              BIT6    /* NACK (both) */
#define CR_STOP             BIT5    /* stop (master) */
#define CR_START            BIT4    /* start (master) */
#define CR_GCEN             BIT3     /* general call support (slave) */
#define CR_MST_EN           BIT2     /* enable clock out (master) */
#define CR_I2C_EN           BIT1     /* enable I2C (both) */
#define CR_I2C_RST          BIT0     /* reset I2C (both) */
#define CR_ENABLE           \
	(CR_ALIRQ | CR_NAKRIRQ | CR_DRIRQ | CR_DTIRQ | CR_MST_EN | CR_I2C_EN)

// I2C Status Register
#define REG_I2C_SR          0x04
#define SR_SBS              BIT23 /* start byte */
#define SR_HSS              BIT22 /* high speed mode */
#define SR_START            BIT11    /* start */
#define SR_AL               BIT10    /* arbitration lost */
#define SR_GC               BIT9    /* general call */
#define SR_SAM              BIT8    /* slave address match */
#define SR_STOP             BIT7     /* stop received */
#define SR_NACK             BIT6     /* NACK received */
#define SR_TD               BIT5    /* transfer done */
#define SR_BB               BIT3     /* bus busy */
#define SR_I2CB             BIT2     /* chip busy */
#define SR_RW               BIT0      /* set when master-rx or slave-tx mode */

// I2C Clock Division Register
#define REG_I2C_CDR         0x08
#define CDR_COUNT_100KHZ \
    ((uint32_t)((APB_CLOCK)/(2*100000) - (GSR_VALUE/2) - 2)) // if APB=100, this will be 496
#define CDR_COUNT_400KHZ \
    ((uint32_t)((APB_CLOCK)/(2*400000) - (GSR_VALUE/2) - 2)) // if APB=100, this will be 121
#define CDR_COUNT_1MHZ \
    ((uint32_t)((APB_CLOCK)/(2*1000000) - (GSR_VALUE/2) - 2)) // if APB=100, this will be 46

// I2C Data Register
#define REG_I2C_DR          0x0C

// I2C Address Register
#define REG_I2C_AR          0x10

// I2C Set/Hold Time and Glitch Suppresion Setting Register
#define REG_I2C_TGSR        0x14
#define GSR_VALUE            0x5
#define TSR_VALUE            0x5

// I2C Bus Monitor Register
#define REG_I2C_BMR        0x18

// I2C Burst Mode Register
#define REG_I2C_BSTMR        0x1C

// I2C Revision Register
#define REG_I2C_REVISION        0x30

#define NUM_I2C_CTRL 4

#define i2c_printf(...) {fLib_printf("[i2c]");fLib_printf( __VA_ARGS__ );}

#ifdef I2C_DEBUG
void kdp_i2c_dump_register(kdp_i2c_ctrl_e ctrl_id)
{
    i2c_printf("APB clock = %d\n", APB_CLOCK);
    i2c_printf("(0x00) control register = 0x%x\n", I2CRegRead(ctrl_id, REG_I2C_CR));
    i2c_printf("(0x04) status register = 0x%x\n", I2CRegRead(ctrl_id, REG_I2C_SR));
    i2c_printf("(0x08) clock division = 0x%x\n", I2CRegRead(ctrl_id, REG_I2C_CDR));
    i2c_printf("(0x0C) data register = 0x%x\n", I2CRegRead(ctrl_id, REG_I2C_DR));
    i2c_printf("(0x10) address register = 0x%x\n", I2CRegRead(ctrl_id, REG_I2C_AR));
    i2c_printf("(0x14) TGSR register = 0x%x\n", I2CRegRead(ctrl_id, REG_I2C_TGSR));
    i2c_printf("(0x18) bus monitor = 0x%x\n", I2CRegRead(ctrl_id, REG_I2C_BMR));
    i2c_printf("(0x1C) burst mode = 0x%x\n", I2CRegRead(ctrl_id, REG_I2C_BSTMR));
    i2c_printf("(0x30) revision = 0x%x\n\n", I2CRegRead(ctrl_id, REG_I2C_REVISION));
}
#endif

kdp_status_e kdp_i2c_init(kdp_i2c_ctrl_e ctrl_id, kdp_i2c_bus_speed_e bus_speed)
{
    if (ctrl_id >= NUM_I2C_CTRL)
        return KDP_STATUS_ERROR_CTRL_NO_EXIST;

    // enable i2c_pclk
    // FIXME: should use internal system clock function (not yet implemented)
    // to manage all peripherals pclk
    uint32_t val = (0x2 << ctrl_id);
    masked_outw(SCU_FTSCU100_PA_BASE + 0x60, val, val);

    // reset i2c controller
    I2CRegMaskedSet(ctrl_id, REG_I2C_CR, CR_I2C_RST);

    int i;
    for (i = 0; i < 500; i++)
        if (!(I2CRegRead(ctrl_id, REG_I2C_CR) & CR_I2C_RST))
            break;

    if (i >= 500){
        i2c_printf("reset failed\n");
        return KDP_STATUS_ERROR_I2C_RESET_FAILED;
    }


    // set bus speed, SCL = PCLK / (2 * COUNT  + GSR + 4)

    // set GSR and TSR value
    I2CRegWrite(ctrl_id, REG_I2C_TGSR, (GSR_VALUE << 10) | TSR_VALUE);

    // select CDR Count value depending on bus speed
    uint32_t count_bits;
    switch(bus_speed)
    {
        case KDP_I2C_SPEED_100K:
            count_bits = CDR_COUNT_100KHZ;
            break;
        case KDP_I2C_SPEED_400K:
            count_bits = CDR_COUNT_400KHZ;
            break;
        case KDP_I2C_SPEED_1M:
            count_bits = CDR_COUNT_1MHZ;
            break;
        default:
            // could be user specified clock in Hz
            count_bits = ((APB_CLOCK)/(2*(uint32_t)bus_speed) - (GSR_VALUE/2) - 2);
            break;
    }

    I2CRegWrite(ctrl_id, REG_I2C_CDR, count_bits); // not care about COUNTH and DUTY bits

    return KDP_STATUS_OK;
}

kdp_status_e kdp_i2c_deinit(kdp_i2c_ctrl_e ctrl_id)
{
    // disable gpio_pclk
    // FIXME: should use system-clock API (not yet implemented) to manage all peripherals pclk
    uint32_t val = (0x2 << ctrl_id);
    masked_outw(SCU_FTSCU100_PA_BASE + 0x60, 0, val);

    return KDP_STATUS_OK;
}

static kdp_status_e i2c_polling_completion(kdp_i2c_ctrl_e ctrl_id, kdp_bool_e ignore_NAK)
{
    // polling SR and take actions correspondingly
    for (int i = 0; i < 10000; i++)
    {
        uint32_t sr_status = I2CRegRead(ctrl_id, REG_I2C_SR);
        I2CRegWrite(ctrl_id, REG_I2C_SR, sr_status);

        if (!ignore_NAK && (sr_status & SR_NACK))
            return KDP_STATUS_ERROR_I2C_DEV_NACK;
        else if (sr_status & (SR_AL))
            return KDP_STATUS_ERROR_I2C_BUS_BUSY;
        else if (sr_status & SR_TD)
            return KDP_STATUS_OK;
    }
    return KDP_STATUS_ERROR_I2C_STATUS_TIMEOUT;
}

static kdp_bool_e waiting_for_bus_available(kdp_i2c_ctrl_e ctrl_id)
{
    for (int i = 0; i < 10000; i++)
    {
        if (I2CRegRead(ctrl_id, REG_I2C_SR) & SR_BB)
            continue;
        else
            return KDP_BOOL_TRUE;
    }
    return KDP_BOOL_FALSE;
}

kdp_status_e i2c_tx_rx(kdp_i2c_ctrl_e ctrl_id, uint16_t slave_addr,
    uint8_t *data, uint32_t num, kdp_bool_e isRead, kdp_bool_e with_STOP)
{
    kdp_status_e status;
    kdp_bool_e ignore_NAK = KDP_BOOL_FALSE;

    if (!waiting_for_bus_available(ctrl_id))
        return KDP_STATUS_ERROR_I2C_BUS_BUSY;

    // send slave address and check status

    // write address byte into DR
    // TODO: support 10-bit address
    uint8_t byte_addr = (uint8_t)((slave_addr << 1) & 0xFE);
    if (isRead)
        byte_addr |= 0x1;

    I2CRegWrite(ctrl_id, REG_I2C_DR, byte_addr);

    // send out slave address with START condition by setting CR
    I2CRegWrite(ctrl_id, REG_I2C_CR, (CR_I2C_EN | CR_MST_EN | CR_TBEN | CR_START));

    // wait for complete status
    status = i2c_polling_completion(ctrl_id, ignore_NAK);
    if (status != KDP_STATUS_OK){
        if (status == KDP_STATUS_ERROR_I2C_DEV_NACK)
            I2CRegMaskedSet(ctrl_id, REG_I2C_CR, CR_STOP);

        return status;
    }

    if (!waiting_for_bus_available(ctrl_id))
        return KDP_STATUS_ERROR_I2C_BUS_BUSY;

    // send or receive data

    for (int i = 0; i < num; i++)
    {
        ignore_NAK = KDP_BOOL_FALSE;

        uint32_t ctrl_flag = (CR_I2C_EN | CR_MST_EN | CR_TBEN);
        if (with_STOP && (i == (num - 1)))
        {
            ctrl_flag |= CR_STOP;
            if (isRead){
                ctrl_flag |= CR_NAK;
                ignore_NAK = KDP_BOOL_TRUE;
            }
        }

        if (!isRead) // write data to the DR
            I2CRegWrite(ctrl_id, REG_I2C_DR, data[i]);

        // start transmission
        I2CRegWrite(ctrl_id, REG_I2C_CR, ctrl_flag);

        // check status
        // wait for complete status
        status = i2c_polling_completion(ctrl_id, ignore_NAK);
        if (status != KDP_STATUS_OK){
            if (status == KDP_STATUS_ERROR_I2C_DEV_NACK)
                I2CRegMaskedSet(ctrl_id, REG_I2C_CR, CR_STOP);

            return status;
        }

        if (isRead) // read data
            data[i] = (uint8_t)(0xFF & I2CRegRead(ctrl_id, REG_I2C_DR));
    }
    return KDP_STATUS_OK;
}

static u8 i2c_buf[32];  // allocate i2c transfer buffer

kdp_status_e kdp_i2c_transmit(
    kdp_i2c_ctrl_e ctrl_id, uint16_t slave_addr, u16 reg, u16 reg_size, u16 len, uint8_t *data)
{
    u8 *bufp = i2c_buf;

    if (reg_size == 2)
        *bufp++ = reg >> 8;  // store MSB of reg
    *bufp++ = reg & 0xff;
    *bufp = *data;  // we do only single byte transfer currently, i.e. len must be 1
    return i2c_tx_rx(ctrl_id, slave_addr, i2c_buf, reg_size+len, KDP_BOOL_FALSE, KDP_BOOL_TRUE);
}

kdp_status_e kdp_i2c_receive(
    kdp_i2c_ctrl_e ctrl_id, uint16_t slave_addr, u16 reg, u16 reg_size, u16 len, uint8_t *data)
{
    kdp_status_e ret;
    u8 *bufp = i2c_buf;

    if (reg_size == 2)
        *bufp++ = reg >> 8;  // store MSB of reg
    *bufp = reg & 0xff;
    ret = i2c_tx_rx(ctrl_id, slave_addr, i2c_buf, reg_size, KDP_BOOL_FALSE, KDP_BOOL_FALSE);
    if (ret == KDP_STATUS_OK) {
        ret = i2c_tx_rx(ctrl_id, slave_addr, data, len, KDP_BOOL_TRUE, KDP_BOOL_TRUE);
    }
    else {
        DSG("I2C read addressing error = %d", ret);
    }
    return ret;
}
