/*
 * Kneron System driver
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include "config/board_kdp520.h"
#include "system.h"
#include <cmsis_os2.h>
#include "scu_reg.h"
#include "scu_extreg.h"
#include "power.h"
#include "clock.h"
#include "ddr.h"
#include "DrvPWMTMR010.h"
#include "dbg.h"
#include "delay.h"
#include "kdp_memxfer.h"
#include "mpu.h"
#include "ota_update.h"
#include "kdp_crc.h"

extern const struct s_kdp_memxfer kdp_memxfer_module;

enum {
    SUBSYS_NPU      = 1,
    SUBSYS_PD_NPU,
    SUBSYS_LCDC,
    SUBSYS_NCPU,
};

u32 bootup_status;
u32 warm_boot;
u32 __sys_int_flag;

#define BOOTUP_STATUS_WARM      (SCU_REG_BTUP_STS_SMR | SCU_REG_BTUP_STS_PMR2)

static void sys_reset(int subsystem)
{
    switch (subsystem) {
        case SUBSYS_NPU:
            SCU_EXTREG_SWRST_SET_NPU_resetn(1);
            break;
        case SUBSYS_PD_NPU:
            SCU_EXTREG_SWRST_SET_PD_NPU_resetn(1);
            break;
        case SUBSYS_LCDC:
            SCU_EXTREG_SWRST_SET_LCDC_resetn(1);
            break;
        case SUBSYS_NCPU:
            SCU_EXTREG_CM4_NCPU_CTRL_SET_wakeup(0);
            SCU_EXTREG_CM4_NCPU_CTRL_SET_wakeup(1);    
    }
}

void system_isr(void)
{
    outw(SCU_REG_INT_STS, 0xffffffff); //clear sleep wakeup interrupt   

    __sys_int_flag = 0x1;
}

static void reset_handler(void)
{
    bootup_status = inw(SCU_REG_BTUP_STS);
    outw(SCU_REG_BTUP_STS, 0xffffffff);  // clear boot-up status
    outw(SCU_REG_BTUP_CTRL, SCU_REG_BTUP_CTRL_RTC_BU_EN | // RTC wakeup allowed
                            SCU_REG_BTUP_CTRL_PWRBTN_EN | // send power button output signal
                            SCU_REG_BTUP_CTRL_GPO_OUT);

    NVIC_ClearPendingIRQ((IRQn_Type)SYS_SYSTEM_IRQ);
    NVIC_EnableIRQ((IRQn_Type)SYS_SYSTEM_IRQ);

    outw(SCU_REG_INT_STS, 0xffffffff); // clear all interrupt status
    outw(SCU_REG_INT_EN, 0xffffffff); // enable all interrupts during boot

    //can't directly write to the PLL control pins, it needs to use the 
    //FCS or PLL_UPDATE command that contained he power-ode register
    clock_mgr_set_scuclkin(scuclkin_pll0div3, TRUE);
    delay_ms(2);

    power_mgr_ops(POWER_MGR_OPS_FCS);
    __WFI();
    do{
    }while((__sys_int_flag)!= 0x1);

//    clock_mgr_set_muxsel(CLOCK_MUXSEL_NCPU_TRACECLK_DEFAULT |
//                         CLOCK_MUXSEL_SCPU_TRACECLK_SRC_PLL0DIV3 |
//                         CLOCK_MUXSEL_CSIRX1_CLK_PLL5 |
//                         CLOCK_MUXSEL_NPU_CLK_PLL0 | /* CLOCK_MUXSEL_NPU_CLK_PLL4 | */ 
//                         CLOCK_MUXSEL_PLL4_FREF_PLL0DIV | 
//                         CLOCK_MUXSEL_UART_0_IRDA_UCLK_UART);
    delay_ms(1);
}

void system_init(void)
{
    delay_ms_enable();
    
    NVIC_SetVector((IRQn_Type)SYS_SYSTEM_IRQ, (u32)system_isr);

    reset_handler();

    if ((bootup_status & BOOTUP_STATUS_WARM) == BOOTUP_STATUS_WARM)
        warm_boot = 1;
    else
        warm_boot = 0;

    clock_mgr_init();

    /* Default power domain is already on */
    {
        //clock_mgr_open_pll4(); // npu
        delay_us(10 * 5);
        clk_enable(CLK_PLL4_FREF_PLL0);
        delay_us(30 * 5);
        clk_enable(CLK_PLL4);
        delay_us(30 * 5);
        clk_enable(CLK_PLL4_OUT);
    }
                
    /* Turn on NPU power domain */
    power_mgr_set_domain(POWER_DOMAIN_NPU, 1);
    if (warm_boot) {
        /* TODO: reload ncpu fw to NiRAM */
    }

    {
        /* Turn on DDR power domain for cold boot */
        power_mgr_set_domain(POWER_DOMAIN_DDRCK, 1);

        clk_enable(CLK_PLL1);
        delay_us(10 * 5);
        clk_enable(CLK_PLL1_OUT);
        delay_us(30 * 5);

        clk_enable(CLK_PLL2);
        delay_us(10 * 5);
        clk_enable(CLK_PLL2_OUT);

        clk_enable(CLK_PLL3);

#if (V2K_ENABLE_TYPE == V2K_ENABLE_HMX2056) || \
    (V2K_ENABLE_TYPE == V2K_ENABLE_OV9286) || \
    (V2K_ENABLE_TYPE == V2K_ENABLE_HMX2056_OV9286) || \
    (V2K_ENABLE_TYPE == V2K_ENABLE_HMXRICA)

        clock_mgr_change_pll3_clock(2, 268, 2, 5, 27, 13, 5, 15, 4);

#elif (V2K_ENABLE_TYPE == V2K_ENABLE_GC2145)
        clock_mgr_change_pll3_clock(2, 242, 2, 4, 11, 5, 4, 7, 1);
#elif (V2K_ENABLE_TYPE == V2K_ENABLE_SC132GS)
        clock_mgr_change_pll3_clock(2, 242, 2, 4, 7, 1, 4, 11, 5);
#elif (V2K_ENABLE_TYPE == V2K_ENABLE_GC2145_SC132GS)
        clock_mgr_change_pll3_clock(2, 242, 2, 4, 11, 5, 4, 7, 1);
#elif (V2K_ENABLE_TYPE == V2K_ENABLE_SC132GS_GC2145)
        clock_mgr_change_pll3_clock(2, 242, 2, 4, 7, 1, 4, 11, 5);
#endif
        delay_us(10 * 5);
        clk_enable(CLK_PLL3_OUT1);
        clk_enable(CLK_PLL3_OUT2);

        clock_mgr_change_pll5_clock(1, 0x63, 2);
    }
    
    //fcs
    {
        delay_us(10 * 5);
        clk_enable(CLK_PLL4_FREF_PLL0);
        delay_us(30 * 5);
        clk_enable(CLK_PLL4);
        delay_us(30 * 5);
        clk_enable(CLK_PLL4_OUT);

        NVIC_ClearPendingIRQ((IRQn_Type)SYS_SYSTEM_IRQ);
        NVIC_EnableIRQ((IRQn_Type)SYS_SYSTEM_IRQ);
        
        outw(SCU_REG_INT_STS, 0xffffffff); // clear all interrupt status

        clk_enable(CLK_FCS_PLL2);
        clk_enable(CLK_FCS_DLL);
        clock_mgr_set_scuclkin(scuclkin_pll0div3, FALSE);
        power_mgr_ops(POWER_MGR_OPS_FCS);

        __WFI();
        do {
        } while((__sys_int_flag)!= 0x1);
    }

    delay_ms_disable();
}

void system_wakeup_ncpu(int boot_loader_flag, unsigned char wakeup_all)
{
    delay_ms_enable();

    if (boot_loader_flag)
        delay_ms(200);

    if (1 == wakeup_all) {    
        clk_enable(CLK_SCPU_TRACE);
        clk_enable(CLK_NCPU);
        clk_enable(CLK_NPU);
        sys_reset(SUBSYS_PD_NPU);
        sys_reset(SUBSYS_NPU);
    }
    
    SCU_EXTREG_CM4_NCPU_CTRL_SET_wakeup(1);

    delay_ms_disable();
}

void system_init_ncpu(void)
{
    clk_enable(CLK_SCPU_TRACE);
    clk_enable(CLK_NCPU);
    clk_enable(CLK_NPU);
    sys_reset(SUBSYS_PD_NPU);
    sys_reset(SUBSYS_NPU);
}

/**
 * @brief RELOAD_NCPU_FW(), Reload NCPU firmware from flash
 */
void reload_ncpu_fw(void)
{
    mpu_niram_enable();   
    kdp_memxfer_module.init(MEMXFER_OPS_DMA, MEMXFER_OPS_DMA);
    kdp_memxfer_module.flash_to_niram();
    mpu_niram_disable();
    sys_reset(SUBSYS_NCPU);
}

/**
*   flag = 0, just launch ncpu
*   flag <> 0, load and launch ncpu
*   flag < 0, not using mpu
*
*   @retVal: 0, new sum32 value matches with another one in FLASH
*           >0, return wrong new sum32 value
*/
u32 load_ncpu_fw(int reset_flag)
{
    u32 ret = 0;
    SCU_EXTREG_CM4_NCPU_CTRL_SET_wakeup(0);  // stop ncpu first
    if (reset_flag) {
        if (reset_flag > 0)
            mpu_niram_enable();
        kdp_memxfer_module.init(MEMXFER_OPS_DMA, MEMXFER_OPS_DMA);
        kdp_memxfer_module.flash_to_niram();
        /*ncpu may change its interrupt vector in NiRAM, which may fail the image check.
        So ncpu check should be done before ncput is waken up and run*/
        ret = system_check_fw_image(NCPU_FW);
        if (reset_flag > 0)
            mpu_niram_disable();
    }
    SCU_EXTREG_CM4_NCPU_CTRL_SET_wakeup(1);  // restart ncpu
    return ret;
}

/**
 * @brief Calculate sum32 of scpu/ncpu fw and compare to the value which resides at the end of flash
 * N-iRAM(64KB) 0x28000000 ~ 0x2800FFFF
 * S-iRAM(88KB) 0x10102000 ~ 0x10117FFF
 * 
 * If sums are the same, return 0  (Note: check value could be zero)
 * Else return wrong calculated value, if value is zero, return 0xffffffff
 */
u32 system_check_fw_image(int fw_type)
{
    u8 *pBase;
    u32 sum32_cal, sum32_in_flash;
    u32 size;

    if(fw_type == SCPU_FW) {
        pBase = (u8 *)SCPU_START_ADDRESS;
        size = (SCPU_IMAGE_SIZE - 4);
    }
    else if(fw_type == NCPU_FW){
        pBase = (u8 *)NCPU_START_ADDRESS;
        size = (NCPU_IMAGE_SIZE - 4);
    }
    
    sum32_cal = kdp_gen_sum32(pBase, size);
    sum32_in_flash = *(u32 *)(pBase + size);
    if (sum32_cal == sum32_in_flash)
            return 0;
    if (sum32_cal == 0)
        sum32_cal--;
    return sum32_cal;
}

