/*
 * Kneron Peripheral API
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include <stdlib.h>
#include "cmsis_os2.h"
#include "scu_extreg.h"
#include "kdp_uart.h"

#include "kdp_io.h"
#include "kdp_usbd.h"
#include "kdp_usbd_def.h"

// an usb device mode object for internal use
usbd_ctrl_block_t usbd_cb =
    {
        .notifyTid = 0,
        .notifyFlag = 0,
        .ep0_halted = FALSE,
        .dev_desc = NULL,
        .config_state = CONFIG_DEFAULT_STATE,
};

// USB ISR
static void usbd_isr(void)
{
    usbd_handle_device_interrupts();
}

void usbd_init_reg_isr(void)
{
    SCU_EXTREG_USB_OTG_CTRL_SET_EXTCTRL_SUSPENDM(1);
    SCU_EXTREG_USB_OTG_CTRL_SET_u_iddig(1);
    SCU_EXTREG_USB_OTG_CTRL_SET_wakeup(0);
    SCU_EXTREG_USB_OTG_CTRL_SET_l1_wakeup(0);
    SCU_EXTREG_USB_OTG_CTRL_SET_OSCOUTEN(0);
    SCU_EXTREG_USB_OTG_CTRL_SET_PLLALIV(0);
    SCU_EXTREG_USB_OTG_CTRL_SET_XTLSEL(0);
    SCU_EXTREG_USB_OTG_CTRL_SET_OUTCLKSEL(0);

    // disable all OTG interrupts
    UsbRegWrite(REG_OTG_IER, ~(0x0));

    // enable only Device interrupt (no Host or OTG)
    UsbRegWrite(REG_GLB_INT, ~(INT_POLARITY | DEV_INT) & 0xF);

    // listen all 4 groups for 0x140
    UsbRegWrite(REG_DEV_MIGR, 0x0);

    // grop 0 interrupt mask
    // FIXME: should care about mroe interrupts
    UsbRegWrite(REG_DEV_MISG0,
                ~(G0_CX_SETUP_INT_RO |
                  G0_CX_COMFAIL_INT_RO |
                  G0_CX_COMABT_INT_RW1C));

    // listen no group 1 interrrupts for 0x148
    UsbRegWrite(REG_DEV_MISG1, 0xFFFFFFFF);

    // enable interested interrupts in group 2 0x14C
    UsbRegWrite(REG_DEV_MISG2,
                ~(G2_RX0BYTE_INT_RW1C |
                  G2_DMA_ERROR_RW1C |
                  G2_SUSP_INT_RW1C |
                  G2_RESM_INT_RW1C |
                  G2_USBRST_INT_RW1C));

    // set device idle counter = 7ms
    UsbRegWrite(REG_DEV_ICR, 0x7);
    // device soft reset
    UsbRegMaskedSet(REG_DEV_CTL, BIT4);
    // clear all FIFO counter
    UsbRegMaskedSet(REG_DEV_TST, BIT0);
    // enable chip
    UsbRegMaskedSet(REG_DEV_CTL, BIT5);

    // clear all interrupts status for RW1C bits
    UsbRegWrite(REG_OTG_ISR, 0xFFFFFFFF);
    UsbRegWrite(REG_DEV_ISG0, 0xFFFFFFFF);
    UsbRegWrite(REG_DEV_ISG2, 0xFFFFFFFF);

    // global interrupt enable
    UsbRegMaskedSet(REG_DEV_CTL, BIT2);

    NVIC_SetVector(OTG_SBS_3_IRQ, (uint32_t)usbd_isr);

    // Clear and Enable SAI IRQ
    NVIC_ClearPendingIRQ(OTG_SBS_3_IRQ);
    NVIC_EnableIRQ(OTG_SBS_3_IRQ);
}

kdp_status_e kdp_usbd_init(void)
{
    usbd_cb.notifyTid = 0;
    usbd_cb.notifyFlag = 0x0;
    usbd_cb.ep0_halted = FALSE;
    usbd_cb.dev_desc = NULL;

    usbd_reset_event_queue();

    usbd_cb.evt_id = osEventFlagsNew(NULL);
    if (usbd_cb.evt_id == 0)
        usbprintf("!!!!!!! create event flag failed\n");

    usbd_init_reg_isr();

    return KDP_STATUS_OK;
}

kdp_status_e kdp_usbd_deinit(void)
{
    usbd_clean_event_queue();

    return KDP_STATUS_ERROR;
}

kdp_status_e kdp_usbd_reset_device()
{
    // if it is not configured, no need to reset
    if (!kdp_usbd_is_dev_configured())
        return KDP_STATUS_ERROR;

    // disable bus
    kdp_usbd_set_enable(KDP_BOOL_FALSE);

    // first reset all endpoints and terminate all in-progress transfer
    usbd_fifo_ctrl_t *fifo_cb = usbd_cb.fifo_cbs;
    for (int fno = 0; fno < FIFO_NUM; fno++)
        if (fifo_cb[fno].enabled)
        {
            kdp_usbd_reset_endpoint(fifo_cb[fno].endpointAddress);
        }

    // some delay for USB bus, FIXME ?
    osDelay(100);

    usbd_cb.ep0_halted = FALSE;
    usbd_cb.config_state = CONFIG_DEFAULT_STATE;

    // re-init registers and isr
    usbd_init_reg_isr();

    usbd_reset_event_queue();

    // re-enable bus
    kdp_usbd_set_enable(KDP_BOOL_TRUE);

    return KDP_STATUS_OK;
}

kdp_status_e kdp_usbd_set_device_descriptor(
    kdp_usbd_speed_e speed,
    kdp_usbd_device_descriptor_t *dev_desc)
{
    // for now support high speed only, FIXME
    if (speed != KDP_USBD_HIGH_SPEED)
        return KDP_STATUS_ERROR;

    if (dev_desc->bNumConfigurations > 1)
        return KDP_STATUS_ERROR;

    usbd_cb.dev_desc = dev_desc;

    return KDP_STATUS_OK;
}

kdp_status_e kdp_usbd_set_device_qualifier_descriptor(
    kdp_usbd_speed_e speed,
    kdp_usbd_device_qualifier_descriptor_t *dev_qual_desc)
{
    usbd_cb.dev_qual_desc = dev_qual_desc;
    return KDP_STATUS_OK;
}

kdp_status_e kdp_usbd_set_enable(kdp_bool_e enable)
{
    if (enable)
        // Make PHY work properly, FIXME ?
        UsbRegMaskedClr(REG_PHY_TST, TST_JSTA);
    else
        // Make PHY not work, FIXME ?
        UsbRegMaskedSet(REG_PHY_TST, TST_JSTA);

    return KDP_STATUS_OK;
}

kdp_bool_e kdp_usbd_is_dev_configured(void)
{
    if (usbd_cb.config_state == CONFIG_CONFIGURED_STATE)
        return KDP_BOOL_TRUE;
    else
        return KDP_BOOL_FALSE;
}

kdp_status_e kdp_usbd_get_event(kdp_usbd_event_t *uevent)
{
    int ret = usbd_pop_event_from_queue(uevent);

    if (ret)
        return KDP_STATUS_OK;
    else
        return KDP_STATUS_ERROR;
}

kdp_status_e kdp_usbd_set_notify(osThreadId_t tid, uint32_t tflag)
{
    usbd_cb.notifyTid = tid;
    usbd_cb.notifyFlag = tflag;
    return KDP_STATUS_OK;
}

kdp_status_e kdp_usbd_control_send_sync(uint8_t *buf, uint32_t size, uint32_t timeout_ms)
{
    uint32_t left_send = size;
    const uint32_t cx_fifo_size = 64;
    uint32_t tryMs = 0;
    bool first_written = TRUE;

    // to make sure cx fifo is empty
    UsbRegMaskedSet(REG_CXCFE, CX_CLR);

    while (left_send > 0)
    {
        bool doTxfer = FALSE;

        if (first_written || (UsbRegRead(REG_DEV_ISG0) & G0_CX_IN_INT_RO))
        {
            uint32_t transfer_size = MIN(cx_fifo_size, left_send);
            // note: there is no G0_CX_IN_INT_RO for first 64-byte write
            // so we need to write it first
            if (usbd_dma_fifo_transfer_sync_try((uint32_t *)buf, transfer_size, DMA_TARGET_ACC_CXF, WRITE_FIFO, 500))
            {
                doTxfer = TRUE;
                tryMs = 0;
                left_send -= transfer_size;
                buf += transfer_size;
                first_written = FALSE;
            }
        }

        if (!doTxfer)
        {
            osDelay(1);
            ++tryMs;
            if (tryMs >= timeout_ms)
                return KDP_STATUS_ERROR_USBD_TRANSFER_TIMEOUT;
        }
    }

    return KDP_STATUS_OK;
}

kdp_status_e kdp_usbd_control_receive_sync(uint8_t *buf, uint32_t *size, uint32_t timeout_ms)
{
    uint32_t wanted_size = *size;
    uint32_t tryMs = 0;
    *size = 0;

    while (wanted_size > 0)
    {
        bool doTxfer = FALSE;

        if (UsbRegRead(REG_DEV_ISG0) & G0_CX_OUT_INT_RO)
        {
            uint32_t fifo_bytes = (UsbRegRead(REG_CXCFE) >> 24) & 0x7F;
            uint32_t transfer_size = MIN(fifo_bytes, wanted_size);

            if (usbd_dma_fifo_transfer_sync_try((uint32_t *)buf, transfer_size, DMA_TARGET_ACC_CXF, READ_FIFO, 500))
            {
                doTxfer = TRUE;
                tryMs = 0;
                wanted_size -= transfer_size;
                buf += transfer_size;
                *size += transfer_size;
            }
        }

        if (!doTxfer)
        {
            osDelay(1);
            ++tryMs;
            if (tryMs >= timeout_ms)
                return KDP_STATUS_ERROR_USBD_TRANSFER_TIMEOUT;
        }
    }

    return KDP_STATUS_OK;
}

kdp_status_e kdp_usbd_control_respond(kdp_usbd_status_report_e status)
{
    if (status == KDP_USBD_RESPOND_OK)
        // respond ACK
        UsbRegMaskedSet(REG_CXCFE, CX_DONE);
    else if (status == KDP_USBD_RESPOND_ERROR)
        // respond STALL
        UsbRegMaskedSet(REG_CXCFE, CX_STL | CX_DONE);
    else
    {
        usbprintf("cx respond: wrong parameter!\n");
        return KDP_STATUS_ERROR;
    }

    return KDP_STATUS_OK;
}

static kdp_status_e usbd_bulk_in_send(uint32_t endpoint, uint32_t *buf, uint32_t txLen, int8_t isBlockingCall)
{
    usbd_fifo_ctrl_t *fifo_cb = usbd_cb.fifo_cbs;

    // find out which FIFO no for the IN endpoint address
    int8_t fno = usbd_endpoint_to_fifo(endpoint);

    // below do some error checking
    {
        if (fno == -1)
            return KDP_STATUS_ERROR_USBD_INVALID_ENDP;

        if (fifo_cb[fno].transferType != TXFER_BULK)
            return KDP_STATUS_ERROR_USBD_INVALID_TRANSFER;

        if (fifo_cb[fno].isTransferring)
            return KDP_STATUS_ERROR_USBD_TRANSFER_IN_PROGRESS;
    }

    // set up fifo cb
    fifo_cb[fno].isTransferring = TRUE;
    fifo_cb[fno].user_buf_addr = usbd_dma_remap_addr((uint32_t)buf);
    fifo_cb[fno].user_buf_len = txLen;
    // check if need to send a zero-length packet at the end of transfer
    fifo_cb[fno].short_or_zl_packet = ((txLen & (fifo_cb[fno].maxPacketSize - 1)) == 0) ? TRUE : FALSE;
    fifo_cb[fno].isBlockingCall = isBlockingCall;

    // reset FIFO content before transmission
    UsbRegMaskedSet(fifo_cb[fno].byteCntReg, FFRST);

    // enable (unmask) the FIFO IN interrupt
    UsbRegMaskedClr(REG_DEV_MISG1, MF0_IN_INT << fno);

    // now leave and let interrupt handler do the transfer work

    return KDP_STATUS_OK;
}

kdp_status_e kdp_usbd_bulk_send_sync(uint32_t endpoint, uint32_t *buf, uint32_t txLen, uint32_t timeout_ms)
{
    kdp_status_e status = usbd_bulk_in_send(endpoint, buf, txLen, TRUE);

    if (status != KDP_STATUS_OK)
        return status;

    usbd_fifo_ctrl_t *fifo_cb = usbd_cb.fifo_cbs;
    int8_t fno = usbd_endpoint_to_fifo(endpoint);

    if (timeout_ms == 0)
        timeout_ms = osWaitForever;

    uint32_t flags = osEventFlagsWait(usbd_cb.evt_id, (0x1 << fno), osFlagsWaitAny, timeout_ms);
    if (flags == osFlagsErrorTimeout)
    {
        usbd_clean_fifo_cb(&fifo_cb[fno]);
        status = KDP_STATUS_ERROR_USBD_TRANSFER_TIMEOUT;
    }
    else
    {
        if (fifo_cb[fno].cur_event == KDP_USBD_EVENT_TRANSFER_DONE)
            status = KDP_STATUS_OK;
        else
            status = KDP_STATUS_ERROR;
    }

    usbd_clean_fifo_cb(&fifo_cb[fno]);
    return status;
}

kdp_status_e kdp_usbd_reset_endpoint(uint32_t endpoint)
{
    int8_t fno = usbd_endpoint_to_fifo(endpoint);

    if (fno == -1)
        return KDP_STATUS_ERROR_USBD_INVALID_ENDP;

    usbd_fifo_ctrl_t *fifo_cb = &usbd_cb.fifo_cbs[fno];

    UsbRegMaskedSet(fifo_cb->byteCntReg, FFRST);

    if (fifo_cb->transferType == TXFER_BULK)
    {
        if (fifo_cb->isTransferring)
        {
            // notify transfer done to user
            if (fifo_cb->isBlockingCall)
            {
                fifo_cb->cur_event = KDP_USBD_EVENT_TRANSFER_TERMINATED;
                osEventFlagsSet(usbd_cb.evt_id, 0x1 << fno);
            }
            else
            {
                usbd_notify_event_to_user(KDP_USBD_EVENT_TRANSFER_TERMINATED, fifo_cb->endpointAddress, 0);
            }
            fifo_cb->isTransferring = FALSE;
        }

        uint8_t enpNo = fifo_cb->enpNo;
        if (fifo_cb->endpointAddress & 0x80)
        {
            // for IN endpoints
            UsbRegMaskedSet(REG_DEV_INMPS_1 + 4 * (enpNo - 1), RSTG_IEPn);
            UsbRegMaskedClr(REG_DEV_INMPS_1 + 4 * (enpNo - 1), RSTG_IEPn);
            UsbRegMaskedSet(REG_DEV_MISG1, MF0_IN_INT << fno);
        }
        else
        {
            // for OUT endpoints
            UsbRegMaskedSet(REG_DEV_OUTMPS_1 + 4 * (enpNo - 1), RSTG_OEPn);
            UsbRegMaskedClr(REG_DEV_OUTMPS_1 + 4 * (enpNo - 1), RSTG_OEPn);
            UsbRegMaskedClr(REG_DEV_MISG1, (MF0_SPK_INT | MF0_OUT_INT) << (fno * 2));
        }
        usbd_clean_fifo_cb(fifo_cb);
    }
    return KDP_STATUS_OK;
}

kdp_status_e kdp_usbd_bulk_send_async(uint32_t endpoint, uint32_t *buf, uint32_t txLen)
{
    return usbd_bulk_in_send(endpoint, buf, txLen, FALSE);
}

static kdp_status_e usbd_bulk_out_receive(uint32_t endpoint, uint32_t *buf, uint32_t blen, int8_t isBlockingCall)
{
    usbd_fifo_ctrl_t *fifo_cb = usbd_cb.fifo_cbs;

    // find out which FIFO no for the IN endpoint address
    int8_t fno = usbd_endpoint_to_fifo(endpoint);

    // below do some error checking
    {
        if (fno == -1)
            return KDP_STATUS_ERROR_USBD_INVALID_ENDP;

        if (fifo_cb[fno].transferType != TXFER_BULK)
            return KDP_STATUS_ERROR_USBD_INVALID_TRANSFER;

        if (fifo_cb[fno].isTransferring)
            return KDP_STATUS_ERROR_USBD_TRANSFER_IN_PROGRESS;
    }

    // set up fifo cb
    fifo_cb[fno].isTransferring = TRUE;
    fifo_cb[fno].user_buf_addr = usbd_dma_remap_addr((uint32_t)buf);
    fifo_cb[fno].user_buf_len = blen;
    fifo_cb[fno].received_length = 0;
    fifo_cb[fno].isBlockingCall = isBlockingCall;

    // the OUT interrupts should have been disabled earlier, re-enable it since buffer is commited
    UsbRegMaskedClr(REG_DEV_MISG1, MF0_OUT_INT << (fno * 2));

    return KDP_STATUS_OK;
}

kdp_status_e kdp_usbd_bulk_receive_sync(uint32_t endpoint, uint32_t *buf, uint32_t *blen, uint32_t timeout_ms)
{
    kdp_status_e status = usbd_bulk_out_receive(endpoint, buf, *blen, TRUE);

    if (status != KDP_STATUS_OK)
        return status;

    usbd_fifo_ctrl_t *fifo_cb = usbd_cb.fifo_cbs;
    int8_t fno = usbd_endpoint_to_fifo(endpoint);

    if (timeout_ms == 0)
        timeout_ms = osWaitForever;

    uint32_t flags = osEventFlagsWait(usbd_cb.evt_id, (0x1 << fno), osFlagsWaitAny, timeout_ms);
    if (flags == osFlagsErrorTimeout)
    {
        status = KDP_STATUS_ERROR_USBD_TRANSFER_TIMEOUT;
    }
    else
    {
        *blen = fifo_cb[fno].received_length;
        if (fifo_cb[fno].cur_event == KDP_USBD_EVENT_TRANSFER_DONE)
            status = KDP_STATUS_OK;
        else
            status = KDP_STATUS_ERROR;
    }

    usbd_clean_fifo_cb(&fifo_cb[fno]);
    return status;
}

kdp_status_e kdp_usbd_bulk_receive_async(uint32_t endpoint, uint32_t *buf, uint32_t blen)
{
    return usbd_bulk_out_receive(endpoint, buf, blen, FALSE);
}

kdp_status_e kdp_usbd_interrupt_send_sync(uint32_t endpoint, uint32_t *buf, uint32_t txLen, uint32_t timeout_ms)
{
    usbd_fifo_ctrl_t *fifo_cb = usbd_cb.fifo_cbs;

    // find out which FIFO no for the IN endpoint address
    int8_t fno = usbd_endpoint_to_fifo(endpoint);

    // below do some error checking
    {
        if (fno == -1)
            return KDP_STATUS_ERROR_USBD_INVALID_ENDP;

        if (fifo_cb[fno].transferType != TXFER_INT)
            return KDP_STATUS_ERROR_USBD_INVALID_TRANSFER;
    }

    // try to wait some time for FIFO empty
    // FIXME: 3000 ?
    for (uint32_t try = 0; try <= 3000; try ++)
        if (UsbRegRead(REG_CXCFE) & (F_EMP_0 << fno))
            break;

    // directly clear FIFO content before transmission
    UsbRegMaskedSet(fifo_cb[fno].byteCntReg, FFRST);

    uint32_t tryMs = 0;
    while (1)
    {
        if (TRUE == usbd_dma_fifo_transfer_sync_try(buf, txLen, 0x1 << fno, WRITE_FIFO, 500))
            break;

        if (timeout_ms != 0 && tryMs >= timeout_ms)
            return KDP_STATUS_ERROR_USBD_TRANSFER_TIMEOUT;

        osDelay(1);
        ++tryMs;
    }

    return KDP_STATUS_OK;
}

kdp_status_e kdp_usbd_interrupt_receive_sync(uint32_t endpoint, uint32_t *buf, uint32_t *rxLen, uint32_t timeout_ms)
{
    usbd_fifo_ctrl_t *fifo_cb = usbd_cb.fifo_cbs;

    // find out which FIFO no for the IN endpoint address
    int8_t fno = usbd_endpoint_to_fifo(endpoint);

    // below do some error checking
    {
        if (fno == -1)
            return KDP_STATUS_ERROR_USBD_INVALID_ENDP;

        if (fifo_cb[fno].transferType != TXFER_INT)
            return KDP_STATUS_ERROR_USBD_INVALID_TRANSFER;
    }

    uint32_t fifo_bytecnt = UsbRegRead(fifo_cb[fno].byteCntReg) & BC_Fn;
    // can transfer only minimum size betwwen FIFO byte count and user buffer residual size
    uint32_t transfer_size = MIN(fifo_bytecnt, *rxLen);

    uint32_t tryMs = 0;
    while (1)
    {
        if (TRUE == usbd_dma_fifo_transfer_sync_try(buf, transfer_size, 0x1 << fno, READ_FIFO, 500))
            break;

        if (timeout_ms != 0 && tryMs >= timeout_ms)
            return KDP_STATUS_ERROR_USBD_TRANSFER_TIMEOUT;

        osDelay(1);
        ++tryMs;
    }

    return KDP_STATUS_OK;
}
