#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cmsis_os2.h"
#include "scu_extreg.h"
#include "kdp_uart.h"

#include "kdp_io.h"
#include "kdp_usbd.h"
#include "kdp_usbd_def.h"

#ifdef USBD_DEBUG

static uint32_t fifo_to_endpoint_address_via_reg(uint32_t fifo_no)
{
    // read specific register to translate fifo no to corresponding endpoint address

    uint32_t fifo_map_val = UsbRegRead(REG_DEV_FMAP);
    uint32_t endpoint = (fifo_map_val >> (fifo_no * 8)) & 0xF;
    uint32_t dir = (fifo_map_val >> ((fifo_no * 8) + 4)) & 0x3;
    if (dir == 0x1)
        endpoint = 0x80 | endpoint;
    return endpoint;
}

static uint32_t endpoint_to_fifo_no_via_reg(uint32_t endpoint)
{
    // read specific registers to translate endpoint address to corresponding fifo no

    uint32_t isOut = !(endpoint & 0x80);
    uint32_t epNo = endpoint & 0xF;
    uint32_t shift_bits = (epNo <= 4) ? (epNo - 1) * 8 : (epNo - 5) * 8;
    uint32_t map_reg = (epNo <= 4) ? 0x1A0 : 0x1A4;

    if (isOut)
        shift_bits += 4;

    return (UsbRegRead(map_reg) >> shift_bits) & 0x3;
}

void usbd_print_status_registers(void)
{
    usbprintf("\n=== status registers ===\n\n");

    usbprintf("(0x80) OTG status = 0x%08x\n", inw(USB_FOTG210_PA_BASE + 0x80));
    usbprintf("(0xE4) HW config = 0x%08x\n", inw(USB_FOTG210_PA_BASE + 0xE4));
    usbprintf("(0x100) Device main ctrl = 0x%08x\n", inw(USB_FOTG210_PA_BASE + 0x100));
    usbprintf("(0x114) PHY Test Mode = 0x%08x\n", inw(USB_FOTG210_PA_BASE + 0x114));
    usbprintf("(0x120) CX Config / FIFO empty = 0x%08x\n", inw(USB_FOTG210_PA_BASE + 0x120));
    usbprintf("(0x1A0) Endpoint map 0 = 0x%08x\n", inw(USB_FOTG210_PA_BASE + 0x1A0));
    usbprintf("(0x1A0) Endpoint map 1 = 0x%08x\n", inw(USB_FOTG210_PA_BASE + 0x1A4));
    usbprintf("(0x1A8) FIFO map = 0x%08x\n", inw(USB_FOTG210_PA_BASE + 0x1A8));
    usbprintf("(0x1AC) FIFO config = 0x%08x\n", inw(USB_FOTG210_PA_BASE + 0x1AC));
    usbprintf("(0x1C0) DMA FIFO select = 0x%08x\n\n", inw(USB_FOTG210_PA_BASE + 0x1C0));
}

void usbd_print_interrupt_status(void)
{
    u32 otg_int_status, otg_int_enable;
    u32 global_int_status, global_int_mask;
    u32 dev_int_status, dev_int_mask;
    u32 grp0_int_status, grp0_int_mask;
    u32 grp1_int_status, grp1_int_mask;
    u32 grp2_int_status, grp2_int_mask;

    otg_int_status = inw(USB_FOTG210_PA_BASE + 0x84);
    otg_int_enable = inw(USB_FOTG210_PA_BASE + 0x88);

    global_int_status = inw(USB_FOTG210_PA_BASE + 0xC0);
    global_int_mask = inw(USB_FOTG210_PA_BASE + 0xC4);

    dev_int_status = inw(USB_FOTG210_PA_BASE + 0x140);
    dev_int_mask = inw(USB_FOTG210_PA_BASE + 0x130);

    grp0_int_status = inw(USB_FOTG210_PA_BASE + 0x144);
    grp0_int_mask = inw(USB_FOTG210_PA_BASE + 0x134);

    grp1_int_status = inw(USB_FOTG210_PA_BASE + 0x148);
    grp1_int_mask = inw(USB_FOTG210_PA_BASE + 0x138);

    grp2_int_status = inw(USB_FOTG210_PA_BASE + 0x14C);
    grp2_int_mask = inw(USB_FOTG210_PA_BASE + 0x13C);

    usbprintf("\n=== interrupt registers ===\n\n");

    usbprintf("(0x84) OTG interrupt = 0x%08x\n", otg_int_status);
    usbprintf("(0x88) interrupt enable = 0x%08x\n", otg_int_enable);
    usbprintf("triggered = 0x%08x\n\n", otg_int_status & otg_int_enable);

    usbprintf("(0xC0) Global interrupt = 0x%08x\n", global_int_status);
    usbprintf("(0xC4) interrupt mask = 0x%08x\n", global_int_mask);
    usbprintf("triggered = 0x%08x\n\n", 0x7 & (global_int_status & ~global_int_mask));

    usbprintf("(0x140) Device interrupt = 0x%08x\n", dev_int_status);
    usbprintf("(0x130) interrupt mask = 0x%08x\n", dev_int_mask);
    usbprintf("triggered = 0x%08x\n\n", 0xF & (dev_int_status & ~dev_int_mask));

    usbprintf("(0x144) Group 0 interrupt = 0x%08x\n", grp0_int_status);
    usbprintf("(0x134) interrupt mask = 0x%08x\n", grp0_int_mask);
    usbprintf("triggered = 0x%08x\n\n", 0x3F & (grp0_int_status & ~grp0_int_mask));

    usbprintf("(0x148) Group 1 interrupt = 0x%08x\n", grp1_int_status);
    usbprintf("(0x138) interrupt mask = 0x%08x\n", grp1_int_mask);
    usbprintf("triggered = 0x%08x\n\n", (grp1_int_status & ~grp1_int_mask));

    usbprintf("(0x14C) Group 2 interrupt = 0x%08x\n", grp2_int_status);
    usbprintf("(0x13C) interrupt mask = 0x%08x\n", grp2_int_mask);
    usbprintf("triggered = 0x%08x\n\n", (grp2_int_status & ~grp2_int_mask));
}

void usbd_print_setup_packet(kdp_usbd_setup_packet_t *setup)
{
    uint8_t *bytes = (uint8_t *)setup;
    usbprintf("SETUP packet hex: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
              bytes[0], bytes[1], bytes[2], bytes[3], bytes[4], bytes[5], bytes[6], bytes[7]);

    char str_type[12];
    char str_recp[12];

    uint8_t bmRequestType_type = ((setup->bmRequestType & 0x60) >> 5);
    switch (bmRequestType_type)
    {
    case 0:
        strcpy(str_type, "Standard");
        break;
    case 1:
        strcpy(str_type, "Class");
        break;
    case 2:
        strcpy(str_type, "Vendor");
        break;
    case 3:
        strcpy(str_type, "Reserved");
        break;
    }

    uint8_t bmRequestType_recipient = (setup->bmRequestType & 0x1F);
    switch (bmRequestType_recipient)
    {
    case 0:
        strcpy(str_recp, "Device");
        break;
    case 1:
        strcpy(str_recp, "Interface");
        break;
    case 2:
        strcpy(str_recp, "Endpoint");
        break;
    case 3:
        strcpy(str_recp, "Other");
        break;
    default:
        strcpy(str_recp, "Reserved");
        break;
    }

    usbprintf("bmRequestType: 0x%02x (%s, %s, %s)\n", setup->bmRequestType,
              (setup->bmRequestType & 0x80) ? "Device to Host" : "Host to Device", str_type, str_recp);

    char str_bReq[20] = "???";
    char str_wValue[64] = "???";

    if (bmRequestType_type == 0)
    {
        switch (setup->bRequest)
        {
        case 0x0:
            strcpy(str_bReq, "GET_STATUS");
            break;
        case 0x1:
        {
            strcpy(str_bReq, "CLEAR_FEATURE");
            switch (setup->wValue)
            {
            case 0:
                strcpy(str_wValue, "endpoint halt");
                break;
            case 1:
                strcpy(str_wValue, "device remote wakeup");
                break;
            case 2:
                strcpy(str_wValue, "test mode");
                break;
            }
        }
        break;
        case 0x3:
            strcpy(str_bReq, "SET_FEATURE");
            break;
        case 0x5:
            strcpy(str_bReq, "SET_ADDRESS");
            strcpy(str_wValue, "Address value");
            break;
        case 0x6:
        {
            strcpy(str_bReq, "GET_DESCRIPTOR");

            // high byte: descriptor type
            switch (setup->wValue >> 8)
            {
            case 1:
                strcpy(str_wValue, "DEVICE");
                break;
            case 2:
                strcpy(str_wValue, "CONFIGURATION");
                break;
            case 3:
                strcpy(str_wValue, "STRING");
                break;
            case 4:
                strcpy(str_wValue, "INTERFACE");
                break;
            case 5:
                strcpy(str_wValue, "ENDPOINT");
                break;
            case 6:
                strcpy(str_wValue, "DEVICE_QUALIFIER");
                break;
            case 7:
                strcpy(str_wValue, "OTHER_SPEED_CONFIGURATION");
                break;
            case 8:
                strcpy(str_wValue, "INTERFACE_POWER");
                break;
            }
            sprintf(str_wValue, "%s, %d", str_wValue, (setup->wValue & 0xFF));
        }
        break;
        case 0x7:
            strcpy(str_bReq, "SET_DESCRIPTOR");
            break;
        case 0x8:
            strcpy(str_bReq, "GET_CONFIGURATION");
            break;
        case 0x9:
            strcpy(str_bReq, "SET_CONFIGURATION");
            strcpy(str_wValue, "Configuration value");
            break;
        }
    }

    usbprintf("bRequest: 0x%02x (%s)\n", setup->bRequest, str_bReq);
    usbprintf("wValue: 0x%x (%s)\n", setup->wValue, str_wValue);
    usbprintf("wIndex: 0x%x\n", setup->wIndex);
    usbprintf("wLength: %d\n\n", setup->wLength);
}

void usbd_print_device_descriptor(kdp_usbd_device_descriptor_t *desc)
{
    uint8_t *byte = (uint8_t *)desc;

    usbprintf("device descp bytes = ");
    for (int i = 0; i < 18; i++)
        fLib_printf("0x%02x ", byte[i]);
    fLib_printf("\n\n\n");
}

void usbd_print_fifo_endpoint_configurations(kdp_usbd_interface_descriptor_t *intf)
{
    usbprintf("=== Endpoint configurations ===\n");
    for (int i = 0; i < intf->bNumEndpoints; i++)
    {
        uint8_t bEndpointAddress = intf->endpoint[i]->bEndpointAddress;
        uint8_t bmAttributes = intf->endpoint[i]->bmAttributes;
        uint16_t wMaxPacketSize = intf->endpoint[i]->wMaxPacketSize;

        char transfer_str[24];
        switch (bmAttributes & 0x3)
        {
        case 0:
            strcpy(transfer_str, "Control");
            break;
        case 1:
            strcpy(transfer_str, "Isochronous");
            break;
        case 2:
            strcpy(transfer_str, "Bulk");
            break;
        case 3:
            strcpy(transfer_str, "Interrupt");
            break;
        }

        usbprintf("Endpoint %s %s %d, max packet = %d\n",
                  transfer_str,
                  (bEndpointAddress & 0x80) ? "IN" : "OUT",
                  bEndpointAddress & 0xF,
                  wMaxPacketSize);
    }

    usbprintf("\n=== FIFO - endpoint configurations ===\n");
    usbprintf("fifo_map_val (0x1A8) = 0x%08x\n", UsbRegRead(REG_DEV_FMAP));
    usbprintf("endp_map0_val (0x1A0) = 0x%08x\n", UsbRegRead(REG_DEV_EPMAP0));
    usbprintf("endp_map1_val (0x1A4) = 0x%08x\n", UsbRegRead(REG_DEV_EPMAP1));
    usbprintf("fifo_config_val (0x1AC) = 0x%08x\n", UsbRegRead(REG_DEV_FCFG));
    usbprintf("fifo_int_mask (0x138)= 0x%08x\n\n", UsbRegRead(REG_DEV_MISG1));

    for (int i = 0; i < intf->bNumEndpoints; i++)
    {
        uint8_t bEndpointAddress = intf->endpoint[i]->bEndpointAddress;
        usbprintf("Endpoint 0x%x is mapping to FIFO %d\n",
                  bEndpointAddress, endpoint_to_fifo_no_via_reg(bEndpointAddress));
    }
    for (int fno = 0; fno < intf->bNumEndpoints; fno++)
    {
        usbprintf("FIFO %d is mapping to 0x%x\n", fno, fifo_to_endpoint_address_via_reg(fno));
    }
}

#endif
