/*
 * Kneron Peripheral API
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#ifndef __KDP_USBD_DEF_H__
#define __KDP_USBD_DEF_H__

#include "kdp_usbd_reg.h"
#include "kdp_usbd.h"

//#define USBD_DEBUG

#define FIFO_NUM 4 // we have 4 FIFOs, each has 1-KB

#define UsbRegRead(reg_offset) inw(USB_FOTG210_PA_BASE + reg_offset)
#define UsbRegWrite(reg_offset, val) outw(USB_FOTG210_PA_BASE + reg_offset, val)
#define UsbRegMaskedSet(reg_offset, val) masked_outw(USB_FOTG210_PA_BASE + reg_offset, val, val)
#define UsbRegMaskedClr(reg_offset, val) masked_outw(USB_FOTG210_PA_BASE + reg_offset, 0, val)

#define MIN(x, y) (((x) < (y)) ? (x) : (y))

#ifdef USBD_DEBUG
#define usbprintf(...)            \
    {                             \
        fLib_printf("[usb] ");    \
        fLib_printf(__VA_ARGS__); \
    }
#else
#define usbprintf(...) \
    {                  \
    }
#endif

enum
{
    CONFIG_DEFAULT_STATE = 0,
    CONFIG_ADDRESS_STATE,
    CONFIG_CONFIGURED_STATE,
};

// for usbd_fifo_ctrl_t:transferType
enum
{
    TXFER_CONTROL = 0,
    TXFER_ISO,
    TXFER_BULK,
    TXFER_INT,
};

// for SETUP packet request
enum
{
    RESP_NACK = 1, /* busy now */
    RESP_ACK,      /* reqeust is done */
    RESP_STALL,    /* request is not supported */
};

// a data struct for FIFO and DMA control
typedef struct
{
    // below are initialized once when endpoint is configured
    uint8_t enabled;          // indicate that this FIFO is enabled
    uint8_t enpNo;            // endpoint no (without direction bit)
    uint32_t endpointAddress; // endpoint address (with direction bit)
    uint32_t maxPacketSize;   // the wMaxPacketSize
    uint8_t transferType;     // Control/Iso/Bulk/Interrupt
    uint32_t byteCntReg;      // FIFO byte count register address

    // below are used while transferring data
    uint8_t isTransferring;   // indicate it is in transferring progress
    uint32_t user_buf_addr;   // user commited buffer address for transfer
    uint32_t user_buf_len;    // user buffer length
    uint32_t received_length; // for Out only

    // below variable represents short packet is coming for bulk out
    // or zero-length packet for bulk in
    // 1: True, 0: False
    uint8_t short_or_zl_packet;

    uint8_t isBlockingCall;
    kdp_usbd_event_name_e cur_event; // internal use for blocking API
} usbd_fifo_ctrl_t;

// define usb device mode control block
typedef struct
{
    osThreadId_t notifyTid;
    osEventFlagsId_t evt_id; // internal use for blocking API
    uint32_t notifyFlag;
    bool ep0_halted;
    kdp_usbd_device_descriptor_t *dev_desc;
    kdp_usbd_device_qualifier_descriptor_t *dev_qual_desc; // is it necessary ?
    uint32_t config_state;
    usbd_fifo_ctrl_t fifo_cbs[FIFO_NUM]; // FIFO control blocks
} usbd_ctrl_block_t;

extern usbd_ctrl_block_t usbd_cb;

enum
{
    READ_FIFO = 0,
    WRITE_FIFO = 1,
};

// kdp_usbd_event.c
extern void usbd_handle_device_interrupts(void);
extern void usbd_reset_event_queue(void);
extern void usbd_clean_event_queue(void);
extern int usbd_pop_event_from_queue(kdp_usbd_event_t *event);
extern void usbd_notify_event_to_user(kdp_usbd_event_name_e ename, uint32_t data1, uint32_t data2);
extern int8_t usbd_endpoint_to_fifo(uint32_t endpoint);
extern void usbd_clean_fifo_cb(usbd_fifo_ctrl_t *fifo_cb);

#ifdef USBD_DEBUG
// kdp_usbd_dbg.c
extern void usbd_print_status_registers(void);
extern void usbd_print_interrupt_status(void);
extern void usbd_print_setup_packet(kdp_usbd_setup_packet_t *setup);
extern void usbd_print_device_descriptor(kdp_usbd_device_descriptor_t *desc);
extern void usbd_print_fifo_endpoint_configurations(kdp_usbd_interface_descriptor_t *intf);
#endif

// kdp_usbd_ep0.c
extern void usbd_handle_cmd_abort(void);
extern void usbd_handle_control_transfer(void);

// kdp_usbd_dma.c
uint32_t usbd_dma_remap_addr(uint32_t addr);
extern bool usbd_dma_fifo_transfer_sync_try(uint32_t *addr, uint32_t len, uint32_t fifo_sel, uint8_t fifo_dir, uint32_t try_count);
extern bool usbd_dma_start_fifo_to_mem(uint32_t fno, usbd_fifo_ctrl_t *fifocb);
extern bool usbd_dma_start_mem_to_fifo(uint32_t fno, usbd_fifo_ctrl_t *fifocb);
extern bool usbd_dma_is_busy(void);

#endif
