#include <stdlib.h>
#include <string.h>
#include "cmsis_os2.h"
#include "scu_extreg.h"
#include "kdp_uart.h"

#include "kdp_io.h"
#include "kdp_usbd.h"
#include "kdp_usbd_def.h"

extern usbd_ctrl_block_t usbd_cb;

inline bool usbd_dma_is_busy()
{
    return (UsbRegRead(REG_DMA_TFN) != DMA_TARGET_ACC_NONE);
}

//  for data SRAM, the address must be remapped
uint32_t usbd_dma_remap_addr(uint32_t addr)
{
    uint32_t tmp;

    if ((addr & (SdRAM_MEM_BASE)) == SdRAM_MEM_BASE)
    {
        tmp = ((addr) & (~0x10000000)) | 0x20000000;
        return tmp;
    }
    return addr;
}

// fifo to memory or memory to fifo transfer
// fifo_dir = 1 : memory -> fifo
// fifo_dir = 0 : fifo -> memory
// dma polling way
bool usbd_dma_fifo_transfer_sync(uint32_t *addr, uint32_t len, uint32_t fifo_sel, uint8_t fifo_dir)
{
    if (usbd_dma_is_busy())
        return FALSE;

    bool status = FALSE;

    // set DMA FIFO selection to accuire DMA
    // FIXME: better use mutex stuff to occupy DMA resource !!
    UsbRegWrite(REG_DMA_TFN, fifo_sel);

    // temporarily disable DMA complt interrupt
    // because we will poll it here
    UsbRegMaskedSet(REG_DEV_MISG2, G2_DMA_CMPLT_RW1C);

    // set DMA address
    UsbRegWrite(REG_DMA_CPS2, usbd_dma_remap_addr((uint32_t)addr));

    // set DMA transfer size
    UsbRegWrite(REG_DMA_CPS1, len << 8);

    if (fifo_dir == WRITE_FIFO)
        UsbRegMaskedSet(REG_DMA_CPS1, BIT1);

    // start DMA
    UsbRegMaskedSet(REG_DMA_CPS1, DMA_START);

    int i;
    // FIXME: why 500 ? just to prevent from being dead forever
    for (i = 0; i < 500; i++)
    {
        // polling DMA completion status
        if (UsbRegRead(REG_DEV_ISG2) & G2_DMA_CMPLT_RW1C)
        {
            UsbRegMaskedSet(REG_DEV_ISG2, G2_DMA_CMPLT_RW1C);
            status = TRUE;
            break;
        }
    }

    // re-enable DMA complt interrupt
    UsbRegMaskedClr(REG_DEV_MISG2, G2_DMA_CMPLT_RW1C);

    // clear DMA FIFO selection
    UsbRegWrite(REG_DMA_TFN, DMA_TARGET_ACC_NONE);

    return status;
}

bool usbd_dma_fifo_transfer_sync_try(uint32_t *addr, uint32_t len, uint32_t fifo_sel, uint8_t fifo_dir, uint32_t try_count)
{
    bool status = FALSE;
    for (int i = 0; i < try_count; i++)
        if (usbd_dma_fifo_transfer_sync(addr, len, fifo_sel, fifo_dir))
        {
            status = TRUE;
            break;
        }
    return status;
}

// out transfer
// configure DMA settings for fifo-to-memory for non-control transfer
// this implementation follows FOTG210 data sheet : 6.2.4 "Programming DMA"
bool usbd_dma_start_fifo_to_mem(uint32_t fno, usbd_fifo_ctrl_t *fifocb)
{
    if (usbd_dma_is_busy())
        return FALSE;

    // select FIFO for DMA
    UsbRegWrite(REG_DMA_TFN, 0x1 << fno);

    uint32_t fifo_bytecnt = UsbRegRead(fifocb->byteCntReg) & BC_Fn;
    // can transfer only minimum size betwwen FIFO byte count and user buffer residual size
    uint32_t transfer_size = MIN(fifo_bytecnt, fifocb->user_buf_len);

    // set DMA memory addr
    UsbRegWrite(REG_DMA_CPS2, fifocb->user_buf_addr);

    // set DMA_LEN and DMA_TYPE = FIFO_to_Memory
    UsbRegWrite(REG_DMA_CPS1, transfer_size << 8);

    // start DMA
    UsbRegMaskedSet(REG_DMA_CPS1, DMA_START);

    fifocb->user_buf_addr += transfer_size;
    fifocb->user_buf_len -= transfer_size;
    fifocb->received_length += transfer_size;

    return TRUE;
}

// in trasnfer
// configure DMA settings for memory-to-fifo for non-control transfer
// this implementation follows FOTG210 data sheet : 6.2.4 "Programming DMA"
bool usbd_dma_start_mem_to_fifo(uint32_t fno, usbd_fifo_ctrl_t *fifocb)
{
    if (usbd_dma_is_busy())
        return FALSE;

    // select FIFO for DMA
    UsbRegWrite(REG_DMA_TFN, 0x1 << fno);

    // can transfer only minimum size betwwen MaxPacketSize and user buffer residual size
    uint32_t transfer_size = MIN(fifocb->maxPacketSize, fifocb->user_buf_len);

    // set DMA memory addr
    UsbRegWrite(REG_DMA_CPS2, fifocb->user_buf_addr);

    // set DMA_LEN and DMA_TYPE = Memory_to_FIFO
    UsbRegWrite(REG_DMA_CPS1, (transfer_size << 8) | 0x2);

    // start DMA
    UsbRegMaskedSet(REG_DMA_CPS1, DMA_START);

    fifocb->user_buf_addr += transfer_size;
    fifocb->user_buf_len -= transfer_size;

    return TRUE;
}
