/*
 * Kneron Peripheral API
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include <stdlib.h>
#include <string.h>
#include "cmsis_os2.h"
#include "scu_extreg.h"
#include "kdp_uart.h"

#include "kdp_io.h"
#include "kdp_usbd.h"
#include "kdp_usbd_def.h"

extern usbd_ctrl_block_t usbd_cb;

static int8_t usbd_send_host_device_descriptor(kdp_usbd_setup_packet_t *setup)
{
    kdp_usbd_device_descriptor_t *desc = usbd_cb.dev_desc;

    // some error checking
    if (!desc)
    {
        usbprintf("GET_DESCRIPTOR is failing becasue device descriptor is not set\n");
        return RESP_STALL;
    }

    bool sts = usbd_dma_fifo_transfer_sync_try((uint32_t *)desc, desc->bLength, DMA_TARGET_ACC_CXF, WRITE_FIFO, 50);
    if (sts)
        return RESP_ACK;
    else
        return RESP_NACK;
}

static int8_t usbd_send_host_device_qual_descriptor(kdp_usbd_setup_packet_t *setup)
{
    kdp_usbd_device_qualifier_descriptor_t *desc = usbd_cb.dev_qual_desc;

    // some error checking
    if (!desc)
    {
        usbprintf("GET_DESCRIPTOR is failing becasue device qualifier descriptor is not set\n");
        return RESP_STALL;
    }

    bool sts = usbd_dma_fifo_transfer_sync_try((uint32_t *)desc, desc->bLength, DMA_TARGET_ACC_CXF, WRITE_FIFO, 50);
    if (sts)
        return RESP_ACK;
    else
        return RESP_NACK;
}

static int8_t usbd_send_host_configuration_descriptors(kdp_usbd_setup_packet_t *setup)
{
    kdp_usbd_device_descriptor_t *dev_desc = usbd_cb.dev_desc;
    uint32_t confIdx = setup->wValue & 0xFF;

    // some error checking
    if (!dev_desc ||
        dev_desc->bNumConfigurations > MAX_USBD_CONFIG ||
        confIdx >= MAX_USBD_CONFIG)
    {
        usbprintf("GET_DESCRIPTOR for configuration is failing due to improper settings\n");
        return RESP_STALL;
    }

    kdp_usbd_config_descriptor_t *conf_desc = dev_desc->config[confIdx];

    // create an temp buffer for combining all sub-descriptors
    uint8_t *buf_desc = malloc(conf_desc->wTotalLength);

    uint16_t txLen = MIN(conf_desc->wTotalLength, setup->wLength);

    // collect all sub-descriptos int one memory
    uint8_t offset = 0;
    memcpy(buf_desc, conf_desc, conf_desc->bLength);
    offset += conf_desc->bLength;

    for (int i = 0; i < conf_desc->bNumInterfaces; i++)
    {
        kdp_usbd_interface_descriptor_t *intf_desc = conf_desc->interface[i];
        memcpy(buf_desc + offset, intf_desc, intf_desc->bLength);
        offset += intf_desc->bLength;
        for (int j = 0; j < intf_desc->bNumEndpoints; j++)
        {
            kdp_usbd_endpoint_descriptor_t *endp_desc = intf_desc->endpoint[j];
            memcpy(buf_desc + offset, endp_desc, endp_desc->bLength);
            offset += endp_desc->bLength;
        }
    }

    bool sts = usbd_dma_fifo_transfer_sync_try((uint32_t *)buf_desc, txLen, DMA_TARGET_ACC_CXF, WRITE_FIFO, 50);

    free(buf_desc);

    if (sts)
        return RESP_ACK;
    else
        return RESP_NACK;
}

static void usbd_init_fifo_configurations(kdp_usbd_interface_descriptor_t *intf)
{
    uint32_t fifo_map_val = 0x0;         // for 0x1A8
    uint32_t endp_map0_val = 0x0;        // for 0x1A0
    uint32_t endp_map1_val = 0x0;        // for 0x1A4
    uint32_t fifo_config_val = 0x0;      // for 0x1AC
    uint32_t fifo_int_mask = 0xFFFFFFFF; // for 0x138, default disable all interrupts

    // also need to init fifo-dma control blocks
    for (int fifo = 0; fifo < FIFO_NUM; fifo++)
        usbd_cb.fifo_cbs[fifo].enabled = FALSE;

    // here assume endpoint number order is ascending
    for (int i = 0; i < intf->bNumEndpoints; i++)
    {
        uint8_t bEndpointAddress = intf->endpoint[i]->bEndpointAddress;
        uint8_t bmAttributes = intf->endpoint[i]->bmAttributes;
        uint16_t wMaxPacketSize = intf->endpoint[i]->wMaxPacketSize;

        // i value implies FIFO number
        uint32_t fifo = i;
        uint8_t isIn = !!(bEndpointAddress & 0x80);
        uint8_t enpNo = bEndpointAddress & 0xF; // retrieve endpoint no without direction
        uint32_t bitfield;

        // set FIFO's direction and corresponding endpoint no.
        bitfield = (isIn << 4) | enpNo;
        fifo_map_val |= (bitfield << (fifo * 8));

        // set endpoint's FIFO no.
        bitfield = isIn ? fifo : (fifo << 4);
        if (enpNo <= 4)
            endp_map0_val |= (bitfield << ((enpNo - 1) * 8));
        else // 5~8
            endp_map1_val |= (bitfield << ((enpNo - 5) * 8));

        // enable the corresponding FIFO and set transfer type
        fifo_config_val |= (BIT5 | (bmAttributes & 0x3)) << (fifo * 8);

        // set max packet size & reset toggle bit
        if (isIn)
        {
            // for IN endpoints
            UsbRegWrite(REG_DEV_INMPS_1 + 4 * (enpNo - 1), wMaxPacketSize & 0x7ff);
            UsbRegMaskedSet(REG_DEV_INMPS_1 + 4 * (enpNo - 1), RSTG_IEPn);
            UsbRegMaskedClr(REG_DEV_INMPS_1 + 4 * (enpNo - 1), RSTG_IEPn);

            // disable interrupt for IN endpoint
            // because when IN fifo is empty, interrupt will be asserted
            // we could enable IN interrupt only when need to watch it
            fifo_int_mask |= (MF0_IN_INT << fifo);
        }
        else
        {
            // for OUT endpoints
            UsbRegWrite(REG_DEV_OUTMPS_1 + 4 * (enpNo - 1), wMaxPacketSize & 0x7ff);
            UsbRegMaskedSet(REG_DEV_OUTMPS_1 + 4 * (enpNo - 1), RSTG_OEPn);
            UsbRegMaskedClr(REG_DEV_OUTMPS_1 + 4 * (enpNo - 1), RSTG_OEPn);

            // enable interrupt for OUT endpoint
            fifo_int_mask &= ~((MF0_SPK_INT | MF0_OUT_INT) << (fifo * 2));
        }

        // init fifo dma control blocks for each enabled fifo
        usbd_cb.fifo_cbs[fifo].enabled = TRUE;
        usbd_cb.fifo_cbs[fifo].enpNo = enpNo;
        usbd_cb.fifo_cbs[fifo].endpointAddress = bEndpointAddress;
        usbd_cb.fifo_cbs[fifo].maxPacketSize = wMaxPacketSize;
        usbd_cb.fifo_cbs[fifo].transferType = bmAttributes & 0x3;
        usbd_cb.fifo_cbs[fifo].byteCntReg = 0x1B0 + 4 * fifo;
        usbd_cb.fifo_cbs[fifo].isTransferring = FALSE;
        usbd_cb.fifo_cbs[fifo].user_buf_addr = 0;
        usbd_cb.fifo_cbs[fifo].user_buf_len = 0;
        usbd_cb.fifo_cbs[fifo].short_or_zl_packet = FALSE;
    }

    // clear all FIFO
    UsbRegMaskedSet(REG_DEV_TST, TST_CLRFF);

    // set FIFO interrupt mask
    UsbRegWrite(REG_DEV_MISG1, fifo_int_mask);

    // endpoint map 0
    UsbRegWrite(REG_DEV_EPMAP0, endp_map0_val);

    // endpoint map1
    UsbRegWrite(REG_DEV_EPMAP1, endp_map1_val);

    // fifo map
    UsbRegWrite(REG_DEV_FMAP, fifo_map_val);

    // fifo config / enable
    UsbRegWrite(REG_DEV_FCFG, fifo_config_val);

    // set Device SOF Mask Timer value as data sheet recommended for HS
    UsbRegWrite(REG_DEV_SMT, 0x44C);

    // set configuration set bit, now allow HW to handle endpoint transfer
    UsbRegMaskedSet(REG_DEV_ADR, AFT_CONF);

#ifdef USBD_DEBUG
    usbd_print_fifo_endpoint_configurations(intf);
#endif
}

static int8_t usbd_set_configuration(kdp_usbd_setup_packet_t *setup)
{
    int8_t resp = RESP_STALL;

    uint8_t config_val = setup->wValue & 0xFF;
    if (config_val == 0)
    {
        usbd_cb.config_state = CONFIG_ADDRESS_STATE;
        // clear configuration set bit
        UsbRegMaskedClr(REG_DEV_ADR, AFT_CONF);
        resp = RESP_ACK;

        // FIXME clear FIFO-endpoint mapping ?
    }
    else
    {
        kdp_usbd_config_descriptor_t *config;

        // compare with all configuration descriptos
        for (int i = 0; i < usbd_cb.dev_desc->bNumConfigurations; i++)
        {
            config = usbd_cb.dev_desc->config[i];
            if (config->bConfigurationValue == config_val)
            {
                usbd_cb.config_state = CONFIG_CONFIGURED_STATE;
                resp = RESP_ACK;
                break;
            }
        }

        // FIXME? for only-one interface, should set up FIFO-endpont mapping now
        if (resp == RESP_ACK && config->bNumInterfaces == 1)
            usbd_init_fifo_configurations(config->interface[0]);

        if (resp == RESP_ACK)
            usbd_notify_event_to_user(KDP_USBD_EVENT_DEV_CONFIGURED, config_val, 0);
    }

    return resp;
}

static int8_t usbd_handle_standard_request(kdp_usbd_setup_packet_t *setup)
{
    int8_t resp = RESP_STALL;

    // handle requests which are not affected by ep0 halt
    switch (setup->bRequest)
    {
    case 0x0: // GET_STATUS
        break;
    case 0x1: // CLEAR_FEATURE
    {
        if (setup->wValue == 0x0)
        {
            // endpoint halt
            kdp_usbd_reset_endpoint(setup->wIndex);
            resp = RESP_ACK;
#ifdef USBD_DEBUG
            usbprintf("reset fno 0x%x endpoint 0x%x\n\n", setup->wIndex);
#endif
        }
        break;
    }
    case 0x3: // SET_FEATURE
        break;
    }

    // if ep0 is halted, some requests should not be done
    if (usbd_cb.ep0_halted)
        return RESP_STALL;

    switch (setup->bRequest)
    {
    case 0x5: // SET_ADDRESS
    {
        // USB2.0 spec says should not be greaten than 127
        if (setup->wValue <= 127)
        {
            // set DEVADR and also clear AFT_CONF
            UsbRegWrite(REG_DEV_ADR, setup->wValue);
            resp = RESP_ACK;
        }
    }
    break;
    case 0x6: // GET_DESCRIPTOR
    {
        // low byte: index of specified descriptor type
        //uint8_t descp_idx = (setup->wValue & 0xFF);

        // high byte: descriptor type
        switch (setup->wValue >> 8)
        {
        case 1: // DEVICE descriptor
            resp = usbd_send_host_device_descriptor(setup);
            break;
        case 2: // CONFIGURATION descriptor
            resp = usbd_send_host_configuration_descriptors(setup);
            break;
        case 3: // STRING descriptor
            break;
        case 4: // INTERFACE descriptor
            break;
        case 5: // ENDPOINT descriptor
            break;
        case 6: // DEVICE_QUALIFIER descriptor
            resp = usbd_send_host_device_qual_descriptor(setup);
            break;
        case 7: // OTHER_SPEED_CONFIGURATION descriptor
            break;
        case 8: // INTERFACE_POWER descriptor
            break;
        }
    }
    break;
    case 0x7: // SET_DESCRIPTOR
        break;
    case 0x8: // GET_CONFIGURATION
        break;
    case 0x9: // SET_CONFIGURATION
        resp = usbd_set_configuration(setup);
        break;
    }

    return resp;
}

void usbd_handle_cmd_abort()
{
    // according to datasheet, this could happen
    // when a new SETUP comes before last one is complete
    // handle it first or the FIFO will be frozen

    usbprintf("SETUP cmd aborted!\n");

    // clear the abort status
    UsbRegMaskedSet(REG_DEV_ISG0, G0_CX_COMABT_INT_RW1C);
}

void usbd_handle_control_transfer()
{
    // ready to read out 8 bytes setup packet
    kdp_usbd_setup_packet_t setup = {0};
    uint8_t *temp = (uint8_t *)&setup;

#ifdef USB_DEBUG
    static int cx_busy_try = 0;
#endif

    // if DMA is busy now, do nothing and will handle in later interrupts
    if (usbd_dma_is_busy())
    {
#ifdef USB_DEBUG
        ++cx_busy_try;
        if ((cx_busy_try % 100) == 0)
            usbprintf("DMA is too busy to do cx transfer!\n");
#endif
        return;
    }

#ifdef USB_DEBUG
    cx_busy_try = 0;
#endif

    // set DMA FIFO selection to CXF
    UsbRegWrite(REG_DMA_TFN, DMA_TARGET_ACC_CXF);

    // directly read DMA_CPS3 twice to get 8-byte setup packet
    *((uint32_t *)temp) = UsbRegRead(REG_DMA_CPS3);
    *((uint32_t *)(temp + 4)) = UsbRegRead(REG_DMA_CPS3);

    // clear DMA FIFO selection
    UsbRegWrite(REG_DMA_TFN, DMA_TARGET_ACC_NONE);

#ifdef USBD_DEBUG
    usbd_print_setup_packet(&setup);
#endif

    // parsing bmRequestType to find out which kind of reqeusts
    int8_t resp = RESP_NACK;
    uint8_t bmRequestType_type = ((setup.bmRequestType & 0x60) >> 5);
    switch (bmRequestType_type)
    {
    case 0: // Standard request
        resp = usbd_handle_standard_request(&setup);
        break;
    case 1: // Class request
    case 2: // Vendor request
        usbd_notify_event_to_user(KDP_USBD_EVENT_SETUP_PACKET,
                                  *((uint32_t *)temp), *((uint32_t *)(temp + 4)));
        break;
    }

    if (resp == RESP_ACK)
        // indicate an OK request to host
        UsbRegMaskedSet(REG_CXCFE, CX_DONE);
    else if (resp == RESP_STALL)
    {
        // indicate a request error to host
        UsbRegMaskedSet(REG_CXCFE, CX_STL | CX_DONE);
        usbprintf("returning a STALL to host for the above cx transfer\n\n");
    }
    else
    {
        // NACK, do nothing for now
        usbprintf("Not handle this cx transfer\n\n");
    }
}
