/*
 * Kneron Peripheral API
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include <stdlib.h>
#include "cmsis_os2.h"
#include "scu_extreg.h"
#include "kdp_uart.h"

#include "kdp_io.h"
#include "kdp_usbd.h"
#include "kdp_usbd_def.h"

#define QLEN 30
static kdp_usbd_event_t event_queue[QLEN]; // a fifo ring queue
static int eqWriteIdx = 0;
static int eqReadIdx = 0;
static osSemaphoreId_t empty_id = 0;
static osSemaphoreId_t filled_id = 0;

extern usbd_ctrl_block_t usbd_cb;

/* this function must not be used in ISR */
void usbd_reset_event_queue()
{
    osStatus_t sts = osOK;

    if (empty_id != 0)
        sts = osSemaphoreRelease(empty_id);

    if (sts == osOK)
        empty_id = osSemaphoreNew(QLEN, QLEN, NULL);
    else
        usbprintf("Failed to release the empty_id semaphore, err = %d\n", sts);

    if (filled_id != 0)
        sts = osSemaphoreRelease(filled_id);

    if (sts == osOK)
        filled_id = osSemaphoreNew(QLEN, 0, NULL);
    else
        usbprintf("Failed to release the filled_id semaphore, err = %d\n", sts);

    // reset read/write indices
    eqWriteIdx = 0;
    eqReadIdx = 0;

    if (empty_id == 0 || filled_id == 0)
        usbprintf("Failed to create semaphores\n");
}

void usbd_clean_event_queue()
{
    if (empty_id != 0)
        osSemaphoreRelease(empty_id);
    empty_id = 0;

    if (filled_id != 0)
        osSemaphoreRelease(filled_id);
    filled_id = 0;
}

static int usbd_push_event_to_queue(kdp_usbd_event_t event)
{
    osStatus_t sts;

    sts = osSemaphoreAcquire(empty_id, 0);
    if (sts != osOK)
        // not overwrite event queue
        return 0;

    event_queue[eqWriteIdx] = event;
    ++eqWriteIdx;
    if (eqWriteIdx >= QLEN)
        eqWriteIdx = 0;

    osSemaphoreRelease(filled_id);
    return 1;
}

int usbd_pop_event_from_queue(kdp_usbd_event_t *event)
{
    osStatus_t sts;

    sts = osSemaphoreAcquire(filled_id, 0);
    // queue is empty
    if (sts != osOK)
        return 0;

    *event = event_queue[eqReadIdx];
    ++eqReadIdx;
    if (eqReadIdx >= QLEN)
        eqReadIdx = 0;

    osSemaphoreRelease(empty_id);
    return 1;
}

// put usb events to an event queue and notify user via thread flag
void usbd_notify_event_to_user(kdp_usbd_event_name_e ename, uint32_t data1, uint32_t data2)
{
    kdp_usbd_event_t uevent;
    uevent.ename = ename;
    uevent.data1 = data1;
    uevent.data2 = data2;
    usbd_push_event_to_queue(uevent);

    if (usbd_cb.notifyTid)
        osThreadFlagsSet(usbd_cb.notifyTid, usbd_cb.notifyFlag);
}

static void usbd_bus_reset_work()
{
    // clear SET_CONFIG state and usb device address
    UsbRegWrite(REG_DEV_ADR, 0x0);

    // clear EP0 STALL bit
    UsbRegMaskedClr(REG_CXCFE, CX_STL);

    // disable (mask) all FIFOs interrupts
    UsbRegWrite(REG_DEV_MISG1, 0xFFFFFFFF);

    // clear all FIFO
    UsbRegMaskedSet(REG_DEV_TST, TST_CLRFF);

    // clear this interrupt bit
    UsbRegMaskedSet(REG_DEV_ISG2, G2_USBRST_INT_RW1C);

    // reset all endpoints, FIXME: better place to do this ?
    {
        usbd_fifo_ctrl_t *fifo_cb = usbd_cb.fifo_cbs;
        for (int fno = 0; fno < FIFO_NUM; fno++)
            if (fifo_cb[fno].enabled)
            {
                kdp_usbd_reset_endpoint(fifo_cb[fno].endpointAddress);
            }
    }

    usbd_notify_event_to_user(KDP_USBD_EVENT_BUS_RESET, 0, 0);
}

static void usbd_handle_dma_error()
{
    // clear this interrupt bit
    UsbRegMaskedSet(REG_DEV_ISG2, G2_DMA_ERROR_RW1C);

    // check which fifo-endpoint is responsible for this

    uint32_t fifoSel = UsbRegRead(REG_DMA_TFN);

    // clear FIFO sel
    UsbRegWrite(REG_DMA_TFN, DMA_TARGET_ACC_NONE);

    uint32_t fno;
    for (fno = 0; fno < FIFO_NUM; fno++)
        if (fifoSel & (0x1 << fno))
            break;

    usbd_fifo_ctrl_t *fifo_cb = &(usbd_cb.fifo_cbs[fno]);

    // reset the endpoint due to DMA error
    kdp_usbd_reset_endpoint(fifo_cb->endpointAddress);

    // also notify user the DMA_ERROR
    usbd_notify_event_to_user(KDP_USBD_EVENT_DMA_ERROR, 0, 0);
}

static void usbd_bus_suspend_work()
{
    // clear this interrupt bit
    UsbRegMaskedSet(REG_DEV_ISG2, G2_SUSP_INT_RW1C);

    // FIXME: should do something here ?

    usbd_notify_event_to_user(KDP_USBD_EVENT_BUS_SUSPEND, 0, 0);
}

static void usbd_bus_resume_work()
{
    // clear this interrupt bit
    UsbRegMaskedSet(REG_DEV_ISG2, G2_RESM_INT_RW1C);

    // FIXME: should do something here ?

    usbd_notify_event_to_user(KDP_USBD_EVENT_BUS_SUSPEND, 0, 0);
}

int8_t usbd_endpoint_to_fifo(uint32_t endpoint)
{
    // check fifo ctrl blocks to get fifo no
    int8_t fno;
    for (fno = 0; fno < FIFO_NUM; fno++)
        if (usbd_cb.fifo_cbs[fno].endpointAddress == endpoint)
            break;

    return (fno < 4) ? fno : -1;
}

void usbd_clean_fifo_cb(usbd_fifo_ctrl_t *fifo_cb)
{
    fifo_cb->user_buf_addr = 0;
    fifo_cb->user_buf_len = 0;
    fifo_cb->isTransferring = FALSE;
    fifo_cb->short_or_zl_packet = FALSE;
}

static inline void usbd_handle_fifo_short_pkt_interrupts(uint32_t interrupt_bits)
{
    for (int fno = 0; fno < FIFO_NUM; fno++)
    {
        if (interrupt_bits & (G1_F0_SPK_INT_RO << (fno * 2)))
        {

            // with a short packet coming means data is less than MaxPacketSize
            // and it is the last packet transferd by host
            // this will be handled in DMA completeion time
            usbd_cb.fifo_cbs[fno].short_or_zl_packet = TRUE;

            // disable short packet interrupt
            // shoudl be re-enabled at next DMA completeion time
            UsbRegMaskedSet(REG_DEV_MISG1, MF0_SPK_INT << (fno * 2));

            break;
        }
    }
}

static inline void usbd_handle_fifo_out_interrupts(uint32_t interrupt_bits)
{
    usbd_fifo_ctrl_t *fifo_cb = usbd_cb.fifo_cbs;

    for (int fno = 0; fno < FIFO_NUM; fno++)
    {
        // handle interrupts for each fifo
        if (interrupt_bits & (G1_F0_OUT_INT_RO << (fno * 2)))
        {
            // if no user buffer is commited, means that this is a start of OUT transfer
            if (fifo_cb[fno].user_buf_addr == 0)
            {
                // disable this OUT interrrupt
                // it should be re-enabled once user commits a buffer for DMA
                UsbRegMaskedSet(REG_DEV_MISG1, MF0_OUT_INT << (fno * 2));

                // then should notify an event to user
                usbd_notify_event_to_user(KDP_USBD_EVENT_TRANSFER_OUT, fifo_cb[fno].enpNo, 0);
            }
            else // configure DMA transfer for FIFO to user buffer
            {
                if (fifo_cb[fno].user_buf_len > 0 &&
                    usbd_dma_start_fifo_to_mem(fno, &fifo_cb[fno]) == TRUE)
                {
                    // DMA has been started, disable this OUT interrupt
                    // and re-enable the interrupt at DMA completion
                    UsbRegMaskedSet(REG_DEV_MISG1, MF0_OUT_INT << (fno * 2));
                }
            }
        }
    }
}

static inline void usbd_handle_fifo_in_interrupts(uint32_t interrupt_bits)
{
    // this handles only one fifo interrupt

    usbd_fifo_ctrl_t *fifo_cb = usbd_cb.fifo_cbs;

    uint32_t fno;
    for (fno = 0; fno < FIFO_NUM; fno++)
        if (interrupt_bits & (G1_F0_IN_INT_RO << fno))
            break;

    // disable (mask) the FIFO IN interrupt immediately to prevent from interrupt re-trigger
    UsbRegMaskedSet(REG_DEV_MISG1, MF0_IN_INT << fno);

    if (fifo_cb[fno].user_buf_len > 0)
    {
        if (usbd_dma_start_mem_to_fifo(fno, &fifo_cb[fno]) == FALSE)
        {
            // re-enable the interrupt for next try
            UsbRegMaskedClr(REG_DEV_MISG1, MF0_IN_INT << fno);
        }
    }
    else
    {
        // FIFO is empty and no more data to send, in other words, the bulk-in transfer is done

        // send zero-length packet if needed
        if (fifo_cb[fno].short_or_zl_packet)
        {
            UsbRegMaskedSet(REG_DEV_INMPS_1 + 4 * (fifo_cb[fno].enpNo - 1), TX0BYTE_IEPn);
            // FIXME: should check G2_TX0BYTE_INT_RW1C in later interrupts
        }

        // notify transfer done to user
        if (fifo_cb[fno].isBlockingCall)
        {
            fifo_cb[fno].cur_event = KDP_USBD_EVENT_TRANSFER_DONE;
            osEventFlagsSet(usbd_cb.evt_id, 0x1 << fno);
        }
        else
        {
            usbd_clean_fifo_cb(&fifo_cb[fno]);
            usbd_notify_event_to_user(KDP_USBD_EVENT_TRANSFER_DONE, fifo_cb[fno].endpointAddress, 0);
        }
    }
}

#define FIFO_SHORT_PKT_INTERRUPTS (G1_F0_SPK_INT_RO | G1_F1_SPK_INT_RO | G1_F2_SPK_INT_RO | G1_F3_SPK_INT_RO)
#define FIFO_OUT_INTERRUPTS (G1_F0_OUT_INT_RO | G1_F1_OUT_INT_RO | G1_F2_OUT_INT_RO | G1_F3_OUT_INT_RO)
#define FIFO_IN_INTERRUPTS (G1_F0_IN_INT_RO | G1_F1_IN_INT_RO | G1_F2_IN_INT_RO | G1_F3_IN_INT_RO)

static void usbd_handle_fifo_interrupts()
{
    uint32_t fifo_interrupts = UsbRegRead(REG_DEV_ISG1) & ~(UsbRegRead(REG_DEV_MISG1));

    // handle OUT short packets interrupts for short packets
    if (fifo_interrupts & FIFO_SHORT_PKT_INTERRUPTS)
        usbd_handle_fifo_short_pkt_interrupts(fifo_interrupts);

    // handle OUT FIFO interrupts
    if (fifo_interrupts & FIFO_OUT_INTERRUPTS)
        usbd_handle_fifo_out_interrupts(fifo_interrupts);

    // handle IN FIFO interrupts
    if (fifo_interrupts & FIFO_IN_INTERRUPTS)
        usbd_handle_fifo_in_interrupts(fifo_interrupts);
}

// FIXME: this does not handle CX DMA complete
static void usbd_handle_dma_complete_interrupt()
{
    // this handles only one DMA complete interrupt
    uint32_t fno;
    uint32_t fifoSel = UsbRegRead(REG_DMA_TFN);
    uint32_t dmaCPS1 = UsbRegRead(REG_DMA_CPS1);

    // clear FIFO selection
    UsbRegWrite(REG_DMA_TFN, DMA_TARGET_ACC_NONE);

    // clear DMA completion interrupt
    UsbRegMaskedSet(REG_DEV_ISG2, G2_DMA_CMPLT_RW1C);

    for (fno = 0; fno < FIFO_NUM; fno++)
        if (fifoSel & (0x1 << fno))
            break;

#ifdef USBD_DEBUG
    if (fno >= FIFO_NUM)
        usbprintf("Error: no fifo for DMA complete interrupts\n");
#endif

    // check DMA-FIFO direction
    if (dmaCPS1 & DMA_TYPE)
    {

        // Memory-to-FIFO, is IN transfer
        // re-enable (unmask) the FIFO IN interrupt here
        // because DMA completion does not mean bulk-in transfer is done
        UsbRegMaskedClr(REG_DEV_MISG1, MF0_IN_INT << fno);
    }
    else
    {
        // FIFO-to-Memory, is OUT transfer

        usbd_fifo_ctrl_t *fifo_cb = &(usbd_cb.fifo_cbs[fno]);

        {
            // re-enable OUT interrupts whatever fifo or short packet
            UsbRegMaskedClr(REG_DEV_MISG1, (MF0_OUT_INT | MF0_SPK_INT) << (fno * 2));

            if (fifo_cb->short_or_zl_packet)
            {
                // this transfer is a short packet, so transfer is done
                // notify transfer done to user
                if (fifo_cb->isBlockingCall)
                {
                    fifo_cb->cur_event = KDP_USBD_EVENT_TRANSFER_DONE;
                    osEventFlagsSet(usbd_cb.evt_id, 0x1 << fno);
                }
                else
                {
                    usbd_clean_fifo_cb(fifo_cb);
                    usbd_notify_event_to_user(KDP_USBD_EVENT_TRANSFER_DONE, fifo_cb->endpointAddress, fifo_cb->received_length);
                }
            }
        }
    }
}

static void usbd_handle_zero_length_packet_interrupt()
{
    uint32_t zl_endp_interrupts = UsbRegRead(REG_DEV_RXZ);
    uint32_t enpNo;
    for (enpNo = 1; enpNo <= 8; enpNo++)
        if (zl_endp_interrupts & (0x1 << (enpNo - 1)))
            break;

    // clean corresponding endpoint's rx zero-length interrupt
    UsbRegMaskedSet(REG_DEV_RXZ, RX0BYTE_EP1 << (enpNo - 1));

    // clear global rx zero-length interrupt
    UsbRegMaskedSet(REG_DEV_ISG2, G2_RX0BYTE_INT_RW1C);

    int8_t fno = usbd_endpoint_to_fifo(enpNo);

#ifdef USBD_DEBUG
    if (fno == -1)
    {
        usbprintf("ERROR: cannot find fifo for zero-length packet endpoint\n");
        return;
    }
#endif

    usbd_fifo_ctrl_t *fifo_cb = &(usbd_cb.fifo_cbs[fno]);
    if (fifo_cb->isTransferring)
    {

        // re-enable OUT interrupts whatever fifo or short packet
        UsbRegMaskedClr(REG_DEV_MISG1, (MF0_OUT_INT | MF0_SPK_INT) << (fno * 2));

        // received a zero-length packet, notify transfer do to user
        // notify transfer done to user
        if (fifo_cb->isBlockingCall)
        {
            fifo_cb->cur_event = KDP_USBD_EVENT_TRANSFER_DONE;
            osEventFlagsSet(usbd_cb.evt_id, 0x1 << fno);
        }
        else
        {
            usbd_clean_fifo_cb(fifo_cb);
            usbd_notify_event_to_user(KDP_USBD_EVENT_TRANSFER_DONE, fifo_cb->endpointAddress, fifo_cb->received_length);
        }

        return;
    }
}

void usbd_handle_device_interrupts()
{
    uint32_t grp_x_int_status = UsbRegRead(REG_DEV_IGR);
    uint32_t grp_0_interrupts = (UsbRegRead(REG_DEV_ISG0) & ~(UsbRegRead(REG_DEV_MISG0)));
    uint32_t grp_2_interrupts = (UsbRegRead(REG_DEV_ISG2) & ~(UsbRegRead(REG_DEV_MISG2)));

    if (grp_x_int_status & GX_INT_G1_RO)
        usbd_handle_fifo_interrupts();

    if (grp_x_int_status & GX_INT_G2_RO)
    {
        if (grp_2_interrupts & G2_DMA_CMPLT_RW1C)
            usbd_handle_dma_complete_interrupt();
        else if (grp_2_interrupts & G2_RX0BYTE_INT_RW1C)
            usbd_handle_zero_length_packet_interrupt();
        else if (grp_2_interrupts & G2_DMA_ERROR_RW1C)
            usbd_handle_dma_error();
        else if (grp_2_interrupts & G2_SUSP_INT_RW1C)
            usbd_bus_suspend_work();
        else if (grp_2_interrupts & G2_RESM_INT_RW1C)
            usbd_bus_resume_work();
        else if (grp_2_interrupts & G2_USBRST_INT_RW1C)
            usbd_bus_reset_work();
    }

    if (grp_x_int_status & GX_INT_G0_RO)
    {
        if (grp_0_interrupts & G0_CX_COMABT_INT_RW1C)
            // first priority
            usbd_handle_cmd_abort();
        else if (grp_0_interrupts & G0_CX_SETUP_INT_RO)
            usbd_handle_control_transfer();
    }

#ifdef USBD_DEBUG
    uint32_t after_grp_x = UsbRegRead(REG_DEV_IGR);
    if (after_grp_x & grp_x_int_status)
    {
        usbprintf("interrupt not handled: 0x140 = 0x%x, 0x%x\n", grp_x_int_status, after_grp_x);
        usbprintf("grp0 int = 0x%x\n", (UsbRegRead(REG_DEV_ISG0) & ~(UsbRegRead(REG_DEV_MISG0))));
        usbprintf("grp1 int = 0x%x\n", (UsbRegRead(REG_DEV_ISG1) & ~(UsbRegRead(REG_DEV_MISG1))));
        usbprintf("grp2 int = 0x%x\n", (UsbRegRead(REG_DEV_ISG2) & ~(UsbRegRead(REG_DEV_MISG2))));
    }
#endif
}
