/*
 * Kneron Application Shared Data
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

/**
 * @file kdp_app.h
 * @brief Public API header of kdp_application generic functions
 */
 
#ifndef __KDP_APP_H_
#define __KDP_APP_H_

#include <stdint.h>
#include "model_type.h"

// To use new FDR. If using old FDR (FDSmallbox+Onet+VGG10), comment this Macro
//#define FDSSD_ONETBLUR_VGG

//for general error code for Kneron applications
enum kdp_app_status_code_e{
    KDP_APP_OK = 0,
    KDP_APP_UNKNOWN_ERR = 1
};

#define KDP_APP_IMAGE_FORMAT_RGBA8888   0x80000000   /**< image format: RGBA8888 */
#define KDP_APP_IMAGE_FORMAT_RAW8       0x80000020   /**< image format: RAW8 */
/* The data byte sequence in DDR memory is [lowest byte]Y0CbY1CrY2CbY3Cr...[highest byte] */
#define KDP_APP_IMAGE_FORMAT_YCBCR422   0x80000037   /**< image format: YCBCR422 */
#define KDP_APP_IMAGE_FORMAT_RGB565     0x80000060   /**< image format: RGB565 */

/**
 * @brief  KDP image configuration structure
 */
struct kdp_img_conf_s {
    int32_t     image_col;              /// column size
    int32_t     image_row;              /// row size
    int32_t     image_ch;               /// channel size
    uint32_t    image_format;           /// image format. KDP_APP_IMAGE_FORMAT_XXXX
    uint32_t    image_memory_address;   /// address of image data
};
typedef struct kdp_img_conf_s kdp_img_conf_t;

/**
 * @brief  KDP dme configuration structure
 */
struct kdp_dme_conf_s {
    int32_t     models_selection;      
    int32_t     output_num;
    struct kdp_img_conf_s img_conf;
};

/**
 * @brief to switch DME mode
 *
 * @param [in] is_dme_p 1=dme mode, 0=non-dme mode
 * @return n/a
 */
void kdp_app_dme_mode(int is_dme_p);

/**
 * @brief To config image setting
 *
 * @param [in] src_index cam0 or cam1
 * @param [in] col column size
 * @param [in] row row size
 * @param [in] ch channel size
 * @param [in] image_format propietory image format
 * @param [in] image_memory_address image DDR address
 * @return n/a
 */
struct kdp_img_conf_s* kdp_app_config_image(
        int src_index, 
        int32_t col, int32_t row, int32_t ch, 
        uint32_t image_format, uint32_t image_memory_address);


/**
 * @brief to update image memory address
 * 
 * !!! MUST be called after kdp_app_config_image() 
 * @param [in] cam_index cam0 or cam1
 * @param [in] image_memory_address new image memory address
 * @return n/a
 */
void kdp_app_config_image_memory_address(int cam_index, int32_t image_memory_address);
                          
/**
 * @brief to get channel size by image format
 * 
 * @param [in] image format
 * @return channel size
 */
uint32_t kdp_app_config_get_channel_size(uint32_t format);

/**
 * @brief to get pixel size by image format
 * 
 * @param [in] image format
 * @return pixel size
 */
uint32_t kdp_app_config_get_pixel_size(uint32_t image_format);

/**
 * @brief to abort model operation  (stop sending request to ncpu)
 *
 */
void kdp_app_abort(void);

/**
 * @brief to update image memory address
 * 
 * @return status
 */
int32_t kdp_app_run_model(void);

/**
 * @brief to load model from flash
 * 
 */
void kdp_app_load_model(void);

/**
 * @brief  config model/input/output information before run model
 * @param [in]  model_type_p (Defined in common/include/model_type.h)
 * @param [in]  kdp_img_config_s for image format information and memory address
 * @param [in]  output memory address
 */
void kdp_app_config_model_input_output(
    uint32_t model_type_p,
    struct kdp_img_conf_s* p_img_config_p,
    void* p_output_p);

#endif
