/*
 * Kneron Application Database APIs
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */
 
/**
 * @file kdp_app_db.h
 * @brief Kneron Application DB APIs
 */

#ifndef __KDP_APP_DB_H_
#define __KDP_APP_DB_H_

#include <stdint.h>
#include "kdp_app.h"

/** @brief flash address of DB */ 
#define KDP_APP_DB_FLASH_ADDR   0x80000

/** @brief MAX user count in DB */
#define MAX_USER                20

/** @brief MAX face count for one user*/
#define MAX_FID                 5

/** @brief DB count. 1 or 2 */
#define DB_NUM					2

/** @brief default threshold for comparision. 
 * 
 *  range: [0 - 1] means [strict -> loose]*/
#define KDP_APP_DB_DEFAULT_GUARD_THRESHOLD 0.425

enum kdp_app_db_type_valid_flag_e{
    TYPE_INVALID = 0,
    TYPE_VALID,
    TYPE_REGISTER,
};

enum kdp_app_db_status_code_e{
    KDP_APP_DB_NO_SPACE = 0x200,   //kdpApp.h
    KDP_APP_DB_ALREADY_SAVED,
    KDP_APP_DB_DEL_NOT_VALID,
    KDP_APP_DB_NO_MATCH,
    KDP_APP_DB_REG_FIRST,
    KDP_APP_DB_USER_NOT_REG,
    KDP_APP_DB_DEL_FAIL
};

typedef struct kdp_app_db_user_data_s {
    uint16_t user_id_in;
    uint16_t user_id_out;
    uint16_t user_idx;
    uint16_t fm_idx;
    uint16_t del_all;
} kdp_app_db_user_data_t;

/**
 * @brief compare user fd/fr result to user data in ddr
 * @param [in] fdr_addr address of fd/fr feature map data
 * @param [out] *user_id recognized user ID
 * @parem [in] thresh_guard comparision threshold for recognition
 * @return compare result and user id where find match feature map data
 */
int32_t kdp_app_db_compare(
        uint32_t fdr_addr, uint16_t *user_id/*output*/, 
        float thresh_guard);

/**
 * @brief compare user fd/fr result to user data in ddr
 * @param [in] rgb_fdr_addr address of fd/fr rgb feature map data
 * @param [in] nir_fdr_addr address of fd/fr nir feature map data
 * @param [out] *user_id recognized user ID
 * @param [in] thresh_guard is the threshold
 * @return compare result and user id where find match feature map data
 */
int32_t kdp_app_db_lw3d_compare(
        uint32_t rgb_fdr_addr, uint32_t nir_fdr_addr, 
        uint16_t *user_id, float thresh_guard);

/**
 * @brief compare user fd/fr result to user data in ddr
 * @param [in] rgb_fdr_addr address of fd/fr rgb feature map data
 * @param [in] rgb_weight weight of rgb portion 
 * @param [in] nir_fdr_addr address of fd/fr nir feature map data
 * @param [in] rgb_weight weight of nir portion 
 * @param [out] *user_id recognized user ID
 * @param [in] thresh_guard is the threshold [0-1] (strict - loose)
 * @return compare result and user id where find match feature map data
 */
int32_t kdp_app_db_lw3d_compare_rgb_nir(
        uint32_t rgb_fdr_addr, float rgb_weight, 
        uint32_t nir_fdr_addr, float nir_weight, 
        uint16_t *user_id, float thresh_guard);

/**
 * @brief register user fm data
 * @param [in] fdr_addr address of fd/fr feature map data
 * @param [in] user_id use id
 * @param [in] fm_idx face index
 * @return status
 */
int32_t kdp_app_db_register(uint32_t fdr_addr, uint16_t user_id, uint16_t fm_idx);

/**
 * @brief register user both fm data
 * @param [in] rgb_fdr_addr address of rgb fd/fr feature map data
 * @param [in] nir_fdr_addr address of nir fd/fr feature map data
 * @param [in] user_id use id
 * @param [in] fm_idx face index
 * @return status
 */
int32_t kdp_app_db_both_register(uint32_t rgb_fdr_addr, uint32_t nir_fdr_addr,  uint16_t user_id, uint16_t fm_idx);

/**
 * @brief Initialize database
 * @param [in] flash_db_addr_p flash address of the database
 * @return registered user count
 */
int32_t kdp_app_db_init(uint32_t flash_db_addr_p);

/**
 * @brief switch between two DB if use two DB setting
 * @param [in] db_index indicate which DB is activated
 */
void    kdp_app_db_switch(uint32_t db_index);

/**
 * @brief use two DB to store 2 kinds of FM data
 */
void    kdp_app_db_split_mode(void);

/**
 * @brief register user data to database
 * @param [in] *p_inout_p p_inout_p->user_id user id
 * @return user index in database
 */
int32_t kdp_app_db_add(kdp_app_db_user_data_t *p_inout_p);

/**
 * @brief delete user data in ddr and flash
 * @param [in] *p_inout_p p_inout_p->user_id user id, "0" means all
 * @return none
 */
int32_t kdp_app_db_delete(kdp_app_db_user_data_t *p_inout_p);

/**
 * @brief output user number and all user id data
 */
int32_t kdp_app_db_list(void);

/**
 * @brief output one feature map data of one user
 * @param [in] *p_inout_p p_inout_p->user_id
 *                        p_inout_p->image_idx
 */
int32_t kdp_app_db_upload(kdp_app_db_user_data_t *p_inout_p);

/**
 * @brief clear all user data not yet valid
 * @return OK or error code
 */
int32_t kdp_app_db_abort_reg(void);

/**
 * @brief check if a db slot is used
 * @param [in] i slot index
 * @return 1: yes  0:no
 */
int32_t kdp_app_db_slot_is_used(uint16_t i);

/**
 * @brief get the next available ID slot for register 
 * @return slot id for register
 */
int kdp_app_db_get_available_ID_slot(void);

/**
 * @brief get feature map 0 count for a user index
 * @param [in] user_idx user index
 * @return count for feature map 0
 */
int kdp_app_db_get_fm0_cnt(int user_idx);

/**
 * @brief get feature map 1 count for a user index
 * @param [in] user_idx user index
 * @return count for feature map 1
 */
int kdp_app_db_get_fm1_cnt(int user_idx);

/**
 * @brief get slot type for a user index
 * @param [in] user_idx user index
 * @return type, 0:invalid, 1:valid, 2:register
 */
int kdp_app_db_get_user_id_slot_type(int user_idx);

/**
 * @brief get user ID for a user index
 * @param [in] user_idx user index
 * @return user ID (could be different from user index)
 */
int kdp_app_db_get_user_id(int user_idx);

/**
 * @brief Calculate similarity of two feature points
 * @param [in] fdr_addr address A of user feature map data
 * @param [in] fdr_addr address B of user feature map data
 * @return similarity score (smaller means more like)
 */
float kdp_app_db_cal_similarity( uint32_t fdr_addr_1, uint32_t fdr_addr_2);

/**
 * @brief Calculate similarity of two feature points
 * @param [out] **db_ddr_p output ddr address of DB temp 
 * @param [out] **db_flash_p flash address of DB
 * @return similarity score (smaller means more like)
 */
int32_t kdp_app_db_query(uint32_t **db_ddr_p, uint32_t **db_flash_p);

#endif /* __KDP_APP_DB_H_ */
