/*
 * Kneron Application FD/FR APIs
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */
 
#ifndef __KDP_APP_FID_H_
#define __KDP_APP_FID_H_

#include <stdint.h>
#include <stdbool.h>
#include "ipc.h"
#include "kdp_app.h"

// To use new FDR. If using old FDR (FDSmallbox+Onet+VGG10), comment this Macro
//#define FDSSD_ONETBLUR_VGG

#define RGB_IMG_SOURCE_H 480
#define RGB_IMG_SOURCE_W 640
#define RGB_DDR_BUF_ADDR KDP_DDR_BASE_IMAGE_BUF

/* ==================
 *  FID application 
 *      kdpAppFID.c
 * ================== */
enum kdp_app_fid_status_code_e {
    KDP_APP_FDR_WRONG_USAGE = 0x100,    ///== KDP_APP_FID_CODES
    KDP_APP_FDR_NO_FACE,                /// error: no face 
    KDP_APP_FDR_BAD_POSE,               /// error: base face pose
    KDP_APP_FDR_NO_ALIVE,               /// error: not alive 
    KDP_APP_FDR_FR_FAIL,                /// error: FR failed
    KDP_APP_LM_LOW_CONFIDENT,           /// error: LM low confident
    KDP_APP_LM_BLUR,                    /// error: LM blur
    KDP_APP_SMALL_FACE                  /// error: face too small
};

enum kdp_app_fid_op_e{
    KDP_APP_FID_OP_EXTRACT,             /// extract feature map only
    KDP_APP_FID_OP_INFERENCE,           /// do inference (do DB comparison)
    KDP_APP_FID_OP_REGISTER             /// do register (insert DB temp data)
};

// data structure 
//---------------------------------
typedef struct kdp_app_fid_inout_s {
    enum kdp_app_fid_op_e op;           ///operation mode
    uint16_t face_id;                   //1-5: faceId (1-5)
    uint16_t user_id;                   ///as input: user id got from Host
                                        ///as output: user id returned by KL520
    float thresh_fid;                   ///threshold value for face recognition
} kdp_app_fid_inout_t;

/** @brief face detectin result
 *
 * facedet_result_s is defined in ipc.h 
 * */
typedef struct facedet_result_s kdp_app_face_det_res_t;

/** @brief landmark result
 *
 * landmark_result_s is defined in ipc.h 
 * */
typedef struct landmark_result_s kdp_app_face_lm_res_t;

/** @brief face recognition result
 *
 * fr_result_s defined in ipc.h 
 * */
typedef struct fr_result_s kdp_app_face_fmap_t;

typedef struct kdp_app_image_buffer_desc_s {
   uint32_t buf_count;                  ///image buffer count
   uint32_t alignment_in_bit;           ///alignment size in bit
   uint32_t rgb_buf_base_addr;          ///rgb image buffer address
   uint32_t rgb_img_size;               ///rgb image size
   uint32_t nir_buf_base_addr;          ///nir image buffer address
   uint32_t nir_img_size;               ///nir image size
} kdp_app_image_buffer_desc_t;

/**
 * @brief do face recognition for inference or register
 * @param [inout] *p_inout_p 
 *          { op;
 *            face_id; //as input
 *            user_id; //as input (or ouput) }
 *        op = KDP_APP_FID_OP_INFERENCE   : fd/fr + 1:n compare
 *             KDP_APP_FID_OP_REGISTER    : fd/fr + save to temp
 *             KDP_APP_FID_OP_EXTRACT     : fd/fr only
 *        mode = 1-5 : capture + fd/fr + register in DDR DB
 *        outputPtr: N/A
 * @return OK or Err code
 */
int32_t kdp_app_fdr(
    kdp_app_fid_inout_t *p_inout_p,
    kdp_app_face_det_res_t *p_fd_out_p,
    kdp_app_face_lm_res_t *p_lm_out_p,
    kdp_app_face_fmap_t *p_fr_out_p);


/**
 * @brief do face detection
 * @param[out] p_out_p: detected face bounding box
 * @return OK or Err code
 */
int32_t kdp_app_fd(kdp_app_face_det_res_t *p_out_p/*out*/);


/**
 * @brief init necessary data, face data base
 * @param(in) n/a
 * @return OK 
 */
int32_t kdp_app_init(void);


/**
 * @brief get image setting for HOST's request 
 * @param[out] image_buf descriptor : image buf count, data addr alignment, base address and size
 * @param[in]  image_config param   : image width, height and format
 * @return OK (0) or error code
 */
int kdp_app_query_img_transfer_setting(kdp_app_image_buffer_desc_t *buf_desc_p,
                                       kdp_img_conf_t *img_config_p);

// /**
//  * @brief (Beta) perform Kneron LW3D algorithm
//  * @return OK or Err code
//  */
// int32_t kdp_app_lw3d_fdr(int distance_flag);




/**
 * @brief (Beta) get 5 pose score for a landmark set
 * @param[out] pose : 5 pose score
 * @param[in]  landmark : 5 points landmark, 10 int value
 */
void kdp_app_get_pose_value(uint32_t* landmarks, float *poses);


/**
 * @brief (Beta) check if it is a bad pose for a landmark set
 * @param[in]  landmark : 5 points landmark, 10 int value
 * @return True for bad pose, False otherwise
 */
bool kdp_app_check_bad_pose(uint32_t *landmarks);

int kdp_app_check_head_pose(kdp_app_face_lm_res_t* landmarks, uint8_t reg_idx, uint8_t force_flag,
                            uint8_t* face_reg_thresh);



#endif
