/*
 * Kneron Host Communication driver
 *
 * Copyright (C) 2018 Kneron, Inc. All rights reserved.
 *
 */
 
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include "cmsis_os2.h"
#include "base.h"
#include "host_com.h"
#include "com.h"
#include "msg.h"
#include "kneron_mozart.h"
#include "dbg.h"
#include "kdp_app.h"
#include "kdp_app_fid.h"
#include "kdp_app_db.h"
#include "usb_com.h"
#include "kdp_memory.h"
#include "kdp_peripheral.h"
#include "kdp_uart.h"
#include "Driver_Flash.h"
#include "DrvSPI020.h"
#include "v2k_cam.h"
#include "system.h"
#include "framework/event.h"
#include "os_tick.h"
#include "ota_update.h"
#include "kdp_ddr_table.h"
#include "ipc.h"
#include "delay.h"
#include "power_manager.h"
#include "power.h"

extern void kdp_uart_com_write(u8* buffer, int size);
extern int  kdp_uart_com_read(u8* address, u32 size);
extern void kdp_uart_com_init(void);
extern void mpu_niram_disable(void);
extern void mpu_niram_enable(void);
extern u32 kdp_timer_read(u32 timer_index);
extern int kdp_timer_init(u32 timer_index);
extern int kdp_memxfer_ddr_to_flash(u32 dst, u32 src, size_t bytes);

#define NiRAM_MEM_BASE 0x28000000
#define TIMER_DEFAULT                   1

// SFID prototype
u32 host_mem_read(u32 address, u32 image_size);
u32 fdlmfrlv_outputs(u32 *check_flag, u16 *msg_out, u32 lv_flag);
void get_lw3d_io_memory(void);
void output_results(u16 *output_buf, void *out_result, u32 count);

#define MSG_APP_BAD_IMAGE 0x100 // 256
#define MSG_APP_BAD_INDEX 0x101
#define MSG_APP_UID_EXIST 0x102
#define MSG_APP_UID_DIFF 0x103
#define MSG_APP_IDX_FIRST 0x104
#define MSG_APP_IDX_MISSING 0x105
#define MSG_APP_DB_NO_USER 0x106
#define MSG_APP_DB_FAIL 0x107   // unknown general failure
#define MSG_APP_LV_FAIL 0x108
#define MSG_APP_DB_NO_MATCH 10

#define FID_THRESHOLD 0.425
#define FID_MAX_USER 20

#define MSG_DB_BAD_PARAM   248  // database action/format error
#define MSG_SFID_OUT_FAIL  249  // data upload failed
#define MSG_SFID_NO_MEM    250  // memory allocation failed
#define MSG_AUTH_FAIL      251  // authentication failed
#define MSG_FLASH_FAIL     252  // flash programming failed (bad sector?)
#define MSG_SFID_DATA_FAIL 253  // data download failed
#define MSG_SFID_USR_ZERO  254  // user id zero error
#define MSG_APP_NOT_CONFIG 255  // no appropriate Start issued previously

#define UPDATE_MODULE_NONE 0
#define UPDATE_MODULE_SCPU 1
#define UPDATE_MODULE_NCPU 2

// message for DME
#define MSG_DME_DATA_FAIL  253  // data download failed
#define MSG_DME_NO_CONFIG  255  // no Start DME issued previously

osThreadId_t tid_host_comm;

void usb_host_comm_init(uint32_t flag);
void uart_host_comm_init(uint32_t flag);

// Global Application Mode Control
u16 APP_id = 0;  // 1 -> 3D (LW3D), 2 -> 2D, 3 -> DME
u16 SYS_sts = 0; // system status

u16 SFID_user;   // SFID user id
u16 SFID_index;  // SFID image index
kdp_app_image_buffer_desc_t SFID_buf_desc;
float RGB_threshold = FID_THRESHOLD;
float NIR_threshold = FID_THRESHOLD;

uint32_t dbg_model_type = 0;

static kdp_img_conf_t rgb_img_config;
static kdp_img_conf_t nir_img_config;



// Flash control
ARM_Flash_SignalEvent_t m_event;
extern ARM_DRIVER_FLASH Driver_Flash;
ARM_DRIVER_FLASH *pdrv_flash = &Driver_Flash;

// flash function prototypes
int flash_sector_merge(u8 *sector_start_addr, u8 *new_data_start, u32 new_data_size);
int flash_wait_ready(int timeout_ms);

/**
 * @brief cmd_parser(), uart/usb message parser
 */

int cmd_parser(u8 *buf, int len)
{
    MsgHdr *msghdr;
    CmdPram *msgcmd_arg;
    RspPram *msgrsp_arg;
    OpPram *opcmdpram;
    u8 *pdata = NULL;
    u32 *pwdata = NULL;
    u8 rstatus = STS_OK;
    int wlen;
    int ret;

    msghdr = (MsgHdr*)buf;
    msgcmd_arg = (CmdPram*)(buf + sizeof(MsgHdr));
    msgrsp_arg = (RspPram*)(buf + sizeof(MsgHdr));  // use the same rcv buffer to format responses
    opcmdpram = (OpPram*)msgcmd_arg;  // used to parse operational commands

    trace_msg("[Host_com] cmd_parser[%d]: cmd %d [addr, len]=[0x%x, %d]\n", len, msghdr->cmd, msgcmd_arg->addr, msgcmd_arg->len);

    pdata = (u8*)(buf + (sizeof(MsgHdr) + sizeof(CmdPram)));  // points to stuff beyond error and byte count
    pwdata = (u32*)(buf + (sizeof(MsgHdr) + sizeof(CmdPram)));
    wlen = (msgcmd_arg->len % 4) == 0 ? msgcmd_arg->len / 4 : msgcmd_arg->len / 4 + 1;  // last bytes is treated as word

    /* Command parsing */
    if (msghdr->cmd < 0x100)  // build-in commands
    {        
        switch (msghdr->cmd) {
        // Test Commands
        case CMD_MEM_WRITE:
#if MSG_MEM_BYTE_WRITE
        /* NPU memory(0x30000000) doest not support byte R/W */
            if ((msgcmd_arg->addr >= NPU_PA_BASE) && (msghdr->addr < 0x40000000)) {
#else
            if (1) {
#endif
                for (int i = 0; i < wlen; i++) {
                    outw(msgcmd_arg->addr + i * 4, (u32) *(pwdata + i));
                }
            }
            else {
                for (int i = 0; i < msgcmd_arg->len; i++) {
                    outb(msgcmd_arg->addr + i, *(pdata + i));
                }
            }
            // msg byte write size is unchanged
            msgrsp_arg->error = NO_ERROR;
            msg_pack(CMD_MEM_WRITE, NULL, 0);
            break;

        case CMD_MEM_READ:
#if MSG_MEM_BYTE_WRITE
        /* NPU memory(0x30000000) doest not support byte R/W */
            if ((msgcmd_arg->addr >= NPU_PA_BASE) && (msghdr->addr < 0x40000000)) {
#else
            if (1) {
#endif
                for (int i = 0; i < wlen; i++) {
                    *(pwdata + i) = inw(msgcmd_arg->addr + 4 * i);
                }
            }
            else {
                for (int i = 0; i < msgcmd_arg->len; i++) {
                    *(pdata + i) = inb(msgcmd_arg->addr + i);
                }
            }
            // msg byte read size is unchanged
            msgrsp_arg->error = NO_ERROR;
            msg_pack(CMD_MEM_READ, (u8 *)pdata, msgcmd_arg->len);
            break;
        case CMD_TEST_ECHO:
            // msg bytes echo'ed is unchanged
            msgrsp_arg->error = NO_ERROR;
            msg_pack(CMD_TEST_ECHO, (u8 *)pdata, msgcmd_arg->len);
            break;

        case CMD_MEM_CLR:
            memset((void *)msgcmd_arg->addr, 0, msgcmd_arg->len);
            // msg clear size is unchanged
            msgrsp_arg->error = NO_ERROR;
            msg_pack(CMD_MEM_CLR, NULL, 0);
            break;

        case CMD_FILE_WRITE:
        {
            u32 bytes_read;
            bytes_read = host_mem_read(msgcmd_arg->addr, msgcmd_arg->len);
            if (bytes_read == msgcmd_arg->len)
                msgrsp_arg->error = NO_ERROR;
            else
                msgrsp_arg->error = FILE_ERROR;
            msgrsp_arg->bytes = bytes_read;
            msg_pack(CMD_FILE_WRITE, NULL, 0);
            break;
        }
        case CMD_FLASH_MEM_WRITE:
        {
           // This code is Winbond specific.  For other vendors, modification might be needed!!!!!!
            u32 offset_addr = msgcmd_arg->addr & 0x0FFF;  // find offset addr into sector
            u8* flash_sector_start = msg_tbuf + 0x40;  // keep away from potential data
            u8* flash_sector_check = msg_rbuf + 0x40;  // reuse the buffer
            u32 i;

            if ((msgcmd_arg->len + offset_addr) > 0x1000) {
                msgrsp_arg->error = 1;  // target addr out of range
                msgrsp_arg->bytes = 0;
            }
            else {  // peform FLASH data compare, merge and write
                // load flash sector into tx buf as temp storage
                pdrv_flash->ReadData(msgcmd_arg->addr & 0xFFFFF000, flash_sector_start, 0x1000);  // read in 4KB
                ret = flash_wait_ready(100);  // wait for the read operation to complete
                dbg_msg("[Host_com] Read Flash Sector Wait Time -> %d\r\n", ret);
                if (ret < 0) {
                    err_msg("[ERROR] host_com: Read Flash Sector failed, ret -> %d\r\n", ret);
                }

                ret = flash_sector_merge(flash_sector_start+offset_addr, msgcmd_arg->data, msgcmd_arg->len);
                info_msg("[Host_com] Flash Merge ret code is %d\r\n", ret);
                
                if (ret < 0) { // erase the sector first before programming the whole 4KB
                    pdrv_flash->EraseSector(msgcmd_arg->addr);
                    ret = flash_wait_ready(100);  // wait for the erase operation to complete
                    if (ret < 0) {
                        dbg_msg("[Host_com] Erase Flash Sector Wait Time -> %d\r\n", ret);
                    }
                    if (ret >= 0) {
                        for (i = 0; i < 0x1000; i += 0x100) {  // program each 256 byte page separately
                            pdrv_flash->ProgramData((msgcmd_arg->addr & 0xFFFFF000)+i, flash_sector_start+i, 0x100);
                            flash_wait_ready(100);
                        }
                        ret = -1;  // restore ret value for later stages
                    }
                }
                
                if (ret > 0) {  // program the new data only, one page at a time
                    u32 first_page, more_page, last_page;
                    first_page = 0x100 - (offset_addr & 0xFF);  // size of first page
                    more_page = msgcmd_arg->len - first_page;
                    last_page = more_page & 0xFF;
                    more_page = more_page & 0xF00;
                    if (first_page) { // do we have partial first page?
                        pdrv_flash->ProgramData(msgcmd_arg->addr, flash_sector_start + offset_addr, first_page);
                        flash_wait_ready(100);
                    }
                    for (i = first_page; i < first_page + more_page; i += 0x100) {  // program complete pages
                        pdrv_flash->ProgramData(msgcmd_arg->addr + i, flash_sector_start + i, 0x100);
                        ret = flash_wait_ready(100);
                        dbg_msg("[Host_com] Flash Sector Write Wait Time -> %d\r\n", ret);
                    }
                    if (last_page) {
                        i = first_page + more_page;
                        pdrv_flash->ProgramData(msgcmd_arg->addr + i, flash_sector_start + i, last_page);
                        flash_wait_ready(100);
                    }
                }

                // do post check
                if (ret != 0) {
                    pdrv_flash->ReadData(msgcmd_arg->addr & 0xFFFFF000, flash_sector_check, 0x1000);  // read the new sector
                    ret = flash_wait_ready(100);  // wait for the read operation to complete
                    if (memcmp(flash_sector_start, flash_sector_check, 0x1000)) {  // if not equal, failed verification
                        msgrsp_arg->error = 2;
                    }
                    else {
                        msgrsp_arg->error = NO_ERROR;
                    }
                }
                else {
                    msgrsp_arg->error = NO_ERROR;
                }
                msgrsp_arg->bytes = msgcmd_arg->len;
            }
            // generate response msg
            msg_pack(CMD_FLASH_MEM_WRITE, NULL, 0);
            break;
        }

        // Operational Commands
        case CMD_RESET:
        {
            u32 reset_mode, reset_flag = 0;  // reset flag: 1->sleep, 2->reset, 3->RTC

            reset_mode = opcmdpram->op_parm1;
#if USB_HOST
            if ((reset_mode < 2) || (reset_mode == 255)  || (reset_mode == 256)
#else
            if (((reset_mode < 5) && (reset_mode != 2)) || (reset_mode == 255) || (reset_mode == 256)
#endif
                || ((reset_mode >> 16) == 0x1000) || ((reset_mode >> 16) == 0x2000))
            {
                if ((reset_mode + opcmdpram->op_parm2) == 0xFFFFFFFF) {
                    if (reset_mode == 3) reset_flag = 1;    // go to sleep
                    if (reset_mode == 255) reset_flag = 2;  // reset
                    if (reset_mode == 256) reset_flag = 3;  // power down
                    if ((reset_mode >> 16) == 0x1000) {
                        log_set_level_scpu(reset_mode & 0x000F);  // set scpu debug output level
                        log_set_level_ncpu((reset_mode & 0x0F00)>> 8);  // set ncpu debug output level

                        // Disable OS tick for power but keep it for profile/debug
                        if ((reset_mode & 0x000F) >= LOG_PROFILE)
                            OS_Tick_Enable();
                        else
                            OS_Tick_Disable();
                    }
                    if ((reset_mode >> 16) == 0x2000) {
                        // switch active boot image, bit 0 -> scpu, bit 1 -> ncpu

                        int ret = ota_update_switch_active_partition(reset_mode & 0x3);
                        msgrsp_arg->error = ret;

                    }
                    msgrsp_arg->error = NO_ERROR;  // not doing anything else for now
                }
                else
                    msgrsp_arg->error = BAD_CONFIRMATION;
            }
            else
                msgrsp_arg->error = BAD_MODE;
            msgrsp_arg->bytes = 0;
            msg_pack(CMD_RESET, NULL, 0);
            if (reset_flag) {
                msg_wait();  // wait till reply message is done
                // ncpu fw needs update to support waking up from sleep
//                if (reset_flag == 1) power_manager_sleep();
                if (reset_flag == 2) power_mgr_sw_reset();
                if (reset_flag == 3) power_manager_shutdown();
            }
            break;
        }

        case CMD_SYSTEM_STATUS:
        {
            u16* status;
            msgrsp_arg->error = DEVICE_ID;   // use "error" parameter for device ID
            msgrsp_arg->bytes = FW_VERSION;  // and "len" parameter for firmware version
            status = (u16*)msgrsp_arg->data;
            *status++ = SYS_sts;
            *status = APP_id;
            msg_pack(CMD_SYSTEM_STATUS, msgrsp_arg->data, 4);
            break;
        }

        case CMD_UPDATE_FW:
        {
            u32 model_id;
            model_id = opcmdpram->op_parm1;

            switch(model_id)
            {
                case UPDATE_MODULE_SCPU:
                {
                    ret = ota_update_scpu();
                    msgrsp_arg->error = ret;
                    msgrsp_arg->bytes = UPDATE_MODULE_SCPU;
                    msg_pack(CMD_UPDATE_FW, NULL, 0);
                    if(ret == OTA_UPDATE_SUCCESS) {
                        delay_us(50*1000);
                        power_mgr_sw_reset();
                    }
                    break;
                }

                case UPDATE_MODULE_NCPU:
                {
                    ret = ota_update_ncpu();
                    msgrsp_arg->error = ret;
                    msgrsp_arg->bytes = UPDATE_MODULE_NCPU;
                    msg_pack(CMD_UPDATE_FW, NULL, 0);
                    if (ret == OTA_UPDATE_SUCCESS) {
                        delay_us(50*1000);
                        power_mgr_sw_reset();
                    }
                    break;
                }
                case UPDATE_MODULE_NONE:
                default:
                    break;
            }
            break;
        }

        case CMD_DOWNLOAD_MODEL:
        {
            u32 model_id, size;
            model_id = opcmdpram->op_parm1;
            size = opcmdpram->op_parm2;
            if(model_id != 0)
            {
                ret = ota_update_model(model_id, size);
                msgrsp_arg->error = ret;
                msgrsp_arg->bytes = model_id;
                msg_pack(CMD_DOWNLOAD_MODEL, NULL, 0);
                if(ret == OTA_UPDATE_SUCCESS)
                {
                    delay_us(50*1000);
                    power_mgr_sw_reset();
                }
            }
            break;
        }

        case CMD_QUERY_APPS:
        case CMD_SELECT_APP:
        case CMD_SET_MODE:
        case CMD_SET_EVENTS:
        case CMD_UPDATE:
        case CMD_ABORT:
        default:
            msgrsp_arg->error = RSP_UKNOWN_CMD;
            msgrsp_arg->bytes = 0;
            msg_pack(msghdr->cmd, NULL, 0);
            break;
        }
    }
    else if (msghdr->cmd >> 8 == 0x20)
    {
#ifdef DBG_ALONE
        switch (msghdr->cmd)
        {
            case CMD_DBG_USB_MEM_WR:
            {
                uint32_t mem_addr = (uint32_t)msgcmd_arg->addr;
                uint32_t buf_len = (uint32_t)msgcmd_arg->len;
                dbg_msg("cmd_parser: USB mem write to addr:0x%x, size:%d\n", mem_addr, buf_len);
                // get data from USB
                ret = usb_mem_read(mem_addr, buf_len);
                if(ret == buf_len) {
                    dbg_msg("cmd_parser: read USB data successfully\n");
                    msgrsp_arg->error = NO_ERROR;
                    msgrsp_arg->bytes = 0;
                    msg_pack(CMD_ACK, NULL, 0);
                }
                else {
                    dbg_msg("Error! cmd_parser: read USB data failed\n");
                    msgrsp_arg->error = USB_ERROR;
                    msgrsp_arg->bytes = 0;
                    msg_pack(CMD_DBG_USB_MEM_WR, NULL, 0);
                }
                break;
            }

            case CMD_DBG_SET_MODEL_TYPE:
            {
                dbg_model_type = (uint32_t)msgcmd_arg->addr;
                msgrsp_arg->error = NO_ERROR;
                msgrsp_arg->bytes = 0;
                msg_pack(CMD_ACK, NULL, 0);

                dbg_msg("cmd_parser: set model_type=%d\n", dbg_model_type);
                break;
            }
            default:
            break;
        }
#endif
    }
    else
    {
        msgrsp_arg->error = RSP_UKNOWN_CMD;
        msgrsp_arg->bytes = 0;
        msg_pack(msghdr->cmd, NULL, 0);
    }
    return rstatus;
}
  

void output_results(u16 *output_buf, void *out_result, u32 count)
{
    u32 *result = out_result;
    while (count) {
        *output_buf++ = (u16) *result++;
        count--;
    }
}

// The ack_packet needs to be static, place it on stack doesn't seem to work
// TODO: encode source id and add crc for uart data mode
u8 ack_packet[20] = {0x35, 0x8A, 12, 0, 4, 0, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

u32 host_mem_read(u32 address, u32 size)
{
#if USB_HOST
    usb_com_write (ack_packet, 16, FLAG_HOST_COMM_USB_DONE, 1);  // use data mode
    return(usb_com_read((u8*) address, size, FLAG_HOST_COMM_USB_DONE, 1));  // use data mode
#else
// NOTE: For uart, we also use msg_rx buffer for data receiving, thus destroying the command data that might
//       be needed later.  So we must preserve it (save it in the msg_tx buffer) for restoration at the end.
    u32 block, residual, offset;
    block = size >> 12;
    residual = size & 0x0FFF;

    save_command();
    offset = 0;
    while (block) {
        kdp_uart_com_write(ack_packet, 16);
        kdp_uart_com_read((u8*) address + offset, 0x1000); // FLAG_HOST_COMM_UART_DONE);
        offset += 0x1000;
        block--;
    }
    if (residual) {
        kdp_uart_com_write(ack_packet, 16);
        kdp_uart_com_read((u8*) address + offset, residual); //, FLAG_HOST_COMM_UART_DONE);
    }
    restore_command();
    return(size);
#endif
}

int flash_wait_ready(int timeout_ms)
{
    ARM_FLASH_STATUS flash_status;
    int i;

    for (i = 0; i < timeout_ms; i++) {
        flash_status = pdrv_flash->GetStatus();
        if (flash_status.busy == 0) break;
        delay_us(1*1000);
    }
    if (i == timeout_ms) i = -1;  // we have timed out
    return i;
}

/* Flash sector merge function operates on a single sector at a time
*  input - starting address of original flash data
*        - starting address of new data to be written
*        - size of new data in bytes
*  output - 0 -> data is identical, no write needed
*         - positive -> data is merged, no need to erase first
*         - negative -> data is merged, needs erase first
*/
int flash_sector_merge(u8* sector_offset, u8* data_start, u32 data_size)
{
    // we don't do data alignment adjustment yet, assume all inputs are 32-bit aligned
    u16 need_erase = 0, need_update = 0;
    u32 i, *pflash_word, *pnew_word;
    
    pflash_word = (u32*)sector_offset;
    pnew_word = (u32*)data_start;
    for (i = data_size >> 2; i > 0; i--)
    {
        if (*pflash_word != *pnew_word) need_update = 1;
        if ((*pflash_word ^ *pnew_word) & *pnew_word) {
            need_erase = 1;
        }
        else {
            pflash_word++;
            pnew_word++;
        }
        if (need_erase && need_update) break;  // early abort
    }
    
    if (need_update) {
        memcpy(sector_offset, data_start, data_size);
        if (need_erase) {
            return -1;
        }
        return 1;
    }
    return 0;  // data are identical
}

static void device_sleep(void)
{
    SYS_sts = 1;
//    trace_msg("[Host_com] Going to sleep....\r\n");
}

static void device_wakeup(void)
{
    SYS_sts = 0;
//    trace_msg("[Host_com] Waking up....\r\n");
}

struct pm_s pm_host_com_device =
{
    .nap = 0,
    .wakeup_nap = 0,
    .deep_nap = 0,
    .wakeup_deep_nap = 0,
    .sleep = (pm_call) device_sleep,
    .wakeup_sleep = (pm_call) device_wakeup,
    .deep_sleep = (pm_call) device_sleep,
    .wakeup_deep_sleep = (pm_call) device_wakeup,
};

static void host_comm_thread(void *argument)
{
    int32_t flags;
    int len;

    //init kdp_app
    kdp_app_init();

    ota_handle_first_time_boot();
    err_msg("SCPU host-com in working ... \n");
	
#if USB_HOST
    usb_host_comm_init(FLAG_HOST_COMM_USB_DONE);
#else
    uart_host_comm_init(FLAG_HOST_COMM_UART_DONE);
#endif
    //reserve DDR memory for image buffer
    rgb_img_config.image_memory_address = kdp_ddr_reserve(640*480*2);    // 16 bpp
    nir_img_config.image_memory_address = kdp_ddr_reserve(640*480);      // 8  bpp

    // register PM_DEVICE_HOST_COM for sleep & wakeup
    power_manager_register(PM_DEVICE_HOST_COM, &pm_host_com_device);
    for (;;) {
        flags = osThreadFlagsWait(FLAG_HOST_COMM_USB_DONE | FLAG_HOST_COMM_UART_DONE | FLAG_HOST_COMM_TIMER, osFlagsWaitAny, osWaitForever);
        osThreadFlagsClear(flags);

        if (flags & FLAG_HOST_COMM_TIMER) {
            if (len > 0) {
                fLib_printf("Get uart_len = %d\n", len);
                msg_read(msg_rbuf, len);
            }
        }

#if USB_HOST
        if (flags & FLAG_HOST_COMM_USB_DONE) {
#else
        if (flags & FLAG_HOST_COMM_UART_DONE) {
#endif
            msg_read(msg_rbuf, msg_rbuf[2]+4);  // process command
        }
    }
}

void host_com_init(void)
{
#if USB_HOST
    usb_init();
#else
    kdp_uart_com_init();
#endif

    tid_host_comm = osThreadNew(host_comm_thread, NULL, NULL);
#if 0
    osStatus_t status;    status = osThreadSetPriority(tid_host_comm, osPriorityNormal);
    if (status != osOK) {
        fLib_printf("osThreadSetPriority (host comm) error: %d\n", status);
    }
#endif
}
