/* -------------------------------------------------------------------------- 
 * @file msg.c 
 *      pc command message threads (Uart/USB)
 *---------------------------------------------------------------------------*/
#include "base.h"
#include "com.h"
#include "msg.h"
#include "io.h"
#include "string.h"
#include "dbg.h"
#include "host_com.h"
#include "usb_com.h"

u8 msg_rbuf[MSG_DATA_BUF_MAX + sizeof(MsgHdr)+ sizeof(RspPram) + 4 + 4];  // allow u16 crc value plus
u8 msg_tbuf[MSG_DATA_BUF_MAX + sizeof(MsgHdr)+ sizeof(RspPram) + 4 + 4];  // workaround needed for usb bug

#define CRC16_CONSTANT 0x8005

u16 gen_crc16(u8 *data, u32 size)
{
	u16 out = 0;
	int bits_read = 0, bit_flag, i;

	/* Sanity check: */
	if (data == NULL)
		return 0;

	while (size > 0)
	{
		bit_flag = out >> 15;

		/* Get next bit: */
		out <<= 1;
		out |= (*data >> bits_read) & 1; // item a) work from the least significant bits

										 /* Increment bit counter: */
		bits_read++;
		if (bits_read > 7)
		{
			bits_read = 0;
			data++;
			size--;
		}

		/* Cycle check: */
		if (bit_flag)
			out ^= CRC16_CONSTANT;

	}

	// push out the last 16 bits
	for (i = 0; i < 16; ++i) {
		bit_flag = out >> 15;
		out <<= 1;
		if (bit_flag)
			out ^= CRC16_CONSTANT;
	}

	// reverse the bits
	u16 crc = 0;
	i = 0x8000;
	int j = 0x0001;
	for (; i != 0; i >>= 1, j <<= 1) {
		if (i & out) crc |= j;
	}

	return crc;
}

/**
 * @brief data_write(), write raw msg to host
 * @param buf buffer data to write
 * @param len data length
 */
int data_write(u8 *buf, int len)
{
#if USB_HOST
    usb_com_write(buf, len, FLAG_HOST_COMM_USB_DONE, 0);  // use msg mode
#else
    kdp_uart_com_write(buf, len);
#endif

    return 0;
}

void no_host_reply(void)
{
#if USB_HOST
    usb_idle();
#endif
}

/*
 * @brief msg_read(), process the incoming uart message and checks the packet format
 * @param buf buffer for uart read
 */
int msg_read(u8 *buf, int size)
{
    MsgHdr *msghdr = (MsgHdr*)msg_rbuf;
    RspPram *msgrsp_arg;
    u32 crc16;
	int length;
    u16 msg_rsp;
	
	if (msghdr->preamble != MSG_HDR_CMD)
	{
        dbg_msg("BAD Preamble %d\r\n", size);
        return -1;
	}
	if (msghdr->pSize & PKT_CRC_FLAG)
    {
	    length = msghdr->pSize & 0x01FFF;  // mask off packet flags
        crc16 = gen_crc16(buf, length);  // compute crc value
//        dbg_msg("Computed CRC16[%d] = %04X\r\n", length, crc16);
        if (crc16 != (buf[length+0]+(buf[length+1]<<8)))
	    {
	        // send CRC ERROR packet, format the message in msg_rbuf first
            dbg_msg("BAD CRC [%d] crc16=%x : %x\r\n", length, crc16, buf[length+3]+(buf[length+3]<<8));
            msg_rsp = msghdr->cmd | 0x8000;
            msghdr = (MsgHdr*) msg_tbuf;
            msghdr->cmd = msg_rsp;
            msghdr->mSize = 8;
            msgrsp_arg = (RspPram*) (msg_tbuf + sizeof(MsgHdr));
            msgrsp_arg->error = CMD_CRC_ERR;
            msgrsp_arg->bytes = 0;
            msg_write(msg_tbuf, sizeof(MsgHdr)+sizeof(RspPram), 0);  // send crc error packet, without crc
	        return -1;
	    }
	}
	cmd_parser(buf, size);  // if there are more than one command, we only process the first one
	return size;
}

/**
 * @brief msg_write(), write msg to PC by uart
 *        also format uart packet here
 * @param buf buffer data to write
 * @param len data length
 */
int msg_write(u8 *buf, int len, int crc_flag)
{
    MsgHdr *msghdr;
    u32 crc16, *crc;

    crc = (u32*)&buf[len];  // identify the crc location
    msghdr = (MsgHdr*) buf;
    msghdr->preamble = MSG_HDR_RSP;
    if (crc_flag)
    {
        msghdr->pSize = len | PKT_CRC_FLAG;  // packet + crc(4B) - pkt header (2B)
        crc16 = gen_crc16(buf, len);
        *crc = crc16;  // store crc
        len += 4;  // adjust len to include crc
    }
    else
        msghdr->pSize = len - 4;

#if USB_HOST
    usb_com_write(buf, len, FLAG_HOST_COMM_USB_DONE, 0);  // use msg mode
#else
    kdp_uart_com_write(buf, len);
#endif

    return 0;    
}

/**
 * @brief msg_pack(), pack message and send out, use crc for UART only
 * @param cmd message command
 * @param buf data buffer
 * @param len length of data buffer, if len < 0, don't flip the cmd
 */
void msg_pack(u16 cmd, u8 *buf, int len)
{
    MsgHdr *msghdr;
	CmdPram *msgcmd_arg;
	RspPram *msgrsp_arg;
    if (len > 0) {  // copy stuff beyond len and address
        memcpy((msg_tbuf + (sizeof(MsgHdr)+sizeof(RspPram))), buf, len);
    };
        
    msghdr = (MsgHdr *) msg_tbuf;  // msghdr now points to tbuf
    msgcmd_arg = (CmdPram*) (msg_rbuf + sizeof(MsgHdr));
    msgrsp_arg = (RspPram*) (msg_tbuf + sizeof(MsgHdr));
     if (len >= 0) {
        msghdr->cmd = cmd | 0x8000;
        msghdr->mSize = len + sizeof(RspPram);  // minimum msg payload include RspPram
        msgrsp_arg->error = msgcmd_arg->addr;
        msgrsp_arg->bytes = msgcmd_arg->len;
        len += sizeof(MsgHdr)+sizeof(RspPram);
    }
    else {
        msghdr->cmd = cmd;
        msghdr->mSize = 0;  // no message payload
        len = sizeof(MsgHdr);  // just send msg header
    }
#if USB_HOST
    msg_write(msg_tbuf, len, 0);  // no crc
#else
    msg_write(msg_tbuf, len, 1);  // use crc
#endif
    return;
}

void msg_wait(void)
{
#if USB_HOST
    while(get_usb_mode() != 1);  // wait till message is done
#endif
}

void save_command(void)
{
    memcpy(msg_tbuf, msg_rbuf, msg_rbuf[2]+4);  // copying the received command (must be < 255 bytes) to tx buf
}

void restore_command(void)
{
    memcpy(msg_rbuf, msg_tbuf, msg_tbuf[2]+4);  // copying the received command to tx buf
}
