/*
 * Kneron Host Communication driver
 *
 * Copyright (C) 2018 Kneron, Inc. All rights reserved.
 *
 */
 
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include "cmsis_os2.h"
#include "base.h"
#include "host_com.h"
#include "com.h"
#include "msg.h"
#include "kneron_mozart.h"
#include "dbg.h"
#include "kdp_app.h"
#include "kdp_app_fid.h"
#include "kdp_app_lw3d.h"
#include "kdp_app_db.h"
#include "kdp_app_dme.h"
#include "usb_com.h"
#include "kdp_memory.h"
#include "kdp_peripheral.h"
#include "kdp_uart.h"
#include "Driver_Flash.h"
#include "DrvSPI020.h"
#include "v2k_cam.h"
#include "system.h"
#include "framework/event.h"
#include "os_tick.h"
#include "ota_update.h"
#include "kdp_ddr_table.h"
#include "ipc.h"
#include "delay.h"
#include "power_manager.h"
#include "power.h"

extern void kdp_uart_com_write(u8* buffer, int size);
extern int  kdp_uart_com_read(u8* address, u32 size);
extern void kdp_uart_com_init(void);
extern void mpu_niram_disable(void);
extern void mpu_niram_enable(void);
extern u32 kdp_timer_read(u32 timer_index);
extern int kdp_timer_init(u32 timer_index);
extern int kdp_memxfer_ddr_to_flash(u32 dst, u32 src, size_t bytes);

#define NiRAM_MEM_BASE 0x28000000
#define TIMER_DEFAULT                   1

// SFID prototype
u32 host_mem_read(u32 address, u32 image_size);
u32 fdlmfrlv_outputs(u32 *check_flag, u16 *msg_out, u32 lv_flag);
void get_lw3d_io_memory(void);
void output_results(u16 *output_buf, void *out_result, u32 count);

#define MSG_APP_BAD_IMAGE 0x100 // 256
#define MSG_APP_BAD_INDEX 0x101
#define MSG_APP_UID_EXIST 0x102
#define MSG_APP_UID_DIFF 0x103
#define MSG_APP_IDX_FIRST 0x104
#define MSG_APP_IDX_MISSING 0x105
#define MSG_APP_DB_NO_USER 0x106
#define MSG_APP_DB_FAIL 0x107   // unknown general failure
#define MSG_APP_LV_FAIL 0x108
#define MSG_APP_DB_NO_MATCH 10

#define FID_THRESHOLD 0.425
#define FID_MAX_USER 20

#define MSG_DB_BAD_PARAM   248  // database action/format error
#define MSG_SFID_OUT_FAIL  249  // data upload failed
#define MSG_SFID_NO_MEM    250  // memory allocation failed
#define MSG_AUTH_FAIL      251  // authentication failed
#define MSG_FLASH_FAIL     252  // flash programming failed (bad sector?)
#define MSG_SFID_DATA_FAIL 253  // data download failed
#define MSG_SFID_USR_ZERO  254  // user id zero error
#define MSG_APP_NOT_CONFIG 255  // no appropriate Start issued previously

#define UPDATE_MODULE_NONE 0
#define UPDATE_MODULE_SCPU 1
#define UPDATE_MODULE_NCPU 2

// message for DME
#define MSG_DME_DATA_FAIL  253  // data download failed
#define MSG_DME_NO_CONFIG  255  // no Start DME issued previously

osThreadId_t tid_host_comm;

void usb_host_comm_init(uint32_t flag);
void uart_host_comm_init(uint32_t flag);

// Global Application Mode Control
u16 APP_id = 0;  // 1 -> 3D (LW3D), 2 -> 2D, 3 -> DME
u16 SYS_sts = 0; // system status

u16 SFID_user;   // SFID user id
u16 SFID_index;  // SFID image index
kdp_app_image_buffer_desc_t SFID_buf_desc;
float RGB_threshold = FID_THRESHOLD;
float NIR_threshold = FID_THRESHOLD;

uint32_t dbg_model_type = 0;

static kdp_img_conf_t rgb_img_config;
static kdp_img_conf_t nir_img_config;

static kdp_app_face_det_res_t       *p_fd_out_s = 0;
static kdp_app_face_lm_res_t        *p_lm_out_s = 0;
static kdp_app_face_fmap_t          *p_fr_out_s = 0;
static kdp_app_fid_inout_t          *p_fdr_out_s = 0;

static kdp_app_lw3d_face_det_res_t  *p_rgb_fd_out_s = 0;
static kdp_app_lw3d_face_lm_res_t   *p_rgb_lm_out_s = 0;
static kdp_app_lw3d_face_fmap_t     *p_rgb_fr_out_s = 0;
static kdp_app_lw3d_face_det_res_t  *p_nir_fd_out_s = 0;
static kdp_app_lw3d_face_lm_res_t   *p_nir_lm_out_s = 0;
static kdp_app_lw3d_face_fmap_t     *p_nir_fr_out_s = 0;
static kdp_app_lw3d_inout_t         *p_lw3d_out_s = 0;
static kdp_app_dme_res_t            *p_od_out_s = 0;

static uint32_t p_raw_out_s = 0;
static struct kdp_dme_conf_s dme_conf;
static uint32_t final_len = 0;

// Flash control
ARM_Flash_SignalEvent_t m_event;
extern ARM_DRIVER_FLASH Driver_Flash;
ARM_DRIVER_FLASH *pdrv_flash = &Driver_Flash;

// flash function prototypes
int flash_sector_merge(u8 *sector_start_addr, u8 *new_data_start, u32 new_data_size);
int flash_wait_ready(int timeout_ms);

/**
 * @brief cmd_parser(), uart/usb message parser
 */

int cmd_parser(u8 *buf, int len)
{
    MsgHdr *msghdr;
    CmdPram *msgcmd_arg;
    RspPram *msgrsp_arg;
    OpPram *opcmdpram;
    u8 *pdata = NULL;
    u32 *pwdata = NULL;
    u8 rstatus = STS_OK;
    int wlen;
    int ret;

    msghdr = (MsgHdr*)buf;
    msgcmd_arg = (CmdPram*)(buf + sizeof(MsgHdr));
    msgrsp_arg = (RspPram*)(buf + sizeof(MsgHdr));  // use the same rcv buffer to format responses
    opcmdpram = (OpPram*)msgcmd_arg;  // used to parse operational commands

    trace_msg("[Host_com] cmd_parser[%d]: cmd %d [addr, len]=[0x%x, %d]\n", len, msghdr->cmd, msgcmd_arg->addr, msgcmd_arg->len);

    pdata = (u8*)(buf + (sizeof(MsgHdr) + sizeof(CmdPram)));  // points to stuff beyond error and byte count
    pwdata = (u32*)(buf + (sizeof(MsgHdr) + sizeof(CmdPram)));
    wlen = (msgcmd_arg->len % 4) == 0 ? msgcmd_arg->len / 4 : msgcmd_arg->len / 4 + 1;  // last bytes is treated as word

    /* Command parsing */
    if (msghdr->cmd < 0x100)  // build-in commands
    {        
        switch (msghdr->cmd) {
        // Test Commands
        case CMD_MEM_WRITE:
#if MSG_MEM_BYTE_WRITE
        /* NPU memory(0x30000000) doest not support byte R/W */
            if ((msgcmd_arg->addr >= NPU_PA_BASE) && (msghdr->addr < 0x40000000)) {
#else
            if (1) {
#endif
                for (int i = 0; i < wlen; i++) {
                    outw(msgcmd_arg->addr + i * 4, (u32) *(pwdata + i));
                }
            }
            else {
                for (int i = 0; i < msgcmd_arg->len; i++) {
                    outb(msgcmd_arg->addr + i, *(pdata + i));
                }
            }
            // msg byte write size is unchanged
            msgrsp_arg->error = NO_ERROR;
            msg_pack(CMD_MEM_WRITE, NULL, 0);
            break;

        case CMD_MEM_READ:
#if MSG_MEM_BYTE_WRITE
        /* NPU memory(0x30000000) doest not support byte R/W */
            if ((msgcmd_arg->addr >= NPU_PA_BASE) && (msghdr->addr < 0x40000000)) {
#else
            if (1) {
#endif
                for (int i = 0; i < wlen; i++) {
                    *(pwdata + i) = inw(msgcmd_arg->addr + 4 * i);
                }
            }
            else {
                for (int i = 0; i < msgcmd_arg->len; i++) {
                    *(pdata + i) = inb(msgcmd_arg->addr + i);
                }
            }
            // msg byte read size is unchanged
            msgrsp_arg->error = NO_ERROR;
            msg_pack(CMD_MEM_READ, (u8 *)pdata, msgcmd_arg->len);
            break;

        case CMD_ACK_NACK:
        {
            dbg_msg("[Host_com] Received ACK/NACK\r\n");
            // Send results back to host
            if (msgcmd_arg->addr) {
                no_host_reply();
                break;  // if NACK, ignore the command and no reply
            }
            // We don't enforce Source ID check yet, since only DME is using this
            if (dme_conf.img_conf.image_format & IMAGE_FORMAT_RAW_OUTPUT){
                data_write((u8 *) p_raw_out_s, final_len);
            } else {
                data_write((u8 *) p_od_out_s, final_len);
            }
            break;
        }
        
        case CMD_TEST_ECHO:
            // msg bytes echo'ed is unchanged
            msgrsp_arg->error = NO_ERROR;
            msg_pack(CMD_TEST_ECHO, (u8 *)pdata, msgcmd_arg->len);
            break;

        case CMD_MEM_CLR:
            memset((void *)msgcmd_arg->addr, 0, msgcmd_arg->len);
            // msg clear size is unchanged
            msgrsp_arg->error = NO_ERROR;
            msg_pack(CMD_MEM_CLR, NULL, 0);
            break;

        case CMD_FILE_WRITE:
        {
            u32 bytes_read;
            bytes_read = host_mem_read(msgcmd_arg->addr, msgcmd_arg->len);
            if (bytes_read == msgcmd_arg->len)
                msgrsp_arg->error = NO_ERROR;
            else
                msgrsp_arg->error = FILE_ERROR;
            msgrsp_arg->bytes = bytes_read;
            msg_pack(CMD_FILE_WRITE, NULL, 0);
            break;
        }
        case CMD_FLASH_MEM_WRITE:
        {
           // This code is Winbond specific.  For other vendors, modification might be needed!!!!!!
            u32 offset_addr = msgcmd_arg->addr & 0x0FFF;  // find offset addr into sector
            u8* flash_sector_start = msg_tbuf + 0x40;  // keep away from potential data
            u8* flash_sector_check = msg_rbuf + 0x40;  // reuse the buffer
            u32 i;

            if ((msgcmd_arg->len + offset_addr) > 0x1000) {
                msgrsp_arg->error = 1;  // target addr out of range
                msgrsp_arg->bytes = 0;
            }
            else {  // peform FLASH data compare, merge and write
                // load flash sector into tx buf as temp storage
                pdrv_flash->ReadData(msgcmd_arg->addr & 0xFFFFF000, flash_sector_start, 0x1000);  // read in 4KB
                ret = flash_wait_ready(100);  // wait for the read operation to complete
                dbg_msg("[Host_com] Read Flash Sector Wait Time -> %d\r\n", ret);
                if (ret < 0) {
                    err_msg("[ERROR] host_com: Read Flash Sector failed, ret -> %d\r\n", ret);
                }

                ret = flash_sector_merge(flash_sector_start+offset_addr, msgcmd_arg->data, msgcmd_arg->len);
                info_msg("[Host_com] Flash Merge ret code is %d\r\n", ret);
                
                if (ret < 0) { // erase the sector first before programming the whole 4KB
                    pdrv_flash->EraseSector(msgcmd_arg->addr);
                    ret = flash_wait_ready(100);  // wait for the erase operation to complete
                    if (ret < 0) {
                        dbg_msg("[Host_com] Erase Flash Sector Wait Time -> %d\r\n", ret);
                    }
                    if (ret >= 0) {
                        for (i = 0; i < 0x1000; i += 0x100) {  // program each 256 byte page separately
                            pdrv_flash->ProgramData((msgcmd_arg->addr & 0xFFFFF000)+i, flash_sector_start+i, 0x100);
                            flash_wait_ready(100);
                        }
                        ret = -1;  // restore ret value for later stages
                    }
                }
                
                if (ret > 0) {  // program the new data only, one page at a time
                    u32 first_page, more_page, last_page;
                    first_page = 0x100 - (offset_addr & 0xFF);  // size of first page
                    more_page = msgcmd_arg->len - first_page;
                    last_page = more_page & 0xFF;
                    more_page = more_page & 0xF00;
                    if (first_page) { // do we have partial first page?
                        pdrv_flash->ProgramData(msgcmd_arg->addr, flash_sector_start + offset_addr, first_page);
                        flash_wait_ready(100);
                    }
                    for (i = first_page; i < first_page + more_page; i += 0x100) {  // program complete pages
                        pdrv_flash->ProgramData(msgcmd_arg->addr + i, flash_sector_start + i, 0x100);
                        ret = flash_wait_ready(100);
                        dbg_msg("[Host_com] Flash Sector Write Wait Time -> %d\r\n", ret);
                    }
                    if (last_page) {
                        i = first_page + more_page;
                        pdrv_flash->ProgramData(msgcmd_arg->addr + i, flash_sector_start + i, last_page);
                        flash_wait_ready(100);
                    }
                }

                // do post check
                if (ret != 0) {
                    pdrv_flash->ReadData(msgcmd_arg->addr & 0xFFFFF000, flash_sector_check, 0x1000);  // read the new sector
                    ret = flash_wait_ready(100);  // wait for the read operation to complete
                    if (memcmp(flash_sector_start, flash_sector_check, 0x1000)) {  // if not equal, failed verification
                        msgrsp_arg->error = 2;
                    }
                    else {
                        msgrsp_arg->error = NO_ERROR;
                    }
                }
                else {
                    msgrsp_arg->error = NO_ERROR;
                }
                msgrsp_arg->bytes = msgcmd_arg->len;
            }
            // generate response msg
            msg_pack(CMD_FLASH_MEM_WRITE, NULL, 0);
            break;
        }

        // Operational Commands
        case CMD_RESET:
        {
            u32 reset_mode, reset_flag = 0;  // reset flag: 1->sleep, 2->reset, 3->RTC

            reset_mode = opcmdpram->op_parm1;
#if USB_HOST
            if ((reset_mode < 2) || (reset_mode == 255)  || (reset_mode == 256)
#else
            if (((reset_mode < 5) && (reset_mode != 2)) || (reset_mode == 255) || (reset_mode == 256)
#endif
                || ((reset_mode >> 16) == 0x1000) || ((reset_mode >> 16) == 0x2000))
            {
                if ((reset_mode + opcmdpram->op_parm2) == 0xFFFFFFFF) {
                    if (reset_mode == 3) reset_flag = 1;    // go to sleep
                    if (reset_mode == 255) reset_flag = 2;  // reset
                    if (reset_mode == 256) reset_flag = 3;  // power down
                    if ((reset_mode >> 16) == 0x1000) {
                        log_set_level_scpu(reset_mode & 0x000F);  // set scpu debug output level
                        log_set_level_ncpu((reset_mode & 0x0F00)>> 8);  // set ncpu debug output level

                        // Disable OS tick for power but keep it for profile/debug
                        if ((reset_mode & 0x000F) >= LOG_PROFILE)
                            OS_Tick_Enable();
                        else
                            OS_Tick_Disable();
                    }
                    if ((reset_mode >> 16) == 0x2000) {
                        // switch active boot image, bit 0 -> scpu, bit 1 -> ncpu

                        int ret = ota_update_switch_active_partition(reset_mode & 0x3);
                        msgrsp_arg->error = ret;

                    }
                    msgrsp_arg->error = NO_ERROR;  // not doing anything else for now
                }
                else
                    msgrsp_arg->error = BAD_CONFIRMATION;
            }
            else
                msgrsp_arg->error = BAD_MODE;
            msgrsp_arg->bytes = 0;
            msg_pack(CMD_RESET, NULL, 0);
            if (reset_flag) {
                msg_wait();  // wait till reply message is done
                // ncpu fw needs update to support waking up from sleep
//                if (reset_flag == 1) power_manager_sleep();
                if (reset_flag == 2) power_mgr_sw_reset();
                if (reset_flag == 3) power_manager_shutdown();
            }
            break;
        }

        case CMD_SYSTEM_STATUS:
        {
            u16* status;
            msgrsp_arg->error = DEVICE_ID;   // use "error" parameter for device ID
            msgrsp_arg->bytes = FW_VERSION;  // and "len" parameter for firmware version
            status = (u16*)msgrsp_arg->data;
            *status++ = SYS_sts;
            *status = APP_id;
            msg_pack(CMD_SYSTEM_STATUS, msgrsp_arg->data, 4);
            break;
        }

        case CMD_UPDATE_FW:
        {
            u32 model_id;
            model_id = opcmdpram->op_parm1;

            switch(model_id)
            {
                case UPDATE_MODULE_SCPU:
                {
                    ret = ota_update_scpu();
                    msgrsp_arg->error = ret;
                    msgrsp_arg->bytes = UPDATE_MODULE_SCPU;
                    msg_pack(CMD_UPDATE_FW, NULL, 0);
                    if(ret == OTA_UPDATE_SUCCESS) {
                        delay_us(50*1000);
                        power_mgr_sw_reset();
                    }
                    break;
                }

                case UPDATE_MODULE_NCPU:
                {
                    ret = ota_update_ncpu();
                    msgrsp_arg->error = ret;
                    msgrsp_arg->bytes = UPDATE_MODULE_NCPU;
                    msg_pack(CMD_UPDATE_FW, NULL, 0);
                    if (ret == OTA_UPDATE_SUCCESS) {
                        delay_us(50*1000);
                        power_mgr_sw_reset();
                    }
                    break;
                }
                case UPDATE_MODULE_NONE:
                default:
                    break;
            }
            break;
        }

        case CMD_DOWNLOAD_MODEL:
        {
            u32 model_id, size;
            model_id = opcmdpram->op_parm1;
            size = opcmdpram->op_parm2;
            if(model_id != 0)
            {
                ret = ota_update_model(model_id, size);
                msgrsp_arg->error = ret;
                msgrsp_arg->bytes = model_id;
                msg_pack(CMD_DOWNLOAD_MODEL, NULL, 0);
                if(ret == OTA_UPDATE_SUCCESS)
                {
                    delay_us(50*1000);
                    power_mgr_sw_reset();
                }
            }
            break;
        }

        case CMD_QUERY_APPS:
        case CMD_SELECT_APP:
        case CMD_SET_MODE:
        case CMD_SET_EVENTS:
        case CMD_UPDATE:
        case CMD_ABORT:
        default:
            msgrsp_arg->error = RSP_UKNOWN_CMD;
            msgrsp_arg->bytes = 0;
            msg_pack(msghdr->cmd, NULL, 0);
            break;
        }
    }
    
    // Application Specific Commands
    else if (msghdr->cmd < 0x200)  // group 1 of applications (0x100-0x1FF)
    {
        switch(msghdr->cmd)
        {
            case CMD_SFID_LW3D_START:
            case CMD_SFID_START:
            {
                u32 command_type;
                if (msghdr->cmd == CMD_SFID_LW3D_START)
                    command_type = 1;  // 3D mode
                else
                    command_type = 2;  // 2D mode
                kdp_app_dme_mode(0);
                // check if we were in ADD USER mode previously
                if (SFID_user) {
                    // abort any previous session
                    ret = kdp_app_db_abort_reg();
                    if (ret != KDP_APP_OK) {
                        err_msg("[Host_com] Start Abort Old Session FAILED.\r\n");
                    }
                }
                info_msg("[Host_com] Start FD/FR/DB.\r\n");
                SFID_user = 0;  // switch to NORMAL FD/FR/DB MODE
                SFID_index = 0;

                u16 *pshort;
                if (msghdr->mSize > 8) { // contains image format data
                    pshort = (u16*)&msgcmd_arg->data;
                    rgb_img_config.image_col = *pshort++;
                    rgb_img_config.image_row = *pshort++;
                    pwdata = (u32*) pshort;
                    rgb_img_config.image_format = *pwdata++;
                    if (command_type == 1) {
                        pshort = (u16*)pwdata;
                        nir_img_config.image_col = *pshort++;
                        nir_img_config.image_row = *pshort++;
                        pwdata = (u32*) pshort;
                        nir_img_config.image_format = *pwdata;
                    }
                }
                if (command_type == 1)  // check image format first
                    ret = kdp_app_lw3d_query_img_transfer_setting(&SFID_buf_desc, &rgb_img_config, &nir_img_config);
                else
                    ret = kdp_app_query_img_transfer_setting(&SFID_buf_desc, &rgb_img_config);
                if (ret) {  // start command failed
                    msgrsp_arg->error = MSG_APP_NOT_CONFIG;
                    msgrsp_arg->bytes = 0;  // image size set to 0
                }
                else {  // image format acceptable, then store threshold and image format data
                    APP_id = command_type;
                    dbg_msg("[Host_com] START count = %d, base = %x, size = %d, alignment = %d\r\n",
                        SFID_buf_desc.buf_count, SFID_buf_desc.rgb_buf_base_addr, SFID_buf_desc.rgb_img_size, SFID_buf_desc.alignment_in_bit);
                    if (msgcmd_arg->addr)
                        RGB_threshold = *(float*)&msgcmd_arg->addr;  // save the user specified threshold value
                    else
                        RGB_threshold = FID_THRESHOLD;  // use default value
                    if (msgcmd_arg->len)
                        NIR_threshold = *(float*)&msgcmd_arg->len;  // save the user specified threshold value
                    else
                        NIR_threshold = FID_THRESHOLD;  // use default value
                    // start continuous inference & DB match
                    msgrsp_arg->error = NO_ERROR;
                    msgrsp_arg->bytes = SFID_buf_desc.rgb_img_size;
                }
                if (command_type == 1) {
                    pwdata = (u32*)&msgcmd_arg->data;  // we use rcv buffer to format responses
                    if (ret)
                        *pwdata = 0;
                    else
                        *pwdata = SFID_buf_desc.nir_img_size;
                    msg_pack(CMD_SFID_LW3D_START, (u8*)pwdata, 4);
                }
                else
                    msg_pack(CMD_SFID_START, NULL, 0);
                // wait for image stream
                break;
            }
            case CMD_SFID_NEW_USER:
            {
                if ((msgcmd_arg->addr == 0) || (APP_id == 0) || (APP_id > 2)) {
                    if (msgcmd_arg->addr == 0)
                        msgrsp_arg->error = MSG_SFID_USR_ZERO; // user id 0 is not allowed
                    else
                        msgrsp_arg->error = MSG_APP_NOT_CONFIG;  // no Start SFID previously
                    msgrsp_arg->bytes = 0;
                    msg_pack(CMD_SFID_NEW_USER, NULL, 0);
                    break;
                }

                if (SFID_user == 0) {  // switch to ADD USER MODE
                    SFID_user = msgcmd_arg->addr;
                    SFID_index = msgcmd_arg->len;
                }
                else {
                    if (SFID_user == msgcmd_arg->addr)
                        SFID_index = msgcmd_arg->len;
                    else {
                        if (msgcmd_arg->len == 1) {
                            //  abort any previous NEW USER session first
                            ret = kdp_app_db_abort_reg();
                            if (ret == KDP_APP_OK) {
                                SFID_user = msgcmd_arg->addr;
                                SFID_index = 1;
                            }
                            // else, this condition shouldn't happen
                            else {
                                err_msg("[Host_com] Add user Abort Old Session %d Failed -> %d\r\n",
                                        msgcmd_arg->addr, ret);
                            // don't know how to handle this, leave to image cmd
                                SFID_user = msgcmd_arg->addr;  // force the case anyway
                            }
                        }
                        else {  // we don't send error code out anymore
                            SFID_user = msgcmd_arg->addr;  // let image cmd deal with the problem
                            SFID_index = msgcmd_arg->len;
                        }
                    }
                }
                info_msg("[Host_com] Add New User %d Indx -> %d\r\n", SFID_user, SFID_index);
                msgrsp_arg->error = NO_ERROR;
                msgrsp_arg->bytes = 0;
                msg_pack(CMD_SFID_NEW_USER, NULL, -1);  // use zero payload response
                // wait for image stream
                break;
            }
            case CMD_SEND_IMAGE:
            {
                u32 out_flag, out_size, lv_flag = 0;  // lv_flag (byte) bit 7-> result_valid, bit 0-> result
                u16 *pfdfr_out;

                out_flag = (u16) msgcmd_arg->len;  // save flags
                out_flag &= 0x0F;  // ---- ---Z vrmd VRMD -> final LV, rsv.fr.lm.fd, rgbLV.FR.LM.FD
                pfdfr_out = (u16*)msgrsp_arg->data;  // point to output data field
                *pfdfr_out++ = 0;  // initialize the response msg data fields
                *pfdfr_out++ = 0;  // now pointing to the optional results field

                if (APP_id != 2) { // check if we have SFID structures set up?
                    msgrsp_arg->error = MSG_APP_NOT_CONFIG;  // image processing not set up yet
                    msgrsp_arg->bytes = CMD_SEND_IMAGE;  // add source id
                    msg_pack(CMD_ACK_NACK, NULL, 0);
                    break;  // early abort
                }

                get_lw3d_io_memory();
                p_fdr_out_s = (kdp_app_fid_inout_t*)p_lw3d_out_s;  // lw3d I/O structure is compatible
                p_fd_out_s = (kdp_app_face_det_res_t*)p_rgb_fd_out_s;  // lw3d uses the same structures
                p_lm_out_s = (kdp_app_face_lm_res_t*)p_rgb_lm_out_s;
                p_fr_out_s = (kdp_app_face_fmap_t*)p_rgb_fr_out_s;

                if(SFID_index == 0){
                    p_fdr_out_s->op = KDP_APP_FID_OP_INFERENCE;
                    p_fdr_out_s->face_id = 0; //inference mode
                }
                else {
                    p_fdr_out_s->op = KDP_APP_FID_OP_REGISTER;
                    p_fdr_out_s->face_id = SFID_index; //reg mode, fm index
                }
                p_fdr_out_s->user_id = SFID_user;
                p_fdr_out_s->thresh_fid = RGB_threshold; //threshold value for FID db compare

                // get image
                ret = host_mem_read(SFID_buf_desc.rgb_buf_base_addr, msgcmd_arg->addr);  // specify how many bytes
                if (ret == msgcmd_arg->addr) { // got image
                    // input range check first
                    if ((SFID_user != 0) && (SFID_index != 0)) {
                        if ((SFID_user > FID_MAX_USER) || (SFID_index > 5)) {
                            // bad user id or image index
                            msgrsp_arg->error = MSG_APP_BAD_INDEX;
                            msgrsp_arg->bytes = 0;
                            msg_pack(CMD_SEND_IMAGE, msgrsp_arg->data, 4);
                            break;
                        }
                    }
                    // now send command to AppMgr
                    *(u32*)p_fd_out_s = 0;  // initialize the results, FD and FR only
                    *(u32*)p_fr_out_s = 0;  // fdr() does not perform LV right now
                    ret = kdp_app_fdr(p_fdr_out_s, p_fd_out_s, p_lm_out_s, p_fr_out_s);

                    u16* mode;
                    mode = (u16*) &msgrsp_arg->bytes;
                    if (ret == KDP_APP_OK) {
                        // got user_id or feature map stored
                        msgrsp_arg->error = NO_ERROR;
                        // we have u16 mode and u16 user_id/fm_indx in a u32 field (Little Endian)
                        if (SFID_user) {
                            *mode = 1;
                            mode++;
                            *mode = SFID_index;
                        }
                        else {
                            *mode = 0;
                            mode++;
                            *mode = p_fdr_out_s->user_id;
                        }
                    }
                    else { // error conditions
                        if (ret == KDP_APP_DB_NO_MATCH)
                            msgrsp_arg->error = MSG_APP_DB_NO_MATCH;
                        else if (ret == KDP_APP_DB_ALREADY_SAVED)
                            msgrsp_arg->error = MSG_APP_UID_EXIST;
                        else if (ret == KDP_APP_DB_REG_FIRST)
                            msgrsp_arg->error = MSG_APP_IDX_FIRST;
                        else if (ret == KDP_APP_DB_USER_NOT_REG)
                            msgrsp_arg->error = MSG_APP_IDX_MISSING;
                        else if (ret == KDP_APP_DB_NO_SPACE)
                            msgrsp_arg->error = MSG_APP_DB_FAIL;
                        else if (ret == KDP_APP_FDR_WRONG_USAGE)
                            msgrsp_arg->error = MSG_APP_BAD_INDEX;
                        else if (ret == KDP_APP_DB_DEL_NOT_VALID)
                            msgrsp_arg->error = MSG_APP_DB_NO_USER;
                        else
                            msgrsp_arg->error = MSG_APP_BAD_IMAGE;  // default case, fd/fr failure
                    }
                    out_size = fdlmfrlv_outputs(&out_flag, pfdfr_out, lv_flag);
                    pfdfr_out = (u16*)&msgrsp_arg->data;
                    *pfdfr_out = out_flag;  // update output flag
                    msg_pack(CMD_SEND_IMAGE, msgrsp_arg->data, 4 + out_size);
                }
                else { // no image
                    msgrsp_arg->error = MSG_APP_BAD_IMAGE;
                    msgrsp_arg->bytes = 0;
                    if (ret < 0)
                        no_host_reply();  // if tx/rx timed out, abort the msg reply
                    else
                        msg_pack(CMD_SEND_IMAGE, msgrsp_arg->data, 4);
                }
                break;
            }

            case CMD_SEND_LW3D:
            {
                u32 out_flag, out_size, lv_flag = 0;  // lv_flag (byte) bit 7-> result_valid, bit 0-> result
                u16 *pfdfr_out;

                out_flag = *(u16*) msgcmd_arg->data;  // save flags
                out_flag &= 0x177;  // ---- ---Z vrmd VRMD -> final LV, rsv.fr.lm.fd, rgbLV.FR.LM.FD
                pfdfr_out = (u16*)msgrsp_arg->data;  // point to output data field
                *pfdfr_out++ = 0;  // initialize the response msg data fields
                *pfdfr_out++ = 0;  // now pointing to the optional results field

                if (APP_id != 1) { // check if we have LW3D structures set up?
                    msgrsp_arg->error = MSG_APP_NOT_CONFIG;  // image processing not set up yet
                    msgrsp_arg->bytes = CMD_SEND_LW3D;  // add source id
                    msg_pack(CMD_ACK_NACK, NULL, 0);
                    break;  // early abort
                }
                get_lw3d_io_memory();
                if(SFID_index == 0){
                    p_lw3d_out_s->op = KDP_APP_LW3D_OP_INFERENCE;
                    p_lw3d_out_s->face_id = 0; //inference mode
                }
                else {
                    p_lw3d_out_s->op = KDP_APP_LW3D_OP_REGISTER;
                    p_lw3d_out_s->face_id = SFID_index; //reg mode, fm index
                }
                p_lw3d_out_s->user_id = SFID_user;
                p_lw3d_out_s->rgb_thresh_fid = RGB_threshold; //threshold value for FID db compare
                p_lw3d_out_s->nir_thresh_fid = NIR_threshold; //threshold value for FID db compare

                // get image
                dbg_msg("[Host_com] Wait for lw3d images\r\n");

                ret = host_mem_read(SFID_buf_desc.rgb_buf_base_addr, msgcmd_arg->addr);

                if (ret == SFID_buf_desc.rgb_img_size) {
                    ret = host_mem_read(SFID_buf_desc.nir_buf_base_addr, msgcmd_arg->len);
                }
                else
                {
                    // send NACK to abort nir image downloading
                    msgrsp_arg->error = MSG_APP_BAD_IMAGE;  // image not expected size
                    msgrsp_arg->bytes = CMD_SEND_LW3D;  // add source id
                    msg_pack(CMD_ACK_NACK, NULL, 0);
                    break;
                }
                if (ret == msgcmd_arg->len) { // got nir image as well
                    // input range check first
                    if ((SFID_user != 0) && (SFID_index != 0)) {
                        if ((SFID_user > FID_MAX_USER) || (SFID_index > 5)) {
                            // bad user id or image index
                            msgrsp_arg->error = MSG_APP_BAD_INDEX;
                            msgrsp_arg->bytes = 0;
                            msg_pack(CMD_SEND_LW3D, msgrsp_arg->data, 4);
                            break;
                        }
                    }
                    *(u32*)p_rgb_fd_out_s = 0;  // initialize the results, FD and FR only
                    *(u32*)p_rgb_fr_out_s = 0;
                    *(u32*)p_nir_fd_out_s = 0;
                    *(u32*)p_nir_fr_out_s = 0;
                    ret = kdp_app_lw3d(p_lw3d_out_s,
                            p_rgb_fd_out_s, p_rgb_lm_out_s, p_rgb_fr_out_s,
                            p_nir_fd_out_s, p_nir_lm_out_s, p_nir_fr_out_s);
                    u16* mode;
                    mode = (u16*) &msgrsp_arg->bytes;  // point to the mode field in the rsp msg
                    if (ret) trace_msg("[host_com] lw3d ret = 0x%x\r\n", ret);
                    if (ret == KDP_APP_OK) {
                        // got user_id or new user feature map stored
                        msgrsp_arg->error = NO_ERROR;
                        lv_flag = 0x81;  // LV is valid and good
                        // we have u16 mode and u16 user_id/fm_indx in a u32 field (Little Endian)
                        if (SFID_user) {
                            *mode = 1;
                            mode++;
                            *mode = SFID_index;
                        }
                        else {
                            *mode = 0;
                            mode++;
                            *mode = p_lw3d_out_s->user_id;
                        }
                    }
                    else { // error conditions
                        msgrsp_arg->bytes = 0;  // the mode fields has no meaning
                        if ((ret < KDP_APP_DB_DEL_FAIL) && (ret >= KDP_APP_DB_NO_SPACE))
                            lv_flag = 0x81;  // DB error means LV passed
                        if (ret == KDP_APP_DB_NO_MATCH)
                            msgrsp_arg->error = MSG_APP_DB_NO_MATCH;
                        else if (ret ==  KDP_APP_LW3D_NO_ALIVE) {
                            msgrsp_arg->error = MSG_APP_BAD_IMAGE;
                            lv_flag = 0x80;
                        }
                        else if (ret == KDP_APP_DB_ALREADY_SAVED)
                            msgrsp_arg->error = MSG_APP_UID_EXIST;
                        else if (ret == KDP_APP_DB_REG_FIRST)
                            msgrsp_arg->error = MSG_APP_IDX_FIRST;
                        else if (ret == KDP_APP_DB_USER_NOT_REG)
                            msgrsp_arg->error = MSG_APP_IDX_MISSING;
                        else if (ret == KDP_APP_DB_NO_SPACE)
                            msgrsp_arg->error = MSG_APP_DB_FAIL;
                        else if (ret == KDP_APP_FDR_WRONG_USAGE)
                            msgrsp_arg->error = MSG_APP_BAD_INDEX;
                        else if (ret == KDP_APP_DB_DEL_NOT_VALID)
                            msgrsp_arg->error = MSG_APP_DB_NO_USER;
                        else
                            msgrsp_arg->error = MSG_APP_BAD_IMAGE;  // default case, fd/fr failure
                    }
                    out_size = fdlmfrlv_outputs(&out_flag, pfdfr_out, lv_flag);
                    pfdfr_out = (u16*)&msgrsp_arg->data;
                    *pfdfr_out = out_flag;  // update output flag
                    msg_pack(CMD_SEND_LW3D, msgrsp_arg->data, 4 + out_size);
                }
                else {
                    msgrsp_arg->error = MSG_APP_BAD_IMAGE;
                    msgrsp_arg->bytes = 0;
                    if (ret < 0)
                        no_host_reply();  // if tx/rx timed out, abort the msg reply
                    else
                        msg_pack(CMD_SEND_LW3D, msgrsp_arg->data, 4);
                }
                break;
            }

            case CMD_SFID_ADD_DB:
            {
                kdp_app_db_user_data_t app_data;
                app_data.user_id_in = SFID_user;
           
                if ((APP_id == 0) || (APP_id > 2)) {
                    msgrsp_arg->error = MSG_APP_NOT_CONFIG;  // not in SFID mode
                    msgrsp_arg->bytes = 0;
                    msg_pack(msghdr->cmd, NULL, 0);
                    break;
                }
                ret = kdp_app_db_add(&app_data);    
                if (ret == KDP_APP_OK) {  // success
                    msgrsp_arg->error = NO_ERROR;
                }
                else { // error conditions
                    switch (ret)
                    {
                        case KDP_APP_DB_ALREADY_SAVED:
                            msgrsp_arg->error = MSG_APP_UID_EXIST;
                            break;
                        case KDP_APP_DB_REG_FIRST:
                            msgrsp_arg->error = MSG_APP_IDX_FIRST;
                            break;
                        case KDP_APP_DB_USER_NOT_REG:
                            msgrsp_arg->error = MSG_APP_IDX_MISSING;
                            break;
                        default:
                            // all other cases, unknown failure
                            msgrsp_arg->error = MSG_APP_DB_FAIL;
                            break;
                    }
                }
                msgrsp_arg->bytes = 0;
                msg_pack(CMD_SFID_ADD_DB, NULL, 0);
                break;
            }
            case CMD_SFID_DELETE_DB:
            {
                kdp_app_db_user_data_t app_data;
                app_data.user_id_in = msgcmd_arg->addr;
                if ((APP_id == 0) || (APP_id > 2)) {
                    msgrsp_arg->error = MSG_APP_NOT_CONFIG;  // not in SFID mode
                    msgrsp_arg->bytes = 0;
                    msg_pack(msghdr->cmd, NULL, 0);
                    break;
                }
                if(msgcmd_arg->addr) 
                    app_data.del_all = FALSE;
                else
                    app_data.del_all = TRUE;

                info_msg("[Host_com] DELETE USER %d\n", msgcmd_arg->addr);

                ret = kdp_app_db_delete(&app_data);    

                if (ret == KDP_APP_OK) // success
                    msgrsp_arg->error = NO_ERROR;
                else {  //user id not found
                    if (ret == KDP_APP_DB_DEL_FAIL)
                        msgrsp_arg->error = MSG_APP_DB_NO_USER;
                    else
                        msgrsp_arg->error = MSG_APP_DB_FAIL;
                }
                msgrsp_arg->bytes = 0;
                msg_pack(CMD_SFID_DELETE_DB, NULL, 0);
                break;
            }
            case CMD_SFID_EDIT_DB:
            {
                u32 *db_ddr_addr, *db_flash_addr, db_size, APP_id_OK;
                db_size = kdp_app_db_query(&db_ddr_addr, &db_flash_addr);  // get DB size & location
                if ((APP_id == 0) || (APP_id > 2)) APP_id_OK = 0;
                else APP_id_OK = 1;
                if(msgcmd_arg->addr == 1) { // import DB
                    if (APP_id_OK == 0) {
                        msgrsp_arg->error = MSG_APP_NOT_CONFIG;  // image processing not set up yet
                        msgrsp_arg->bytes = CMD_SFID_EDIT_DB;  // add source id
                        msg_pack(CMD_ACK_NACK, NULL, 0);
                        break;  // abort
                    }
                    if (db_size == *(u32*)msgcmd_arg->data) {
                        ret = host_mem_read((u32)db_ddr_addr, db_size);  // get DB binary
                        if (ret < 0) {
                            no_host_reply();  // if tx/rx timed out, abort the msg processing
                            kdp_app_db_init((u32)db_flash_addr); // reload database from flash, addr needed if db was flushed
                            break;
                        }
                        msgrsp_arg->bytes = 0;
                        if (ret == db_size) {
//                            kdp_timer_init(0);
                            kdp_memxfer_ddr_to_flash((u32)db_flash_addr, (u32) db_ddr_addr, db_size);  // update db in flash
//                            dbg_msg("[Host_com] DB Flash time = %d\r\n", kdp_timer_read(0)/100);
                            msgrsp_arg->error = NO_ERROR;
                        }
                        else {
                            msgrsp_arg->error = MSG_SFID_DATA_FAIL;
                            kdp_app_db_init((u32)db_flash_addr);  // need to reload database
                        }
                    }
                    else {
                        msgrsp_arg->error = MSG_DB_BAD_PARAM;
                        msgrsp_arg->bytes = CMD_SFID_EDIT_DB;  // add source id
                        msg_pack(CMD_ACK_NACK, NULL, 0);
                        break;  // abort
                    }
                }
                else if (msgcmd_arg->addr == 2) { // export DB
                    if (APP_id_OK == 1) {
                        msgrsp_arg->error = 0;
                        msgrsp_arg->bytes = db_size;  // set size of DB for export
                        msg_pack(CMD_SFID_EDIT_DB, NULL, 0);  // send reply now
                        usb_com_write ((u8*)db_ddr_addr, db_size, FLAG_HOST_COMM_USB_DONE, 2);  // use alternate data mode
                        break;
                    }
                    else {
                        msgrsp_arg->error = MSG_APP_NOT_CONFIG;  // image processing not set up yet
                        msgrsp_arg->bytes = 0;
                    }
               }
                else {  // unknown DB action
                    msgrsp_arg->error = MSG_DB_BAD_PARAM;
                    msgrsp_arg->bytes = 0;
                }
                msg_pack(CMD_SFID_EDIT_DB, NULL, 0);
                break;
            }
            case CMD_DME_START:
            {
                // NOTE: Need to determine if DME structure needs to be set up first before processing images
                info_msg("[Host_com] Start DME mode.\r\n");
                //Get size of inference mode (4 bytes)
                uint32_t model_size = msgcmd_arg->addr;

                //Load fw info
                memcpy((uint32_t*)KDP_DDR_MODEL_INFO_TEMP, &(msgcmd_arg->len), msghdr->mSize - 4);

                uint32_t cmd_addr_offset = 9*sizeof(uint32_t); //cmd addr offset in fw_info
                uint32_t cmd_start_addr = *(uint32_t*)(KDP_DDR_MODEL_INFO_TEMP + cmd_addr_offset);

                if (model_size <= 0) {
                    msgrsp_arg->error = MSG_DME_DATA_FAIL;
                    msgrsp_arg->bytes = CMD_DME_START;
                    msg_pack(CMD_ACK_NACK, NULL, 0);
                    break;  // early abort
                }

                //Load setup, weights and commands
                ret = host_mem_read(cmd_start_addr, model_size);  // specify how many bytes
                if (ret == model_size) {
                    APP_id = 3;  // enter DME mode
                    msgrsp_arg->error = NO_ERROR;
                    msgrsp_arg->bytes = ret;
                    msg_pack(CMD_DME_START, NULL, 0);
                } else {
                    msgrsp_arg->error = MSG_DME_DATA_FAIL;
                    msgrsp_arg->bytes = 0;
                    msg_pack(CMD_DME_START, NULL, 0);
                }
								
                kdp_app_dme_mode(1/*is_dme*/);
                break;
            }
            case CMD_DME_CONFIG:
            {
                info_msg("[Host_com] DME Config.\r\n");
                if (APP_id == 3) {
                    memcpy(&dme_conf, &(msgcmd_arg->addr), sizeof(struct kdp_dme_conf_s));
                    msgrsp_arg->error = NO_ERROR;
                }
                else
                    msgrsp_arg->error = MSG_APP_NOT_CONFIG;  // not in DME mode
                msgrsp_arg->bytes = 0;
                msg_pack(CMD_DME_CONFIG, NULL, 0);  // use zero payload response
                break;
            }
            case CMD_DME_SEND_IMAGE:
            {
                dbg_msg("Received cmd: CMD_DME_SEND_IMAGE\n");

                if ((0 == dme_conf.output_num) || (APP_id != 3)) { // check if we have DME structures set up?
                    if (APP_id == 3)
                        msgrsp_arg->error = MSG_DME_NO_CONFIG;  // image processing not set up yet
                    else
                        msgrsp_arg->error = MSG_APP_NOT_CONFIG;  // not in DME mode
                    msgrsp_arg->bytes = CMD_DME_SEND_IMAGE;  // add source id
                    msg_pack(CMD_ACK_NACK, NULL, 0);
                    break;  // early abort
                }

                if (!(dme_conf.img_conf.image_format & IMAGE_FORMAT_BYPASS_PRE)) {
                    // write RGB565 image into DDR
                    ret = host_mem_read(KDP_DDR_BASE_IMAGE_BUF, msgcmd_arg->addr);  // specify how many bytes
                }  else {
                    // write RGBA image into DDR
                    ret = host_mem_read(DDR_MEM_BASE, msgcmd_arg->addr);  // specify how many bytes
                }

                uint32_t raw_len = 0; //For allocating memory in advance
                uint32_t output_num = dme_conf.output_num;
                if (dme_conf.img_conf.image_format & IMAGE_FORMAT_RAW_OUTPUT) {
                    //Get output length in fw_info
                    uint32_t output_len = *(uint32_t*)(KDP_DDR_MODEL_INFO_TEMP + 6*sizeof(uint32_t));
                    raw_len = (1 + 3*output_num + output_len)*sizeof(uint32_t);
                    if (p_raw_out_s == 0) {
                        p_raw_out_s = kdp_ddr_reserve(raw_len);
                    }
                    ret = kdp_app_dme((kdp_app_dme_res_t *)p_raw_out_s, dme_conf);
                } else {
                    if(p_od_out_s == 0) {
                        p_od_out_s = (kdp_app_dme_res_t *) kdp_ddr_reserve(
                                sizeof(kdp_app_dme_res_t));
                    }
                    raw_len = sizeof(kdp_app_dme_res_t);
                    ret = kdp_app_dme(p_od_out_s, dme_conf);
                }

                //For sending results to HOST
                if (dme_conf.img_conf.image_format & IMAGE_FORMAT_RAW_OUTPUT) {
                    int output_len = 0;
                    for (int i = 0; i < output_num; i++){
                        int h = *(uint32_t *)(p_raw_out_s + (3*i+1)*sizeof(uint32_t));
                        int c = *(uint32_t *)(p_raw_out_s + (3*i+2)*sizeof(uint32_t));
                        int w = *(uint32_t *)(p_raw_out_s + (3*i+3)*sizeof(uint32_t));
                        output_len = output_len + h * c * w;
                    }

                    //output_num+HCW+H2C2W2+DATA1+DATA2
                    final_len = (1 + 3*output_num + output_len)*sizeof(uint32_t);
                } else {
                    int det_num = p_od_out_s->box_count;
                    final_len = 2*sizeof(uint32_t) + det_num * sizeof(kdp_app_bounding_box_t);
                }
                msgrsp_arg->error = NO_ERROR;
                msgrsp_arg->bytes = final_len;

                //Msg for the length of raw results
                msg_pack(CMD_DME_SEND_IMAGE, NULL, 0);
                break;
            }
            default:
                msgrsp_arg->error = RSP_UKNOWN_CMD;
                msgrsp_arg->bytes = 0;
                msg_pack(msghdr->cmd, NULL, 0);
                break;
        }
    }
    else if (msghdr->cmd >> 8 == 0x20)
    {
#ifdef DBG_ALONE
        switch (msghdr->cmd)
        {
            case CMD_DBG_USB_MEM_WR:
            {
                uint32_t mem_addr = (uint32_t)msgcmd_arg->addr;
                uint32_t buf_len = (uint32_t)msgcmd_arg->len;
                dbg_msg("cmd_parser: USB mem write to addr:0x%x, size:%d\n", mem_addr, buf_len);
                // get data from USB
                ret = usb_mem_read(mem_addr, buf_len);
                if(ret == buf_len) {
                    dbg_msg("cmd_parser: read USB data successfully\n");
                    msgrsp_arg->error = NO_ERROR;
                    msgrsp_arg->bytes = 0;
                    msg_pack(CMD_ACK, NULL, 0);
                }
                else {
                    dbg_msg("Error! cmd_parser: read USB data failed\n");
                    msgrsp_arg->error = USB_ERROR;
                    msgrsp_arg->bytes = 0;
                    msg_pack(CMD_DBG_USB_MEM_WR, NULL, 0);
                }
                break;
            }

            case CMD_DBG_SET_MODEL_TYPE:
            {
                dbg_model_type = (uint32_t)msgcmd_arg->addr;
                msgrsp_arg->error = NO_ERROR;
                msgrsp_arg->bytes = 0;
                msg_pack(CMD_ACK, NULL, 0);

                dbg_msg("cmd_parser: set model_type=%d\n", dbg_model_type);
                break;
            }
            default:
            break;
        }
#endif
    }
    else
    {
        msgrsp_arg->error = RSP_UKNOWN_CMD;
        msgrsp_arg->bytes = 0;
        msg_pack(msghdr->cmd, NULL, 0);
    }
    return rstatus;
}

void get_lw3d_io_memory(void)
{
    if(p_lw3d_out_s == 0)
        p_lw3d_out_s = (kdp_app_lw3d_inout_t*)kdp_ddr_reserve(sizeof(kdp_app_lw3d_inout_t));
    if(p_rgb_fd_out_s == 0)
        p_rgb_fd_out_s = (kdp_app_lw3d_face_det_res_t*)kdp_ddr_reserve(sizeof(kdp_app_lw3d_face_det_res_t));
    if(p_rgb_lm_out_s == 0)
        p_rgb_lm_out_s = (kdp_app_lw3d_face_lm_res_t*)kdp_ddr_reserve(sizeof(kdp_app_lw3d_face_lm_res_t));
    if(p_rgb_fr_out_s == 0)
        p_rgb_fr_out_s = (kdp_app_lw3d_face_fmap_t*)kdp_ddr_reserve(sizeof(kdp_app_lw3d_face_fmap_t));
    if(p_nir_fd_out_s == 0)
        p_nir_fd_out_s = (kdp_app_lw3d_face_det_res_t*)kdp_ddr_reserve(sizeof(kdp_app_lw3d_face_det_res_t));
    if(p_nir_lm_out_s == 0)
        p_nir_lm_out_s = (kdp_app_lw3d_face_lm_res_t*)kdp_ddr_reserve(sizeof(kdp_app_lw3d_face_lm_res_t));
    if(p_nir_fr_out_s == 0)
        p_nir_fr_out_s = (kdp_app_lw3d_face_fmap_t*)kdp_ddr_reserve(sizeof(kdp_app_lw3d_face_fmap_t));
}

/* This function checks the output data and if valid, will append them to the message
** input:  output flags to check, pointer to start of output data in message buffer
** output: number of bytes added to the message buffer
*/
u32 fdlmfrlv_outputs(u32 *check_flag, u16 *pfdfr_out, u32 lv_flag)
{
    u32 out_size = 0;
    u32 out_flag = *check_flag;
    if (*(u32*)p_rgb_fd_out_s) { // FD valid?
        if (out_flag & 1) {  // do we export FD?
            output_results(pfdfr_out, &p_rgb_fd_out_s->xywh[0], 4); // u16 output, 4 members
            out_size += 8;
            pfdfr_out += 4;
        }
    }
    else
        out_flag &= 0x0fff0;  // no FD, then no LM, FR, LV either
    if (out_flag & 2) {  // export LM?
        if (*(u32*)p_rgb_fd_out_s) { // if FD is good, assume LM is also good
            output_results(pfdfr_out, &p_rgb_lm_out_s->marks[0].x, 10); // 10 members
            out_size += 20;
            pfdfr_out += 10;
        }
    }
    if (out_flag & 4) {  // export FR?
        if (*(u32*)p_rgb_fr_out_s) { // FR valid?
            memcpy((u8*)pfdfr_out, (u8*)p_rgb_fr_out_s->feature_map, 2048);
            out_size += 2048;
            pfdfr_out += 1024;
        }
        else
            out_flag &= 0x0fffb;
    }
    if (out_flag & 8) {  // export LV?
        if (lv_flag & 0x80) { // LV valid?
            *(u32*)pfdfr_out = lv_flag & 1;  // store the results
            out_size += 4;
            pfdfr_out += 2;
        }
        else
            out_flag &= 0x0fff7;
    }
    if (*(u32*)p_nir_fd_out_s) { // nir FD valid?
       if (out_flag & 0x10) {  // do we export nir FD?
            output_results(pfdfr_out, &p_nir_fd_out_s->xywh[0], 4); // u16 output, 4 members
            out_size += 8;
            pfdfr_out += 4;
        }
    }
    else
        out_flag &= 0x0fe0f;  // no FD, then no LM, FR, LV, Final LV either
    if (out_flag & 0x20) {  // export nir LM?
        if (*(u32*)p_nir_fd_out_s) { // if FD is good, assume LM is also good
            output_results(pfdfr_out, &p_nir_lm_out_s->marks[0].x, 10); // 10 members
            out_size += 20;
            pfdfr_out += 10;
        }
    }
    if (out_flag & 0x40) {  // export FR?
        if (*(u32*)p_nir_fr_out_s) { // FR valid?
            memcpy((u8*)pfdfr_out, (u8*)p_nir_fr_out_s->feature_map, 2048);
            out_size += 2048;
            pfdfr_out += 1024;
        }
        else
            out_flag &= 0x0ffbf;
    }
     if (out_flag & 0x100) {  // export final LV?
        if (lv_flag & 0x80) { // NIR LV result valid?
            *(u32*)pfdfr_out = lv_flag & 1;  // store the results
            out_size += 4;
            pfdfr_out += 2;
        }
        else
            out_flag &= 0x0feff;
    }
    *check_flag = out_flag;
    return out_size;
}

void output_results(u16 *output_buf, void *out_result, u32 count)
{
    u32 *result = out_result;
    while (count) {
        *output_buf++ = (u16) *result++;
        count--;
    }
}

// The ack_packet needs to be static, place it on stack doesn't seem to work
// TODO: encode source id and add crc for uart data mode
u8 ack_packet[20] = {0x35, 0x8A, 12, 0, 4, 0, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

u32 host_mem_read(u32 address, u32 size)
{
#if USB_HOST
    usb_com_write (ack_packet, 16, FLAG_HOST_COMM_USB_DONE, 1);  // use data mode
    return(usb_com_read((u8*) address, size, FLAG_HOST_COMM_USB_DONE, 1));  // use data mode
#else
// NOTE: For uart, we also use msg_rx buffer for data receiving, thus destroying the command data that might
//       be needed later.  So we must preserve it (save it in the msg_tx buffer) for restoration at the end.
    u32 block, residual, offset;
    block = size >> 12;
    residual = size & 0x0FFF;

    save_command();
    offset = 0;
    while (block) {
        kdp_uart_com_write(ack_packet, 16);
        kdp_uart_com_read((u8*) address + offset, 0x1000); // FLAG_HOST_COMM_UART_DONE);
        offset += 0x1000;
        block--;
    }
    if (residual) {
        kdp_uart_com_write(ack_packet, 16);
        kdp_uart_com_read((u8*) address + offset, residual); //, FLAG_HOST_COMM_UART_DONE);
    }
    restore_command();
    return(size);
#endif
}

int flash_wait_ready(int timeout_ms)
{
    ARM_FLASH_STATUS flash_status;
    int i;

    for (i = 0; i < timeout_ms; i++) {
        flash_status = pdrv_flash->GetStatus();
        if (flash_status.busy == 0) break;
        delay_us(1*1000);
    }
    if (i == timeout_ms) i = -1;  // we have timed out
    return i;
}

/* Flash sector merge function operates on a single sector at a time
*  input - starting address of original flash data
*        - starting address of new data to be written
*        - size of new data in bytes
*  output - 0 -> data is identical, no write needed
*         - positive -> data is merged, no need to erase first
*         - negative -> data is merged, needs erase first
*/
int flash_sector_merge(u8* sector_offset, u8* data_start, u32 data_size)
{
    // we don't do data alignment adjustment yet, assume all inputs are 32-bit aligned
    u16 need_erase = 0, need_update = 0;
    u32 i, *pflash_word, *pnew_word;
    
    pflash_word = (u32*)sector_offset;
    pnew_word = (u32*)data_start;
    for (i = data_size >> 2; i > 0; i--)
    {
        if (*pflash_word != *pnew_word) need_update = 1;
        if ((*pflash_word ^ *pnew_word) & *pnew_word) {
            need_erase = 1;
        }
        else {
            pflash_word++;
            pnew_word++;
        }
        if (need_erase && need_update) break;  // early abort
    }
    
    if (need_update) {
        memcpy(sector_offset, data_start, data_size);
        if (need_erase) {
            return -1;
        }
        return 1;
    }
    return 0;  // data are identical
}

static void device_sleep(void)
{
    SYS_sts = 1;
//    trace_msg("[Host_com] Going to sleep....\r\n");
}

static void device_wakeup(void)
{
    SYS_sts = 0;
//    trace_msg("[Host_com] Waking up....\r\n");
}

struct pm_s pm_host_com_device =
{
    .nap = 0,
    .wakeup_nap = 0,
    .deep_nap = 0,
    .wakeup_deep_nap = 0,
    .sleep = (pm_call) device_sleep,
    .wakeup_sleep = (pm_call) device_wakeup,
    .deep_sleep = (pm_call) device_sleep,
    .wakeup_deep_sleep = (pm_call) device_wakeup,
};

static void host_comm_thread(void *argument)
{
    int32_t flags;
    int len;

    //init kdp_app
    kdp_app_init();

    ota_handle_first_time_boot();
    err_msg("SCPU working ver 2 \n");
	
#if USB_HOST
    usb_host_comm_init(FLAG_HOST_COMM_USB_DONE);
#else
    uart_host_comm_init(FLAG_HOST_COMM_UART_DONE);
#endif
    // register PM_DEVICE_HOST_COM for sleep & wakeup
    power_manager_register(PM_DEVICE_HOST_COM, &pm_host_com_device);
    for (;;) {
        flags = osThreadFlagsWait(FLAG_HOST_COMM_USB_DONE | FLAG_HOST_COMM_UART_DONE | FLAG_HOST_COMM_TIMER, osFlagsWaitAny, osWaitForever);
        osThreadFlagsClear(flags);

        if (flags & FLAG_HOST_COMM_TIMER) {
            if (len > 0) {
                fLib_printf("Get uart_len = %d\n", len);
                msg_read(msg_rbuf, len);
            }
        }

#if USB_HOST
        if (flags & FLAG_HOST_COMM_USB_DONE) {
#else
        if (flags & FLAG_HOST_COMM_UART_DONE) {
#endif
            msg_read(msg_rbuf, msg_rbuf[2]+4);  // process command
        }
    }
}

void host_com_init(void)
{
#if USB_HOST
    usb_init();
#else
    kdp_uart_com_init();
#endif

    tid_host_comm = osThreadNew(host_comm_thread, NULL, NULL);
#if 0
    osStatus_t status;    status = osThreadSetPriority(tid_host_comm, osPriorityNormal);
    if (status != osOK) {
        fLib_printf("osThreadSetPriority (host comm) error: %d\n", status);
    }
#endif
}
