#include <stdio.h>
#include <string.h>

#include "cmsis_os2.h"
#include "ota_update.h"
#include "kdp_crc.h"
#include "usb_com.h"
#include "Driver_Flash.h"
#include "DrvSPI020.h"
#include "msg.h"
#include "kneron_mozart.h"
#include "dbg.h"
#include "com.h"
#include "delay.h"
#include "io.h"
#include "power_manager.h"
#include "kdp_ddr_table.h"

#define MODEL_INFO_FLASH_ADDR       0x00100000
#define MODEL_INFO_SIZE             0XE7
#define MODEL_ALL_BIN_FLASH_ADDR    0x00101000

#define POLY 0x8408

// disable compile optimization
//#pragma GCC push_options
//#pragma GCC optimize ("O0")


extern ARM_DRIVER_FLASH Driver_Flash;
static ARM_DRIVER_FLASH *pdrv_flash = &Driver_Flash;

extern u8 msg_rbuf[MSG_DATA_BUF_MAX + sizeof(MsgHdr) + sizeof(RspPram) + 4 + 4];
extern u8 msg_tbuf[MSG_DATA_BUF_MAX + sizeof(MsgHdr) + sizeof(RspPram) + 4 + 4];

extern u32 host_mem_read(u32 address, u32 image_size);

#define  BOOT_STATE_CONFIRMED    0x1
#define  BOOT_STATE_FIRST_BOOT   0x2
#define  BOOT_STATE_POST_FIRST_BOOT  0x4
#define  BOOT_STATE_NOT_CONFIRMED    0x8

#define MAX(a, b) ((a < b) ? (b) : (a))

#define MAX_BOOT_SEQ  0x7ffffff0

extern int flash_wait_ready(int timeout_ms);

typedef struct {
    u32 partition_id;
    u32 seq;
    u32 flag;
} ota_boot_cfg_item_t;

typedef struct {
    ota_boot_cfg_item_t scpu_cfg;
    ota_boot_cfg_item_t ncpu_cfg;
} ota_boot_cfg_t;

static ota_boot_cfg_t boot_cfg_0, boot_cfg_1;

int flashing;

int ota_update_sleep(enum pm_device_id dev_id)
{
    while (flashing == 1) {
        err_msg("ota_update_sleep: stop for flashing.\n");
        osThreadFlagsWait(BIT27, osFlagsWaitAll, osWaitForever);
    }
    err_msg("ota_update_sleep: ok\n");

    return 0;
}

int ota_update_deep_sleep(enum pm_device_id dev_id)
{
    while (flashing == 1) {
        err_msg("ota_update_deep_sleep: stop for flashing.\n");
        osThreadFlagsWait(BIT27, osFlagsWaitAll, osWaitForever);
    }
    err_msg("ota_update_deep_sleep: ok\n");

    return 0;
}

struct pm_s ota_update_pm = {
    .sleep          = ota_update_sleep,
    .deep_sleep     = ota_update_deep_sleep,
};

/*
Compare flash content with buffer content, the size shall be more than 4k,
     because small block (<4k) was verified already, here is to verify the
     consistency/integrity among 4k blocks
Return:
    0 - success
    -1 - fail
*/

static int ota_post_flash_verify_4kblock(u32 flash_addr, u32 size, u8 *pbuf)
{
    int remainder, loop, i, ret;

    if (size < 0x1000) {
        dbg_msg("size is less than 4k\n");
        return -1;
    }

    loop = size / 0x1000;
    remainder = size % 0x1000;
    u8* flash_sector_check = msg_rbuf + 0x40;

    for (i = 0; i < loop; i++) {
        pdrv_flash->ReadData((flash_addr + i * 0x1000) & 0xFFFFF000,
            flash_sector_check, 0x1000);  // read the new sector
        ret = flash_wait_ready(300);
        if (ret == -1)
        {
            err_msg("Flash read failure, timeout\n");
            return -1;
        }

        ret = memcmp(flash_sector_check, pbuf + i * 0x1000, 0x1000);
        if (ret != 0)
        {
            err_msg("Found diff, flash failed\n");
            return -1;
        }
    }

    if (remainder != 0) {
        pdrv_flash->ReadData(flash_addr + loop * 0x1000, flash_sector_check, remainder);
        ret = flash_wait_ready(300);
        if (ret == -1)
        {
            err_msg("Flash read failure, timeout\n");
            return -1;
        }

        ret = memcmp(flash_sector_check, pbuf + loop * 0x1000, size);
        if (ret != 0)
        {
            err_msg("Flash compare failed\n");
            return -1;
        }

    }

    return 0;
}


/*
Write memory data to flash, the size is 4k blocks (i.e. n*4k)

Return:
    0 - success
    -1 - fail
*/
static int ota_mem_to_flash_4k_blocks(u32 mem_addr, u32 flash_addr, u32 size)
{
    int ret;

    /* validate parameters */
    if (size % 0x1000 != 0)
    {
        err_msg("Wrong size, not n*4K");
        return -1;
    }

    if (mem_addr & 0x1f != 0)
    {
        err_msg("memory address does not align to 32bit boundary");
        return -1;
    }

    if (flash_addr & 0x1f != 0)
    {
        err_msg("flash address does not align to 32bit boundary");
        return -1;
    }

    /* erase flash sectors firstly */

    u16 sect_num = size / 0x1000;     /* sector size = 4K */
    u32 offset = 0;
    for (int sect = 0; sect < sect_num; sect++)
    {
        pdrv_flash->EraseSector(flash_addr + offset);
        ret = flash_wait_ready(300);  // wait for the erase operation to complete
        if (ret < 0) {
            dbg_msg("Erase Flash Sector Timeout\n");
        }

        for (int i = 0; i < 0x1000; i += 0x100) {
            pdrv_flash->ProgramData((uint32_t)(flash_addr + offset + i),
                (void *)(mem_addr + offset + i), 0x100);
            flash_wait_ready(500);
        }

        /* read back for confirmation*/
        u8* flash_sector_check = msg_rbuf + 0x40;
        pdrv_flash->ReadData(flash_addr + offset, flash_sector_check, 0x1000);
        ret = flash_wait_ready(300);
        if (memcmp((void *)(mem_addr + offset), (void *)flash_sector_check, 0x1000)) {
            err_msg("Flash readback verification fail at flash addr=%x\n",
                (flash_addr + offset));
            return -1;
        }
        offset += 0x1000;
    }

    delay_us(500 * 1000);
    return 0;
}

/*
Write memory data to flash, the size is less than 4k

Return:
    0 - success
    -1 - fail
*/
static int ota_mem_to_flash_small_block(u32 mem_addr, u32 flash_addr, u32 size)
{
    int ret;

    /* validate parameters */
    if (size >= 0x1000)
    {
        err_msg("Wrong size, bigger than 4K");
        return -1;
    }

    if (mem_addr & 0x1f != 0)
    {
        err_msg("memory address does not align to 32bit boundary");
        return -1;
    }

    if (flash_addr & 0x1f != 0)
    {
        err_msg("flash address does not align to 32bit boundary");
        return -1;
    }

    /* erase flash sectors firstly */

    pdrv_flash->EraseSector(flash_addr);
    ret = flash_wait_ready(300);  // wait for the erase operation to complete
    if (ret < 0) {
        dbg_msg("Erase Flash Sector Timeout\n");
    }

    pdrv_flash->ProgramData(flash_addr, (void *)mem_addr, size);
    flash_wait_ready(500);

    /* read back for confirmation*/
    u8* flash_sector_check = msg_rbuf + 0x40;
    pdrv_flash->ReadData(flash_addr, flash_sector_check, size);
    ret = flash_wait_ready(300);
    if (memcmp((void *)mem_addr, (void *)flash_sector_check, size)) {
        err_msg("Flash readback verification fail at flash addr=%x\n", flash_addr);
        return -1;
    }

    delay_us(200 * 1000);

    return 0;
}

static void ota_init_partition_boot_cfg()
{
    int ret;
    boot_cfg_0.scpu_cfg.partition_id = 0;
    boot_cfg_0.scpu_cfg.seq = 1;
    boot_cfg_0.scpu_cfg.flag = BOOT_STATE_CONFIRMED;

    boot_cfg_0.ncpu_cfg.partition_id = 0;
    boot_cfg_0.ncpu_cfg.seq = 1;
    boot_cfg_0.ncpu_cfg.flag = BOOT_STATE_CONFIRMED;

    boot_cfg_1.scpu_cfg.partition_id = 1;
    boot_cfg_1.scpu_cfg.seq = 0;
    boot_cfg_1.scpu_cfg.flag = BOOT_STATE_NOT_CONFIRMED;

    boot_cfg_1.ncpu_cfg.partition_id = 1;
    boot_cfg_1.ncpu_cfg.seq = 0;
    boot_cfg_1.ncpu_cfg.flag = BOOT_STATE_NOT_CONFIRMED;

    pdrv_flash->EraseSector(PARTITION_0_CFG_START_IN_FLASH);

    ret = flash_wait_ready(200);
    if (ret < 0) {
        err_msg("Error: Erase Flash Sector Timeout\n");
    }

    pdrv_flash->EraseSector(PARTITION_1_CFG_START_IN_FLASH);

    ret = flash_wait_ready(200);
    if (ret < 0) {
        err_msg("Error: Erase Flash Sector Timeout\n");
    }

    pdrv_flash->ProgramData(PARTITION_0_CFG_START_IN_FLASH, &boot_cfg_0, 32);
    ret = flash_wait_ready(200);
    if (ret < 0) {
        err_msg("Error: Flash partition 0 config Timeout\n");
    }

    pdrv_flash->ProgramData(PARTITION_1_CFG_START_IN_FLASH, &boot_cfg_1, 32);
    ret = flash_wait_ready(200);
    if (ret < 0) {
        err_msg("Error: Flash partition 1 config Timeout\n");
    }

}

/* get active SCPU partition ID

Return:
    0 - partition 0
    1 - partition 1
    -1 - error condition (2 active partitions)
*/

int ota_get_active_scpu_partition()
{
    pdrv_flash->ReadData(PARTITION_0_CFG_START_IN_FLASH, &boot_cfg_0, sizeof(boot_cfg_0));
    flash_wait_ready(300);
    pdrv_flash->ReadData(PARTITION_1_CFG_START_IN_FLASH, &boot_cfg_1, sizeof(boot_cfg_1));
    flash_wait_ready(300);

    if ((boot_cfg_0.scpu_cfg.flag == 0xffffffff) && (boot_cfg_1.scpu_cfg.flag == 0xffffffff)) {
        // no config data is there, need to create them for partition 0/1
        ota_init_partition_boot_cfg();
        return 0;
    }

    if (boot_cfg_0.scpu_cfg.flag & boot_cfg_1.scpu_cfg.flag & BOOT_STATE_CONFIRMED)
    {
        err_msg("Critical Error: 2 active SCPU boot config\n");
        return -1;
    }

    if ((boot_cfg_0.scpu_cfg.partition_id == boot_cfg_1.scpu_cfg.partition_id)
        && (boot_cfg_0.scpu_cfg.seq == boot_cfg_1.scpu_cfg.seq))
    {
        // no config data is there, need to create them for partition 0/1
        ota_init_partition_boot_cfg();
        return 0;
    }

    if ((boot_cfg_0.scpu_cfg.flag & BOOT_STATE_CONFIRMED) == BOOT_STATE_CONFIRMED)
        return 0;

    if ((boot_cfg_1.scpu_cfg.flag & BOOT_STATE_CONFIRMED) == BOOT_STATE_CONFIRMED)
        return 1;
    return 0;
}

/* get active NCPU partition ID

Return:
    0 - partition 0
    1 - partition 1
    -1 - error condition (2 active partitions)
*/

int ota_get_active_ncpu_partition()
{
    pdrv_flash->ReadData(PARTITION_0_CFG_START_IN_FLASH, &boot_cfg_0, sizeof(boot_cfg_0));
    flash_wait_ready(300);

    pdrv_flash->ReadData(PARTITION_1_CFG_START_IN_FLASH, &boot_cfg_1, sizeof(boot_cfg_1));
    flash_wait_ready(300);

    if ((boot_cfg_0.ncpu_cfg.flag == 0xffffffff) && (boot_cfg_1.ncpu_cfg.flag == 0xffffffff)) {
        // no config data is there, need to create them for partition 0/1
        ota_init_partition_boot_cfg();
        return 0;
    }

    if (boot_cfg_0.ncpu_cfg.flag & boot_cfg_1.ncpu_cfg.flag & BOOT_STATE_CONFIRMED)
    {
        err_msg("Critical Error: 2 active NCPU boot config\n");
        return -1;
    }

    if ((boot_cfg_0.ncpu_cfg.partition_id == boot_cfg_1.ncpu_cfg.partition_id) &&
        (boot_cfg_0.ncpu_cfg.seq == boot_cfg_1.ncpu_cfg.seq))
    {
        // no config data is there, need to create them for partition 0/1
        ota_init_partition_boot_cfg();
        return 0;
    }

    if ((boot_cfg_0.ncpu_cfg.flag & BOOT_STATE_CONFIRMED) == BOOT_STATE_CONFIRMED)
        return 0;

    if ((boot_cfg_1.ncpu_cfg.flag & BOOT_STATE_CONFIRMED) == BOOT_STATE_CONFIRMED)
        return 1;
    return 0;
}

int ota_update_scpu()
{
    int ret;
    u8 *pBase;
    u32 local_sum32, remote_sum32;
    u8  pre_active_partition;
    u32 flash_cfg_addr, flash_data_addr;
    ota_boot_cfg_t ota_cfg;
    u32 seq;

    flashing = 1;
    err_msg("ota_update_scpu: flashing ...\n");

    NVIC_DisableIRQ(UART_FTUART010_0_IRQ);       //UART0
    NVIC_DisableIRQ(UART_FTUART010_1_IRQ);       //UART1
    NVIC_DisableIRQ(UART_FTUART010_1_1_IRQ);     //UART2
    NVIC_DisableIRQ(UART_FTUART010_1_2_IRQ);     //UART3
    NVIC_DisableIRQ(UART_FTUART010_1_3_IRQ);     //UART4
    NVIC_DisableIRQ(NPU_NPU_IRQ);                //NPU

    pBase = (u8 *)KDP_DDR_BASE;
    memset(pBase, 0, SCPU_IMAGE_SIZE);
    ret = host_mem_read((u32)pBase, SCPU_IMAGE_SIZE);
    if (ret == SCPU_IMAGE_SIZE)
    {
        local_sum32 = kdp_gen_sum32(pBase, SCPU_IMAGE_SIZE - 4);
        remote_sum32 = *(u32 *)(pBase + SCPU_IMAGE_SIZE - 4);
        if (local_sum32 != remote_sum32) {
            ret = OTA_AUTH_FAIL;
            goto exit;
        }

    }
    else {
        ret = OTA_DOWNLOAD_FAIL;
        goto exit;
    }

    pre_active_partition = ota_get_active_scpu_partition();
    if (pre_active_partition == 0) {
        // active partition is 0, flash to partition 1
        flash_cfg_addr = PARTITION_1_CFG_START_IN_FLASH;
        flash_data_addr = SCPU_PARTITION1_START_IN_FLASH;
    }
    else if(pre_active_partition == 1){
        // active partition is 1, flash to partition 0
        flash_cfg_addr = PARTITION_0_CFG_START_IN_FLASH;
        flash_data_addr = SCPU_PARTITION0_START_IN_FLASH;
    } else {
        ret = OTA_FLASH_FAIL;
        goto exit;
    }

    // flash data
    ret = ota_mem_to_flash_4k_blocks((u32)pBase, (u32)flash_data_addr, SCPU_IMAGE_SIZE);
    if (ret == -1)
    {
        ret = OTA_FLASH_FAIL;
        goto exit;
    }

    ret = ota_post_flash_verify_4kblock(flash_data_addr, SCPU_IMAGE_SIZE, pBase);
    if (ret != 0) {
        err_msg("Error: post flash verification failed\n");
        ret = OTA_FLASH_FAIL;
        goto exit;
    }

    // update boot cfg

    if (pre_active_partition == 0) {
        seq = MAX(boot_cfg_0.scpu_cfg.seq, boot_cfg_1.scpu_cfg.seq);

        if(seq > MAX_BOOT_SEQ){
            boot_cfg_0.scpu_cfg.seq = 0;
            boot_cfg_1.scpu_cfg.seq = 1;
        } else {
            boot_cfg_1.scpu_cfg.seq = seq + 1;
        }
        boot_cfg_1.scpu_cfg.flag = BOOT_STATE_FIRST_BOOT;
        ota_cfg = boot_cfg_1;
    }
    else {
        seq = MAX(boot_cfg_0.scpu_cfg.seq, boot_cfg_1.scpu_cfg.seq);
 
        if(seq > MAX_BOOT_SEQ){
            boot_cfg_0.scpu_cfg.seq = 1;
            boot_cfg_1.scpu_cfg.seq = 0;
        } else {
            boot_cfg_0.scpu_cfg.seq = seq + 1;
        }
        boot_cfg_0.scpu_cfg.flag = BOOT_STATE_FIRST_BOOT;
        ota_cfg = boot_cfg_0;
    }

    ret = ota_mem_to_flash_small_block((u32)&ota_cfg, flash_cfg_addr, sizeof(ota_cfg));

    pdrv_flash->ReadData(PARTITION_0_CFG_START_IN_FLASH, &boot_cfg_0, sizeof(boot_cfg_0));
    flash_wait_ready(300);

    pdrv_flash->ReadData(PARTITION_1_CFG_START_IN_FLASH, &boot_cfg_1, sizeof(boot_cfg_1));
    flash_wait_ready(300);

    if (pre_active_partition == 0) {
        ret = memcmp(&boot_cfg_1, &ota_cfg, sizeof(ota_cfg));
    }
    else {
        ret = memcmp(&boot_cfg_0, &ota_cfg, sizeof(ota_cfg));
    }

    if (ret == -1)
        ret = OTA_FLASH_FAIL;
    else
        ret = OTA_UPDATE_SUCCESS;

exit:
    err_msg("ota_update_scpu: flashing done\n");
    flashing = 0;
    return ret;
}

int ota_update_ncpu()
{
    int ret;
    u8 *pBase;
    u32 local_sum32, remote_sum32;
    u8  pre_active_partition;
    u32 flash_cfg_addr, flash_data_addr;
    ota_boot_cfg_t ota_cfg;
    u32 seq;

    flashing = 1;
    err_msg("ota_update_ncpu: flashing ...\n");

    NVIC_DisableIRQ(UART_FTUART010_0_IRQ);       //UART0
    NVIC_DisableIRQ(UART_FTUART010_1_IRQ);       //UART1
    NVIC_DisableIRQ(UART_FTUART010_1_1_IRQ);     //UART2
    NVIC_DisableIRQ(UART_FTUART010_1_2_IRQ);     //UART3
    NVIC_DisableIRQ(UART_FTUART010_1_3_IRQ);     //UART4
    NVIC_DisableIRQ(NPU_NPU_IRQ);                //NPU

    pBase = (u8 *)KDP_DDR_BASE;
    memset(pBase, 0, NCPU_IMAGE_SIZE);
    ret = host_mem_read((u32)pBase, (u32)NCPU_IMAGE_SIZE);
    if (ret == NCPU_IMAGE_SIZE)
    {
        local_sum32 = kdp_gen_sum32(pBase, NCPU_IMAGE_SIZE - 4);
        remote_sum32 = *(u32 *)(pBase + NCPU_IMAGE_SIZE - 4);
        if (local_sum32 != remote_sum32) {
            ret = OTA_AUTH_FAIL;
            goto exit;
        }

    }
    else {
        ret = OTA_DOWNLOAD_FAIL;
        goto exit;
    }

    pre_active_partition = ota_get_active_ncpu_partition();
    if (pre_active_partition == 0) {
        // active partition is 0, flash to partition 1
        flash_cfg_addr = PARTITION_1_CFG_START_IN_FLASH;
        flash_data_addr = NCPU_PARTITION1_START_IN_FLASH;
    }
    else if (pre_active_partition == 1){
        // active partition is 1, flash to partition 0
        flash_cfg_addr = PARTITION_0_CFG_START_IN_FLASH;
        flash_data_addr = NCPU_PARTITION0_START_IN_FLASH;
    } else {
        ret = OTA_FLASH_FAIL;
        goto exit;
    }

    // flash data
    ret = ota_mem_to_flash_4k_blocks((u32)pBase, (u32)flash_data_addr, NCPU_IMAGE_SIZE);
    if (ret == -1)
    {
        ret = OTA_FLASH_FAIL;
        goto exit;
    }

    ret = ota_post_flash_verify_4kblock(flash_data_addr, NCPU_IMAGE_SIZE, pBase);
    if (ret != 0) {
        err_msg("Error: post flash verification failed\n");
        ret = OTA_FLASH_FAIL;
        goto exit;
    }

    // update boot cfg

    if (pre_active_partition == 0) {
        seq = MAX(boot_cfg_0.ncpu_cfg.seq, boot_cfg_1.ncpu_cfg.seq);

        if(seq > MAX_BOOT_SEQ){
            boot_cfg_0.ncpu_cfg.seq = 0;
            boot_cfg_1.ncpu_cfg.seq = 1;
        } else {
            boot_cfg_1.ncpu_cfg.seq = seq + 1;
        }

        boot_cfg_1.ncpu_cfg.flag = BOOT_STATE_FIRST_BOOT;
        ota_cfg = boot_cfg_1;
    }
    else {
        seq = MAX(boot_cfg_0.ncpu_cfg.seq, boot_cfg_1.ncpu_cfg.seq);

        if(seq > MAX_BOOT_SEQ){
            boot_cfg_0.ncpu_cfg.seq = 1;
            boot_cfg_1.ncpu_cfg.seq = 0;
        } else {
            boot_cfg_0.ncpu_cfg.seq = seq + 1;
        }

        boot_cfg_0.ncpu_cfg.flag = BOOT_STATE_FIRST_BOOT;
        ota_cfg = boot_cfg_0;
    }

    ret = ota_mem_to_flash_small_block((u32)&ota_cfg, flash_cfg_addr, sizeof(ota_cfg));
    if (ret == -1)
    {
        ret = OTA_FLASH_FAIL;
        goto exit;
    }

    pdrv_flash->ReadData(PARTITION_0_CFG_START_IN_FLASH, &boot_cfg_0, sizeof(boot_cfg_0));
    flash_wait_ready(300);

    pdrv_flash->ReadData(PARTITION_1_CFG_START_IN_FLASH, &boot_cfg_1, sizeof(boot_cfg_1));
    flash_wait_ready(300);

    if (pre_active_partition == 0) {
        ret = memcmp(&boot_cfg_1, &ota_cfg, sizeof(ota_cfg));
    }
    else {
        ret = memcmp(&boot_cfg_0, &ota_cfg, sizeof(ota_cfg));
    }

    if (ret != 0) {
        err_msg("Error: partition config read back compare fail\n");
        ret = OTA_FLASH_FAIL;
    } else {
        ret = OTA_UPDATE_SUCCESS;
    }

exit:
    err_msg("ota_update_scpu: flashing done\n");
    flashing = 0;
    return ret;
}

int ota_update_model(u32 model_id, u32 size)
{
    u32 sum32_download, sum32_embedded;
    u32 ddr_buf;
    int ret;
    u16 tmp;

    NVIC_DisableIRQ(UART_FTUART010_0_IRQ);       //UART0
    NVIC_DisableIRQ(UART_FTUART010_1_IRQ);       //UART1
    NVIC_DisableIRQ(UART_FTUART010_1_1_IRQ);     //UART2
    NVIC_DisableIRQ(UART_FTUART010_1_2_IRQ);     //UART3
    NVIC_DisableIRQ(UART_FTUART010_1_3_IRQ);     //UART4
    NVIC_DisableIRQ(NPU_NPU_IRQ);                //NPU

    ddr_buf = KDP_DDR_BASE;

    ret = host_mem_read(ddr_buf, size);
    if (ret == size)
    {
        sum32_embedded = *(u32 *)(ddr_buf + size - 4);
        sum32_download = kdp_gen_sum32((u8 *)ddr_buf, size - 4);
        if (sum32_embedded != sum32_download)
        {
            return OTA_AUTH_FAIL;
        }

    }
    else {
        return OTA_DOWNLOAD_FAIL;
    }

    if ((size % 0x1000) != 0)
    {
        tmp = (size + 0x1000 - 1) / 0x1000;
        size = tmp * 0x1000;
    }

    // flash data
    ret = ota_mem_to_flash_4k_blocks(ddr_buf, MODEL_INFO_FLASH_ADDR, size);
    if (ret == -1)
    {
        return OTA_FLASH_FAIL;
    }

    ret = ota_post_flash_verify_4kblock((u32)MODEL_INFO_FLASH_ADDR, size, (u8 *)ddr_buf);
    if (ret != 0) {
        err_msg("Error: post flash verification failed\n");
        return -1;
    }

    return OTA_UPDATE_SUCCESS;

}

int ota_update_switch_active_partition(u32 partition)
{
    u32 seq;
    int ret;

    if((partition != 1) && (partition != 2)) {
        err_msg("Error: wrong partition number\n");
        return -1;
    }

    pdrv_flash->ReadData(PARTITION_0_CFG_START_IN_FLASH, &boot_cfg_0, sizeof(boot_cfg_0));
    flash_wait_ready(300);

    pdrv_flash->ReadData(PARTITION_1_CFG_START_IN_FLASH, &boot_cfg_1, sizeof(boot_cfg_1));
    flash_wait_ready(300);

    if ((boot_cfg_0.scpu_cfg.flag == 0xffffffff) && (boot_cfg_1.scpu_cfg.flag == 0xffffffff)) {
        // no config data is there, need to create them for partition 0/1
        ota_init_partition_boot_cfg();
        return 0;
    }

    if (boot_cfg_0.ncpu_cfg.flag & boot_cfg_1.ncpu_cfg.flag & BOOT_STATE_CONFIRMED)
    {
        err_msg("Critical Error: 2 active NCPU boot config\n");
        return -1;
    }

    if ((boot_cfg_0.ncpu_cfg.partition_id == boot_cfg_1.ncpu_cfg.partition_id) 
        && (boot_cfg_0.ncpu_cfg.seq == boot_cfg_1.ncpu_cfg.seq))
    {
        // no config data is there, need to create them for partition 0/1
        ota_init_partition_boot_cfg();
        err_msg("only one partition, cannot switch\n");
        return -1;
    }

    if (partition == 1)     //switch SCPU
    {
        /* determine if any update ever happened, if not, just return */
        if ((boot_cfg_0.scpu_cfg.flag == BOOT_STATE_CONFIRMED) && 
           (boot_cfg_0.scpu_cfg.seq == 1) &&
           (boot_cfg_1.scpu_cfg.flag == BOOT_STATE_NOT_CONFIRMED) && 
           (boot_cfg_1.scpu_cfg.seq == 0))
        {
            err_msg("Never have SCPU firmware updated, cannot switch\n");
            return -1;
        }

        if (boot_cfg_0.scpu_cfg.flag == BOOT_STATE_CONFIRMED)
        {
            boot_cfg_1.scpu_cfg.flag = BOOT_STATE_CONFIRMED;
            seq = MAX(boot_cfg_0.scpu_cfg.seq, boot_cfg_1.scpu_cfg.seq);
            boot_cfg_1.scpu_cfg.seq = seq + 1;
            boot_cfg_0.scpu_cfg.flag = BOOT_STATE_NOT_CONFIRMED;

        } else if (boot_cfg_1.scpu_cfg.flag == BOOT_STATE_CONFIRMED)
        {
            boot_cfg_0.scpu_cfg.flag = BOOT_STATE_CONFIRMED;
            seq = MAX(boot_cfg_0.scpu_cfg.seq, boot_cfg_1.scpu_cfg.seq);
            boot_cfg_0.scpu_cfg.seq = seq + 1;
            boot_cfg_1.scpu_cfg.flag = BOOT_STATE_NOT_CONFIRMED;
        }
    }

    if (partition == 2)     //switch NCPU
    {

        /* determine if any update ever happened, if not, just return */
        if ((boot_cfg_0.ncpu_cfg.flag == BOOT_STATE_CONFIRMED) && 
           (boot_cfg_0.ncpu_cfg.seq == 1) &&
           (boot_cfg_1.ncpu_cfg.flag == BOOT_STATE_NOT_CONFIRMED) && 
           (boot_cfg_1.ncpu_cfg.seq == 0))
        {
            err_msg("Never have NCPU firmware updated, cannot switch\n");
            return -1;
        }

        if (boot_cfg_0.ncpu_cfg.flag == BOOT_STATE_CONFIRMED)
        {
            boot_cfg_1.ncpu_cfg.flag = BOOT_STATE_CONFIRMED;
            seq = MAX(boot_cfg_0.ncpu_cfg.seq, boot_cfg_1.ncpu_cfg.seq);
            boot_cfg_1.ncpu_cfg.seq = seq + 1;
            boot_cfg_0.ncpu_cfg.flag = BOOT_STATE_NOT_CONFIRMED;

        } else if (boot_cfg_1.ncpu_cfg.flag == BOOT_STATE_CONFIRMED)
        {
            boot_cfg_0.ncpu_cfg.flag = BOOT_STATE_CONFIRMED;
            seq = MAX(boot_cfg_0.ncpu_cfg.seq, boot_cfg_1.ncpu_cfg.seq);
            boot_cfg_0.ncpu_cfg.seq = seq + 1;
            boot_cfg_1.ncpu_cfg.flag = BOOT_STATE_NOT_CONFIRMED;
        }

    }

    ret = ota_mem_to_flash_small_block((u32)&boot_cfg_0, PARTITION_0_CFG_START_IN_FLASH, sizeof(boot_cfg_0));
    if (ret == -1)
    {
        err_msg("Flash write fail on %x\n", PARTITION_0_CFG_START_IN_FLASH);
        return OTA_FLASH_FAIL;
    }

    ret = ota_mem_to_flash_small_block((u32)&boot_cfg_1, PARTITION_1_CFG_START_IN_FLASH, sizeof(boot_cfg_1));
    if (ret == -1)
    {
        err_msg("Flash write fail on %x\n", PARTITION_1_CFG_START_IN_FLASH);
        return OTA_FLASH_FAIL;
    }
    return OTA_UPDATE_SUCCESS;
}

int ota_handle_first_time_boot(void)
{
    int ret;

    /* read flash to mem */

    pdrv_flash->ReadData(PARTITION_0_CFG_START_IN_FLASH, &boot_cfg_0, sizeof(boot_cfg_0));
    flash_wait_ready(300);

    pdrv_flash->ReadData(PARTITION_1_CFG_START_IN_FLASH, &boot_cfg_1, sizeof(boot_cfg_1));
    flash_wait_ready(300);

    if ((boot_cfg_0.scpu_cfg.flag == 0xffffffff) && (boot_cfg_1.scpu_cfg.flag == 0xffffffff)) {
        ret = -1;
        goto exit;
    }

    if ((boot_cfg_0.scpu_cfg.flag == BOOT_STATE_FIRST_BOOT) ||
        (boot_cfg_0.ncpu_cfg.flag == BOOT_STATE_FIRST_BOOT) ||
        (boot_cfg_1.scpu_cfg.flag == BOOT_STATE_FIRST_BOOT) ||
        (boot_cfg_1.ncpu_cfg.flag == BOOT_STATE_FIRST_BOOT)) {
        err_msg("Error: wrong state, BOOT_STATE_FIRST_BOOT shall not be here\n");
        ret = -1;
        goto exit;
    }

    /* determine if necessary to read flash */

    if ((boot_cfg_0.scpu_cfg.flag != BOOT_STATE_POST_FIRST_BOOT) &&
        (boot_cfg_0.ncpu_cfg.flag != BOOT_STATE_POST_FIRST_BOOT) &&
        (boot_cfg_1.scpu_cfg.flag != BOOT_STATE_POST_FIRST_BOOT) &&
        (boot_cfg_1.ncpu_cfg.flag != BOOT_STATE_POST_FIRST_BOOT)) {

        ret = 0;
        goto exit;
    }

    if (boot_cfg_0.scpu_cfg.flag == BOOT_STATE_POST_FIRST_BOOT) {
        boot_cfg_0.scpu_cfg.flag = BOOT_STATE_CONFIRMED;
        if(boot_cfg_1.scpu_cfg.flag == BOOT_STATE_CONFIRMED) {
            boot_cfg_1.scpu_cfg.flag = BOOT_STATE_NOT_CONFIRMED;
        }
        err_msg("SCPU partition 0 was confirmed\n");

    }

    if (boot_cfg_0.ncpu_cfg.flag == BOOT_STATE_POST_FIRST_BOOT) {
        boot_cfg_0.ncpu_cfg.flag = BOOT_STATE_CONFIRMED;
        if(boot_cfg_1.ncpu_cfg.flag == BOOT_STATE_CONFIRMED) {
            boot_cfg_1.ncpu_cfg.flag = BOOT_STATE_NOT_CONFIRMED;
        }
        err_msg("NCPU partition 0 was confirmed\n");
    }

    if (boot_cfg_1.scpu_cfg.flag == BOOT_STATE_POST_FIRST_BOOT) {
        boot_cfg_1.scpu_cfg.flag = BOOT_STATE_CONFIRMED;
        if(boot_cfg_0.scpu_cfg.flag == BOOT_STATE_CONFIRMED) {
            boot_cfg_0.scpu_cfg.flag = BOOT_STATE_NOT_CONFIRMED;
        }
        err_msg("SCPU partition 1 was confirmed\n");
    }

    if (boot_cfg_1.ncpu_cfg.flag == BOOT_STATE_POST_FIRST_BOOT) {
        boot_cfg_1.ncpu_cfg.flag = BOOT_STATE_CONFIRMED;
        if(boot_cfg_0.ncpu_cfg.flag == BOOT_STATE_CONFIRMED) {
            boot_cfg_0.ncpu_cfg.flag = BOOT_STATE_NOT_CONFIRMED;
        }
        err_msg("NCPU partition 1 was confirmed\n");
    }

    /* write back to flash */
    ret = ota_mem_to_flash_small_block((u32)&boot_cfg_0, PARTITION_0_CFG_START_IN_FLASH, sizeof(boot_cfg_0));
    if (ret == -1)
    {
        err_msg("Flash write fail on %x\n", PARTITION_0_CFG_START_IN_FLASH);
        ret = OTA_FLASH_FAIL;
        goto exit;
    }

    ret = ota_mem_to_flash_small_block((u32)&boot_cfg_1, PARTITION_1_CFG_START_IN_FLASH, sizeof(boot_cfg_1));
    if (ret == -1)
    {
        err_msg("Flash write fail on %x\n", PARTITION_1_CFG_START_IN_FLASH);
        ret = OTA_FLASH_FAIL;
    } else {
        ret = 0;
    }

exit:
    power_manager_register(PM_DEVICE_OTA_UPDATE, &ota_update_pm);

    return ret;
}

// restore optimization
//#pragma GCC pop_options
