#ifndef __V2K_CAM_H__
#define __V2K_CAM_H__


#include <cmsis_os2.h>
#include "framework/v2k.h"
#include "base.h"


#define KDP_CAMERA_START_WITH_PREVIEW   0x00000001L

enum camera_state {
    CAMERA_STATE_IDLE = 0,
    CAMERA_STATE_INITED,
    CAMERA_STATE_RUNNING,
    CAMERA_STATE_IN_FDR_INFERENCE,
    CAMERA_STATE_IN_FDR_REGISTRATION,
    CAMERA_STATE_IN_FDR_AUTO_REGISTRATION,
    CAMERA_STATE_IN_FDR_REGISTRATION_CONFIRM,
    CAMERA_STATE_IN_FDR_REGISTRATION_POSE_JUSTIFY,
};

enum display_state {
    DISPLAY_STATE_CLOSED = 0,
    DISPLAY_STATE_OPENED,
    DISPLAY_STATE_RUNNING,
};

struct v2k_cam_context {
    //int dev_idx;
    int handle;
    unsigned int width;
    unsigned int height;
    unsigned int pixelformat;
    unsigned int old_addr;
    unsigned int curr_addr;
    struct camera_buffer_info *app_buf_infos;
    struct v2k_buffer api_buf;
};

// struct v2k_cam_context_2 {
//     osThreadId_t tid_v2k_cam;
// };


int v2k_camera_init(bool);

/* 
 * When uncomment macro : PERPETUAL_RUN, please refer to the following code snippet to implement the v2k_cam 
 * application program.
 *
    int ret;
    unsigned int frame_addr;
    unsigned int frame_size;
    
    //will open cam_test_thread
    kdp_camera_open();
    kdp_camera_start(KDP_CAMERA_START_WITH_PREVIEW);

    for (;;) {
        ret = kdp_camera_buffer_capture(&frame_addr, &frame_size);
        if (0 != ret) break;//continue;


        DSG("frame_addr=%x frame_size=%d waiting", frame_addr, frame_size);
        delay_ms(10);

        if (0 != kdp_camera_buffer_prepare())
            break;
    }

    kdp_camera_stop();
    kdp_camera_close();
*/

/* 
 * Sample code that shows how to register a call back function that 
 * captures a v2k_cam image.                                        
 *
static void sample_on_newimage(
    struct v2k_buffer *buf, void *user_start, unsigned int user_length) 
{
    static int m_frame_number = 0;
    DSG("<%s> buffer[%d]: type=%x start=0x%x len=%u frames=%d ts=%llu", 
        __func__, 
        buf->index,
        buf->type, 
        (unsigned int)user_start, 
        user_length,
        ++m_frame_number,
        buf->timestamp);
}
register_v2k_cam_newimage(sample_on_newimage);
*/
typedef void (*fn_v2k_cam_newimage)(
    unsigned int cam_idx, struct v2k_buffer *, void *user_start, unsigned int user_length);

void register_v2k_cam_newimage(fn_v2k_cam_newimage fn);

//unsigned int kdp_camera_get_buffer_cnt(struct v2k_cam_context *v2k_ctx);
//int kdp_camera_capture_frame(struct v2k_cam_context *v2k_ctx, unsigned int *addr, unsigned int *size);
int kdp_camera_capture_frame(unsigned int cam_idx, unsigned int *addr, unsigned int *size);



/**
 * @brief Create a camera instance to access a hardware camera.
 * @param [in] cam_idx : camera device index
 * @return device handle, if kdp_camera_open was successfully, or
 *         -1,            if an error occurred
 */
int kdp_camera_open(unsigned int cam_idx);
//int kdp_camera_open(struct v2k_cam_context *v2k_ctx);

/**
 * @brief Release the camera instance.
 * @param [in] cam_idx : camera device index
 * @return 0,       if kdp_camera_close was successfully, or
 *         others,  if an error occurred
 */
int kdp_camera_close(unsigned int cam_idx);
//int kdp_camera_close(struct v2k_cam_context *v2k_ctx);
/**
 * @brief Query camera device capabilities
 * @param [in] cam_idx : camera device index
 * @param [out] lp_cap : pointer to struct v2k_capability
 * @return 0,       if kdp_camera_get_device_info was successfully, or
 *         others,  if an error occurred
 */
int kdp_camera_get_device_info(unsigned int cam_idx, struct v2k_capability *lp_cap);
/**
 * @brief Set frame data format
 * @param [in] cam_idx : camera device index
 * @param [in] width : camera frame width
 * @param [in] height : camera frame height
 * @param [in] pixelformat : camera frame pixel format
 * @return 0,       if kdp_camera_set_frame_format was successfully, or
 *         others,  if an error occurred
 */
int kdp_camera_set_frame_format(unsigned int cam_idx,
                                unsigned int width,
                                unsigned int height,
                                unsigned int pixelformat);
/**
 * @brief Get frame data format
 * @param [in] cam_idx : camera device index
 * @param [out] pointer to struct v2k_format.
 * @return 0,       if kdp_camera_get_frame_format was successfully, or
 *         others,  if an error occurred
 */
int kdp_camera_get_frame_format(unsigned int cam_idx, struct v2k_format *lpfmt);
/**
 * @brief Initial camera buffer
 * @param [in] cam_idx : camera device index
 * @return 0,       if kdp_camera_buffer_init was successfully, or
 *         others,  if an error occurred
 */
int kdp_camera_buffer_init(unsigned int cam_idx);
/**
 * @brief Start capturing.
 * @param [in] cam_idx : camera device index
 * @param flags KDP_CAMERA_START_WITH_PREVIEW, capture and preview frames to lcdc 
 * @return 0,       if kdp_camera_start was successfully, or
 *         others,  if an error occurred
 */
int kdp_camera_start(unsigned int cam_idx, unsigned int flags);
//int kdp_camera_start(struct v2k_cam_context *v2k_ctx, unsigned int flags);
/**
 * @brief Stop capturing.
 * @param [in] cam_idx : camera device index
 * @return 0,       if kdp_camera_stop was successfully, or
 *         others,  if an error occurred
 */
int kdp_camera_stop(unsigned int cam_idx);
//int kdp_camera_stop(struct v2k_cam_context *v2k_ctx);
/**
 * @brief Prepare a buffer from camera sdk.
 * @param [in] cam_idx : camera device index
 * @return 0,       if kdp_camera_buffer_prepare was successfully, or
 *         others,  if an error occurred
 */
int kdp_camera_buffer_prepare(unsigned int cam_idx);
//int kdp_camera_buffer_prepare(struct v2k_cam_context *v2k_ctx);
/**
 * @brief Capture a buffer to the caller or application.
 * @param [in] cam_idx : camera device index
 * @return 0,       if kdp_camera_buffer_capture was successfully, or
 *         others,  if an error occurred
 */
int kdp_camera_buffer_capture(unsigned int cam_idx, unsigned int *addr, unsigned int *size);
//int kdp_camera_buffer_capture(struct v2k_cam_context *v2k_ctx, unsigned int *addr, unsigned int *size);

/**
 * @brief Read a camera buffer simply.
 * @param [in] cam_idx : camera device index
 * @return 0,       if kdp_camera_buffer_read was successfully, or
 *         others,  if an error occurred
 */
int kdp_camera_buffer_read(unsigned int cam_idx, unsigned int *addr, unsigned int *size);
//int kdp_camera_buffer_read(struct v2k_cam_context *v2k_ctx, unsigned int *addr, unsigned int *size);


#endif
